
typedef struct input_file
{
  struct input_file *next;
  unsigned char     *data;
  size_t             size;
  char              *file_name;
} input_file_t;

static input_file_t *input_files;

int ifile_read     (unsigned char **out, size_t *olen,
                    const char *fn, size_t pad)
{
  input_file_t *FF=NULL;
  int eno;
  int f= -1;
  ssize_t R;
  size_t off;
  char *op;
  struct stat sbuf;

  FF= xallocz(sizeof(*FF));
  FF->next= input_files;
  FF->file_name= xstrdup(fn);

  op= "open";
  f= open(fn, O_RDONLY);
  if (f<0) goto fail;

  if (fstat(f,&sbuf)) goto fail;
  errno= EBADF;
  if ((sbuf.st_mode&S_IFMT) != S_IFREG) goto fail;

  FF->size= sbuf.st_size;
  FF->data= xalloc(FF->size+pad);
  off= 0;
  op= "read";
  while(off!=FF->size)
  {
    R= read(f, FF->data+off, FF->size-off);
    if (R<=0) goto fail;
    off+= R;
  }
  input_files= FF;
  if (pad) memset(FF->data+FF->size, 0, pad);
  *out= FF->data;
  *olen= FF->size;
  return 0;

fail:
  eno= errno;
  if (f>=0) close(f);
 
  if (FF) 
  {
    xfree(FF->data);
    xfree(FF->file_name);
    free(FF);
  }

  fprintf(stderr,"error in %s(%s): %s\n", op, fn, strerror(eno));
  exit(1);
}

static input_file_t *ifile_find(const unsigned char *x)
{
  input_file_t *F;
  for(F=input_files;F;F=F->next)
    if (x>=F->data && x<F->data+F->size) 
       return F;
  return NULL;
}

static void ifile_report_file(char *fn, int lno)
{
  char buf[20];
  int lno_len, fn_len;

  sprintf(buf,"%d", lno);
  lno_len= strlen(buf) + 1;
  fn_len= strlen(fn);
  if (lno_len+fn_len<80)
  {
    fprintf(stderr,"%s:%d\n", fn, lno);
  }
  else
  {
    fn= fn + fn_len - 80 + lno_len;
    fprintf(stderr,"%s:%d\n", fn, lno);
  }
}

static void ifile_get_line 
   (unsigned char *data, const unsigned char *pos,
    unsigned char **lstart, int *lno)
{
  unsigned char *S;
  int L;
  L= 1;
  S= data;
  while(data<pos)
  {
    if (*data=='\n') { L++; S= data+1; }
    data++;
  }
  *lstart= S;
  *lno= L;
}

static unsigned char *ifile_line_end(const unsigned char *pos)
{
  while(1)
  {
    if (*pos=='\n' || *pos==0) return (unsigned char*) pos;
    pos++;
  }
}

static void ifile_mark(int n)
{
  int i;
  for(i=0;i<n;i++) fprintf(stderr," ");
  fprintf(stderr,"^\n");
}

static void ifile_print(const unsigned char *start, int len, int ln)
{
  while(len && ((start[0]&0xc0)==0x80)) { start++; len--; }
  fwrite(start, 1, len, stderr);
  if (ln) fprintf(stderr,"\n");
}

static void ifile_print_sep(char x)
{
  int i;
  for(i=0;i<80;i++) fprintf(stderr,"%c",x);
  fprintf(stderr,"\n");
}

static void ifile_report_pos(unsigned char *start, const unsigned char *pos)
{
  int line_length, delta;
  const unsigned char *end, *pstart, *pend;

  end= ifile_line_end(pos);

  ifile_print_sep('-');
  line_length= end - start; 
  if (line_length<=80)
  {
    ifile_print(start, line_length, 1);
    ifile_mark(pos-start);
  }
  else
  {
    pstart= pos-40;
    pend= pos+40;

    if (pend>end) { delta= pend-end; pend= end; pstart += delta; }
    if (pstart<start) { delta= start-pstart; pend+= delta; pstart= start; }
    if (pend>end) { pend= end; }

    ifile_print(pstart, pend-pstart, 1);
    ifile_mark(pos-pstart);
  }
  ifile_print_sep('=');
}

void ifile_report  (const unsigned char *pos)
{
  input_file_t *FF;
  unsigned char *lstart;
  int lno;
  FF= ifile_find(pos);
  if (!FF)
  {
    ifile_print_sep('=');
    fprintf(stderr,"Unknown input position\n");
    ifile_print_sep('=');
    return ;
  } 
  ifile_get_line(FF->data, pos, &lstart, &lno);
  ifile_print_sep('=');
  ifile_report_file(FF->file_name, lno);
  ifile_report_pos(lstart, pos);
}

char *get_pos_ifile(const unsigned char *pos)
{
  input_file_t *FF;
  FF= ifile_find(pos);
  if (!FF)
  {
    fprintf(stderr,"Internal error: Unknown input file\n");
    exit(1);
  }
  return FF->file_name;
}

void irror(uchar *pos, const char *fmt, ...)
{
  va_list ap;
  ifile_report(pos);
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  exit(1);
}

void irror2(uchar *pos1, uchar *pos2, const char *fmt, ...)
{
  va_list ap;
  ifile_report(pos1);
  ifile_report(pos2);
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  exit(1);
}

void iwarn(uchar *pos, const char *fmt, ... )
{
  va_list ap;
  ifile_report(pos);
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
}
