
/***
  This file defines a bytecode buffer.

  It's capable of gradually building bytecode from different sources
  such as direct insertions of bytes, referrals to files and containing
  other bytecode structures within.
 ***/

enum { BYCO_HEAD, BYCO_SUB, BYCO_MEM, BYCO_FILE };

typedef struct byco_node
{
  int type;
  struct byco_node *prev, *next;
  union {
    struct
    {
      struct byco    *sub;
      int       own_sub;
    };
    struct 
    {
      uint32_t  data_size;
      uint32_t  data_len;
      uint8_t  *data;
      uint32_t  own_data;
    };
    struct
    {
      char     *file_name;
      uint32_t  file_size;
    };
  };
} byco_node_t;

struct byco 
{
  int nvars;           // place-holder for number of local variables used
                       // this is set externally.
  byco_node_t *head;
};

static byco_node_t *byco_node(int type)
{
  byco_node_t *N;
  N= xallocz(sizeof(*N));
  N->type= type;
  return N;
}

struct byco *byco_new()
{
  struct byco *B;
  byco_node_t *H;
  B= xallocz(sizeof(*B));
  H= byco_node(BYCO_HEAD);
  H->prev= H->next= H;
  B->head= H;
  return B;
}

static void byco_append_node(struct byco *B, byco_node_t *N)
{
  N->prev= B->head->prev;
  N->next= B->head;
  B->head->prev->next= N;
  B->head->prev= N;
}

/**
  returns the last node in the buffer if it's a memory node
  and there is space inside it.

  otherwise, creates a new memory node and appends to the buffer.
 **/
static byco_node_t *byco_node_mem(struct byco *B)
{
  byco_node_t *N;

  N= B->head->prev;
  if (N->type==BYCO_MEM && N->data_len<N->data_size)
      return N;
  
  N= byco_node(BYCO_MEM);
  N->data_size= 1024;
  N->data= xallocz(N->data_size);
  N->own_data= 1;
  byco_append_node(B,N);
  return N;
}

/***
  This is the most used function of this structure.
  Copies the given amount of data into the buffer, 
  creating MEM type nodes.
 ***/
void byco_cpy(struct byco *B, const void *data, size_t len)
{
  byco_node_t *N;
  const uint8_t *D;
  size_t cp;
  D= data;
  while(len)
  {
    N= byco_node_mem(B);
    cp= imin2(N->data_size - N->data_len, len);
    memcpy(N->data+N->data_len, D, cp);
    N->data_len += cp;
    D+= cp;
    len-= cp;
  }
}

void byco_byte(struct byco *B, uint8_t V)
{
  byco_cpy(B, &V, 1);
}

 // encodes an opcode.
void byco_op(struct byco *B, uint8_t V)
{
  byco_byte(B, V);
}

 // yığına atılmak üzere bir tamsayı sabitini kodlar.
 // tamsayının kaç bayt kapladığına bağlı olarak farklı
 // komutlar kullanılabilir.
void byco_int   (struct byco *B, uint32_t V)
{
  uint8_t D[5];
  int     i;
  i= 0;
  if (V&(0xff<<24))
  {
    D[i++]= GUII_C4;
    D[i++]= V>>24;
    D[i++]= V>>16;
    D[i++]= V>>8;
    D[i++]= V;
  }
  else if (V&(0xff<<16))
  {
    D[i++]= GUII_C3;
    D[i++]= V>>16;
    D[i++]= V>>8;
    D[i++]= V;
  }
  else if (V&(0xff<<8))
  {
    D[i++]= GUII_C2;
    D[i++]= V>>8;
    D[i++]= V;
  }
  else if ( V & 0x80 )
  {
    D[i++]= GUII_C1;
    D[i++]= V;
  }
  else
  {
    D[i++]= 0x80 | V;
  }
  byco_cpy(B, D, i);
}

void byco_str(struct byco *B, const char* str)
{
  uint32_t  L;
  if (!str) str= "";
  L= strlen(str)+1;
  byco_int(B,L);
  byco_byte(B, GUII_STR);
  byco_cpy(B, str, L);
}

/** 
  floating point values are always encoded with 4 bytes.
 **/
void byco_float (struct byco *B, float F)
{
  int32_t V;
  uint32_t U;
  uint8_t D[5];
  int i;
  V= F*65536;
  U= V;
  i= 0;
  D[i++]= GUII_CF;
  D[i++]= U>>24;
  D[i++]= U>>16;
  D[i++]= U>>8;
  D[i++]= U;
  byco_cpy(B, D, 5); 
}

   // Appends a buffer to another one.
void byco_byco(struct byco *dst, struct byco *src,int own)
{
  byco_node_t *N;
  N= byco_node(BYCO_SUB);
  N->sub= src;
  N->own_sub= own;
  byco_append_node(dst, N);
}

void ioerror(const char *op, const char *fn)
{
  fprintf(stderr,"%s(%s): %s\n", op, fn, strerror(errno));
  exit(1);
}


  // appends a file to the buffer. it's not read at this time,
  // only a reference to it is maintained.
void byco_file(struct byco *B, const char *fn)
{
  byco_node_t *N;
  struct stat sb;

  if (stat(fn, &sb)) ioerror("stat", fn);
  N= byco_node(BYCO_FILE);
  N->file_name= strdup(fn);
  N->file_size= sb.st_size;
  byco_append_node(B, N);
}

// We now have some instructions with special encodings.

void byco_blob(struct byco *B, const void *data,uint32_t len)
{
  byco_int(B,len);
  byco_op(B, GUII_BLOB);
  byco_cpy(B, data, len);
}

size_t byco_size(struct byco *B);

size_t byco_file_blob(struct byco *B, const char *fn)
{
  struct byco *file;
  size_t file_size;

  file= byco_new();
  byco_file(file, fn);
  file_size= byco_size(file);

  byco_int(B, file_size);
  byco_op(B, GUII_BLOB);
  byco_byco(B, file, 1);

  return file_size;
}

void byco_setg_i(struct byco *B, uint8_t varno, uint32_t value)
{
  byco_int(B, value);
  byco_op(B, GUII_SETG);
  byco_byte(B, varno);
}

void byco_setg_s(struct byco *B, uint8_t varno, char* value)
{
  byco_str(B, value);
  byco_op(B, GUII_SETG);
  byco_byte(B, varno);
}

void byco_free(struct byco *B);

void byco_remove_node(byco_node_t *N)
{
  N->prev->next= N->next;
  N->next->prev= N->prev;
}

void byco_free_node(byco_node_t *N)
{
  if (!N) return ;
  switch(N->type)
  {
  case BYCO_HEAD: break;
  case BYCO_MEM: if (N->own_data) free(N->data); break;
  case BYCO_FILE: if (N->file_name) free(N->file_name); break;
  case BYCO_SUB: if (N->own_sub) byco_free(N->sub); break;
  }
  free(N);
}

void byco_free_nodes(struct byco *B)
{
  byco_node_t *N;
  while((N=B->head->next)!=B->head)
  {
    byco_remove_node(N);
    byco_free_node(N);
  }
}

void byco_free(struct byco *B)
{
  byco_free_nodes(B);
  byco_free_node(B->head);
  free(B);
}


/***
   Read interface
 ***/
size_t byco_size(struct byco *B)
{
  byco_node_t *N;
  size_t sz;
  sz= 0;
  for(N=B->head->next;N!=B->head;N=N->next)
  switch(N->type)
  {
  case BYCO_HEAD: break;
  case BYCO_FILE: sz+= N->file_size; break;
  case BYCO_MEM:  sz+= N->data_len; break;
  case BYCO_SUB:  sz+= byco_size(N->sub); break;
  }
  return sz;
}

struct byco_ptr 
{
  byco_node_t *node;
  size_t offset;
  int file;
  struct byco_ptr *sub;
};

struct byco_ptr *byco_read_start(struct byco *B);
void byco_read_end(struct byco_ptr *P);

void bycop_set_node(struct byco_ptr *P, byco_node_t *N)
{
  if (!P->node) goto done_unset;
  switch(P->node->type)
  {
  case BYCO_HEAD: break ;
  case BYCO_MEM: P->offset= 0; break;
  case BYCO_SUB: byco_read_end(P->sub); P->sub= NULL; break;
  case BYCO_FILE: 
     if (P->file>=0) close(P->file);
     P->file= -1;
     break;
  }
done_unset:
  P->node= N;
  if (!P->node) goto done_set;
  switch(N->type)
  {
  case BYCO_HEAD: return ;
  case BYCO_MEM: P->offset= 0; break;
  case BYCO_SUB: P->sub= byco_read_start(N->sub); break;
  case BYCO_FILE: 
       P->file= open(N->file_name, O_RDONLY);
       if (P->file<0) ioerror("open", N->file_name);
       break;
  }
done_set:
  return ;
}

struct byco_ptr *byco_read_start(struct byco *B)
{
  struct byco_ptr *P;
  P= xallocz(sizeof(*P));
  bycop_set_node(P, B->head->next);
  return P;
}

void byco_read_end(struct byco_ptr *P)
{
  bycop_set_node(P, NULL);
  free(P);
}

size_t byco_read(struct byco_ptr *P, void *out, size_t len)
{
  size_t R, cp;
  ssize_t S;
again:
  if (!P->node) return 0;
  switch(P->node->type)
  {
  case BYCO_HEAD:
     return 0;
  case BYCO_SUB:
     R= byco_read(P->sub, out, len);
     if (R) return R;
     break;
  case BYCO_FILE:
     S= read(P->file, out, len);
     if (S>0) return S;
     if (S<0) ioerror("read", P->node->file_name);
     break;
  case BYCO_MEM:
     cp= imin2(len, P->node->data_len-P->offset);
     if (!cp) break;
     memcpy(out, P->node->data+P->offset, cp);
     P->offset+= cp;
     return cp;
  }
  bycop_set_node(P, P->node->next);
  goto again;
}

void byco_collect(struct byco *B, unsigned char **Rdata, size_t *Rlen)
{
  byco_node_t *F;
  size_t cp, R;
  struct byco_ptr *P;

  F= B->head->next;
  if (F->next==B->head && F->type==BYCO_MEM) goto retf;

  F= byco_node(BYCO_MEM);
  F->data_size= byco_size(B);
  F->data_len= 0;
  F->data= xalloc(F->data_size);

  P= byco_read_start(B);
  while ((cp=F->data_size-F->data_len))
  {
    R= byco_read(P, F->data+F->data_len, cp);
    if (!R) die("internal, size mismatch in byco\n");
    F->data_len += R;
  }
  byco_read_end(P);

  byco_free_nodes(B);
  byco_append_node(B, F);

retf:
  *Rdata= F->data;
  *Rlen= F->data_len;
}

