
enum {
   node_number, node_ident, node_string,
   node_sub, node_add, node_div, node_mod, node_mul,
   node_assign, node_comma, node_negate,

   node_call, node_field, node_self_field, node_list,
};

typedef struct expr expr_t;

struct expr
{
  int type;
  unsigned char *data;
  unsigned char *pos;
  int nsub;
  struct expr **sub;
};

int expr_match(expr_t *E, int type, uchar *data, int nsub)
{
  if (!E) return 0;
  if (E->type!=type) return 0;
  if (nsub>=0 && E->nsub!=nsub) return 0;
  if (data && (!E->data || strcmp((char*)E->data, (char*)data))) return 0; 
  return 1;
}

expr_t *expr_new
    (int type, unsigned char *data, unsigned char *pos, int nsub)
{
  expr_t *E;
  E= xallocz(sizeof(*E));
  E->type= type;
  E->pos= pos;
  if (data) E->data= xstrdupu(data);
  E->nsub= nsub;
  if (nsub) E->sub= xallocz(nsub * sizeof(E->sub[0]));
  return E;
}

expr_t *expr_dup(expr_t *E, unsigned char *pos)
{
  expr_t *C;
  int i;
  C= expr_new(E->type, E->data, pos, E->nsub);
  for(i=0;i<C->nsub;i++)
    C->sub[i]= expr_dup(E->sub[i], pos);
  return C;
}

expr_t *expr_new_leaf
    (int type, unsigned char *data, unsigned char *pos)
{
  return expr_new(type, data, pos, 0);
}

expr_t *expr_new_subs
    (int type, unsigned char *pos, int nsub, ... )
{
  expr_t *E;
  int i;
  va_list ap;
  E= expr_new(type, NULL, pos, nsub);
  va_start(ap,nsub);
  for(i=0;i<nsub;i++)
    E->sub[i]= va_arg(ap,expr_t*);
  va_end(ap);
  return E;
}

void exprbuf_free(ptrbuf_t *P, int *N, expr_t ***R)
{
  size_t sz;
  ptrbuf_free(P, (void***) R, &sz);
  *N= sz;
}

expr_t *expr_new_binop(int type, uchar *pos, expr_t *L, expr_t *R)
{
  return expr_new_subs(type, pos, 2, L, R);
}

expr_t *expr_new_field(uchar *pos, expr_t *L, expr_t *R)
{
  return expr_new_binop(node_field, pos, L, R);
}

expr_t *expr_new_ident(uchar *data, uchar *pos)
{
  return expr_new_leaf(node_ident, data, pos);
}

expr_t *expr_new_field_ed(uchar *pos, expr_t *L, uchar *rdata)
{
  return expr_new_field(pos, L, expr_new_ident(rdata, pos));
}

expr_t *expr_new_self_field(uchar *pos, uchar *data)
{
  return expr_new_leaf(node_self_field, data, pos); 
}

expr_t *expr_new_assign(uchar *pos, expr_t *L, expr_t *R)
{
  return expr_new_binop(node_assign, pos, L, R);
}


int is_self_field(expr_t *E, uchar *data)
{
  //
  // ident check is redundant, the parser always outputs this
  //
  return expr_match(E, node_self_field, data, 0);
}

//
//   returns the data part of E, if E is of the form .xx
//
uchar *is_self_field_x(expr_t *E)
{
  if (expr_match(E, node_self_field, NULL, 0))
     return E->data;
  else
     return NULL;
}

//
//  returns the left child of E, if E is of the form L = R 
//
expr_t* is_assign(expr_t *E)
{
  if (expr_match(E, node_assign, NULL, 2)) return E->sub[0];
                                      else return NULL;
}

int is_ident(expr_t *E, uchar *data)
{
  return expr_match(E, node_ident, data, 0);
}

int is_field(expr_t *E, uchar *obj, uchar *field)
{
  if (!expr_match(E, node_field, NULL, 2)) return 0;
  if (!is_ident(E->sub[0], obj)) return 0;
  if (!is_ident(E->sub[1], field)) return 0;
  return 1;
}


void psubs(expr_t *E, char *mid);

void pprint_expr(expr_t *E)
{
  char *op;
  switch(E->type)
  {
   case node_sub: op= "-"; goto binop;
   case node_add: op= "+"; goto binop;
   case node_div: op= "/"; goto binop;
   case node_mod: op= "%"; goto binop;
   case node_mul: op= "*"; goto binop;
   case node_assign: op= "="; goto binop;
   case node_comma: op= ","; goto binop;
   case node_self_field: printf(".%s", E->data); break;
   case node_field: psubs(E,"."); break;
   case node_list: psubs(E,"::"); break;  
   case node_call: psubs(E,"#"); break;  
   case node_ident:
   case node_number: printf("%s",E->data); break;
   case node_string: printf("'%s'", E->data); break;
   default:
      printf("{%d", E->type); 
      if (E->data) printf("<%s>", E->data);
      psubs(E,"@");
      printf("}");
      break;
  }
  return ;

binop:
  if (E->nsub==2)
  {
    printf("("); pprint_expr(E->sub[0]); printf("%s",op);
                 pprint_expr(E->sub[1]); printf(")");
  }
  else
  {
    printf("[%s ", op);
    psubs(E,"@");
    printf("]");
  }
}

void psubs(expr_t *E, char *mid)
{
  int i;
  for(i=0;i<E->nsub;i++) { if(i) printf("%s",mid); pprint_expr(E->sub[i]); }
}

