

typedef struct varset varset_t;

struct varset
{
  int type;
  strmap_t *map;
  int count;
  value_t **values;
  int n_values;
  FPA list;
};

/**
  A varset is like a scope in programming languages. It holds a list
  of variables.

  It maintains them in two different places:
  - a map
  - a list

  The map is used for looking up the variable.

  The list is used for iterating over the set. Such iteration is done
  for global variables. Local variables are computed lazily. However,
  this is not so for global variables since their uses may occur in
  any kreen and we can't know which use is first. Therefore, global
  variables are computed at the start of the bytecode program. In
  order to make sure that all of them are computed, we need to 
  iterate over them using the list.
 **/

static char *varset_key(varset_t *T, char *name)
{
  return T->type==at_global ? vkey_global(name) : vkey_local(name);
}

static int varset_find(varset_t *T, char *name, value_t **RV, char **RK)
{
  char *key;
  value_t *V;

  if (!T) return 0; 

  key= varset_key(T, name);
  V= strmap_find(T->map, key);

  if (V)
  {
    free(key);
  }
  else
  {
    if (RK) *RK= key;
       else free(key);
  }

  *RV= V;
  return V!=NULL;
}


static value_t* varset_insert(varset_t *T, value_t *V)
{
  V->Atype= T->type;        // access type
  V->Aarg[0]= T->count++;   // idx
  strmap_insert(T->map, V->key, V);
  fpa_put(&T->list, V);
  return V;
}


static varset_t *varset_new(int type)
{
  varset_t *S;
  S= xallocz(sizeof(*S));
  S->map= strmap_new();
  S->type= type;
  S->count= 0;
  fpa_init(&S->list, &S->n_values, (void***) &S->values, 100);
  return S;
}

static void insert_system_globals(varset_t *S)
{
  int i;
  value_t *V;
  static char *global_names[]= {
    [GUIGV_FONT1HEIGHT]           "font1h",
    [GUIGV_FONT2HEIGHT]           "font2h",
    [GUIGV_FONT3HEIGHT]           "font3h",
    [GUIGV_FONT4HEIGHT]           "font4h",
    [GUIGV_FONT5HEIGHT]           "font5h",
    [GUIGV_GUI_SCROLLBAR_WIDTH]   "scroll_width",
    [GUIGV_GUI_SCROLLBAR_HEIGHT]  "scroll_height",
    [GUIGV_GUI_SYSBAR_HEIGHT]     "sysbar_height",
    [GUIGV_GUI_NAVBAR_HEIGHT]     "navbar_height"
  };
  for(i=0;i<sizeof(global_names)/sizeof(char*);i++)
    if (global_names[i])
    {
      V= value_new(v_var, vkey_global(global_names[i]), NULL);
      V->state= vs_known;
      varset_insert(S, V);
    }
}
