
typedef struct econtext econtext_t;

struct econtext 
{
  int type;
  strmap_t *wimap;
  varset_t *globals, *locals, *varset;
  strmap_t *value_map;
};

void invalid_field(uchar *pos)
{
  irror(pos, "invalid field reference\n");
}

void con_insert(econtext_t *C, value_t *V)
{
  strmap_insert(C->value_map, V->key, V);
}

/**
  A context is used for determining how an expression is going
  to be evaluated.

  Each value is stored in the map, so that we know whether it
  has been computed or not.

  However, local and global values are not stored in there, they
  are stored in the 'locals' and 'globals' varsets. 

  Only coordinate expressions are stored in the value_map.
 **/

enum { CON_GLOBAL, CON_SLAY, CON_WLAY };

 /**
   There are three contexts as listed above. In each context, the
   set of available features are different. For instance, while
   we're evaluating global variables, we may not refer to any
   kidget or kreen coordinates.
 **/

static int con_find_fk(econtext_t *C, char *key, value_t **RV)
{
  *RV= strmap_find(C->value_map, key);
  if (*RV) { free(key); return 1; }
  return 0;
}

// 
// This is called only for a kreen layout context. cc is always valid
//
static value_t *con_lookup_scoord_nc(econtext_t *C, int cc)
{
  value_t *V;
  char *key;
  if (con_find_fk(C, key= vkey_scoord(cc), &V)) return V;
  free(key);
  return NULL;
}

static value_t *con_lookup_scoord(econtext_t *C, int cc, uchar *pos)
{
  value_t *V;
  char *key;

  if (cc==-1) invalid_field(pos);
  switch(C->type)
  {
  case CON_SLAY:
  case CON_WLAY:
    if (con_find_fk(C, key= vkey_scoord(cc), &V)) return V;
    V= value_new(v_coord, key, pos);
    V->cc= cc;
    V->winame= (uchar*) "screen";
    break;
  default:
    invalid_field(pos);
  }

              switch(C->type)
              {
  // when computing a kreen layout, all dimensions of a kreen
  // are accessible. the right and bottom coordinates are stored
  // in local variables. the rest are at the coordinate array 
  // position 0.
              case CON_SLAY:
  switch(cc)
  {
  case wc_left:
  case wc_top:
  case wc_width:
  case wc_height:   make_coord_accessor(V, 0, cc); break;
  case wc_right:    make_local_accessor(V, 2);     break;
  case wc_bottom:   make_local_accessor(V, 3);     break;
  }
              break;
  // when computing a kidget layout, x and y coordinates of
  // a kreen read 0 and they can't be assigned to.
  // bottom and right coordinates are the same as height and
  // width. these are stored in local variables 0 and 1.
              case CON_WLAY:
  switch(cc)
  {
  case wc_left:
  case wc_top:      make_const0_accessor(V);       break;
  case wc_width:
  case wc_right:    make_local_accessor(V, 0);     break;
  case wc_height: 
  case wc_bottom:   make_local_accessor(V, 1);     break;
  }
  V->state= vs_known;
              break;
  }
  
  con_insert(C, V);
  return V;
}

//
//  display coordinates are only within kreen layout code
//  and there, only the width and height are accessible 
//  as local variables 0 and 1
//
static value_t *con_lookup_dcoord(econtext_t *C, int cc, uchar *pos)
{
  int vn;
  value_t *V;
  char *key;
  switch(cc)
  {
  case wc_width:   vn= 0; break;
  case wc_height:  vn= 1; break;
         default:  invalid_field(pos);
  }

  if (con_find_fk(C, key= vkey_scoord(cc), &V)) return V;
  V= value_new(v_var, key, pos);
  make_local_accessor(V, vn);
  V->state= vs_known;
  con_insert(C, V);
  return V;
}

static value_t *con_lookup_wcoord
    (econtext_t *C, uchar *obj, int cc, uchar *pos)
{
  kidget_t *W;
  value_t *V;
  char *key;

  if (con_find_fk(C, key= vkey_wcoord(obj,cc), &V)) return V;
  V= value_new(v_coord, key, pos);
  V->cc= cc;
  V->winame= obj;

  W= strmap_find(C->wimap, (char*) obj);
  if (!W) irror(pos, "unknown kidget\n");

  switch(cc)
  {
  case wc_left:
  case wc_top:
  case wc_width:
  case wc_height:
     make_coord_accessor(V, W->index_in_parent, cc);
     break;
  case wc_bottom:
  case wc_right:
     varset_insert(C->locals, V); // this simply allocates a new local var
     break;
  }

  con_insert(C, V);
  return V;
}

static value_t *con_lookup_field
   (econtext_t *C, uchar *obj, int cc, uchar *pos)
{
  switch(C->type)
  {
  case CON_GLOBAL: invalid_field(pos); 
  case CON_SLAY:
    if (!ustrcmp(obj, "display")) return con_lookup_dcoord(C, cc, pos);
    if (!ustrcmp(obj, "screen"))  return con_lookup_scoord(C, cc, pos);
    break;
  case CON_WLAY:
    if (!ustrcmp(obj, "display")) break;
    if (!ustrcmp(obj, "screen"))  return con_lookup_scoord(C, cc, pos);
    return con_lookup_wcoord(C, obj, cc, pos);
  }
  invalid_field(pos);
  return NULL;
}

static value_t *con_lookup_ident(econtext_t *C, expr_t *E)
{
  char *name;
  value_t *V;
  char *key;
  name= (char*) E->data;

  switch(C->type)
  {
  case CON_GLOBAL:
    if (varset_find(C->globals, name, &V, &key)) return V;
    return varset_insert(C->globals, value_new(v_var, key, E->pos));
  case CON_SLAY:
    if (varset_find(C->globals, name,&V, NULL)) return V;
    irror(E->pos, "unknown variable\n");
  case CON_WLAY:
    if (varset_find(C->globals, name, &V, NULL)) return V;
    if (varset_find(C->locals,  name, &V, &key)) return V;
    return varset_insert(C->locals, value_new(v_var, key, E->pos)); 
  }

  irror(E->pos, "internal error: bad context\n");
  return NULL;
}

static value_t *con_lookup_self_field(econtext_t *C, int cc, uchar *pos)
{
  if (C->type==CON_SLAY) return con_lookup_scoord(C, cc, pos); 
  invalid_field(pos);
  return NULL;
}

//
//  This is the entry function. E is an l-value. 
//  The returned value is its representation.
//  If the l-value is encountered for the first time,
//  it's value object inserted into the map.
//
static value_t *con_lookup(econtext_t *C, expr_t *E)
{
  switch(E->type)
  {
  case node_field:      
    if (E->nsub!=2) invalid_field(E->pos); 
    return con_lookup_field(C, E->sub[0]->data, ce2cc(E->sub[1]), E->pos);
  case node_self_field: 
    return con_lookup_self_field(C, ce2cc(E), E->pos);
  case node_ident:
    return con_lookup_ident(C, E);
  }
  irror(E->pos, "unrecognized value\n");
  return NULL;
}

static void insert_predef_locals(varset_t *S, char **pd)
{
  int i;
  char *name;
  for(i=0;(name=pd[i]);i++)
    varset_insert(S, value_new(v_var, vkey_local(name), NULL));
  //
  // The values created here are never used.
  // New values for these are made when they are
  // referred to in an expression. 
  // The only function of varset_insert above
  // is to increment S->count :P
  // 
  // I could simply do that, but this might be
  // useful for debugging.
  //
}

static econtext_t* econtext_new(int type, varset_t *G, kreen_t *S)
{
  int i;
  econtext_t *C;

  static char *predef_locals_slay[]= { ":disp-w:", ":disp-h:",
                                       ":scr-r:", ":scr-b:", NULL};

  static char *predef_locals_wlay[]= { ":scr-w:", ":scr-h:", NULL };

  C= xallocz(sizeof(*C));
  C->type= type;
  C->globals= G;
  switch(type)
  {
  case CON_GLOBAL: break;
  case CON_SLAY: 
     C->locals= varset_new(at_local);
     C->value_map= strmap_new();
     insert_predef_locals(C->locals, predef_locals_slay);
     break;
  case CON_WLAY:
     C->locals= varset_new(at_local); 
     insert_predef_locals(C->locals, predef_locals_wlay);
     C->value_map= strmap_new();
     C->wimap= strmap_new();
     for(i=0;i<S->n_allkidget;i++)
       strmap_insert(C->wimap, S->allkidget[i]->ident, S->allkidget[i]);
     break;
  }
  return C;
}

