


/****************************************************************************
                           LEXER

  The lexer is used only by the parser. Therefore, it's interface is 
  imported here. It's not relevant outside the input module.
****************************************************************************/



typedef struct lex lex_t;

enum {
  lex_eof,
  lex_number,     lex_ident,       lex_string,
  lex_plus,       lex_minus,       lex_slash,     lex_star,    lex_equal,
  lex_percent,
  lex_round_open, lex_round_close,
  lex_square_open, lex_square_close,
  lex_curly_open, lex_curly_close,
  lex_comma,      lex_dot,         lex_semicolon, lex_colon, 

  lex_last= lex_colon
};

lex_t    *lex_new       (const char *path);
int       lex_get       (lex_t *L, unsigned char **data, unsigned char **pos); 
int       lex_advance   (lex_t *L, unsigned char **data, unsigned char **pos);
lex_t    *lex_destroy   (lex_t *L);


struct lex { 
  char          *path;
  unsigned char *input;
  size_t         input_size;
  unsigned char *ptr,        *end;

  unsigned char *tokbuf,      tokbuf_store[1024];
  int            toklen,      tokbuf_size;
  unsigned char *tokpos;
  int            token;
  int            tokval_s25f6;

  char           dbuf[256];

  struct lex    *parent;
};


lex_t *lex_destroy(lex_t *L)
{
  if (!L) return NULL;
 
  if (L->tokbuf!=L->tokbuf_store) xfree(L->tokbuf);
  xfree(L->path);
  xfree(L->input);
  free(L);
  return NULL;
}

enum 
{
  lexc_space= lex_last+1,
  lexc_newline,
  lexc_letter,
  lexc_digit,
  lexc_comment,
};

static const unsigned char lex_class[256] =
{
  ['+'] = lex_plus,        ['-'] = lex_minus,
  ['*'] = lex_star,        ['/'] = lex_slash,
  ['='] = lex_equal,       ['%'] = lex_percent,
  ['{'] = lex_curly_open,  ['}'] = lex_curly_close,
  ['('] = lex_round_open,  [')'] = lex_round_close,
  ['['] = lex_square_open, [']'] = lex_square_close,
  ['.'] = lex_dot,         [','] = lex_comma,
  [';'] = lex_semicolon,   [':'] = lex_colon,
  ['"'] = lex_string, 

  ['#'] = lexc_comment,

  ['a'] = lexc_letter,     ['b'] = lexc_letter,     ['c'] = lexc_letter,
  ['d'] = lexc_letter,     ['e'] = lexc_letter,     ['f'] = lexc_letter,
  ['g'] = lexc_letter,     ['h'] = lexc_letter,     ['i'] = lexc_letter,
  ['j'] = lexc_letter,     ['k'] = lexc_letter,     ['l'] = lexc_letter,
  ['m'] = lexc_letter,     ['n'] = lexc_letter,     ['o'] = lexc_letter,
  ['p'] = lexc_letter,     ['q'] = lexc_letter,     ['r'] = lexc_letter,
  ['s'] = lexc_letter,     ['t'] = lexc_letter,     ['u'] = lexc_letter,
  ['v'] = lexc_letter,     ['w'] = lexc_letter,     ['x'] = lexc_letter,
  ['y'] = lexc_letter,     ['z'] = lexc_letter,

  ['A'] = lexc_letter,     ['B'] = lexc_letter,    ['C'] = lexc_letter,
  ['D'] = lexc_letter,     ['E'] = lexc_letter,    ['F'] = lexc_letter,
  ['G'] = lexc_letter,     ['H'] = lexc_letter,    ['I'] = lexc_letter,
  ['J'] = lexc_letter,     ['K'] = lexc_letter,    ['L'] = lexc_letter,
  ['M'] = lexc_letter,     ['N'] = lexc_letter,    ['O'] = lexc_letter,
  ['P'] = lexc_letter,     ['Q'] = lexc_letter,    ['R'] = lexc_letter,
  ['S'] = lexc_letter,     ['T'] = lexc_letter,    ['U'] = lexc_letter,
  ['V'] = lexc_letter,     ['W'] = lexc_letter,    ['X'] = lexc_letter,
  ['Y'] = lexc_letter,     ['Z'] = lexc_letter,    ['_'] = lexc_letter,

  ['0'] = lexc_digit,      ['5'] = lexc_digit,
  ['1'] = lexc_digit,      ['6'] = lexc_digit,
  ['2'] = lexc_digit,      ['7'] = lexc_digit,
  ['3'] = lexc_digit,      ['8'] = lexc_digit,
  ['4'] = lexc_digit,      ['9'] = lexc_digit,

  [  1] = lexc_space,      [  2] = lexc_space,     [  3] = lexc_space,     
  [  4] = lexc_space,      [  5] = lexc_space,     [  6] = lexc_space,      
  [  7] = lexc_space,      [  8] = lexc_space,     [  9] = lexc_space,
  [ 10] = lexc_newline,    [ 11] = lexc_space,     [ 12] = lexc_space,
  [ 13] = lexc_space,      [ 14] = lexc_space,     [ 15] = lexc_space,
  [ 16] = lexc_space,      [ 17] = lexc_space,     [ 18] = lexc_space,
  [ 19] = lexc_space,      [ 20] = lexc_space,     [ 21] = lexc_space,
  [ 22] = lexc_space,      [ 23] = lexc_space,     [ 24] = lexc_space,
  [ 25] = lexc_space,      [ 26] = lexc_space,     [ 27] = lexc_space,
  [ 28] = lexc_space,      [ 29] = lexc_space,     [ 30] = lexc_space,
  [ 31] = lexc_space,      [ 32] = lexc_space,     [127] = lexc_space,
};

static void lex_error
  (lex_t *L, const unsigned char *pos, const char *fmt, ...)
{
  va_list ap;
  if (pos) ifile_report(pos);
  va_start(ap,fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  exit(1);
}

static char *lex_report_char(lex_t *L, unsigned int x)
{
  if (x>32 && x<127)
    sprintf(L->dbuf,"%c",x);
  else
    sprintf(L->dbuf,"(x%02x)", x&0xff);
  return L->dbuf;
}

static void lex_tokbuf_single(lex_t *L, unsigned char x)
{
  L->tokbuf[0]= x;
  L->tokbuf[1]= 0;
  L->toklen= 1;
}

static void lex_grow_tokbuf(lex_t *L)
{
  L->tokbuf_size+= L->tokbuf_size/2;
  if (L->tokbuf==L->tokbuf_store)
  {
    L->tokbuf= xalloc(L->tokbuf_size);
    memcpy(L->tokbuf, L->tokbuf_store, L->toklen);
  }
  else
  {
    L->tokbuf= xallocr(L->tokbuf, L->tokbuf_size);
  }
}

static void lex_put_chr(lex_t *L, unsigned char x)
{
  if (L->toklen==L->tokbuf_size) lex_grow_tokbuf(L);
  L->tokbuf[L->toklen++]= x;
}

static void lex_begin_token(lex_t *L, unsigned char token)
{
  L->tokpos= L->ptr;
  L->toklen= 0;
  L->token= token;
}

static void lex_end_token(lex_t *L)
{
  if (L->toklen==L->tokbuf_size) lex_grow_tokbuf(L);
  L->tokbuf[L->toklen]= 0;
}

static void lex_get_ident(lex_t *L)
{
  unsigned char V, C;
  lex_begin_token(L, lex_ident);
  V= *L->ptr;
  while(1)
  {
    lex_put_chr(L, V);          
    L->ptr++;
    if (L->ptr>=L->end) break;
    V= *L->ptr;
    C= lex_class[V];
    if (C!=lexc_letter && C!=lexc_digit) break;
  }
  lex_end_token(L);
}

static int lex_hexval(unsigned char V)
{
  if (V>='a' && V<='f') return V-'a'+10;
  if (V>='A' && V<='F') return V-'A'+10;
  if (V>='0' && V<='9') return V-'0';
  return -1;
}

//
// we get here when we see the 'x' in the hexadecimal number
// so, L->ptr points to the 'x'
//
static void lex_get_hex(lex_t *L)
{
  unsigned char V;
  V= *L->ptr;
  while(1)
  {
    lex_put_chr(L, V); 
    if (++L->ptr>=L->end) break;
    V= *L->ptr;
    if (lex_hexval(V)==-1) break;
  }
  lex_end_token(L);
  if (L->toklen==2) 
     irror(L->tokpos, "invalid hexadecimal number\n");
  return ;
}

static void lex_get_number(lex_t *L)
{
  unsigned char V, C;
  lex_begin_token(L, lex_number);
  V= *L->ptr;
  while(1)
  {
    lex_put_chr(L, V); 
    if (++L->ptr>=L->end) goto done; 
    V= *L->ptr;
    C= lex_class[V];
    if (C!=lexc_digit)
    {
      if ((V=='x' || V=='X') && L->toklen==1 && L->tokbuf[0]=='0')
      {
        lex_get_hex(L);
        return ;
      }
      break;
    } 
  }
  if (V!='.') goto done;

  lex_put_chr(L, V); 
  if (++L->ptr>=L->end) goto bad_fraction;
  V= *L->ptr;
  C= lex_class[V];
  if (C!=lexc_digit) goto bad_fraction;
  while(1)
  {
    lex_put_chr(L, V);          
    if (++L->ptr>=L->end) break;
    V= *L->ptr;
    C= lex_class[V];
    if (C!=lexc_digit) break;
  }
done:
  lex_end_token(L);
  return ;
bad_fraction:
  lex_error(L,L->ptr,"bad character in number: %s", lex_report_char(L,V));
}


static void lex_get_string(lex_t *L)
{
  unsigned char V;
  int Hi,Lo;
  lex_begin_token(L, lex_string);
  if (++L->ptr>=L->end) goto unterminated;

restart:
  V= *L->ptr;
  if (V=='"') goto done;
  if (V!='%') 
  {
    lex_put_chr(L,V);
    if (++L->ptr>=L->end) goto unterminated;
    goto restart;
  }
  if (++L->ptr>=L->end) goto unterminated;
  V= *L->ptr;
  switch(V)
  {
  case 'n': lex_put_chr(L, '\n'); break;
  case 'p': lex_put_chr(L, '%'); break;
  case 'Q': lex_put_chr(L, '"'); break;
  case 'q': lex_put_chr(L, '\''); break;
  case 't': lex_put_chr(L, '\t'); break;
  default:
     Hi= lex_hexval(V);
     if (Hi!=-1) goto low_nibble;
     lex_error(L,L->ptr,"unknown escape sequence %%%s", lex_report_char(L,V));
     return ;
  }
  if (++L->ptr>=L->end) goto unterminated;
  goto restart;
low_nibble:
  if (++L->ptr>=L->end) goto unterminated;
  V= *L->ptr;
  Lo= lex_hexval(V);
  if (Lo==-1)
  {
    lex_error(L,L->ptr,"bad escape sequence %%.%s", lex_report_char(L,V));
    return ;
  }
  lex_put_chr(L, (Hi<<4) | Lo);
  if (++L->ptr>=L->end) goto unterminated;
  goto restart;
done:
  L->ptr++;
  lex_end_token(L);
  return ;
unterminated:
  lex_error(L,L->tokpos,"unterminated string literal\n");
}

static void lex_single_chr_token
    (lex_t *L, unsigned char token, unsigned char chr)
{
   lex_tokbuf_single(L,chr);
   L->token= token;
   L->tokpos= L->ptr;
}

int lex_get(lex_t *L, unsigned char **data, unsigned char **pos)
{
  if (data) *data= L->tokbuf;
  if (pos)  *pos= L->tokpos;
  return L->token;
}

int lex_advance(lex_t *L, unsigned char **data, unsigned char **pos)
{
restart:
  if (L->ptr>=L->end) 
  {
    lex_single_chr_token(L, lex_eof, 0);
    if (data) *data= L->end-1;
    if (pos) *pos= L->end-1;
    return lex_eof;
  }
  switch (lex_class[*L->ptr])
  {
    case lexc_comment: while(L->ptr<L->end && *L->ptr!='\n') L->ptr++;
                       goto restart;
    case lexc_space:   L->ptr++; goto restart;
    case lexc_newline: L->ptr++; goto restart;
    case lexc_letter:  lex_get_ident(L); goto tokret;
    case lexc_digit:   lex_get_number(L); goto tokret;
    case lex_string:   lex_get_string(L); goto tokret;
    case lex_plus:
    case lex_minus:
    case lex_slash:
    case lex_star:
    case lex_percent:
    case lex_equal:
    case lex_round_open: 
    case lex_round_close:
    case lex_curly_open:
    case lex_curly_close:
    case lex_square_open:
    case lex_square_close:
    case lex_comma: 
    case lex_dot: 
    case lex_semicolon: 
    case lex_colon: 
      lex_single_chr_token(L, lex_class[*L->ptr], *L->ptr);
      L->ptr++;
      goto tokret;
  }
  lex_tokbuf_single(L,*L->ptr);
  lex_error(L, L->ptr, "unknown character %s\n", lex_report_char(L,*L->ptr));
  return 0;
tokret:
  return lex_get(L, data, pos);
}

lex_t *lex_new(const char *path)
{
  lex_t *L;
  L= xalloc(sizeof(*L));
  L->path= xstrdup(path);
  L->tokbuf= L->tokbuf_store;
  L->tokbuf_size= sizeof(L->tokbuf_store);

  if (ifile_read(&L->input, &L->input_size, path, 1)) exit(1);

  L->ptr= L->input;
  L->end= L->input+L->input_size;

  lex_advance(L, NULL, NULL);
  return L;
}


#ifdef SELF_TEST


int main(int argc, char **argv)
{
  lex_t *L;
  int t;
  unsigned char *data;
  L= lex_new(argv[1]);
  if (!L) die("out of memory\n");
  while((t=lex_get(L, &data, NULL))!=lex_eof)
  {
    printf("%d<%s>\n", t, (char*) data);
    lex_advance(L, NULL, NULL);
  }
  lex_destroy(L);
  return 0;
}
#endif
