
/**
  This struct contains the data necessary for parsing a screen.
  Information here is not relevant after parsing. Therefore, I 
  keep them here to prevent access to them.
 **/
typedef struct preen
{
  kreen_t *kreen;
  int n_sub;    kreen_t **sub;

  FPA Passign, Palign, Psub, Pkidget; 
} preen_t;

enum {
  pint_brace= node_list + 1,
};

static const struct
{
  int prec, assoc;
} prectab[]=
{
  [node_negate] { 200, 1 },
  [node_add] { 100, 0 },
  [node_sub] { 100, 0 },
  [node_div] { 110, 0 },
  [node_mod] { 110, 0 },
  [node_mul] { 110, 0 },
  [node_comma] { 50, 1 },
  [node_assign] { 60, 1 },
  [pint_brace] { 0, 0 },
  [node_call] { 10, 0 },
};

typedef struct
{
  int otos, osize;    expr_t **ostack;
  int atos, asize;    expr_t **astack;

  lex_t *lex;
  unsigned char *token_data, *token_pos;
  int token;

  unsigned char *mark_pos;
  int mark_atos;

  input_t *ipt;
  preen_t *preen;
} parser_t;

typedef struct
{
  char *name;
  void (*func)(parser_t *);
} parser_kw_t;

static const struct 
{
  char *sname; char *syntax;
} wclassdef[]=
{
  [GFXCLASS_LABEL]       { "label",         "TA"   },
  [GFXCLASS_LOGVIEW]     { "logview",       "A"    },
  [GFXCLASS_SPINNER]     { "spinner",       "A"    },
  [GFXCLASS_PROGRESSBAR] { "progressbar",   "A"    },
  [GFXCLASS_RADIO]       { "radio",         "TA"   },
  [GFXCLASS_PANEL]       { "panel",         "A"    },
  [GFXCLASS_BUTTON]      { "button",        "tiA"  },
  [GFXCLASS_TOGGLE]      { "toggle",        "tiA"  },
  [GFXCLASS_PIXGRID]     { "pixgrid",       "A"    },
  [GFXCLASS_CHECKBOX]    { "checkbox",      "TA"   },
  [GFXCLASS_DIRLIST]     { "dirlist",       "A"    },
  [GFXCLASS_FILELIST]    { "filelist",      "A"    },
  [GFXCLASS_TEXTENTRY]   { "textentry",     "tA"   },
  [GFXCLASS_SCROLLBAR]   { "scrollbar",     "A"    },
  [GFXCLASS_LIST]        { "list",          "A"    },
  [GFXCLASS_ICONLIST]    { "iconlist",      "A"    },
  [GFXCLASS_SLIDER]      { "slider",        "A"    },
  [GFXCLASS_DROPDOWN]    { "dropdown",      "A"    },
  [GFXCLASS_CANVAS]      { "canvas",        "A"    },
};
static int kidget_class_find(char *name)
{
  int i;
  for(i=0;i<sizeof(wclassdef)/sizeof(wclassdef[0]);i++)
    if (wclassdef[i].sname && !strcmp(wclassdef[i].sname,name)) return i;
  return -1;
}


static void parser_error(parser_t *P, const char *fmt, ...)
{
  ifile_report(P->token_pos);
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  exit(1);
}

static void parser_advance(parser_t *P)
{
  P->token= lex_advance(P->lex, &P->token_data, &P->token_pos);
}

static void parser_mark(parser_t *P)
{
  P->mark_atos= P->atos;
  P->mark_pos= P->token_pos;
}

static void parser_expect(parser_t *P, int advance, int token, char *name)
{
  if (advance)
    parser_advance(P);
  if (P->token!=token) 
       parser_error(P, "%s expected\n", name);
}

static void parse_kidget(parser_t *P);

static void parser_dispatch_kw
  (parser_t *P, const parser_kw_t *KW, int allow_widgets)
{
  int i;
  parser_expect(P, 0, lex_ident, "keyword");
  for(i=0;KW[i].name;i++)
    if (!strcmp(KW[i].name, (char*)P->token_data))
    {
      (*KW[i].func)(P);
      return ;
    }
  if (allow_widgets)
  {
     if (kidget_class_find((char*) P->token_data)!=-1)
     {
       parse_kidget(P);
       return ;
     }
  }
  parser_error(P, "unknown keyword\n");
}

static void parser_push_o(parser_t *P, expr_t *E)
{
  if (P->otos==P->osize-1) 
    { P->osize += P->osize/2; 
      P->ostack= xallocr(P->ostack, sizeof(P->astack[0])*P->osize); }
  P->otos++;
  P->ostack[P->otos]= E;
}

static void parser_push_a(parser_t *P, expr_t *E)
{
  if (P->atos==P->asize-1) 
    { P->asize += P->asize/2; 
      P->astack= xallocr(P->astack, sizeof(P->astack[0])*P->asize); }
  P->atos++;
  P->astack[P->atos]= E;
}


static void parser_cpush_a(parser_t *P, int type,int advance)
{
  expr_t *E;
  E= expr_new(type, P->token_data, P->token_pos, 0);
  parser_push_a(P, E);
  if (advance) parser_advance(P);
}

static void parser_cpush_o(parser_t *P, int type, int nsub, int advance)
{
  expr_t *E;
  E= expr_new(type, NULL, P->token_pos, nsub);
  parser_push_o(P, E);
  if (advance) parser_advance(P);
}


static expr_t *parser_pop_a(parser_t *P)
{
  if (P->atos==-1) parser_error(P, "stack underflow in expr\n");
  P->atos--;
  return P->astack[P->atos+1];
}

static expr_t *parser_pop_o(parser_t *P)
{
  if (P->atos==-1) parser_error(P, "stack underflow in expr\n");
  P->otos--;
  return P->ostack[P->otos+1];
}

// pop arguments from the arg stack. puts them into the sub-expression
// array of E
static void parser_get_args(parser_t *P, expr_t *E)
{
  int N, i;
  int K;
  N= E->nsub;
  K= P->atos-N+1;
  for(i=0;i<N;i++) E->sub[i]= P->astack[K+i];
  P->atos= K-1;
}

static const int parser_lex2node[lex_last]=
{
  [lex_number] node_number,
  [lex_ident] node_ident,
  [lex_string] node_string,
  [lex_minus]  node_sub,
  [lex_plus]  node_add,
  [lex_slash] node_div,
  [lex_percent] node_mod,
  [lex_star]  node_mul,

  [lex_equal] node_assign,
  [lex_comma] node_comma,
  [lex_dot]   node_field,
};

static void parser_push_multipart_arg(parser_t *P, int type)
{
  expr_t *E;
  E= expr_new(type, NULL, P->mark_pos, P->atos-P->mark_atos);
  parser_get_args(P,E); 
  parser_push_a(P, E);
}


// parses an identifier or a field reference in the form kidget.field
// even though it's not necessary, I will also parse longer sequences  
// such as x.y.z.t etc.
static void parse_ident(parser_t *P)
{
  parser_mark(P);
  parser_cpush_a(P, node_ident, 1); 
  if (P->token!=lex_dot) return ;

  while (P->token==lex_dot)
  {
    parser_expect(P, 1, lex_ident, "identifier");
    parser_cpush_a(P, node_ident, 1);
  }
  parser_push_multipart_arg(P, node_field);
}

static void parse_list(parser_t *P)
{
  parser_mark(P);
  parser_advance(P);
  while(P->token!=lex_square_close)
  {
    parser_expect(P, 0, lex_ident, "identifier");
    parser_cpush_a(P, node_ident, 1);
  }
  parser_push_multipart_arg(P, node_list);
  parser_advance(P);
}

static void parse_term(parser_t *P)
{
again:
  switch(P->token)
  {
  case lex_string:     
  case lex_number:     
    parser_cpush_a(P, parser_lex2node[P->token],1);
    break;
  case lex_minus:
    parser_cpush_o(P, node_negate, 1, 1);
    goto again;
  case lex_ident:
    parse_ident(P);
    break;
  case lex_dot: 
    parser_advance(P);
    if (P->token!=lex_ident) 
       parser_error(P, "identifier expected\n");
    parser_cpush_a(P, node_self_field, 1); 
    break;
  case lex_round_open:
    parser_cpush_o(P, pint_brace, 0, 1); 
    goto again;
  case lex_square_open:
    parse_list(P); 
    break;
  default:
    parser_error(P, "term expected\n");
  }
}

static void parser_reduce_once(parser_t *P)
{
  expr_t *E;
  E= parser_pop_o(P);
  parser_get_args(P, E);
  parser_push_a(P, E);
}

static void parser_reduce_until(parser_t *P, int op)
{
  int oprec, oassoc;
  oprec= prectab[op].prec;
  oassoc= prectab[op].assoc;
  while(P->otos!=-1)
  {
    int top, tprec;
    top= P->ostack[P->otos]->type;
    tprec= prectab[top].prec;
    if (tprec<oprec) return ;
    if (tprec>oprec) goto reduce; 
    if (top!=op)     goto reduce;
    if (oassoc)      return ;
reduce:
    parser_reduce_once(P); 
  }
}

static int parse_op(parser_t *P)
{
  int op;
again:
  switch(P->token)
  {
  case lex_minus:
  case lex_plus:
  case lex_slash:
  case lex_percent:
  case lex_star:
  case lex_equal:
  case lex_comma:
    op= parser_lex2node[P->token];
    parser_reduce_until(P, op);
    parser_cpush_o(P, op, 2, 1);
    return 0;
  case lex_round_close:
    while(P->otos!=-1)
    {
      int ttype;
      ttype= P->ostack[P->otos]->type;
      if (ttype==pint_brace)
      {
        P->otos--; // FIXME: leaking the expr here
        parser_advance(P);
        goto again;
      }
      if (ttype==node_call)
      {
        parser_advance(P);
        parser_reduce_once(P);
        goto again;
      }
      parser_reduce_once(P);
    }
    parser_error(P, "dangling )\n");
    break;
  case lex_semicolon:
  case lex_eof:
  case lex_curly_open:
    return 1;
  case lex_round_open: // a '(' is an fcall operator when
                       // an operator is expected
    parser_cpush_o(P, node_call, 2, 1);
    return 0;
  }
  parser_error(P, "operator expected\n");
  return 1;
}

static expr_t *parse_expr(parser_t *P)
{
  do 
    parse_term(P);
  while(!parse_op(P));
  while (P->otos!=-1)
  {
    expr_t *E;
    E= P->ostack[P->otos];
    if (E->type==pint_brace)
    {
       P->token_pos= E->pos;
       parser_error(P, "unmatched (\n");
    }
    parser_reduce_once(P);
  }
  if (P->atos!=0)
     parser_error(P, "extra elements\n");
  return parser_pop_a(P);
}

parser_t *parser_new(const char *path)
{
  parser_t *P;
  input_t *ipt;
  P= xallocz(sizeof(*P));
  P->asize= 200;
  P->osize= 200;
  P->atos= -1;
  P->otos= -1;
  P->astack= xalloc(sizeof(expr_t*)*P->asize);
  P->ostack= xalloc(sizeof(expr_t*)*P->osize);
  P->lex= lex_new(path);
  P->token= lex_get(P->lex, &P->token_data, &P->token_pos);

  ipt= P->ipt= xallocz(sizeof(*P->ipt));
  fpa_init(&ipt->Pkreen, &ipt->n_kreen, (void***) &ipt->kreen, 128);
  fpa_init(&ipt->Passign, &ipt->n_assign, (void***) &ipt->assign, 128);
  fpa_init(&ipt->Picon,   &ipt->n_icon,   (void***) &ipt->icon, 16);
  fpa_init(&ipt->Pfont,   &ipt->n_font,   (void***) &ipt->font, 8);
  return P;
}

static void parser_flatten_comma(expr_t *E, ptrbuf_t *B)
{
  if (E->type!=node_comma)
  {
    ptrbuf_put(B, E);
    return ;
  }
  // normally, the tree is right-heavy, since the comma
  // operator is right associative.
  parser_flatten_comma(E->sub[0],B); 
  parser_flatten_comma(E->sub[1],B);
  free(E->sub);
  free(E);
}

static void parser_comma_collect(expr_t *E, int *nO, expr_t ***O)
{
  ptrbuf_t *P;
  size_t L;
  P= ptrbuf_new(256);
  parser_flatten_comma(E, P);
  ptrbuf_free(P, (void***) O, &L);
  *nO= L;
}

static void parse_let(parser_t *P, FPA *dst)
{
  expr_t *E;
  ptrbuf_t *B;

  parser_advance(P);
  E= parse_expr(P);
  parser_expect(P, 0, lex_semicolon, ";");
  parser_advance(P);

  B= ptrbuf_new(256);
  parser_flatten_comma(E, B);
  while ((E=ptrbuf_get(B))) fpa_put(dst,E);
  ptrbuf_free(B, NULL, NULL);
}

static void parse_kreen_let(parser_t *P)
{
  parse_let(P,&P->preen->Passign);
}

static void parse_top_let(parser_t *P)
{
  parse_let(P,&P->ipt->Passign);
}

static int parse_align_options(parser_t *P)
{
  int O, i,V;
  static const struct {
    char *name;
    int value;
  } optnames[]=
  {
    { "lr", ALIGNOPT_LR }, { "rl", ALIGNOPT_RL },
    { "td", ALIGNOPT_TD }, { "bu", ALIGNOPT_BU },
    { "ew", ALIGNOPT_EW }, { "eh", ALIGNOPT_EH },
    { "et", ALIGNOPT_ET }, { "eb", ALIGNOPT_EB },
    { "el", ALIGNOPT_EL }, { "er", ALIGNOPT_ER },
    { NULL, -1 }
  };
  O= 0;
  for(parser_advance(P);P->token==lex_dot;parser_advance(P))
  {
    parser_expect(P, 1, lex_ident, "identifier");
    for(i=0;optnames[i].name;i++)
       if (!strcmp((char*) P->token_data, optnames[i].name))
          break;
    V= optnames[i].value;
    if (V==-1) parser_error(P, "unknown alignment option\n");
    O |= V;
  }
  if (!O)
     parser_error(P, "alignment with no options\n");
  return O;
}

static void parse_align(parser_t *P)
{
  align_t *A;
  expr_t *E;
  A= xallocz(sizeof(*A));
  A->options= parse_align_options(P);
  A->pos= P->token_pos;

  switch (P->token)
  {
  case lex_ident:  parse_ident(P); break;
  case lex_square_open: parse_list(P); break;
  default: parser_error(P,"identifier or list expected\n");
  }
  A->kidget_list= parser_pop_a(P);

  if (P->token!=lex_semicolon)
  {
    E= parse_expr(P);
    parser_comma_collect(E, &A->n_assign, &A->assign);
    parser_expect(P, 0, lex_semicolon, "';'");
  }
  parser_advance(P);
  fpa_put(&P->preen->Palign, A);
}


static void parse_kreen_name(parser_t *P, kreen_t *S)
{
  parser_expect(P, 1, lex_ident, "ident");
  S->name= xstrdupu(P->token_data);
  parser_advance(P);
  if (P->token==lex_dot)
  {
    parser_expect(P, 1, lex_number, "number");
    S->subname= atoi((char*) P->token_data);
    parser_advance(P);
  }
}

static kreen_t *kreen_new(parser_t *P)
{
  preen_t *U;
  kreen_t *S;
  S= xalloc(sizeof(*S));
  S->subname= -1;
  U= xalloc(sizeof(*U));
  fpa_init(&U->Passign, &S->n_assign, (void***) &S->assign, 256);
  fpa_init(&U->Palign, &S->n_align, (void***) &S->align, 256);
  fpa_init(&U->Pkidget, &S->n_kidget, (void***) &S->kidget, 256);
  fpa_init(&U->Psub, &U->n_sub, (void***) &U->sub, 256);
  U->kreen= S;
  P->preen= U;
  return S;
}

static expr_t *parser_construct_attrib(parser_t *P, char *field, int vtype)
{
  expr_t *T;
  unsigned char *pos;

  pos= P->token_pos;
  T= expr_new(node_assign, NULL, pos, 2);
  T->sub[0]= expr_new(node_self_field, (unsigned char *) field, pos, 0);
  T->sub[1]= expr_new(vtype, P->token_data, pos, 0);
  return T;
}

static void parse_kidget(parser_t *P)
{
  ptrbuf_t *assign;
  char *syn;
  kidget_t *W;
  int q;
  size_t N;
  int need_semicolon;

  W= xallocz(sizeof(*W));
  W->pos= P->token_pos;

  W->klass= kidget_class_find((char*) P->token_data);
  if (W->klass<0)
    parser_error(P, "internal inconsistency, unknown kidget class\n");

  parser_expect(P, 1, lex_ident, "identifier");
  W->ident= xstrdup((char*)P->token_data);
  parser_advance(P);

  assign= ptrbuf_new(16);

  syn= wclassdef[W->klass].syntax;
  need_semicolon= 1;
  for(q=0;!q;syn++) switch(*syn)
  {
  case 't': if (P->token!=lex_string) continue;
  case 'T': parser_expect(P, 0, lex_string, "string"); 
     ptrbuf_put(assign, parser_construct_attrib(P, "text", node_string));
     parser_advance(P);
     continue;
  case 'i': if (P->token!=lex_ident) continue;
  case 'I': parser_expect(P, 0, lex_ident, "icon identifier");
     ptrbuf_put(assign, parser_construct_attrib(P,"icon", node_ident));
     parser_advance(P);
     continue;
  case 'A':
     if (P->token!=lex_semicolon && P->token!=lex_curly_open)
        parser_flatten_comma(parse_expr(P), assign);
     continue;
  case 0:
     q= 1;
     continue;
  default:
     parser_error(P, "internal error, unknown class syntax spec %c\n",
                      syn[0]);
  }
  if (need_semicolon)
  {
    parser_expect(P, 0, lex_semicolon, "';'");
    parser_advance(P);
  }

  ptrbuf_free(assign, (void***) &W->assign, &N);
  W->n_assign= N;

  fpa_put(&P->preen->Pkidget, W);
}

static void parse_sub_kreen(parser_t *P);




static int compare_kreen_subname(void *_a, void *_b)
{
  kreen_t *A= _a, *B= _b;
  return A->subname - B->subname;
}

static void move_subscreens_into(preen_t *U, kidget_t *W)
{
  ptrbuf_t *B;
  int j,i;
  size_t sz;
  kreen_t *S;

  B= ptrbuf_new(20);
  for(j=0;j<U->n_sub;j++)
  {
    if (!ustrcmp(U->sub[j]->name,W->ident)) 
    {
      ptrbuf_put(B, U->sub[j]);
      U->sub[j--]= U->sub[--U->n_sub];
    }
  }
  ptrbuf_free(B, (void***) &W->sub, &sz); W->n_sub= sz;
  if (!sz) 
     irror(W->pos, "panel doesn't have any subscreens\n");

  bsort_ptr((void**) W->sub, W->n_sub, compare_kreen_subname);

  for(i=0;i<W->n_sub;i++)
  {
    S= W->sub[i];
    if (S->subname>i)
       irror(S->pos, "screen #%d is missing in panel kidget '%s'\n",
               S->subname-1, W->ident);
    else if (S->subname<i)
       irror(S->pos, "duplicate definition of screen %s.%d\n",
               W->ident, S->subname);
  }
}
//
// We move subscreens from the preen to the
// corresponding panel widget. 
//
static void move_subscreens(preen_t *U)
{
  int i;
  kreen_t *S;
  kidget_t *W;
  if (!U->n_sub) return ;

  S= U->kreen;
  for(i=0;i<S->n_kidget;i++)
  {
    W= S->kidget[i];
    if (W->klass==GFXCLASS_PANEL) 
       move_subscreens_into(U, W);
  }

  if (U->n_sub)
     irror(U->sub[0]->pos, "sub-screen does not belong to any panel widget\n");
}

static kreen_t* parse_kreen(parser_t *P)
{
  static const parser_kw_t kw[]=
  { { "let",  parse_kreen_let },
    { "screen", parse_sub_kreen },
    { "align", parse_align },
    { NULL, NULL } };
  kreen_t *S;
  S= kreen_new(P);
  parse_kreen_name(P, S);
  parser_expect(P, 0, lex_curly_open, "{");
  parser_advance(P);
  while(P->token!=lex_curly_close) parser_dispatch_kw(P, kw, 1);
  parser_advance(P);

  move_subscreens(P->preen);

  return S;
}

static void parse_sub_kreen(parser_t *P)
{
  preen_t *U;
  U= P->preen;
  fpa_put(&U->Psub, parse_kreen(P)); 
  P->preen= U;
}

static void parse_top_kreen(parser_t *P)
{
  fpa_put(&P->ipt->Pkreen, parse_kreen(P));
}

static void parse_icon(parser_t *P)
{
  FPA Pfile;
  icondef_t *IC;
  iconfile_t *F;

  IC= xallocz(sizeof(*IC));
  fpa_init(&Pfile, &IC->n_file, (void***) &IC->file, 10);

  parser_expect(P, 1, lex_ident, "identifier");
  IC->pos= P->token_pos;
  IC->ident= xstrdup((char*) P->token_data);

  while(P->token!=lex_semicolon)
  {
    F= xallocz(sizeof(*F));
    parser_expect(P, 1, lex_number, "dpi");
    F->dpi= atoi((char*) P->token_data);
    if (F->dpi<0) parser_error(P, "bad dpi\n");
    parser_expect(P, 1, lex_string, "file name");   
    F->file= xstrdup((char*) P->token_data);
    fpa_put(&Pfile, F);
  }
  parser_advance(P);

  fpa_put(&P->ipt->Picon, IC);
}

static void parse_font(parser_t *P)
{
  kontdef_t *FD;
  FD= xallocz(sizeof(*FD));
  FD->pos= P->token_pos;
  parser_expect(P, 1, lex_number, "font number\n");
  FD->number= atoi((char*) P->token_data);
  parser_expect(P, 1, lex_string, "font file\n");
  FD->file= xstrdup((char*) P->token_data);
  parser_expect(P, 1, lex_number, "font size\n");
  FD->size= strtod((char*) P->token_data, NULL);
  parser_expect(P, 1, lex_semicolon, "';'");
  parser_advance(P);
  fpa_put(&P->ipt->Pfont, FD);
}

static void parse_top(parser_t *P)
{
  static const parser_kw_t kw[]=
  { { "screen", parse_top_kreen }, 
    { "let",    parse_top_let },
  //  { "include", parse_include },
    { "font",    parse_font },
    { "icon",    parse_icon },
    {  NULL, NULL } };
  while(P->token!=lex_eof) parser_dispatch_kw(P, kw, 0);
}

input_t *parse_file(const char *fn)
{
  parser_t *P;
  input_t *IPT;
  P= parser_new(fn);
  parse_top(P);
  IPT= P->ipt;
  // FIXME: free P
  return IPT;
}


void print_expr(expr_t *E)
{

static const char *nodeprint[]=
{
   [node_sub] "SUB",
   [node_add] "ADD",
   [node_div] "DIV",
   [node_mod] "MOD",
   [node_mul] "MUL",
   [node_assign] "ASGN",
   [node_comma] "COMMA",

   [node_field] "FIELD",
   [node_self_field] "SFIELD",
   [node_list] "LIST",
   [node_call] "CALL",
   [pint_brace] "BRACE",

   [node_ident] "IDENT",
   [node_number] "NUMBER",
   [node_string] "STRING",
};

  if (!E) { printf("NULL"); return ; }
  if (nodeprint[E->type]) printf("%s", nodeprint[E->type]);
  if (E->data) printf("<%s>", E->data);
  if (E->nsub)
  {
    int i;
    printf("(");
    for(i=0;i<E->nsub;i++) { if (i) printf(","); print_expr(E->sub[i]); }
    printf(")");
  }
}

void p_expr(expr_t *E) { print_expr(E); printf("\n"); }

#ifdef SELF_TEST

void print_stacks(parser_t *P)
{
  int i;
  printf("     A-Stack\n");
  for(i=P->atos;i>-1;i--) { print_expr(P->astack[i]); printf("\n"); }
  printf("     O-Stack\n");
  for(i=P->otos;i>-1;i--) { print_expr(P->ostack[i]); printf("\n"); }
}

void print_assign(int n, expr_t **A)
{
  int i;
  printf("assignments\n");
  for(i=0;i<n;i++) { print_expr(A[i]); printf("\n"); }
}

void print_kidget(kidget_t *W)
{
  printf("kidget(%s,%d) {\n", W->ident, W->klass);
  print_assign(W->n_assign, W->assign);
  printf("}\n");
}

void print_align(align_t *A)
{
  printf("align(%d) {\n", A->options);
  printf("kidgets= ");
  print_expr(A->kidget_list);
  printf("\n");
  print_assign(A->n_assign, A->assign);
  printf("}\n");
}

void print_font(kontdef_t *F)
{
  printf("font(%d)= <%s>.%f\n", F->number, F->file, F->size);
}

void print_icon(icondef_t *C)
{
  int i;
  printf("icon(%s) {\n", C->ident);
  for(i=0;i<C->n_file;i++)
    printf("<%s>.%d\n", C->file[i]->file, C->file[i]->dpi);
  printf("}\n");
}

void print_kreen(kreen_t *S)
{
  int i;
  printf("kreen(%s.%d) {\n", S->name, S->subname);
  print_assign(S->n_assign, S->assign);
  for(i=0;i<S->n_align;i++) print_align(S->align[i]);
  for(i=0;i<S->n_kidget;i++) print_kidget(S->kidget[i]);
  for(i=0;i<S->n_sub;i++) print_kreen(S->sub[i]);
  printf("}\n");
}

void print_ipt(input_t *ipt)
{
  int i;
  for(i=0;i<ipt->n_kreen;i++) print_kreen(ipt->kreen[i]);
  print_assign(ipt->n_assign, ipt->assign);
  for(i=0;i<ipt->n_icon;i++) print_icon(ipt->icon[i]);
  for(i=0;i<ipt->n_font;i++) print_font(ipt->font[i]);
}

int main(int argc,char **argv)
{
  print_ipt( parse_file(argv[1]) );
  return 0;
}
#endif
