/***********************************************************************
 *                                                                     *
 *                            System Timers                            *
 *                                                                     *
 ***********************************************************************
  Timers are necessary for repeating actions. When a scrollbar button
  is pressed, it should continue scrolling if the user keeps the mouse
  button pressed.

  Initially, I will implement only internal timers. Later on, I will
  export the functionality to the user and remove the timeout argument
  from the gfx_event function.

  When a timer expires, the system will call a function with a stored
  argument.  Timers are one-shot only. Users of this functionality should
  schedule another one when needed.
 ***********************************************************************/

static int gui_timer_schedule
  (unsigned int msecs, int (*f)(void*,gfx_event_t*), void *a)
{
  static int last_timer;
  if (!screen) return -1;

  if (screen->n_timers==screen->s_timers)
  {
    if (!screen->s_timers) 
    {
      screen->s_timers= 10;
      screen->timers= malloc(sizeof(screen->timers[0])*screen->s_timers);
    }
    else
    {
      screen->s_timers+= screen->s_timers /2; 
      screen->timers= realloc(screen->timers,
                sizeof(screen->timers[0])*screen->s_timers);
    }
  }

  gui_timer_t *tt;
  tt= screen->timers + screen->n_timers;
  screen->n_timers++;
  tt->func= f;
  tt->arg= a;
  tt->value= msecs;
  tt->ident= last_timer++;
  return tt->ident;
}

 // this function returns 0 if there are no timers
 // it returns 1 if there is a timer and the remaining
 // time for it is stored in Rmsecs
static int gui_timer_get(unsigned int *Rmsecs)
{
  int i;
  unsigned int min;
  if (!screen || !screen->n_timers) return 0;
  min= screen->timers[0].value;
  for(i=1;i<screen->n_timers;i++)
     min = min < screen->timers[i].value ? min : screen->timers[i].value;
  *Rmsecs= min;
  return 1;
}

static int gui_timer_elapsed(unsigned int elap,gfx_event_t *event)
{
  int i,R;
  int T;
  if (!screen || !screen->n_timers) return 0;
  R= 0;
  T= screen->n_timers;
  for(i=0;i<T;i++)
  {
    gui_timer_t *tt;
    tt= screen->timers+i;
    if (elap<tt->value) { tt->value-= elap; continue; }
    if (!R)
    {
      int (*F)(void *,gfx_event_t*);
      void *A;
      F= tt->func;
      A= tt->arg;
      screen->timers[i]= screen->timers[T-1];
      i--;
      screen->timers[T-1]= screen->timers[screen->n_timers-1];
      T--;
      screen->n_timers--;
      if (F)
        R= (*F)(A,event); 
      continue;
    }
    tt->value= 0;
  }
  return R;
}

static void gui_timer_cancel(int ident)
{
  int i;
  if (ident==-1 || !screen || !screen->n_timers) return ;
  for(i=0;i<screen->n_timers;i++)
   if (screen->timers[i].ident==ident)
   {
    screen->timers[i]= screen->timers[screen->n_timers-1];
    screen->n_timers--;
    break;
   }
}

