
/***********************************************************************
 * #ANDRO                                                              *
 *                          Android Subsystem                          *
 *                                                                     *
 ***********************************************************************

                        .--._
                         '. ='-._
                           '.= =('-.
                             \ = (( '.
                      ,       | =((  _\.___
                     /=`"--.,_;..--'`     _`'.
                    /  =      .------.   (o) )>
                   /_=___,.../= `((   |_.--'`
                          _.'  = ((   /
             jgs     _.-'`=  = _((   /
                  .-'__=___=___((_,.'
            
 ***********************************************************************
  The android system is very weird in the way it handles events. First
  of all, when you make a native application, it's not really this native
  program running directly on the processor. A Java class is loaded first,
  which sets up the environment in which to talk to the rest of the system.
  This Java program then loads some methods from your native shared library
  and calls them. While it's doing these things, some threads are created
  to do various stuff, I don't really understand when and how many. 

  The system provides events thru several channels. There are two main
  methods used. One group of events are 'system events'. These are not
  related to user input etc. but are generated by the system in order to
  manage the application. For example, when the application is sent to
  background, a 'Pause' event is sent. Such system events are not reported
  thru a file descriptor but are processed directly by the Java wrapper
  around the native shared object. The application object exposed to the
  Java interface contains some function pointers, one for each type of
  system event. The Java part calls these function in *SOME* thread when
  the corresponding event occurs. This causes system events to be reported
  asynchronously. Your event loop in the main thread never sees these
  things, they get handled by another thread. In order to make them 
  synchronous, we report these events to the main thread via a pipe from
  the event handler functions.

  Another group of events are related to user input. These get reported
  thru a file descriptor, much like X11 events. The system has a struct
  called a 'looper' which can watch multiple file descriptors for events
  or 'ready to read' status. The looper is actually a simple wrapper 
  around the 'poll' system call. When you're ready to process user input,
  you add the input queue to the looper and then go into a loop, telling
  the looper to report activity on any file descriptor. 

  Fortunately, the looper can also watch regular file descriptors such
  as pipes or sockets. This makes it possible to synchronize system
  events with user events. 
 ***********************************************************************/

#ifdef GFX_ANDROID

/***********************************************************************
 *                             Definitions                             *
 ***********************************************************************
  The PIPE_* constants indicate which channel is ready for reading.
  They are given as identifiers to the looper thingy.

    PIPE_CMD     indicates that some asynchronous event has happened and
                 the callback has written some event on the command pipe.
    PIPE_EVENT   indicates that user input is available.
    PIPE_SENSOR  will be used for the sensor channel. It's not used yet.
 ***********************************************************************/

#define  PIPE_CMD    1
#define  PIPE_EVENT  2
#define  PIPE_SENSOR 3

/***********************************************************************
  CMD_* constants identify which asynchronous callback has occured.
 ***********************************************************************/

enum
{
  CMD_APP_START,
  CMD_APP_STOP,
  CMD_APP_RESUME,
  CMD_APP_PAUSE,
  CMD_WINDOW_DESTROYED,
  CMD_WINDOW_CREATED,
  CMD_WINDOW_RESIZED,
  CMD_WINDOW_REDRAW,
  CMD_WINDOW_CONTENTRECT,
  CMD_SAVE_STATE,
  CMD_FOCUS_CHANGE,
  CMD_LOW_MEMORY,
  CMD_CONFIG_CHANGE,
  CMD_INPUTQ_DESTROYED,
  CMD_INPUTQ_CREATED,
  CMD_JAVARESULT,
  CMD_QUIT
};

static uint32_t *Gpixels;
static int display_dpi;

#define x_red_mask    0x0000ff
#define x_green_mask  0x00ff00
#define x_blue_mask   0xff0000
#define x_red_shift   0
#define x_green_shift 8
#define x_blue_shift  16

void (*gfx_app_stop)();
void (*gfx_app_get_state)(unsigned char **Rdata, int *Rlen);
static int getsysbar();
static int getnavbar();


/***********************************************************************
 * #A.CALLBACK                                                         *
 *                          System Callbacks                           *
 *                                                                     *
 ***********************************************************************
  System callbacks are things like onResume, onWindowCreated etc. These
  are supposed to be handled in a separate thread but that makes things
  complicated. So, I'm going to serialize them into a pipe and let the
  main thread do everything. The google guys made sure that these events
  are acknowledged by the main thread before returning from the callback.
  I'm going to do the same thing and include an 'acknowledgement' field
  in each request. This might be important for adding event queues to
  loopers etc. 

  Requests are represented by the following struct. Only a pointer to 
  this struct is passed around since everything is in the same address
  space.
 ***********************************************************************/

struct req {
  int type;
  int complete;
  union {
    ANativeWindow *window;
    AInputQueue   *inputq;
    struct {
       void* ptr;
       size_t size;
    }             buffer; 
    int           intv;
  } d;
};

/***********************************************************************
  The following functions send, receive and acknowledge requests.
 ***********************************************************************/

static void make_request(app_t *A, request_t *R)
{
   R->complete= 0;
   if (write(A->cmdpipe_w, &R, sizeof(R))!=sizeof(R))
   {
     logg("Couldn't write to command pipe: %s\n", strerror(errno));
     return ;
   }
   
   pthread_mutex_lock(&A->mutex); 
   while(!R->complete) pthread_cond_wait(&A->cond, &A->mutex); 
   pthread_mutex_unlock(&A->mutex); 
}

static int get_request(app_t *A,request_t **R)
{
   request_t *T;
   if (read(A->cmdpipe_r, &T, sizeof(T))!=sizeof(T))
   {
     logg("Couldn't read from command pipe: %s\n", strerror(errno));
     return 1;
   }
   *R= T;
   return 0;
}

 // called by the main thread so that the java thread can chug on
static void complete_request(app_t *A, request_t *R)
{
   pthread_mutex_lock(&A->mutex); 
   R->complete= 1;
   pthread_cond_broadcast(&A->cond);
   pthread_mutex_unlock(&A->mutex); 
}

/***********************************************************************
  Following are the functions which get called from the java world. These
  simply create some struct and send it over the pipe to the native
  thread we're using.
 ***********************************************************************
 * #A.CB.APPSTATE                                                      *
 *                          Application State                          *
 *                                                                     *
 ***********************************************************************/

static void simple_command(ANativeActivity *ac, int cmd)
{
   request_t R;
   R.type= cmd;
   make_request(ac->instance, &R);
}

static void onStart(ANativeActivity* activity)
{
   logg("Start: %p\n", activity);
   simple_command(activity, CMD_APP_START);
}

static void onResume(ANativeActivity* activity)
{
   logg("Resume: %p\n", activity);
   simple_command(activity, CMD_APP_RESUME);
}

static void onPause(ANativeActivity* activity)
{
   logg("Pause: %p\n", activity);
   simple_command(activity, CMD_APP_PAUSE);
}

static void onStop(ANativeActivity* activity)
{
   logg("Stop: %p\n", activity);
   simple_command(activity, CMD_APP_STOP);
}

/***********************************************************************
 * #A.CB.SYSOBJ                                                        *
 *                       System Object Handling                        *
 *                                                                     *
 ***********************************************************************
  The system notifies the availability of various objects such as the
  native window, input queue etc. thru callbacks. It's quite stupid, but
  what can you do? The destruction of these objects are also reported
  in the same way.
 
  It looks like the default glue code doesn't handle all of the callbacks.
  These are: 

  void (*onNativeWindowResized)
        (ANativeActivity* activity, ANativeWindow* window);
  void (*onContentRectChanged)
        (ANativeActivity* activity, const ARect* rect);

  WindowResized and ContentRectChanged are called when the device is
  rotated, as far as I understand. WindowRedrawNeeded is called when
  the application is brought back to foreground. I will do these after 
  I have the graphics system.
 ***********************************************************************/

static void onNativeWindowCreated
   (ANativeActivity* activity, ANativeWindow* window)
{
  request_t R;
  logg("NativeWindowCreated: %p -- %p\n", activity, window);
  R.type= CMD_WINDOW_CREATED;
  R.d.window= window;
  make_request(activity->instance, &R);
}

static void onNativeWindowDestroyed
   (ANativeActivity* activity, ANativeWindow* window)
{
  request_t R;
  logg("NativeWindowDestroyed: %p -- %p\n", activity, window);
  R.type= CMD_WINDOW_DESTROYED;
  R.d.window= window;
  make_request(activity->instance, &R);
}

static void onNativeWindowRedrawNeeded
              (ANativeActivity* activity, ANativeWindow* window)
{
  request_t R;
  logg("NativeWindowRedraw: %p -- %p\n", activity, window);
  R.type= CMD_WINDOW_REDRAW;
  R.d.window= window;
  make_request(activity->instance, &R);
}

static void onInputQueueCreated
   (ANativeActivity* activity, AInputQueue* queue)
{
  request_t R;
  logg("InputQueueCreated: %p -- %p\n", activity, queue);
  R.type= CMD_INPUTQ_CREATED;
  R.d.inputq= queue;
  make_request(activity->instance, &R);
}

static void onInputQueueDestroyed
   (ANativeActivity* activity, AInputQueue* queue)
{
  request_t R;
  logg("InputQueueDestroyed: %p -- %p\n", activity, queue);
  R.type= CMD_INPUTQ_DESTROYED;
  R.d.inputq= queue;
  make_request(activity->instance, &R);
}

static void* onSaveInstanceState
   (ANativeActivity* activity, size_t* outLen)
{
  request_t R;
  logg("SaveInstanceState: %p\n", activity);
  R.type= CMD_SAVE_STATE;
  R.d.buffer.ptr= NULL;
  R.d.buffer.size= 0;
  make_request(activity->instance, &R);
  if (R.d.buffer.ptr)
    *outLen= R.d.buffer.size;
  return R.d.buffer.ptr;
}

static void onWindowFocusChanged
   (ANativeActivity* activity, int focused)
{
  request_t R;
  logg("WindowFocusChanged: %p -- %d\n", activity, focused);
  R.type= CMD_FOCUS_CHANGE;
  R.d.intv= focused;
  make_request(activity->instance, &R);
}

static void onLowMemory(ANativeActivity* activity)
{
  request_t R;
  logg("LowMemory: %p\n", activity);
  R.type= CMD_LOW_MEMORY;
  make_request(activity->instance, &R);
}

static void onConfigurationChanged(ANativeActivity* activity)
{
  request_t R;
  logg("ConfigurationChanged: %p\n", activity);
  R.type= CMD_CONFIG_CHANGE;
  make_request(activity->instance, &R);
}

static void onDestroy(ANativeActivity* activity)
{
  request_t R;
  app_t *A;

  logg("Destroy: %p\n", activity);

  A= activity->instance;
  R.type= CMD_QUIT;
  make_request(A, &R);

  close(A->cmdpipe_r);
  close(A->cmdpipe_w);
  pthread_cond_destroy(&A->cond);
  pthread_mutex_destroy(&A->mutex);
  free(A);
}

/***********************************************************************
 *                                                                     *
 *                             Java Bridge                             *
 *                                                                     *
 ***********************************************************************
  Now it's time to make a bridge between the Java world and native world.

  The following function gets called when the shared library gets loaded.
  This registers the native function for getting results from the Java
  environment. It also gets a reference to the activity class so that I
  don't have to look it up later. Looking it up later apperantly doesn't
  work because of some stupid class loader shit and such.
 ***********************************************************************/

#ifdef GFX_ANDROID_ACTIVITY

static void have_java_result(JNIEnv* env, jobject thiz, jint rcode)
{
  request_t R;
  R.type= CMD_JAVARESULT;
  R.d.intv= rcode;
  make_request(my_app, &R);
}


jint JNI_OnLoad(JavaVM *vm, void *reserved)
{
  static JNINativeMethod methods[]=
  {
   { "haveresult", "(I)V", (void*) have_java_result },
  };
  JNIEnv* env;
  int R,att;
  att= 0;
  R= (*vm)->GetEnv(vm, (void**) &env, JNI_VERSION_1_6); 
  if (R==JNI_EDETACHED)
  {
    R= (*vm)->AttachCurrentThread(vm, &env, NULL);
    if (R) return -1;
    att= 1;
  }
#define VSTRF(X) #X
#define STRFY(X) VSTRF(X)

  jclass cls= (*env)->FindClass(env, STRFY(GFX_ANDROID_ACTIVITY));
  (*env)->RegisterNatives(env, cls, methods,
                          sizeof(methods)/sizeof(methods[0]));
  return JNI_VERSION_1_6;
}
#endif

/***********************************************************************
 * #A.JNI                                                              *
 *                             JNI Helpers                             *
 *                                                                     *
 ***********************************************************************/
static jclass Jcls(JNIEnv *env, const char *name)
{
  return (*env)->FindClass(env, name);
}

static jclass JclsO(JNIEnv *env, jobject obj)
{
  return (*env)->GetObjectClass(env, obj);
}

static jobject JstaticfieldO
    (JNIEnv *env, jclass cls, const char *name,const char *type)
{
  jfieldID fid;
  fid= (*env)->GetStaticFieldID(env, cls, name, type);
  return (*env)->GetStaticObjectField(env, cls, fid);
}

  // returns and object (O) and takes an object (o)
static jobject JmethodO_o
  (JNIEnv *env, jobject obj, 
   const char *name, const char *proto,
   jobject arg1)
{
  jmethodID mid;
  jclass cls;

  cls= (*env)->GetObjectClass(env, obj);
  mid= (*env)->GetMethodID(env, cls, name, proto);
  return (*env)->CallObjectMethod(env, obj, mid, arg1);
}

  // returns and object (O) and takes no arguments
static jobject JmethodO
  (JNIEnv *env, jobject obj, const char *name, const char *proto)
{
  jmethodID mid;
  jclass cls;

  cls= (*env)->GetObjectClass(env, obj);
  mid= (*env)->GetMethodID(env, cls, name, proto);
  return (*env)->CallObjectMethod(env, obj, mid);
}

  // returns a boolean (B) and takes an object (o) and an integer (i)
static jboolean JmethodB_oi
  (JNIEnv *env, jobject obj, 
   const char *name, const char *proto,
   jobject arg1, int arg2)
{
  jmethodID mid;
  jclass cls;
  cls= (*env)->GetObjectClass(env, obj);
  mid= (*env)->GetMethodID(env, cls, name, proto);
  return (*env)->CallBooleanMethod(env, obj, mid, arg1, arg2);
}

static jint JmethodI
  (JNIEnv *env, jobject obj, const char *name) 
{
  jmethodID mid;
  jclass cls;
  cls= (*env)->GetObjectClass(env, obj);
  mid= (*env)->GetMethodID(env, cls, name, "()I");
  return (*env)->CallIntMethod(env, obj, mid);
}

/***********************************************************************
 * #A.COMMANDS                                                         *
 *                           Command Handler                           *
 *                                                                     *
 ***********************************************************************
  The following function handles the various commands generated by the
  callbacks. Some things are still missing, such as reporting the 
  application state etc. I will do these later on.

  The function returns
   - 1 if the received command causes us to quit the application
     This happens only if we get a destroy callback. 
     This callback isn't implemented yet.
   - 2 if the received command tells us to save the state of the
     application. 
   - 0 otherwise
 ***********************************************************************/

static void graphics_window_destroyed(app_t *A, ANativeWindow *window);
static void graphics_window_created(app_t *A, ANativeWindow *window);

static int handle_command(app_t *A,int *Rvalue)
{
  request_t *R;
  unsigned char *Sdata; int Slen;

  if (get_request(A, &R)) return 0;
  switch(R->type)
  {
  case CMD_APP_STOP:
     if (gfx_app_stop) (*gfx_app_stop)();
     complete_request(A,R);
     break;
  case CMD_APP_START:
  case CMD_APP_PAUSE:
  case CMD_APP_RESUME:
     // FIXME: inform the client, maybe??
     complete_request(A,R);
     break;
  case CMD_WINDOW_DESTROYED:
     graphics_window_destroyed(A, R->d.window);
     complete_request(A,R);
     break;
  case CMD_WINDOW_CREATED:
     graphics_window_created(A, R->d.window);
     complete_request(A,R);
     gfx_update_all();
     break;
  case CMD_WINDOW_REDRAW:
     gfx_update_all();
     break;
  case CMD_FOCUS_CHANGE:
  case CMD_LOW_MEMORY:
     complete_request(A,R);
     break;
  case CMD_CONFIG_CHANGE:
     logg("APP_CMD_CONFIG_CHANGED\n");
     AConfiguration_fromAssetManager(A->config, A->activity->assetManager);
     complete_request(A,R);
     break;
  case CMD_INPUTQ_DESTROYED:
     if (A->inputq!=R->d.inputq)
     {
       complete_request(A,R);
       break;
     }
     AInputQueue_detachLooper(A->inputq);
     A->inputq= NULL;
     complete_request(A,R);
     break;
  case CMD_INPUTQ_CREATED:
     if (A->inputq!=NULL)
        AInputQueue_detachLooper(A->inputq);
     A->inputq= R->d.inputq;
     AInputQueue_attachLooper(A->inputq, A->looper, PIPE_EVENT, NULL,A);
     complete_request(A,R);
     break;
  case CMD_QUIT:
     // FIXME: when do we get this?
     logg("APP_CMD_QUIT\n");
     A->quit_request= R;
     return 1;
  case CMD_SAVE_STATE:
     // This request gets completed when the save_state function gets
     // called. it MUST be called otherwise the main thread will hang 
     if (gfx_app_get_state) 
     {
       (*gfx_app_get_state)(&Sdata, &Slen);
     }
     else
     {
       Sdata= malloc(1); Sdata[0]= 0;
       Slen= 1;
     }
     R->d.buffer.size= Slen;
     R->d.buffer.ptr= Sdata;
     complete_request(A, R);
     break;
  case CMD_JAVARESULT:
     if (Rvalue) *Rvalue= R->d.intv;
     complete_request(A,R);
     return 3;
  }
  return 0;
}


/***********************************************************************
 *#A.EVENTS                                                            *
 *                      Android  Events                                *
 *                                                                     *
 ***********************************************************************/
static int millitime()
{
  struct timespec tv;
  clock_gettime(CLOCK_MONOTONIC, &tv);
  return tv.tv_sec*1000+tv.tv_nsec/1000/1000;
}

static int android_lookup_key
  (int meta_s, int meta_c, int meta_l, int kc, gfx_event_t *ev)
{
  int p, b;
  static char buf[2];
  char *s;
  int k;

  switch(kc)
  {
    case AKEYCODE_A: case AKEYCODE_B: case AKEYCODE_C: case AKEYCODE_D:
    case AKEYCODE_E: case AKEYCODE_F: case AKEYCODE_G: case AKEYCODE_H:
    case AKEYCODE_I: case AKEYCODE_J: case AKEYCODE_K: case AKEYCODE_L:
    case AKEYCODE_M: case AKEYCODE_N: case AKEYCODE_O: case AKEYCODE_P:
    case AKEYCODE_Q: case AKEYCODE_R: case AKEYCODE_S: case AKEYCODE_T:
    case AKEYCODE_U: case AKEYCODE_V: case AKEYCODE_W: case AKEYCODE_X:
    case AKEYCODE_Y: case AKEYCODE_Z:
       if (!meta_c && (meta_l || meta_s)) b= 'A'; else b= 'a';
       p= b + kc-AKEYCODE_A;
       goto prt;

          /***** Buttons at the bottom of the display ****/
    case AKEYCODE_HOME: 
    case AKEYCODE_BACK: k= GFX_K_ESCAPE; goto ctl;

    case AKEYCODE_DPAD_UP: k= GFX_K_UP; goto ctl;
    case AKEYCODE_DPAD_DOWN: k= GFX_K_DOWN; goto ctl;
    case AKEYCODE_DPAD_LEFT: k= GFX_K_LEFT; goto ctl;
    case AKEYCODE_DPAD_RIGHT: k= GFX_K_RIGHT; goto ctl;

    case AKEYCODE_INSERT: k= GFX_K_INSERT; goto ctl;
    case AKEYCODE_FORWARD_DEL: k= GFX_K_DELETE; goto ctl;
    case AKEYCODE_MOVE_HOME: k= GFX_K_HOME; goto ctl;
    case AKEYCODE_MOVE_END: k= GFX_K_END; goto ctl;
    case AKEYCODE_PAGE_UP: k= GFX_K_PAGEUP; goto ctl;
    case AKEYCODE_PAGE_DOWN: k= GFX_K_PAGEDOWN; goto ctl;

    case AKEYCODE_0: case AKEYCODE_1: case AKEYCODE_2: case AKEYCODE_3:
    case AKEYCODE_4: case AKEYCODE_5: case AKEYCODE_6: case AKEYCODE_7:
    case AKEYCODE_8: case AKEYCODE_9:
       if (meta_s) p= ")!@#$%^&*("[kc-AKEYCODE_0];
              else p= "0123456789"[kc-AKEYCODE_0];
       goto prt;

    case AKEYCODE_DEL: k= GFX_K_BACKSPACE; goto ctl;
    case AKEYCODE_TAB: k= GFX_K_TAB; goto ctl;
    case AKEYCODE_NUMPAD_ENTER:
    case AKEYCODE_ENTER: k= GFX_K_RETURN; goto ctl;

    /*** keycodes without shift meanings ***/
    case AKEYCODE_SPACE:               p= ' '; goto prt;
    case AKEYCODE_STAR:                p= '*'; goto prt;
    case AKEYCODE_POUND:               p= '#'; goto prt;
    case AKEYCODE_AT:                  p= '@'; goto prt;
    case AKEYCODE_PLUS:                p= '+'; goto prt;
    case AKEYCODE_NUMPAD_DIVIDE:       p= '/'; goto prt;
    case AKEYCODE_NUMPAD_MULTIPLY:     p= '*'; goto prt;
    case AKEYCODE_NUMPAD_SUBTRACT:     p= '-'; goto prt;
    case AKEYCODE_NUMPAD_ADD:          p= '+'; goto prt;
    case AKEYCODE_NUMPAD_DOT:          p= '.'; goto prt;
    case AKEYCODE_NUMPAD_COMMA:        p= ','; goto prt;
    case AKEYCODE_NUMPAD_EQUALS:       p= '='; goto prt;
    case AKEYCODE_NUMPAD_LEFT_PAREN:   p= '('; goto prt;
    case AKEYCODE_NUMPAD_RIGHT_PAREN:  p= ')'; goto prt;

    case AKEYCODE_NUMPAD_0:  p= '0'; goto prt;
    case AKEYCODE_NUMPAD_1:  p= '1'; goto prt;
    case AKEYCODE_NUMPAD_2:  p= '2'; goto prt;
    case AKEYCODE_NUMPAD_3:  p= '3'; goto prt;
    case AKEYCODE_NUMPAD_4:  p= '4'; goto prt;
    case AKEYCODE_NUMPAD_5:  p= '5'; goto prt;
    case AKEYCODE_NUMPAD_6:  p= '6'; goto prt;
    case AKEYCODE_NUMPAD_7:  p= '7'; goto prt;
    case AKEYCODE_NUMPAD_8:  p= '8'; goto prt;
    case AKEYCODE_NUMPAD_9:  p= '9'; goto prt;

    case AKEYCODE_COMMA:           s=",<"; goto shif;
    case AKEYCODE_PERIOD:          s=".>"; goto shif;
    case AKEYCODE_MINUS:           s="-_"; goto shif;
    case AKEYCODE_EQUALS:          s="=+"; goto shif;
    case AKEYCODE_LEFT_BRACKET:    s="[{"; goto shif;
    case AKEYCODE_RIGHT_BRACKET:   s="]}"; goto shif;
    case AKEYCODE_BACKSLASH:       s="\\|"; goto shif;
    case AKEYCODE_SEMICOLON:       s=";:"; goto shif;
    case AKEYCODE_APOSTROPHE:      s="'\""; goto shif;
    case AKEYCODE_SLASH:           s="/?"; goto shif;

    case AKEYCODE_F1: case AKEYCODE_F2: case AKEYCODE_F3: case AKEYCODE_F4:
    case AKEYCODE_F5: case AKEYCODE_F6: case AKEYCODE_F7: case AKEYCODE_F8:
    case AKEYCODE_F9: case AKEYCODE_F10: case AKEYCODE_F11: case AKEYCODE_F12:
    default:
        break;
  } 
  return 1;

shif:
  if (meta_s) p= s[1]; else p= s[0];
prt:
  buf[0]= p;
  buf[1]= 0;
  ev->string= buf;
  ev->wide= p;
  if (meta_c) ev->key= p | GFX_K_CONTROL;
         else ev->key= GFX_K_PRT;
  return 0;
ctl:
  buf[0]= 0;
  ev->string= buf;
  ev->wide= 0;
  ev->key= k;
  if (meta_c) ev->key |= GFX_K_CONTROL;
  if (meta_s) ev->key |= GFX_K_SHIFT;
  return 0;
}

static int handle_input(app_t *A, gfx_event_t *ev)
{
  AInputEvent* event;
  int type;
  uint32_t action,index, i, npointers;
  gfx_event_t dummy;
  event = NULL;
  if (!ev) ev= &dummy;
again:
     // Is this an error? should I log it or somehow
     // return an indication? FIXME

  if (AInputQueue_getEvent(A->inputq, &event)<0) return 1;

  type= AInputEvent_getType(event);
  // logg("New input event: type=%d\n", type);

  if (AInputQueue_preDispatchEvent(A->inputq, event)) goto again;

  switch(type)
  {
  case AINPUT_EVENT_TYPE_KEY: goto key_event;
  case AINPUT_EVENT_TYPE_MOTION: goto touch_event;
  }

next_event:
  AInputQueue_finishEvent(A->inputq, event, 1);
  goto again;

touch_event:
  action= AMotionEvent_getAction(event);
  index= (action & AMOTION_EVENT_ACTION_POINTER_INDEX_MASK)
              >>AMOTION_EVENT_ACTION_POINTER_INDEX_SHIFT;
  action= (action & AMOTION_EVENT_ACTION_MASK);

  switch(action)
  {
  case AMOTION_EVENT_ACTION_UP:
     ev->type= GFXEV_BUTTON_RELEASE;
     ev->button= 1;
     ev->x= AMotionEvent_getX(event,index);
     ev->y= AMotionEvent_getY(event,index);
     ev->key= AMotionEvent_getTouchMajor(event,index);
     ev->wide= AMotionEvent_getTouchMinor(event,index);
     A->last_pointer_id= -1;
     break;
  case AMOTION_EVENT_ACTION_DOWN:
     ev->type= GFXEV_BUTTON_PRESS;
     ev->x= AMotionEvent_getX(event,index);
     ev->y= AMotionEvent_getY(event,index);
     ev->key= AMotionEvent_getTouchMajor(event,index);
     ev->wide= AMotionEvent_getTouchMinor(event,index);
     ev->button= 1;
     A->last_pointer_id= AMotionEvent_getPointerId(event, index);    
     break;
  case AMOTION_EVENT_ACTION_MOVE:
     npointers= AMotionEvent_getPointerCount(event);
     for(i=0;i<npointers;i++)
        if (AMotionEvent_getPointerId(event, i)==A->last_pointer_id)
        { 
          ev->type= GFXEV_POINTER_MOVE;
          ev->x= AMotionEvent_getX(event,i);
          ev->y= AMotionEvent_getY(event,i);
          ev->key= AMotionEvent_getTouchMajor(event,i);
          ev->wide= AMotionEvent_getTouchMinor(event,i);
          goto done;
        }
  default:
     goto next_event;
  }

done:
  AInputQueue_finishEvent(A->inputq, event, 1);
  return 0;

key_event:
  switch(AKeyEvent_getAction(event))
  {
  case AKEY_STATE_UP: ev->type= GFXEV_KEY_RELEASE; break;
  case AKEY_STATE_DOWN: ev->type= GFXEV_KEY_PRESS; break;
  default: goto next_event;
  }
  int32_t meta, meta_s, meta_c, meta_l, keycode;
  meta= AKeyEvent_getMetaState(event);
  meta_s=meta & (AMETA_SHIFT_ON|AMETA_SHIFT_LEFT_ON|AMETA_SHIFT_RIGHT_ON);
  meta_c=meta & (AMETA_CTRL_ON|AMETA_CTRL_LEFT_ON|AMETA_CTRL_RIGHT_ON);
  meta_l=meta & AMETA_CAPS_LOCK_ON;
  keycode= AKeyEvent_getKeyCode(event);
  if (android_lookup_key(meta_s, meta_c, meta_l, keycode, ev)) goto next_event;
  goto done;
}


static int get_event(gfx_event_t *E)
{
   int ident;
   int value;
   int dFD, dEV;
   app_t *app;
   int elapsed, have_event;
   int have_timer, initime, delay;

again:
   have_timer= have_user_timer(&delay);
   if (delay<0) have_timer= 0;
   if (have_timer&&(delay==0))
     return elapsed_user_timer(0, 1, E);
   if (have_timer) initime= millitime();

   ident= ALooper_pollAll(have_timer? delay: -1,  &dFD, &dEV, (void**) &app);

   have_event= 0;

   switch(ident)
   {
     // somebody might have woke us up because the looper
     // has changed and we need to poll again.
   case ALOOPER_POLL_WAKE:
      break;
   case PIPE_CMD:
      switch(handle_command(app, &value))
      {
      case 1: E->type= GFXEV_QUIT; have_event=1;   break;
      case 3: E->type= GFXEV(value,JAVARESULT); have_event= 1; break;
      }
      break;
   case PIPE_EVENT:
      if (!handle_input(app, E) && app->window)
        have_event= 1;
      break;
   case ALOOPER_POLL_TIMEOUT:
      // the timeout has elapsed completely
      // this should always return 1, because we got the
      // delay from there..
      if (elapsed_user_timer(delay, 1, E)) return 1;
      goto again;
   default:
      if ((ident&0xffff0000)==0x00010000)
      {
        ident ^= 0x00010000;
        E->key= ident; E->wide= dFD;
        if (dEV&(ALOOPER_EVENT_HANGUP|ALOOPER_EVENT_ERROR|
                 ALOOPER_EVENT_INVALID))
           E->type= GFXEV_READY_EXCEPT;
        else if (dEV&ALOOPER_EVENT_INPUT)
           E->type= GFXEV_READY_READ;
        else if (dEV&ALOOPER_EVENT_OUTPUT)
           E->type= GFXEV_READY_WRITE;
        else 
           E->type= GFXEV_READY_EXCEPT;
        have_event= 1;
        break;
      }
   }

   elapsed= millitime()- initime;
   if (have_event)
   {
     if (have_timer)
        elapsed_user_timer(elapsed, 0, NULL);
     return 1;
   }
   if (have_timer && elapsed_user_timer(elapsed,1,E))
     return 1;
   goto again;
}

/***********************************************************************
 *                            Command Loop                             *
 ***********************************************************************
  We need a function which simply waits for commands and ignores events.
  This is necessary when we're initalizing or destroying the application
  object. The function returns 1 if a request to quit the application was
  received. Otherwise, it returns 0.
 ***********************************************************************/

static int execute_commands(app_t *app)
{
   int ident;
   int dFD, dEV;
   void *dPTR;

again:
   ident= ALooper_pollAll(-1,  &dFD, &dEV,  &dPTR);

   switch(ident)
   {
     // somebody might have woke us up because the looper
     // has changed and we need to poll again.
   case ALOOPER_POLL_WAKE:
      goto again;
   case PIPE_CMD:
      if (handle_command(app,NULL)==1) return 1;
      return 0;
   case PIPE_EVENT:
     handle_input(app, NULL);
     goto again;
   }

   // other things are not possible. maybe I can add some
   // sort of a time-out later.
   // so, unhandled stuff corresponds to an error, which
   // makes the application quit. I think it's OK.
   return 1;
}

/***********************************************************************
 * #A.INIT                                                             *
 *                          Application Init                           *
 *                                                                     *
 ***********************************************************************
  During the initialization, we need to do the following:
   - create a pipe for the asynchronous messages coming in
   - create a mutex and a condition variable for protecting shared memory
   - create a worker thread

  The onCreate function is the entry point from the Java world. We 
  initialize the 'activity' object with the system event handler functions.
 ***********************************************************************/

static void *worker_entry(void*);

void ANativeActivity_onCreate
      (ANativeActivity* activity, void* state, size_t state_size)
{
  int cmdpipe[2];
  pthread_attr_t attr; 
  app_t *app;

  logg("Creating: %p\n", activity);

  // #C.1: Activity callback installation
  activity->callbacks->onDestroy = onDestroy;
  activity->callbacks->onStart = onStart;
  activity->callbacks->onResume = onResume;
  activity->callbacks->onPause = onPause;
  activity->callbacks->onStop = onStop;
  activity->callbacks->onNativeWindowCreated = onNativeWindowCreated;
  activity->callbacks->onNativeWindowDestroyed = onNativeWindowDestroyed;
  activity->callbacks->onInputQueueCreated = onInputQueueCreated;
  activity->callbacks->onInputQueueDestroyed = onInputQueueDestroyed;
  activity->callbacks->onSaveInstanceState = onSaveInstanceState;
  activity->callbacks->onWindowFocusChanged = onWindowFocusChanged;
  activity->callbacks->onLowMemory = onLowMemory;
  activity->callbacks->onConfigurationChanged = onConfigurationChanged;

  ANativeActivity_setWindowFormat(activity, WINDOW_FORMAT_RGBX_8888);
  ANativeActivity_setWindowFlags(activity, 
    AWINDOW_FLAG_FULLSCREEN  ,
    AWINDOW_FLAG_FORCE_NOT_FULLSCREEN |
    AWINDOW_FLAG_LAYOUT_INSET_DECOR |
    AWINDOW_FLAG_LAYOUT_IN_SCREEN |
    AWINDOW_FLAG_LAYOUT_NO_LIMITS
  );

  activity->instance= my_app= app= calloc(1,sizeof(*app));
  app->activity= activity;
  app->last_pointer_id= -1;

  if (state && state_size)
  {
    app->init_state= malloc(state_size);
    memcpy(app->init_state, state, state_size);
    app->init_state_size= state_size;
  }

  pthread_mutex_init(&app->mutex, NULL);
  pthread_cond_init(&app->cond,NULL);

  if (pipe(cmdpipe))
  {
    logg("Couldn't create pipe: %s\n",strerror(errno));
    return ;
  }
  app->cmdpipe_r= cmdpipe[0];
  app->cmdpipe_w= cmdpipe[1];

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  pthread_create(&app->worker_thread, &attr, worker_entry, app);

  pthread_mutex_lock(&app->mutex); 
  while(!app->running) pthread_cond_wait(&app->cond, &app->mutex); 
  pthread_mutex_unlock(&app->mutex);
}

static void* worker_entry(void* arg)
{
  app_t *app;

#ifndef GFX_ANDROID_RELEASE
 if (android_log_target) locon(android_log_target);
#endif

  app= arg;

  app->config= AConfiguration_new();
  AConfiguration_fromAssetManager(app->config, app->activity->assetManager);

  app->looper= ALooper_prepare(ALOOPER_PREPARE_ALLOW_NON_CALLBACKS);
  ALooper_addFd(app->looper, app->cmdpipe_r,
                PIPE_CMD, ALOOPER_EVENT_INPUT, NULL, app);

  pthread_mutex_lock(&app->mutex);
  app->running=1;
  pthread_cond_broadcast(&app->cond);
  pthread_mutex_unlock(&app->mutex);

  getsysbar();
  getnavbar();

  while(!app->inputq || !app->window)
    if (execute_commands(app)) goto done;

  gfx_init_module(gfx_module,gfx_module_len);

  init_data();
  android_main(app->init_state, app->init_state_size);

  while(!app->quit_request)
     if (execute_commands(app))
       break;

done:
  //free_saved_state(A);
  if (app->inputq)
      { AInputQueue_detachLooper(app->inputq); app->inputq= NULL; }
  AConfiguration_delete(app->config);
  app->config= NULL;

  if (app->quit_request)
     complete_request(app, app->quit_request);
  return NULL;
}
 


/***********************************************************************
 * #A.GRAPHICS                                                         *
 *                       Android Graphics                              *
 *                                                                     *
 ***********************************************************************
  In contrast to the input system, the output system is quite simple.
  When you want to put something on the screen, you
   - lock the corresponding rectangle
   - fill the returned buffer
   - unlock and post the buffer
  Since android doesn't keep the contents of a window, I need to do it
  myself, just like I did it for the X11 version, using an X11 image
  and pixmap combo. I will store the image as an array of uint32_t pixels,
  but I'm planning to update the screen using other formats as well.
 
  So basicly, the image will be stored in RGBX_8888 format but if the
  unlock stuff returns a different format, I will convert the pixels
  to that format when I'm writing to the pixel array of the buffer.
 ***********************************************************************/

static void graphics_window_destroyed(app_t *A, ANativeWindow *window)
{
  if (A->window!=window) return ;
  A->window= NULL;
}

static void graphics_window_created(app_t *A, ANativeWindow *window)
{
   int W,H;
   A->window= window;
   W= ANativeWindow_getWidth(A->window);
   H= ANativeWindow_getHeight(A->window);
   if (W<=0) W= 1;
   if (H<=0) H= 1;

   if (W!=x_window_width || H!=x_window_height || !Gpixels)
   {
     int y;
     if (Gpixels) free(Gpixels);
     Gpixels= malloc(sizeof(uint32_t)*W*H);
     if (x_image_data) free(x_image_data);
     x_image_data= malloc(sizeof(x_image_data[0])*H);
     for(y=0;y<H;y++)
       x_image_data[y]= y*W + Gpixels;
     x_window_width= W;
     x_window_height= H;
   }
}


static void gfx_update(int x,int y,int w,int h)
{
   ARect rect;
   ANativeWindow_Buffer buffer;
   int R;
   uint32_t *spixels;
   uint16_t *wpixels;
   int eX, eY;

   if (!updates_on) return ;
   if (!Gpixels) return ; // display not initialized yet
   if (!my_app->window) return ;

   eX= x+w; eY= y+h;

   if (x>=x_window_width || eX<=0 || y>=x_window_height || eY<=0) return ;

   eX= imin2(eX, x_window_width); eY= imin2(eY, x_window_height);
   x= imax2(x, 0); y= imax2(y,0);

   rect.left= x; rect.top= y;
   rect.right= eX; rect.bottom= eY;

   R= ANativeWindow_lock(my_app->window, &buffer, &rect);
   if (R<0)
   {
      logg("Unable to lock buffer: %d\n", R);
      return ;
   }

   // logg("GFX::Buffer width=%d height=%d stride=%d format=%d\n", buffer.width, buffer.height,  buffer.stride, buffer.format);

   // logg("GFX::Rect left=%d top=%d right=%d bottom=%d\n", rect.left,
           // rect.top, rect.right, rect.bottom);
   
  switch(buffer.format)
  {
   case WINDOW_FORMAT_RGBX_8888:
     spixels= buffer.bits;
     for(y=rect.top;y<rect.bottom;y++)
       for(x=rect.left;x<rect.right;x++)
         spixels[y*buffer.stride+x]= Gpixels[x+ y*x_window_width];
     break;
   case WINDOW_FORMAT_RGBA_8888:
     spixels= buffer.bits;
     for(y=rect.top;y<rect.bottom;y++)
       for(x=rect.left;x<rect.right;x++)
         spixels[y*buffer.stride+x]= Gpixels[x+ y*x_window_width] | 0xff000000;
     break;
   case WINDOW_FORMAT_RGB_565:
     wpixels= buffer.bits;
     for(y=rect.top;y<rect.bottom;y++)
       for(x=rect.left;x<rect.right;x++)
       {
         uint32_t P,R,G,B;
         P= x_image_data[y][x];
         R= (P>>16)&0xff;  R>>= 3;
         G= (P>>8) &0xff;  G>>= 2;
         B= P&0xff;        B>>= 3;
         wpixels[y*buffer.stride+x]= (R<<11) | (G<<5) | B;
       }
      break;
    default:
      logg("unknown buffer format %d",buffer.format);
      break;
  }
  ANativeWindow_unlockAndPost(my_app->window);
}

static int get_java_value(char *fname, int defval)
{
  JNIEnv* J;
  JavaVM *vm;
  ANativeActivity* activity;
  int R, att;
  int result;

  jobject jo_activity;
 
  activity= my_app->activity;
   // fuck this shit. clazz is actually a pointer to the
   // activity object.
  vm= my_app->activity->vm;
  
  att= 0;
  R= (*vm)->GetEnv(vm, (void**) &J, JNI_VERSION_1_6); 
  if (R==JNI_EDETACHED)
  {
    R= (*vm)->AttachCurrentThread(vm, &J, NULL);
    if (R) {
      logg("java_val(%s): Can not attach thread\n",fname);
      return defval;
    }
    att= 1;
  }

  jo_activity= activity->clazz;

  result= JmethodI(J, jo_activity, fname);

  if (att) (*vm)->DetachCurrentThread(vm);
  logg("java_val(%s) = %d\n", fname, result);
  return result;
}

static int getdpi()
{
  if (!display_dpi) display_dpi= get_java_value("getDisplayDPI", 86);
  return display_dpi;
}

static int getsysbar()
{
  if (!gui_sysbar_height) 
     gui_sysbar_height= get_java_value("getStatusBarHeight", 25);
  return gui_sysbar_height;
}

static int getnavbar()
{
  if (!gui_navbar_height)
     gui_navbar_height= get_java_value("getNavigationBarHeight", 25);
  return gui_navbar_height;
}


/***********************************************************************
 * #A.SOFTKEY                                                          *
 *                            Soft Keyboard                            *
 *                                                                     *
 ***********************************************************************
  This is from:

https://groups.google.com/forum/?fromgroups=#!topic/android-ndk/Tk3g00wLKhk

  It uses the C++ interface. The corresponding Java code is this:

    imm= activity.getSystemService(Context.INPUT_METHOD_SERVICE);
    decorview= activity.getWindow().getDecorView();

    if (show) 
      result= imm.showSoftInput(decorview, 0);
    else 
      result= imm.hideSoftInputFromWindow(decorview.getWindowToken(), 0);

  void displayKeyboard(bool pShow) 
  {
    // Attaches the current thread to the JVM.
    jint lResult;
    jint lFlags = 0;

    JavaVM* lJavaVM = mApplication->activity->vm;
    JNIEnv* lJNIEnv = mApplication->activity->env;

    JavaVMAttachArgs lJavaVMAttachArgs;
    lJavaVMAttachArgs.version = JNI_VERSION_1_6;
    lJavaVMAttachArgs.name = "NativeThread";
    lJavaVMAttachArgs.group = NULL;

    lResult=lJavaVM->AttachCurrentThread(&lJNIEnv, &lJavaVMAttachArgs);
    if (lResult == JNI_ERR) {
        return;
    }

    // Retrieves NativeActivity.
    jobject lNativeActivity = mApplication->activity->clazz;
    jclass ClassNativeActivity = lJNIEnv->GetObjectClass(lNativeActivity);

    // Retrieves Context.INPUT_METHOD_SERVICE.
    jclass ClassContext = lJNIEnv->FindClass("android/content/Context");
    jfieldID FieldINPUT_METHOD_SERVICE =
        lJNIEnv->GetStaticFieldID(ClassContext,
            "INPUT_METHOD_SERVICE", "Ljava/lang/String;");
    jobject INPUT_METHOD_SERVICE =
        lJNIEnv->GetStaticObjectField(ClassContext,
            FieldINPUT_METHOD_SERVICE);
    jniCheck(INPUT_METHOD_SERVICE);

    // Runs getSystemService(Context.INPUT_METHOD_SERVICE).
    jclass ClassInputMethodManager = lJNIEnv->FindClass(
        "android/view/inputmethod/InputMethodManager");
    jmethodID MethodGetSystemService = lJNIEnv->GetMethodID(
        ClassNativeActivity, "getSystemService",
        "(Ljava/lang/String;)Ljava/lang/Object;");
    jobject lInputMethodManager = lJNIEnv->CallObjectMethod(
        lNativeActivity, MethodGetSystemService,
        INPUT_METHOD_SERVICE);

    // Runs getWindow().getDecorView().
    jmethodID MethodGetWindow = lJNIEnv->GetMethodID(
        ClassNativeActivity, "getWindow",
        "()Landroid/view/Window;");
    jobject lWindow = lJNIEnv->CallObjectMethod(lNativeActivity,
        MethodGetWindow);
    jclass ClassWindow = lJNIEnv->FindClass(
        "android/view/Window");
    jmethodID MethodGetDecorView = lJNIEnv->GetMethodID(
        ClassWindow, "getDecorView", "()Landroid/view/View;");
    jobject lDecorView = lJNIEnv->CallObjectMethod(lWindow,
        MethodGetDecorView);

    if (pShow) {
        // Runs lInputMethodManager.showSoftInput(...).
        jmethodID MethodShowSoftInput = lJNIEnv->GetMethodID(
            ClassInputMethodManager, "showSoftInput",
            "(Landroid/view/View;I)Z");
        jboolean lResult = lJNIEnv->CallBooleanMethod(
            lInputMethodManager, MethodShowSoftInput,
            lDecorView, lFlags);
    } else {
        // Runs lWindow.getViewToken()
        jclass ClassView = lJNIEnv->FindClass(
            "android/view/View");
        jmethodID MethodGetWindowToken = lJNIEnv->GetMethodID(
            ClassView, "getWindowToken", "()Landroid/os/IBinder;");
        jobject lBinder = lJNIEnv->CallObjectMethod(lDecorView,
            MethodGetWindowToken);

        // lInputMethodManager.hideSoftInput(...).
        jmethodID MethodHideSoftInput = lJNIEnv->GetMethodID(
            ClassInputMethodManager, "hideSoftInputFromWindow",
            "(Landroid/os/IBinder;I)Z");
        jboolean lRes = lJNIEnv->CallBooleanMethod(
            lInputMethodManager, MethodHideSoftInput,
            lBinder, lFlags);
    }

    // Finished with the JVM.
    lJavaVM->DetachCurrentThread(); 
   }
 ***********************************************************************/
void gfx_soft_keyboard(int show)
{
  JNIEnv* J;
  JavaVM *vm;
  ANativeActivity* activity;
  int R, att;
  int flags;

  jobject jo_activity;
  jobject jo_imm;
  jobject jo_decorview;
  jboolean result;
 
  activity= my_app->activity;
   // fuck this shit. clazz is actually a pointer to the
   // activity object.
  vm= my_app->activity->vm;
  
  att= 0;
  R= (*vm)->GetEnv(vm, (void**) &J, JNI_VERSION_1_6); 
  if (R==JNI_EDETACHED)
  {
    R= (*vm)->AttachCurrentThread(vm, &J, NULL);
    if (R) return ;
    att= 1;
  }

  jo_activity= activity->clazz;
  flags= 0;

  jo_imm= JmethodO_o(J, jo_activity, 
             "getSystemService", "(Ljava/lang/String;)Ljava/lang/Object;",
             JstaticfieldO(J, 
               Jcls(J,"android/content/Context"),
               "INPUT_METHOD_SERVICE", "Ljava/lang/String;"));

  jo_decorview= JmethodO(J,
                  JmethodO(J, jo_activity,
                    "getWindow", "()Landroid/view/Window;"),
                  "getDecorView", "()Landroid/view/View;"); 
  if (show) 
     result= JmethodB_oi(J, jo_imm, 
               "showSoftInput", "(Landroid/view/View;I)Z",
               jo_decorview, flags);
  else 
     result= JmethodB_oi(J, jo_imm,
               "hideSoftInputFromWindow", "(Landroid/os/IBinder;I)Z",
               JmethodO(J, jo_decorview,  
                 "getWindowToken", "()Landroid/os/IBinder;"),
               flags);
  if (att) (*vm)->DetachCurrentThread(vm);
}



// Miscellaneous Functions
JavaVM *gfx_get_javavm()
{
  return my_app->activity->vm;
}

ANativeActivity* gfx_get_activity()
{
  return my_app->activity;
}

#endif   // GFX_ANDROID

