
/***********************************************************************
 *                                                                     *
 *                         Graphics Operations                         *
 *                                                                     *
 ***********************************************************************
 Most graphics operations involve converting pixels from user format to
 the server format. User format is xRGB32.

 Other operations are quite simple, like drawing lines, rectangles etc.
 ***********************************************************************/

/***********************************************************************
 * User color format is xRGB32. The following are just some accessors. *
 ***********************************************************************/
static uint32_t u_red(uint32_t C) { return (C>>16)&0xff; }
static uint32_t u_green(uint32_t C) { return (C>>8)&0xff; }
static uint32_t u_blue(uint32_t C) { return (C)&0xff; }

static uint32_t color_u(uint32_t uR, uint32_t uG, uint32_t uB)
{
   return ((uR&0xff)<<16)| ((uG&0xff)<<8)| (uB&0xff);
}

static unsigned int s_red(unsigned int V)
   { return (V&x_red_mask)>>x_red_shift; }
static unsigned int s_green(unsigned int V)
   { return (V&x_green_mask)>>x_green_shift; }
static unsigned int s_blue(unsigned int V)
   { return (V&x_blue_mask)>>x_blue_shift; }

static uint32_t color_s(uint32_t uR, uint32_t uG, uint32_t uB)
{
  return ((uR<<x_red_shift)&x_red_mask)
        |((uG<<x_green_shift)&x_green_mask)
        |((uB<<x_blue_shift)&x_blue_mask);
}

static uint32_t color_u_to_s(uint32_t U)
{
  return color_s(u_red(U), u_green(U), u_blue(U));
}

/***********************************************************************
 * I could use fixed point arithmetic for this easily. I don't need a  *
 * double, there are only 256 levels.                                  *
 ***********************************************************************/
static unsigned int blend_chan(unsigned int F, double A, unsigned int B)
{
  return  (unsigned int) ((F&0xff)*A + (B&0xff)*(1-A));
}

/***********************************************************************
 The following functions are used to blend two colors using an alpha
 value. The name of function is as follows: blend_xyZ
 x,y,z are either u or s. u stands for user and s stands for server.
 x and y are the input color type and z is the output color type.
 x is always the foreground and y is the background.
 ***********************************************************************/
static uint32_t blend_uuS(uint32_t alpha, uint32_t fg, uint32_t bg)
{
  double A;

  if (alpha==0xff) return color_u_to_s(fg);
  if (alpha==0) return color_u_to_s(bg);

  A= (double) alpha / 255.0;

  return color_s( blend_chan( u_red(fg), A, u_red(bg) ),
                  blend_chan( u_green(fg), A, u_green(bg) ),
                  blend_chan( u_blue(fg), A, u_blue(bg)));
}

static uint32_t blend_uuU(uint32_t alpha, uint32_t fg, uint32_t bg)
{
  double A;

  if (alpha==0xff)  return fg; 
  if (alpha==0)    return bg;  

  A= (double) alpha / 255.0;
  return color_u( blend_chan( u_red(fg), A, u_red(bg) ),
                  blend_chan( u_green(fg), A, u_green(bg) ),
                  blend_chan( u_blue(fg), A, u_blue(bg)));
}

static uint32_t blend_usS(uint32_t alpha, uint32_t fg, uint32_t bg)
{
  double A;

  if (alpha==0) return bg;
  if (alpha==0xff) return color_u_to_s(fg);

  A= (double) alpha / 255.0;
  return color_s( blend_chan( u_red(fg), A, s_red(bg) ),
                  blend_chan( u_green(fg), A, s_green(bg) ),
                  blend_chan( u_blue(fg), A, s_blue(bg)));
}

static unsigned int blend_ssS(uint32_t alpha, uint32_t F, uint32_t B)
{
  double A;
  if (alpha==0xff) return F;
  if (alpha==0)    return B;
  A= alpha/255.0;
  return color_s( blend_chan( s_red(F), A, s_red(B)),
                  blend_chan( s_green(F), A, s_green(B)),
                  blend_chan( s_blue(F), A, s_blue(B)));
}

static unsigned int blend_FuuS(uint32_t F, double A, uint32_t B)
{
  return color_s( blend_chan( s_red(F), A, s_red(B)),
                  blend_chan( s_green(F), A, s_green(B)),
                  blend_chan( s_blue(F), A, s_blue(B)));
}

static void gfx_fill_rect(int flags, int dst_x,int dst_y,
                   int w,int h, unsigned int color)
{
  int x,y;
  uint32_t sv;
  clip_for_rect(&dst_x, &w, x_window_width);
  clip_for_rect(&dst_y, &h, x_window_height);
  if (w<=0 || h<=0) return ;
  sv= color_u_to_s(color);
  for(y=0;y<h;y++)
   for(x=0;x<w;x++)
     x_image_data[y+dst_y][x+dst_x]= sv;
  if (flags&GFX_UPDATE) gfx_update(dst_x, dst_y, w,h);
}

static void fill_clipped_rect
               (int Cx,int Cy,int Cw, int Ch, 
                int dx,int dy,int ow, int oh, unsigned int color)
{
  if (ow<=0 || oh<=0) return ;
  if (dx>=(Cx+Cw) || dy>=(Cy+Ch) || (dx+ow<=Cx) || (dy+oh<=Cy)) return ;

  if (dx<Cx)
  {
    ow-= (Cx-dx);
    dx= Cx;
  }

  if (dy<Cy)
  {
    oh-= (Cy-dy);
    dy= Cy;
  }
  if (ow<=0 || oh<=0) return ;

  if ((dy+oh)>=(Cy+Ch))
  {
    oh= Cy+Ch-dy;
  }
  if ((dx+ow)>=(Cx+Cw))
  {
    ow= Cx+Cw-dx;
  }
  if (ow<=0 || oh<=0) return ;
  gfx_fill_rect(0,dx,dy,ow,oh,color);
}

static void blend_graymap(int flags, int fg, int bg,
                          const unsigned char *src, int src_w, int src_h,
                          int src_x, int src_y,
                          int dst_x, int dst_y, int width, int height)
{
  int x,y;
  uint32_t A;

  clip_for_rect( &src_x, &width, src_w);
  clip_for_rect( &src_y, &height, src_h);
  clip_for_rect( &dst_x, &width, x_window_width);
  clip_for_rect( &dst_y, &height, x_window_height);
  if (width<=0 || height<=0) return ;

  if (flags&GFX_BACKGND) // blend with fg,bg pair
  {
    for(y=0;y<height;y++)
      for(x=0;x<width;x++)   
      {
        A= src[(x+src_x) + (y+src_y)*src_w];
        x_image_data[y+dst_y][x+dst_x]= blend_uuS(A,fg, bg);
      }
  } 
  else
  {
    for(y=0;y<height;y++)
      for(x=0;x<width;x++)   
      {
        A= src[(x+src_x) + (y+src_y)*src_w];
        x_image_data[y+dst_y][x+dst_x]= 
             blend_usS(A,fg, x_image_data[y+dst_y][x+dst_x]);
      }
  } 

  if (flags&GFX_UPDATE)
    gfx_update(dst_x, dst_y, width, height);
}


static void hline(int x1,int x2, int y,uint32_t color)
{
  if (y<0 || y>=x_window_height || x1>x2 || x1>=x_window_width) return ;
  if (x1<0) x1= 0;
  if (x2>=x_window_width) x2= x_window_width-1;

  for(;x1<=x2;x1++)
    x_image_data[y][x1]= color;
}

static void vline(int x,int y1, int y2,uint32_t color)
{
  if (x<0 || x>=x_window_width || y1>y2 || y1>=x_window_height) return ;
  if (y1<0) y1= 0;
  if (y2>=x_window_height) y2= x_window_height-1;

  for(;y1<=y2;y1++)
    x_image_data[y1][x]= color;
}


static void gfx_draw_rect(int flags, int dst_x,int dst_y,
                   int w,int h, unsigned int color)
{
  uint32_t sv;
  sv= color_u_to_s(color);
  hline(dst_x, dst_x+w-1, dst_y, sv);
  hline(dst_x, dst_x+w-1, dst_y+h-1, sv);
  vline(dst_x, dst_y, dst_y+h-1, sv);
  vline(dst_x+w-1, dst_y, dst_y+h-1, sv);
  if (flags&GFX_UPDATE) gfx_update(dst_x, dst_y, w,h);
}

static void gfx_draw_rect_in
   (unsigned int color, int x,int y, int w,int h,
     int Ax, int Ay, int Aw, int Ah)
{
  if (clip_for_rect(&Ax, &Aw, x_window_width)) return ;
  if (clip_for_rect(&Ay, &Ah, x_window_height)) return ;
  color= color_u_to_s(color);

  if (x<Ax) { w -= Ax- x; x= Ax; }
  if (y<Ay) { h -= Ay -y; y= Ay; }
  if (w<=0 || h<=0) return ;
  if (x+w>Ax+Aw) w= Ax+Aw-x; 
  if (y+h>Ay+Ah) h= Ay+Ah-y; 
  if (w<=0 || h<=0) return ;
  hline(    x, x+w-1,     y, color);
  hline(    x, x+w-1, y+h-1, color);
  vline(    x,     y, y+h-1, color);
  vline(x+w-1,     y, y+h-1, color);
}



/***********************************************************************
  The following functions are used for drawing scroll bars. Eventually,
  I will use them for drawing buttons as well.
 ***********************************************************************/
static void blend_graymap_buffer
        (unsigned int fg, const unsigned char *src, int src_w, int src_h,
                          int src_x, int src_y,
                          unsigned int *dst, int dst_w, int dst_h,
                          int dst_x, int dst_y, int width, int height)
{
  int x,y;
  uint32_t A,D;

  clip_for_rect( &src_x, &width, src_w);
  clip_for_rect( &src_y, &height, src_h);
  clip_for_rect( &dst_x, &width, dst_w);
  clip_for_rect( &dst_y, &height, dst_h);
  if (width<=0 || height<=0) return ;

  for(y=0;y<height;y++)
  {
    for(x=0;x<width;x++)   
    {
       A= src[(x+src_x) + (y+src_y)*src_w];
       D= (y+dst_y)*dst_w + (x+dst_x);
       dst[D]= blend_uuU(A,fg, dst[D]);
    }
  }
}

static void gfx_put_alpha
            (const unsigned int *buffer, int src_w, int src_h,
             int src_x, int src_y,
             int dst_x, int dst_y,
             int width, int height)
{
  int x,y;

  clip_for_rect( &src_x, &width, src_w);
  clip_for_rect( &src_y, &height, src_h);
  clip_for_rect( &dst_x, &width, x_window_width);
  clip_for_rect( &dst_y, &height, x_window_height);
  if (width<=0 || height<=0) return ;

  for(y=0;y<height;y++)
    for(x=0;x<width;x++)
  {
    uint32_t sA, sP;

    sP= buffer[ (x+src_x) + (y+src_y)*src_w ];
    sA= (sP >> 24)&0xff;
    if (!sA) continue;
    x_image_data[y+dst_y][x+dst_x]= blend_usS
                               (sA, sP, x_image_data[y+dst_y][x+dst_x]);
  }
}

static void gfx_put(int flags,
             const unsigned int *buffer, int src_w, int src_h,
             int src_x, int src_y,
             int dst_x, int dst_y,
             int width, int height)
{
  int x,y;

  clip_for_rect( &src_x, &width, src_w);
  clip_for_rect( &src_y, &height, src_h);
  clip_for_rect( &dst_x, &width, x_window_width);
  clip_for_rect( &dst_y, &height, x_window_height);
  if (width<=0 || height<=0) return ;

  for(y=0;y<height;y++)
    for(x=0;x<width;x++)
         x_image_data[y+dst_y][x+dst_x]= color_u_to_s
                          (buffer[ (x+src_x) + (y+src_y)*src_w ]);
  if (flags & GFX_UPDATE)
      gfx_update(dst_x,dst_y,width,height);
}

static void gfx_put_gray(int flags,
             const unsigned char *buffer, int src_w, int src_h,
             int src_x, int src_y,
             int dst_x, int dst_y,
             int width, int height)
{
  int x,y;
  static int gray_palette[256];

  clip_for_rect( &src_x, &width, src_w);
  clip_for_rect( &src_y, &height, src_h);
  clip_for_rect( &dst_x, &width, x_window_width);
  clip_for_rect( &dst_y, &height, x_window_height);
  if (width<=0 || height<=0) return ;

  if ((x_red_mask | x_green_mask | x_blue_mask)== 0xffFFff)
  {
    for(y=0;y<height;y++)
      for(x=0;x<width;x++)
      {  unsigned char P= buffer[ (x+src_x) + (y+src_y)*src_w ];
         x_image_data[y+dst_y][x+dst_x]= (P) | (P<<8) | (P<<16); }
  }
  else
  {
    if (!gray_palette[255])
    {
      int i;
      for(i=0;i<256;i++)
        gray_palette[i]= ((i<<x_red_shift)&x_red_mask) |
                         ((i<<x_green_shift)&x_green_mask) |
                         ((i<<x_blue_shift)&x_blue_mask);
    }

    for(y=0;y<height;y++)
      for(x=0;x<width;x++)
         x_image_data[y+dst_y][x+dst_x]= 
           gray_palette[ buffer[ (x+src_x) + (y+src_y)*src_w ] ];
  }

  if (flags & GFX_UPDATE)
      gfx_update(dst_x,dst_y,width,height);
}

/***********************************************************************
  I need to make a function which draws a 3d box. This will be handy for
  buttons etc. First I will make a renderer for the screen. Later on,
  I can make one for buffers. This one will make a box with light coming
  in from the top left.
 ***********************************************************************/

static void gui_draw_icon_centered(int x,int y,int w,int h, gui_icon_t *icon)
{
  int iw,ih, copy_w, copy_h, src_x, src_y, dst_x, dst_y;
  iw= icon->w;
  ih= icon->h;
  if (iw<=w)
  {
    src_x= 0;
    dst_x= (w-iw)/2;
    copy_w= iw;
  }
  else
  {
    src_x= (iw-w)/2;
    dst_x= 0;
    copy_w= w;
  }
  if (ih<=h)
  {
    src_y= 0;
    dst_y= (h-ih)/2;
    copy_h= ih;
  }
  else
  {
    src_y= (ih-h)/2;
    dst_y= 0;
    copy_h= h;
  }
  gfx_put_alpha(icon->data, iw, ih, src_x, src_y, x+dst_x,y+dst_y,
                                            copy_w, copy_h);
}


/***********************************************************************
 *                                                                     *
 *                           Text Processing                           *
 *                                                                     *
 ***********************************************************************/


static void get_text_data(int font, const char *str, 
                          int *Rw, int *Rh,int *Ra,unsigned char **Rd)
{
  uint32_t *wtr;
  size_t L= strlen(str);
  wtr= alloca((L+1)*sizeof(uint32_t));
  utf8_decode_string((uint8_t*) str, L, wtr, &L);

  if (font<1) font=1;
  if (font>gui_nfonts) font=gui_nfonts;
  font_render(gui_fonts+font-1, wtr, L, Rw, Rh, Ra, Rd);
}

static void text_to_graypix(int font, const char *str, graypix_t *gp, int *Ra)
{
  get_text_data(font, str, &gp->w, &gp->h, Ra, &gp->data);
  gp->stride= gp->w;
}

static void gfx_text_size(int font, char *str, int *W, int *H)
{
  get_text_data(font, str, W, H, NULL, NULL);
}

static void txt_print(int flags, int font, int fg, int bg, int x, int y, 
               char *str,int *Rw, int *Rh)
{
  int w,h; unsigned char *data;
  get_text_data(font, str, &w, &h, NULL, &data);

  blend_graymap(flags, fg, bg, data, w,h, 0,0, x,y,w,h);
  if (data) free(data);
  if (Rw) *Rw= w;
  if (Rh) *Rh= h;
}

static int max_text_width(int font, ...)
{
  char *str;
  va_list ap;
  int mw, tw;
  va_start(ap,font);
  mw= 0;
  while((str=va_arg(ap,char*)))
  {
    get_text_data(font,str,&tw, NULL,NULL,NULL);
    if (tw>mw) mw= tw;
  }
  va_end(ap);
  return mw;
}

/***********************************************************************
  The following is used quite frequently in implementing widgets. Maybe
  I should put it in a function by itself. The _print function isn't
  always sufficient. Sometimes I need to fit some text into a given
  area. The first argument will determine how the given text will be
  positioned in the horizontal and the second will do the vertical axis.
    - B means, align it towards the beginning (top or left)
    - C means, center it
    - E means, align it towards the end (right or bottom)
  In case the text doesn't fit, it should be cut somehow. The following
  letters control that.
    - < cuts from the beginning (top or left of the text)
    - > cuts from the end
    - . cuts from both sides equally
  So, a specification which cuts from both sides and centers in both
  directions would look like:
 
     "C.","C."
 ***********************************************************************/

static void txt_print_fit(const char *horz,const char *vert,
                   int x,int y, int w, int h,
                   int font, int alpha, int fg, int bg,
                   const char *str)
{
  int tw,th; unsigned char *data;
  int src_x, src_y, dst_x, dst_y, cp_w, cp_h;
  int flags;

  get_text_data(font, str, &tw, &th, NULL, &data);

  if (!horz) horz= "C.";
  if (!vert) vert= "C.";

  if (tw>w)
  {
    if (strchr(horz,'<'))       src_x= tw-w;
    else if (strchr(horz,'>'))  src_x= 0;
    else                        src_x= (tw-w)/2; 
    cp_w= w;
    dst_x= x;
  }
  else
  {
    src_x= 0;
    cp_w= tw;
    if (strchr(horz,'B'))       dst_x= x;
    else if (strchr(horz,'E'))  dst_x= x+w-tw;
    else                        dst_x= x+(w-tw)/2;
  }

  if (th>h)
  {
    if (strchr(vert,'<'))       src_y= th-h;
    else if (strchr(vert,'>'))  src_y= 0;
    else                        src_y= (th-h)/2;
    cp_h= h;
    dst_y= y;
  }
  else
  {
    src_y= 0;
    cp_h= th;
    if (strchr(vert,'B'))       dst_y= y;
    else if (strchr(vert,'E'))  dst_y= y+h-th;
    else                        dst_y= y+(h-th)/2;
  }

  if (alpha) flags= GFX_ALPHA;
        else flags= 0;
  blend_graymap(flags, fg, bg, data, tw,th, src_x,src_y, dst_x,dst_y,cp_w,cp_h);
  if (data) free(data);
}

void gfx_text_data(int font, const char *str, 
                   int *Rw, int *Rh,int *Ra,unsigned char **Rd)
{
  get_text_data(font,str,Rw,Rh,Ra,Rd);
}
static void print_clipped(int Cx,int Cy,int Cw,int Ch,
                          int x, int y, int font, const char* str)
{
  unsigned char *tdata;
  int tw,th;

  int sx,sy, dx,dy, ow, oh;

  if (x>=Cx+Cw || y>=Cy+Ch || Cw<=0 || Ch<=0) return ;

  get_text_data(font, (char*) str,&tw,&th,NULL,&tdata);

  sx= 0; sy=0; dx= x; dy= y; ow= tw; oh= th;
  if (dx<Cx)
  {
    sx+= (Cx-dx);
    ow-= (Cx-dx);
    dx= Cx;
  }

  if (dy<Cy)
  {
    sy+= (Cy-dy);
    oh-= (Cy-dy);
    dy= Cy;
  }

  if (ow<=0 || oh<=0) goto done;

  if (dx+ow>Cx+Cw)
  {
    ow= (Cx+Cw-dx);
  }
  if (dy+oh>Cy+Ch)
  {
    oh= (Cy+Ch-dy);
  }
  if (ow<=0 || oh<=0) goto done;

  blend_graymap(0, 0xff000000, 0xffFFFFFF, tdata, tw,th, sx, sy, dx,dy,ow,oh);
done:
  free(tdata);
}

static void gui_draw_icon_and_text
                 (int x,int y,int w,int h,
                  int iconpos, gui_icon_t *icon,
                  int space,
                  uint32_t text_fg, char *text)
{
  int iw,ih,tw,th,ix,iy,tx,ty,total_w,total_h;
  total_w= total_h= 0; // warnUNINVAR
  if (!icon) 
  {
    if (text) goto text_only;
    return ;
  }
  if (!text)
  {
     gui_draw_icon_centered(x,y,w,h,icon);
     return ;
  }

  iw= icon->w;
  ih= icon->h;
  get_text_data(1, text, &tw, &th,NULL,NULL);

  switch(iconpos)
  {
  case 'L': case 'R':
     total_w= iw+tw; 
     if (total_w>w) goto text_only;
     
     if (w-total_w<space) space= w-total_w;
     total_w+= space;
     break;
  case 'T': 
  case 'B': 
  default:
     total_h= ih+th;
     if (total_h>h) goto text_only;
     if (h-total_h<space) space= h-total_h;
     total_h+= space;
     break;
  }
  switch(iconpos)
  {
   case 'L':
     ix= (w - total_w)/2;
     tx= ix+iw+space;
     goto horz_render;
   case 'R':
     tx= (w - total_w)/2;
     ix= tx+tw+space;
     goto horz_render;
   case 'T':
     iy= (h-total_h)/2;
     ty= iy+ih+space;
     break;
   default:
   case 'B':
     iy= (h-total_h)/2;
     ty= iy+ih+space;
     break;
  }
  gui_draw_icon_centered(x, y+iy, w, ih, icon);
  txt_print_fit("C.", "L>", x, y+ty, w, th, 1,1, text_fg,0xFFffFF, text);
  return ;

text_only:
  txt_print_fit("C.","C.",x,y,w,h,1,1,text_fg, 0xffffff, text);
  return ;
horz_render:
  gui_draw_icon_centered(x+ix, y, iw, h, icon);
  txt_print_fit("L>", "C.", x+tx, y, tw, h, 1,1,text_fg,0xffFFff,text);
  return ;
}


/***********************************************************************
  Following functions convert icons to server format. These are used
  mainly for resources.
 ***********************************************************************/

static unsigned int* R_GRAYIMG_toserver
        (unsigned int fg, unsigned int bg, R_GRAYIMG *img)
{
  unsigned int *buffer;
  int i;
  int w,h;
  w= img->w;
  h= img->h;
  buffer= malloc(sizeof(*buffer)*(w*h+2));
  for(i=0;i<w*h;i++) buffer[i+2]= bg;

  // FIXME: there is something fishy with this thing..
  // The buffer is now totally in user format..
  blend_graymap_buffer(fg, img->data, w,h, 0,0, buffer+2, w,h, 0,0, w,h);
  buffer[0]= w;
  buffer[1]= h;
  return buffer;
}

static unsigned int* gui_R_COLORIMG_toicon(R_COLORIMG *img)
{
  unsigned int *buffer;
  int i;
  int w,h;
  w= img->w;
  h= img->h;
  buffer= malloc(sizeof(*buffer)*(w*h+2));
  for(i=0;i<w*h;i++) buffer[i+2]= img->data[i];
  buffer[0]= w;
  buffer[1]= h;
  return buffer;
}

static unsigned int* gui_R_COLORIMG_toserver(R_COLORIMG *img)
{
  unsigned int *buffer;
  int i;
  int w,h;
  w= img->w;
  h= img->h;
  buffer= malloc(sizeof(*buffer)*(w*h+2));
  for(i=0;i<w*h;i++) buffer[i+2]= color_u_to_s(img->data[i]);
  buffer[0]= w;
  buffer[1]= h;
  return buffer;
}

static void gui_R_COLORIMG_getmask(R_COLORIMG *img, uint8_t **Rmask)
{
  unsigned int W,H,sz,i;
  uint8_t *m;
  if (*Rmask) return ;
  W= img->w;
  H= img->h;
  sz= W*H;
  *Rmask= m= malloc(sz);
  for(i=0;i<sz;i++) m[i]= img->data[i]>>24;
}

static unsigned int *gui_rotate_icon_cw(uint32_t *K)
{
  int w,h,x,y;
  unsigned int *R;
  w= K[0];
  h= K[1];
  K+= 2;

  R= malloc(sizeof(*R)*(w*h+2));
  R[0]= h;
  R[1]= w;
  R+= 2;

  for(y=0;y<h;y++) for(x=0;x<w;x++)
    R[ (h-1-y)+ (x)*h ]= K[x+y*w];

  return R-2;
}


static uint32_t *get_server_pix(int src_x,int src_y,int w,int h)
{
  uint32_t *D; int dw;
  int x,y;
  D= calloc(1,sizeof(uint32_t*)*w*h);
  dw= w;
  clip_for_rect(&src_x, &w, x_window_width);
  clip_for_rect(&src_y, &h, x_window_height);

  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      D[x+y*dw]= x_image_data[y+src_y][x+src_x];
  return D;
}

static void put_server_pix(uint32_t *D,int dst_x,int dst_y,int w, int h)
{
  int dw;
  int x,y;
  dw= w;
  clip_for_rect(&dst_x, &w, x_window_width);
  clip_for_rect(&dst_y, &h, x_window_height);

  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      x_image_data[y+dst_y][x+dst_x]= D[x+y*dw];
}

static void put_server_pix_alphamask
   (uint32_t *D,uint8_t *mask,int dst_x,int dst_y,int w, int h)
{
  int dw;
  int x,y;

  dw= w;
  clip_for_rect(&dst_x, &w, x_window_width);
  clip_for_rect(&dst_y, &h, x_window_height);

  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
      if (mask[x+y*dw])
         x_image_data[y+dst_y][x+dst_x]= D[x+y*dw];
}



/***********************************************************************
 *                                                                     *
 *                        New Graphics Methods                         *
 *                                                                     *
 ***********************************************************************
 Instead of passing around dozens of arguments, I'm going to combine them
 into structs. These are called graypix_t and colorpix_t. These contain
 image data as well as dimensions. For example, when you want to draw
 something for a widget, you make a colorpix out of the screen image 
 data, where the data field points to the first pixel of the widget and
 the stride is set to the screen width. This way, you can easily draw
 things in it without carrying the information around. This can also
 be used to draw things in buffers, in case you're rendering things ahead
 of time.
 ***********************************************************************/

static graypix_t graypix_sub(graypix_t src, int ofx, int ofy)
{
  if (ofx>=src.w)  src.w= 0; 
  else if (ofx>0)  { src.w-= ofx; src.data+= ofx; }

  if (ofy>=src.h)  src.h= 0; 
  else if (ofy>0)  { src.h-= ofy; src.data+= ofy*src.stride; }

  return src; 
}

static colorpix_t pix_sub(colorpix_t src, int ofx, int ofy)
{
  if (ofx>=src.w)  src.w= 0; 
  else if (ofx>0)  { src.w-= ofx; src.data+= ofx; }

  if (ofy>=src.h)  src.h= 0; 
  else if (ofy>0)  { src.h-= ofy; src.data+= ofy*src.stride; }

  return src; 
}

static void blend_graypix_bg
   (int fg, int bg, graypix_t src, 
    colorpix_t dst, int posx, int posy)
{
  int x,y;

  clip_for_rect( &posx, &src.w, dst.w);
  clip_for_rect( &posy, &src.h, dst.h);
  if (src.w<=0 || src.h<=0) return ;

  for(y=0;y<src.h;y++)
    for(x=0;x<src.w;x++)   
      PIX(dst,x+posx,y+posy)= blend_uuS(PIX(src,x,y),fg, bg);
}

static void blend_graypix
   (int fg, graypix_t src, 
    colorpix_t dst, int posx, int posy)
{
  int x,y;

  clip_for_rect( &posx, &src.w, dst.w);
  clip_for_rect( &posy, &src.h, dst.h);
  if (src.w<=0 || src.h<=0) return ;

  for(y=0;y<src.h;y++)
    for(x=0;x<src.w;x++) 
      PIX(dst,x+posx,y+posy)= blend_usS(PIX(src,x,y),fg, 
                                        PIX(dst,x+posx,y+posy));
}

static void colorpix_fill_rect
       (colorpix_t dst, int x,int y, int w, int h, int color)
{
  int xx,yy;
  clip_for_rect(&x, &w, dst.w);
  clip_for_rect(&y, &h, dst.h);
  if (w<=0 || h<=0) return ;
  color= color_u_to_s(color);
  for(yy=0;yy<h;yy++)
   for(xx=0;xx<w;xx++)
     PIX(dst, x+xx,y+yy)= color;
}

static void colorpix_hline(colorpix_t *P,int x1,int x2, int y,uint32_t color)
{
  if (x1<0) x1= 0; 
  if (x2>=P->w) x2= P->w-1;
  if (y<0 || y>=P->h || x1>x2) return ;
  color= color_u_to_s(color);
  for(;x1<=x2;x1++) PIX(*P,x1,y)= color;
}

static void colorpix_vline(colorpix_t *P,int x,int y1, int y2,uint32_t color)
{
  if (y1<0) y1= 0;
  if (y2>=P->h) y2= P->h-1;
  if (x<0 || x>=P->w || y1>y2) return ;
  color= color_u_to_s(color);
  for(;y1<=y2;y1++) PIX(*P,x,y1)= color;
}

static void colorpix_gradfill
   (colorpix_t *P,int x0,int y0, int w, int h, uint32_t start, uint32_t end)
{
  int D, dx, dy;
  D= (w+h);
  if (!D) D= 1;

  if (clip_for_rect(&x0, &w, P->w) || clip_for_rect(&y0, &h, P->h)) return ;

  for(dy=0;dy<h;dy++)
    for(dx=0;dx<w;dx++)
      PIX(*P,dx+x0,dy+y0)= blend_FuuS(start, 1.0-(double)(dx+dy)/D, end);
}

static void gui_draw_3dbox(colorpix_t *P, R_3DBOX *B)
{
  int sX,sY,eX,eY;

  sX= 0; eX= P->w-1;
  sY= 0; eY= P->h-1;
                                
  colorpix_hline(P,sX, eX, sY,     B->border);
  colorpix_hline(P,sX, eX, eY,     B->border);
  colorpix_vline(P,sX, sY+1, eY-1, B->border);
  colorpix_vline(P,eX, sY+1, eY-1, B->border);

  colorpix_hline(P,sX+1, eX-2, sY+1, B->light);
  colorpix_vline(P,sX+1, sY+1, eY-2, B->light);
  colorpix_hline(P,sX+1, eX-1, eY-1, B->dark);
  colorpix_vline(P,eX-1, sY+1, eY-1, B->dark);

  colorpix_gradfill(P,sX+2, sY+2, P->w-4, P->h-4, B->start, B->end);   
}

static void colorpix_put_graypix_centered
   (colorpix_t *P, graypix_t *G, int fg)
{
  int sx, sy, dx, dy, cw, ch, cx, cy;
  int cc;
  cc= (G->w-P->w)/2;
  if (cc>=0) { sx= cc; dx= 0; cw= P->w; }
        else { sx= 0; dx= -cc; cw= G->w; }
  cc= (G->h-P->h)/2;
  if (cc>=0) { sy= cc; dy= 0; ch= P->h; }
        else { sy= 0; dy= -cc; ch= G->h; }

  for(cy=0;cy<ch;cy++)
  for(cx=0;cx<cw;cx++)
    PIX(*P, cx+dx, cy+dy)= blend_ssS(255-PIX(*G,cx+sx,cy+sy),
                                     fg, PIX(*P,cx+dx,cy+dy)); 
}

static void put_pix_in_area
    (colorpix_t pix, int x, int y, int Ax, int Ay, int Aw, int Ah)
{
  int ux,uy;
  if (clip_for_rect(&Ax, &Aw, x_window_width)) return ;
  if (clip_for_rect(&Ay, &Ah, x_window_height)) return ;

  if (x<Ax) { pix= pix_sub(pix, Ax-x, 0); x= Ax; }
  if (y<Ay) { pix= pix_sub(pix, 0, Ay-y); y= Ay; }
  if (pix.w<=0 || pix.h<=0) return ;
  
  if (x+pix.w>Ax+Aw) { pix.w= Ax+Aw-x; }
  if (y+pix.h>Ay+Ah) { pix.h= Ay+Ah-y; }
  
  if (pix.w<=0 || pix.h<=0) return ;

  for(uy=0;uy<pix.h;uy++)
    for(ux=0;ux<pix.w;ux++)
     x_image_data[uy+y][ux+x]= pix.data[ux+uy*pix.stride];
}

static void pix_of_widget(widget_t *W, colorpix_t *C)
{
//
// FIXME: we should return the intersection of the widget with
// the screen, for cases where the widget has some corners
// outside of the screen.
//
  static uint32_t dummy;
  int x1,y1,x2,y2;

  x1= W->x; x2= W->x + W->w-1;
  y1= W->y; y2= W->y + W->h-1;

  x1= imax2(x1, 0);
  y1= imax2(y1, 0);
  x2= imin2(x2, x_window_width-1);
  y2= imin2(y2, x_window_height-1);

  if (x2<x1 || y2<y1) goto no_intersection;

  C->w= x2-x1+1;
  C->h= y2-y1+1;
  C->stride= x_window_width;
  C->data= x_image_data[y1]+x1;
  return ;

no_intersection:
  C->w= 0;
  C->h= 0;
  C->stride= 0;
  C->data= &dummy;
}

static void gui_blend_user_to_server_array
  (uint32_t uBG, uint32_t uFG, uint32_t *sOut)
{
  int i;
  for(i=0;i<256;i++)
    sOut[i]= blend_uuS(i, uFG, uBG); // the parameter order is reverse.. sigh.
}

static void gui_blend_user_to_user_array
  (uint32_t uBG, uint32_t uFG, uint32_t *sOut)
{
  int i;
  for(i=0;i<256;i++)
    sOut[i]= blend_uuU(i, uBG, uFG);
}

int gfx_color_s(int color)
{
  return color_u_to_s(color);
}
