
/***********************************************************************
 *                                                                     *
 *                           Screen Handling                           *
 *                                                                     *
 ***********************************************************************
 A screen stores the contents of the previous screen. When the current
 screen is left, those contents are restored.
 ***********************************************************************/

static void screen_store_contents()
{
  if (!screen) return ;

  screen->backstore= get_server_pix(0,0,x_window_width,x_window_height);
  screen->back_w= x_window_width;
  screen->back_h= x_window_height;
}

static void screen_restore_contents(screen_t *S)
{
  if (!S || !S->backstore) return ;
  put_server_pix(S->backstore, 0,0,S->back_w,S->back_h);
  gfx_update(0,0,x_window_width,x_window_height);
  free(S->backstore);
  S->backstore= NULL;
}

/***********************************************************************
 When a screen is created, we also create the necessary widgets. This is
 done in the screen function. The following function is used for entering
 a screen.
 ***********************************************************************/
static void enter_screen
   (int n_widgets, int *widgets,int *co)
{
  screen_t *S;

  screen_op_count++;

  if (screen)
  {
    widget_t *W;
    W= screen->focus_widget;
    if (W && W->focus && W->has_focus)
    {
      (*W->focus)(W,0);
      W->has_focus= 0; 
    }
  }
  gfx_hide_virtual_keyboard();

  screen_store_contents();

  S= calloc(1,sizeof(*S));
  S->down= screen;
  if (screen) screen->up= S;
  screen= S;

  if (co[0]<0) co[0]= 0;
  if (co[1]<0) co[1]= 0;
  if (co[2]<=0) co[2]= x_window_width;
  if (co[3]<=0) co[3]= x_window_height;

  S->x= co[0]; S->y= co[1];
  S->w= co[2]; S->h= co[3]; 

  if (S->x>=x_window_width) S->x= 0;
  if (S->y>=x_window_height) S->y= 0;
  if (S->x+S->w>x_window_width) S->w= x_window_width-S->x;
  if (S->y+S->h>x_window_height) S->h= x_window_height-S->y;

  S->n_widgets= n_widgets;
  S->widgets= widgets;

  gfx_fill_rect(0,S->x,S->y,S->w,S->h,0xededed);
  if (S->x>0 || S->y>0)
    gfx_draw_rect(0, S->x-1,S->y-1,S->w+2,S->h+2, 0x00);
}

static void gui_leave_screen()
{
  screen_t *S;
  int i;

  screen_op_count++;

  if (!screen) return ;
  gfx_hide_virtual_keyboard();

  S= screen;

  for(i=0;i<S->n_widgets;i++)
    if (widgets[S->widgets[i]])
    {
      widget_destroy(widgets[S->widgets[i]]);
      widgets[S->widgets[i]]= NULL;
    }
  if (S->click_handlers) free(S->click_handlers);
  if (S->backstore)      free(S->backstore);
  free_rgroups(S);

  screen= S->down;
  free(S);
  if (screen) screen->up= NULL;

  screen_restore_contents(screen);

  if (screen)
  {
    widget_t *W;
    W= screen->focus_widget;
    if (W && W->focus)
    {
      (*W->focus)(W,1); 
      W->has_focus= 1;
    }
  }
}


/***********************************************************************
 *                                                                     *
 *                            Mouse Events                             *
 *                                                                     *
 ***********************************************************************
  There are three categories of mouse events: 
   - click/release
   - motion
   - grabbed events

  When a widget is created, it registers areas of interest as click 
  handlers. When the click is eventually done, the widget can register
  a release handler for that click. The release event than calls this
  release handler and the release handler is removed from the screen.

  For motion events, a motion handler can be installed/removed. 

  You can also grab the mouse using the grab_mouse() function. This
  function installs an event handler which handles all mouse input
  until ungrab_mouse() is called.
 ***********************************************************************/

static void set_click_handler(int click,int x,int y,int w,int h)
{
  click_handler_t *C;
  if (!screen) return ;
  if (click<0 || click>=screen->n_click_handlers) return ;
  C= screen->click_handlers+click;
  C->x= x;
  C->y= y;
  C->w= w;
  C->h= h;
}

static int install_click_handler
        (int x,int y,int w,int h,int *enabled,
         void *arg, int (*func)(void*,gfx_event_t*),int button)
{
  click_handler_t *C;
  int R;
  if (!screen) return 0;

  if (screen->n_click_handlers==screen->s_click_handlers)
  {
    int delta;
    if (screen->s_click_handlers) delta= 1+screen->s_click_handlers/2;
                             else delta= 1+screen->n_widgets;
    screen->s_click_handlers+= delta;
    screen->click_handlers= realloc(screen->click_handlers, 
                                    sizeof(click_handler_t)*
                                       screen->s_click_handlers);
  }
  R= screen->n_click_handlers;
  screen->n_click_handlers++;
  C= screen->click_handlers + R;
  C->x= x;
  C->y= y;
  C->w= w;
  C->h= h;
  C->enabled= enabled;
  C->arg= arg;
  C->func= func;
  C->button= button;
  return R;
}

static int install_or_modify_click
  (int ch, int *enabled, int (*func)(void*,gfx_event_t*), void* arg,
   int x,int y,int w,int h,int button)
{
  if (ch==-1)
   return install_click_handler(x,y,w,h,enabled,arg, func,button);

  set_click_handler(ch, x,y,w,h);
  return ch;
}

static void remove_click_handler(int H)
{
  int i;
  if (!screen) return;
  if (H<0 || H>=screen->n_click_handlers) return ;
  screen->click_handlers[H].func= NULL;
  for(i=H;i<screen->n_click_handlers;i++)
    if (screen->click_handlers[i].func!=NULL) return ;
  screen->n_click_handlers= H;
}

static void install_release_handler
        (void *arg, int (*func)(void*,gfx_event_t*))
{
  if (!screen) return ;
  screen->release_handler.func= func;
  screen->release_handler.arg= arg;
}

static void install_motion_handler(void *arg,int (*func)(void *,gfx_event_t*))
{
  if (!screen) return ;
  screen->motion_handler.func= func;
  screen->motion_handler.arg= arg;
}

static void uninstall_motion_handler()
{
  if (!screen) return ;
  screen->motion_handler.func= NULL;
  screen->motion_handler.arg= NULL;
}

/***********************************************************************
  Sometimes, it's necessary to grab all mouse events. For example, the
  dropdown widget does this. If the user clicks outside the dropdown list,
  the list gets hidden.
 ***********************************************************************/
static void grab_mouse(int (*func)(void*,gfx_event_t*), void *arg)
{
  if (!screen) return ;
  screen->grabber.func= func;
  screen->grabber.arg= arg;
}

static void ungrab_mouse()
{
  if (!screen) return ;
  screen->grabber.func= NULL;
  screen->grabber.arg= NULL;
}

/***********************************************************************
                        Mouse  Event Handlers
 ***********************************************************************
  The handler functions will return
   0: if the event wasn't handled or, it was handled and a new event
      was generated in its place
   1: if the event was handled and the processing should stop.
 ***********************************************************************/
static int handle_click_event(gfx_event_t *event)
{
  int i;
  click_handler_t *C;

  if (!screen || event->type!=GFXEV_BUTTON_PRESS) return 0;
  for(i=0;i<screen->n_click_handlers;i++)
  {
    C= screen->click_handlers+i;
    if (C->button!=event->button) continue;
    if (range_test(event->x,C->x, C->w) && range_test(event->y,C->y, C->h))
    {
             // if the widget was disabled, look for something else
      if ((!C->enabled || C->enabled[0])&& C->func)
           return (*C->func)(C->arg, event);
    }
  }
  return 0;
}

static int handle_release_event(gfx_event_t *event)
{
  void *A; int (*F)(void *,gfx_event_t*);
  if (!screen || event->type!=GFXEV_BUTTON_RELEASE || event->button!=1)
      return 0;
  if (screen->release_handler.func)
  {
    A= screen->release_handler.arg;
    F= screen->release_handler.func;
    screen->release_handler.arg= NULL;
    screen->release_handler.func= NULL;
    return (*F)(A, event);
  }
  return 0;
}

static int handle_motion_event(gfx_event_t *event)
{
  if (!screen || event->type!=GFXEV_POINTER_MOVE) return 0;
  if (screen->motion_handler.func)
     return (*screen->motion_handler.func)(screen->motion_handler.arg,event);
  return 0;
}

static int handle_grabbed_event(gfx_event_t *event)
{
  if (!screen || !screen->grabber.func) return 0;
  return (*screen->grabber.func)(screen->grabber.arg, event);
}

/***********************************************************************
 *                           Focus Handling                            *
 ***********************************************************************
  When there is a widget which can take keyboard input, we need to
  route key press/release events to that widget. This happens only if
  the widget has focus.

  In order to give focus to a widget, the following can be done:
    - By creating a new widget with focus flag set.  - press Tab to move
    the cursor into it - click in the widget area

  So, each widget object will have three fields regarding the focus
  system:
    - can_focus - has_focus - focus()
  When a focus change is requested, can_focus and has_focus fields will
  go thru the widget list and handle the event. The current focussed
  object will handle key events thru the handle_key function.
 ***********************************************************************/

/***********************************************************************
 The following is used when a widget is created and wants keyboard focus.
 ***********************************************************************/
static void set_focus(widget_t *N)
{
  widget_t *W;
  if (!screen || !N) return ;
  if (!N->can_focus || N->has_focus) return ;

  W= screen->focus_widget;
  if (W && W->can_focus && W->has_focus)
  {
    if (W->focus) (*W->focus)(W, 0);
    W->has_focus= 0;
  }

  if (N->focus) (*N->focus)(N,1);
  N->has_focus= 1;
  screen->focus_widget= N;
}

static void remove_focus()
{
  widget_t *W;
  if (!screen) return ;

  W= screen->focus_widget;
  if (W && W->can_focus && W->has_focus)
  {
    if (W->focus) (*W->focus)(W, 0);
    W->has_focus= 0;
  }
  screen->focus_widget= NULL;
}

 // returns zero if successful
static int make_focus_event(gfx_event_t *E)
{
  widget_t *W;
  if (!screen) return 1;
  W= screen->focus_widget;
  if (!W) return 1;
  if (W->number<GFXEV_LAST+GFXRESERVED) return 1;
  E->type= GFXEV(W->number,FOCUS);
  return 0;
}

static int tabs_before(widget_t *A,widget_t *B)
{
  if (A->y<B->y) return 1;
  if (A->y==B->y && A->x<B->x) return 1;
  return 0;
}

/***********************************************************************
The following function returns 1 if the event was consumed by the focus system.
 ***********************************************************************/
static int focusable_event(gfx_event_t *event)
{
  int i;
  widget_t *W;
  W= NULL; // warnUNINVAR

  if (!screen) return 0; 
  if (event->type==GFXEV_BUTTON_PRESS) 
  { 
    // FIXME: check for modifiers
    if (event->button!=1) return 0;
    for(i=0;i<screen->n_widgets;i++)
    {
      W= widgets[screen->widgets[i]];
      if (!W) continue;
      if (W->can_focus &&
          range_test(event->x, W->x, W->w) &&
          range_test(event->y, W->y, W->h)) break;
    }
    if (i==screen->n_widgets || W==screen->focus_widget) return 0;
    screen->focus_clicked= W;
    return 1;
  }
  if (event->type==GFXEV_BUTTON_RELEASE)
  {
    // FIXME: check for modifiers
    if (event->button!=1 || !screen->focus_clicked) return 0;
    set_focus(screen->focus_clicked);
    screen->focus_clicked= NULL;
    return make_focus_event(event);
  }
  if (event->type==GFXEV_KEY_PRESS)
  {
    widget_t *target, *minpos,*current;

    W= screen->focus_widget;
    if (!W) return 0;
    if (W->handle_key && (*W->handle_key)(W, event)) return 1;

    // FIXME: handle shift-tab
    if (event->key!=GFX_K_TAB)  return 0;
    
    for(i=0;i<screen->n_widgets;i++)
      if (W->number==screen->widgets[i]) break;

    if (i==screen->n_widgets) return 0;
    target= NULL;
    minpos= NULL;
    current= widgets[screen->widgets[i]];
  
    for(i=0;i<screen->n_widgets;i++)
    {
      W= widgets[screen->widgets[i]];
      if (W==current || !W || !W->can_focus) continue;

                   // FIXME: check for visibility also..
      if (!minpos) minpos= W;
      else if (tabs_before(W,minpos)) minpos= W; 

      if (tabs_before(W,current)) continue;
      if (target && tabs_before(target,W)) continue;
      target= W;
    }
    if (!target) target= minpos;
    if (target)
    {
      set_focus(target);
      return make_focus_event(event);
    }
  }
  return 0;
}


/***********************************************************************
  This function should be called when a widget with focus is deleted.
 ***********************************************************************/
static void focus_lost()
{
  if (!screen) return ;
  screen->focus_widget= NULL;
}

void gfx_focus(int wno)
{
  widget_t *W;
  W= get_widget(wno);
  if (!W) remove_focus();
     else set_focus(W);
}

/***********************************************************************
  Sometimes, it's necessary to update a widget while it's in the
  background.  i.e. a new screen has been entered but the widget in the
  old screen is being updated. This is useful for a log viewer widget,
  for example.

  In order to handle that situation, I need to make sure only the visible
  parts of the widget is updated. For this, I should compute a visible
  parts list for the screen and update the display according to that.

  The array is algorithmically quite expensive at O(n*n). It shouldn't
  be a problem because we aren't going to have thousands of screens
  opened on top of each other. However, I'd still like to defer doing
  this until necessary. So, this computation is done only if requested
  (ie. thru the logview widget etc).
 ***********************************************************************/

static void get_screen_rect(screen_t *S, rect_t *A)
{
  A->top= S->y; A->bottom= S->y+S->h-1;
  A->left= S->x; A->right= S->x+S->w-1;
}

static void screen_compute_visible_parts(screen_t *S)
{
  rect_t A, B, D[4];
  rectque_t *Q;
  int i,N,L;
  screen_t *K;

  if (S->visible_parts_time==screen_op_count) return ;
  S->visible_parts_time= screen_op_count;

  if (S->visible_parts) free(S->visible_parts);

  Q= rectque_new();
  get_screen_rect(S, &A);
  rectque_put(Q, &A);
  
  for(K=S->up;K;K=K->up)
  {
    get_screen_rect(K,&B);
    L= rectque_len(Q);
    for(i=0;i<L;i++)
    {
      rectque_get(Q, &A);
      N= subtract_rectangle(&A, &B, D);
      rectque_putn(Q, D, N);
    }
  }

  rectque_get_data(Q, &S->visible_parts, &S->n_visible_parts);
  rectque_free(Q);
}

/***********************************************************************
  Now, it's time to use this thing. A widget will get the pixels
  associated with the screen using the get_screen_pixels() function. After
  making modifications on the data, update_screen_pixels() will be
  called. This function will update the display according to the computed
  list of visible parts.
 ***********************************************************************/

static colorpix_t get_screen_pix(screen_t *S)
{
  colorpix_t P;
  if (S->up)
  {
    P.w= S->back_w;
    P.h= S->back_h;
    P.stride= S->back_w;
    P.data= S->backstore;
  }
  else
  {
    P.w= x_window_width;
    P.h= x_window_height;
    P.stride= x_window_width;
    P.data= x_image_data[0];
  }
  return P;
}

static void screen_update_display(screen_t *S, int x,int y,int w, int h)
{
  int xx,yy;
  clip_for_rect(&x,&w, x_window_width);
  clip_for_rect(&y,&h, x_window_height);
  if (w<=0 || h<=0) return ;
  for(yy=y;yy<y+h;yy++)
    for(xx=x;xx<x+w;xx++)
      x_image_data[yy][xx]= S->backstore[xx+yy*S->back_w];
  gfx_update(x,y,w,h);
}


static void update_screen_pixels(screen_t *S, int x,int y, int w, int h)
{
  rect_t K, E;
  int i;
  if (!S->up)
  {
    gfx_update(x,y,w,h);
    return ;
  }
  screen_compute_visible_parts(S);
  E.top= y; E.bottom= y+h-1;
  E.left= x; E.right= x+w-1;
  for(i=0;i<S->n_visible_parts;i++)
  {
    if (rectangle_intersect(S->visible_parts+i, &E, &K))
      screen_update_display(S, K.left, K.top, 
                                K.right-K.left+1, K.bottom-K.top+1);
  }
}


/***
  It's possible to inject events into the queue. This is useful for
  implementing things like cut-paste easily.
 ***/
static struct
{
   int n, h, s;
   gfx_event_t *data;
} gui_inj;

static void gui_alloc_inj(int sz)
{
  gfx_event_t *J;
  int i;
  if (!sz) sz= 128;
  J= malloc(sz*sizeof(gfx_event_t));
  for(i=0;i<gui_inj.n;i++)    J[i]= gui_inj.data[(gui_inj.h+i)%gui_inj.s];
  if (gui_inj.data) free(gui_inj.data);
  gui_inj.data= J;
  gui_inj.h= 0;
  gui_inj.s= sz;
}

static void gui_inject(gfx_event_t *ev)
{
  if (gui_inj.n==gui_inj.s) gui_alloc_inj(gui_inj.s*1.5);
  gui_inj.data[(gui_inj.h+gui_inj.n)%gui_inj.s]= *ev;
  gui_inj.n++;
}

static int gui_deinject(gfx_event_t *Rev)
{
  if (!gui_inj.n) return 0;
  *Rev= gui_inj.data[gui_inj.h];
  gui_inj.h= (gui_inj.h+1)%gui_inj.s;
  gui_inj.n--;
  return 1;
}


/***********************************************************************
   The return value is one if an event is returned. It's zero in case
   of a failure. FIXME: make a failure event and return it anyway.
 ***********************************************************************/
int gfx_event(gfx_event_t *Revent)
{
  if (gui_deinject(Revent)) return 1;
again:
  if (!get_event(Revent)) return 0;
  if (vkey_event(Revent)) goto again;
  switch(Revent->type)
  {
  case GFXEV_BUTTON_PRESS:
  case GFXEV_BUTTON_RELEASE:
  case GFXEV_POINTER_MOVE:
    if (handle_grabbed_event(Revent)) goto again; else break;
  }

  switch(Revent->type)
  {
  case GFXEV_BUTTON_PRESS:
     if (handle_click_event(Revent)) goto again; else break;
  case GFXEV_BUTTON_RELEASE:
     if (handle_release_event(Revent)) goto again; else break;
  case GFXEV_POINTER_MOVE:
     if (handle_motion_event(Revent)) goto again; else break;
  }

  if (focusable_event(Revent)) goto again;

  return 1; 
}

void gfx_event_relative(int wno, gfx_event_t* e)
{
  widget_t *w;
  w= get_widget(wno);
  if (!w) return ;
  e->x -= w->x;
  e->y -= w->y; 
}
