
/***********************************************************************
 *                                                                     *
 *                         BYTECODE INTERFACE                          *
 *                                                                     *
 ***********************************************************************
   In order for remote GUIs to work, I need to encode the GUI as data
   rather than code. This data will simply be bytecode.
   
   The execution environment will be a stack machine. Each instruction
   is one byte, with some optional data following it.
 ***********************************************************************/

static int gui_run_bytecode
  (const unsigned char *code, int length,int *locals,int *layout);

/** BEGIN: BYTECODE  DEFINITIONS **/
#define GUII_C4          0x01
#define GUII_C3          0x02
#define GUII_C2          0x03
#define GUII_C1          0x04
#define GUII_ADD         0x05
#define GUII_SUB         0x06
#define GUII_MUL         0x07
#define GUII_DIV         0x08
#define GUII_MOD         0x09
#define GUII_STOREINT    0x0A
#define GUII_LOADINT     0x0B
#define GUII_LOCAL       0x0C
#define GUII_WLAY_X      0x0D
#define GUII_WLAY_Y      0x0E
#define GUII_WLAY_W      0x0F
#define GUII_WLAY_H      0x10
#define GUII_CF          0x11
#define GUII_GLOBAL      0x12
#define GUII_STR         0x13
#define GUII_CONSTRUCT   0x14
#define GUII_INTARRAY    0x15    
#define GUII_IPARAM      0x16
#define GUII_DONE        0x17
#define GUII_NOP         0x18
#define GUII_PPARAM      0x19

#define GUII_SETG        0x20
/** This instruction has a suffix byte which identifies the top-level
    value to be set. It can set things like n_screens, window_width, etc.
    IPARAM etc will be erased. **/

#define GUIG_WINDOW_WIDTH   0
#define GUIG_WINDOW_HEIGHT  1
#define GUIG_WINDOW_TITLE   2
#define GUIG_NFONT          3
#define GUIG_NFONTFILE      4
#define GUIG_NICON          5
#define GUIG_NSCREEN        6
#define GUIG_FIRST_WIDGET   7
#define GUIG_LAST_WIDGET    8
#define GUIG_NGLOBAL        9  // number of global variables in guibc_global
                               // in addition to the default ones

#define GUII_FONTFILE       0x21
/**
  Takes three arguments from top to bottom:
   - index into the fontfiles array, 
   - size of the data
   - data, encoded as a blob
 **/
#define GUII_FONT           0x22
/**
  Takes three arguments from top to bottom:
   - index into the fontdesc array, 
   - font file index
   - size in points
 **/
#define GUII_ICON           0x23
/**
  Takes two arguments, from top to bottom:
   - icon number
   - number of images in it
 **/
#define GUII_ICONIMAGE      0x24
/**
  Arguments:
   - icon number
   - image number
   - dpi
   - size of data
   - data encoded as a blob
 **/

#define GUII_BLOB           0x25
/**
  This is essentially the same thing as a string, but the dissassembler
  doesn't try to print it.
 **/

#define GUII_SECTION        0x26
/**
  Defines a section of code. It's followed by one of the GUISEC identifiers
  and then the associated code itself.
  Arguments:
  - screen number
  - number of local variables
  - code size
 **/
#define GUISEC_SCREENDIM      0
#define GUISEC_CONSTRUCTOR    1
#define GUISEC_WIDGETLAYOUT   2

#define GUII_SCRINFO        0x27
/**
  Stores various information about a screen definition. 
  Arguments:
  - screen number
  - nwidgets
  - ndirectwidgets
  - widget ids [integer array]: a mapping from local widget numbers to
    global widget numbers. 
 **/
#define GUII_SUBLAYOUT      0x28
/**
  Lays out a subscreen, which is part of a panel widget.
  Arguments:
  - pindex : index of the parent object within the parent screen
  - start:   index of the first widget to be laid out in the parent screen
  - end:     one plus index of the last widget to be laid out in
             the parent screen
  - nlocals: number of local variables to be used in sub-computation
  - code_length: length of the code that follows this instruction
  The instruction is followed by layout code.
 **/
#define GUII_DUP             0x29
    // duplicates the element at the top of the stack
#define GUII_DROP            0x30
    // removes the element at the top of the stack

#define GUILV_WIDTH                   0
#define GUILV_HEIGHT                  1
#define n_predef_locals               2

#define GUIGV_FONT1HEIGHT             0
#define GUIGV_FONT2HEIGHT             1
#define GUIGV_FONT3HEIGHT             2
#define GUIGV_FONT4HEIGHT             3
#define GUIGV_FONT5HEIGHT             4
#define GUIGV_GUI_SCROLLBAR_WIDTH     5
#define GUIGV_GUI_SCROLLBAR_HEIGHT    6
#define GUIGV_GUI_SYSBAR_HEIGHT       7
#define GUIGV_GUI_NAVBAR_HEIGHT       8
// FIXME:  put radio width/height as well as checkbox width/height here.
#define n_global_vars                 9

#define GUII_GLOBALS_SECTION         0x31
/**
  This defines the globals section which initializes the elements of
  the guibc_global array. The section is stored in a global variable
  and is executed after a connection to the display. It has only one
  argument on the stack: the size of the code section.
 **/

#define GUII_LOADX                   0x32
//
// loads the integer at TOS into the X register
//
#define GUII_XADD                    0x33
//
// adds the value of X register to the integer pointed by AP.
//

//
// here is the numbering for the construction functions
//
enum { 
  GUIWC_BUTTON=1,     GUIWC_LABEL,              GUIWC_RADIO, 
  GUIWC_CHECKBOX,     GUIWC_PROGRESSBAR,        GUIWC_LIST, 
  GUIWC_PIXGRID,      GUIWC_TEXTENTRY,          GUIWC_PANEL,
  GUIWC_SPINNER,      GUIWC_DROPDOWN,           GUIWC_LOGVIEW,
  GUIWC_SLIDER,       GUIWC_CANVAS,             GUIWC_FILELIST,
  GUIWC_DIRLIST,      GUIWC_ICONLIST,           GUIWC_SCROLLBAR,
};


/** END: BYTECODE DEFINITIONS **/

typedef struct
{
  int type;
  union { uint32_t uint;
	  int32_t  sint;
	  float    flo;
	  void    *ptr;
	} d;
} gui_cell_t;


static int guis_tos, guis_size;
static gui_cell_t *guis;

static int guibc_global_default[n_global_vars];
static int guibc_nglobal= n_global_vars;
static int *guibc_global= guibc_global_default;
static const unsigned char *guibc_global_init_code;
static int guibc_global_init_size;

#define GUIC_INT  1
#define GUIC_UNT  2
#define GUIC_FLO  3
#define GUIC_PTR  4

static void init_guibc_global()
{
  guibc_global[GUIGV_FONT1HEIGHT]=  font1height;
  guibc_global[GUIGV_FONT2HEIGHT]=  font2height;
  guibc_global[GUIGV_FONT3HEIGHT]=  font3height;
  guibc_global[GUIGV_FONT4HEIGHT]=  font4height;
  guibc_global[GUIGV_FONT5HEIGHT]=  font5height;
  guibc_global[GUIGV_GUI_SCROLLBAR_WIDTH]= gui_scrollbar_width;
  guibc_global[GUIGV_GUI_SCROLLBAR_HEIGHT]= gui_scrollbar_height;
  guibc_global[GUIGV_GUI_SYSBAR_HEIGHT]= gui_sysbar_height;
  guibc_global[GUIGV_GUI_NAVBAR_HEIGHT]= gui_navbar_height;
  if (guibc_global_init_code)
     gui_run_bytecode(guibc_global_init_code, guibc_global_init_size,
                       NULL, NULL);
}

static int guis_push(gui_cell_t *K)
{
  if (guis_tos==guis_size) return 1;
  guis[guis_tos]= *K;
  guis_tos++;
  return 0;
}

static int guis_dup()
{
  if (guis_tos==guis_size || guis_tos==0) return 1;
  guis[guis_tos]= guis[guis_tos-1];
  guis_tos++;
  return 0;
}


static int guis_pop(gui_cell_t *K)
{
  if (guis_tos==0) return 1;
  guis_tos--;
  *K= guis[guis_tos];
  return 0;
}

static void guis_init()
{
  if (!guis)
  {
    guis_size= 1000;
    guis= malloc(sizeof(guis[0])*guis_size); 
  }
  guis_tos= 0;
}

static void guis_Pint(int32_t k)
{
  gui_cell_t K;
  K.type= GUIC_INT;
  K.d.sint= k;
  guis_push(&K);
}

static void guis_Pptr(void *ptr)
{
  gui_cell_t K;
  K.type= GUIC_PTR;
  K.d.ptr= ptr;
  guis_push(&K);
}

static void* guis_Gptr()
{
  gui_cell_t K;
  if (guis_pop(&K)) return NULL;
  if (K.type==GUIC_PTR) return K.d.ptr;
  return NULL;
}

static void guis_Pflo(float k)
{
  gui_cell_t K;
  K.type= GUIC_FLO;
  K.d.flo= k;
  guis_push(&K);
}

static int32_t guis_Gint()
{ 
  gui_cell_t K;
  if (guis_pop(&K)) return 0;
  if (K.type==GUIC_FLO) return K.d.flo;
  return K.d.sint;
}

static float guis_Gflo()
{ 
  gui_cell_t K;
  if (guis_pop(&K)) return 0;
  if (K.type==GUIC_FLO) return K.d.flo;
  return K.d.sint;
}

typedef struct 
{
  int code_length, nlocals;
  const unsigned char *code; 
} bytecode_t;

static int n_screen_defs;

typedef struct
{
  bytecode_t code[3];
  int nwidgets, ndirect;
  int *widget_ids;
} screen_def_t;

static screen_def_t *screen_defs;

static void construct_widget(int widget_no, int cons_func)
{
  widget_t *W;

  int i1, i2, i3, i4, i5, i6;
  float f1, f2, f3;
  char *s1;
  void *p1;
  
  W= NULL;
  switch(cons_func)
  {
  case GUIWC_DIRLIST: W=  gui_dirlist_new(NULL,NULL,NULL); break; 
  case GUIWC_FILELIST:
    i1= guis_Gint(); 
    i2= guis_Gint(); 
    W= gui_filelist_new(i1,i2,NULL,NULL,NULL);
    break;
  case GUIWC_BUTTON:
    s1= guis_Gptr();
    i1= guis_Gint(); 
    i2= guis_Gint(); 
    i3= guis_Gint();
    W= gfx_button(s1, i1, i2, i3);
    break;
  case GUIWC_LABEL:
    i1= guis_Gint();
    i2= guis_Gint();
    i3= guis_Gint();
    i4= guis_Gint();
    s1= guis_Gptr();
    W= gfx_label(i1,i2,i3,i4,s1);
    break;
  case GUIWC_RADIO:
    s1= guis_Gptr();
    i1= guis_Gint();
    W= radio_new(s1, i1);
    break;
  case GUIWC_CHECKBOX:
    s1= guis_Gptr();
    W= gui_checkbox_new(s1,NULL,NULL);
    break;
  case GUIWC_PROGRESSBAR:
    i1= guis_Gint();
    i2= guis_Gint();
    i3= guis_Gint();
    i4= guis_Gint();
    W= gui_progressbar_new(i1,i2,i3,i4);
    break;
  case GUIWC_LIST: W= gui_list_new(0,0,NULL,NULL,NULL); break;
  case GUIWC_PIXGRID: W= gfx_pixgrid(); break;
  case GUIWC_TEXTENTRY: 
    s1= guis_Gptr();
    W= text_entry_new(s1);
    break;
  case GUIWC_PANEL:
    p1= guis_Gptr();
    W= panel_new(p1);
    break;
  case GUIWC_SPINNER:
    i1= guis_Gint();
    i2= guis_Gint();
    W= gui_spinner(i1,i2);
    break;
  case GUIWC_DROPDOWN: W= dropdown_new(NULL,NULL); break;
  case GUIWC_LOGVIEW: W= gui_logview(); break;
  case GUIWC_SLIDER:
    i1= guis_Gint();
    W= gui_slider_new(i1);
    break;
  case GUIWC_CANVAS:
    i1= guis_Gint();
    i2= guis_Gint();
    W= gui_canvas_new(i1,i2);
    break;
  case GUIWC_ICONLIST:
    i1= guis_Gint(); i2= guis_Gint(); i3= guis_Gint();
    i4= guis_Gint(); i5= guis_Gint(); i6= guis_Gint();
    W= gui_iconlist_new(i1,i2,i3,i4,i5,i6);
    break;
  case GUIWC_SCROLLBAR:
    i1= guis_Gint(); f1= guis_Gflo();
    f2= guis_Gflo(); f3= guis_Gflo();
    W= scrollbar_new(i1,f1,f2,f3,0,NULL,NULL);
    break;
  }
  if (!W) return ;
  widgets[widget_no]= W;
  W->number= widget_no;
}

static void gui_setg(int sel)
{
  uint32_t iV;
  char *sV;
  sV= "";
  iV= 0;
  switch(sel)
  {
  case GUIG_WINDOW_TITLE: sV= guis_Gptr(); break;
  default:                iV= guis_Gint(); 
  }
  switch(sel)
  {
#ifndef GFX_ANDROID
  case GUIG_WINDOW_WIDTH: 
     x_window_width=  iV;
     break;
  case GUIG_WINDOW_HEIGHT:
     x_window_height=  iV;
     break;
  case GUIG_WINDOW_TITLE:  x_window_title= sV? sV: ""; break;
#endif
  case GUIG_NSCREEN:
    n_screen_defs= iV;
    screen_defs= calloc(n_screen_defs, sizeof(screen_defs[0]));
    break;
  case GUIG_FIRST_WIDGET:  first_widget_no= iV; break;
  case GUIG_LAST_WIDGET:   
    last_widget_no= iV; 
    widgets=calloc(last_widget_no+1, sizeof(widgets[0]));
    break;
  case GUIG_NICON:
    gui_nicons= iV;
    if (gui_nicons) gui_icons= calloc(gui_nicons, sizeof(gui_icons[0]));
    break;
  case GUIG_NFONT: 
    if (iV)
    {
      gui_nfonts= iV;
      gui_fontdefs= calloc(iV, sizeof(gui_fontdefs[0]));
      gui_fonts= calloc(iV, sizeof(gui_fonts[0]));
    }
    break;
  case GUIG_NFONTFILE: 
    gui_nfontfiles= iV;
    gui_fontfiles= calloc(iV, sizeof(gui_fontfiles[0]));
    gui_fontfiles[0].data= fontdata0;
    gui_fontfiles[0].size= fontdata0_size;
    gui_fontfiles[1].data= fontdata1;
    gui_fontfiles[1].size= fontdata1_size;
    break;
  case GUIG_NGLOBAL:
    if (iV>0)
    {
      guibc_nglobal= iV + n_global_vars;
      guibc_global= calloc(guibc_nglobal, sizeof(int));
    }
    break;
  }
}

static void gui_layout_widgets
  (int WIDTH, int HEIGHT, int nlocals, int length,
   const unsigned char *code, int *Wcoor);

#ifdef GFX_DEBUG_BYTECODE
#define BD(pd, ...) do { printf("DBG> %6d: ",pc+pd); printf(__VA_ARGS__); printf("\n"); } while(0)
#else
#define BD(...) do { } while(0)
#endif

static int gui_run_bytecode
  (const unsigned char *code, int length,int *locals,int *layout)
{
  int widget_no, cons_func;
  int pc;
  gui_cell_t A1, A2;
  unsigned char C;
  gui_icon_t *IC;
  gui_iconimage_t *IM;
  bytecode_t *BC;
  char *Z;
  int *AP;
  int i1, i2;
  double f1, f2;
  int iV, *iA, i, iS;
  void *pV;
  int pindex, start, end, nlocals, code_size;
  screen_def_t *SD;
  int regX;

  guis_init();
  pc= 0;

next:
  if (pc>=length) return length;
  C= code[pc++];
  if (C&0x80)
  {
    BD(-1, "(push-int) %d", C&0x7f);
    guis_Pint(C&0x7f);
    goto next;
  }
  iV= 0;
  switch(C)
  {
  case GUII_CF:  
  case GUII_C4: iV= code[pc++];
  case GUII_C3: iV= (iV<<8) | code[pc++];
  case GUII_C2: iV= (iV<<8) | code[pc++];
  case GUII_C1: iV= (iV<<8) | code[pc++]; 
     if (C==GUII_CF)
     {
        BD(-5, "(push-float) %f", (double) iV/65536);
        guis_Pflo( (double)iV / 65536); 
     }
     else
     {
        BD(C-4-2, "(push-int) %d", iV);
        guis_Pint(iV); 
     }
     goto next;
  case GUII_STOREINT: if (AP) *AP= guis_Gint(); BD(-1,"STORE-INT"); break;
  case GUII_LOADINT: if (AP) guis_Pint(*AP); BD(-1,"LOAD-INT"); break;
  case GUII_LOADX: regX= guis_Gint(); BD(-1,"LOAD-X"); break;
  case GUII_XADD:  if (AP) *AP += regX; BD(-1,"XADD"); break;
  case GUII_LOCAL: AP= locals + guis_Gint(); BD(-1,"LOCAL"); break;
  case GUII_GLOBAL: AP= guibc_global + guis_Gint(); BD(-1,"GLOBAL"); break;
  case GUII_WLAY_X: AP= layout + 4*guis_Gint() + 0; BD(-1,"WLAY_X"); break;
  case GUII_WLAY_Y: AP= layout + 4*guis_Gint() + 1; BD(-1,"WLAY_Y"); break;
  case GUII_WLAY_W: AP= layout + 4*guis_Gint() + 2; BD(-1,"WLAY_W"); break;
  case GUII_WLAY_H: AP= layout + 4*guis_Gint() + 3; BD(-1,"WLAY_H"); break;
  case GUII_STR: 
       iV= guis_Gint();
       Z= (char*) code+pc;
       BD(-1,"(push-string) %s", Z);
       pc+= iV;
       guis_Pptr(Z);
       break;
  case GUII_DUP: guis_dup(); BD(-1,"DUP"); break;
  case GUII_DROP: guis_pop(&A1); BD(-1,"DROP"); break;
  case GUII_ADD:
  case GUII_SUB:       
  case GUII_MUL:      
  case GUII_DIV:     
  case GUII_MOD:  
     goto arith;
  case GUII_CONSTRUCT:
     BD(-1,"CONSTRUCT");
     widget_no= guis_Gint();
     cons_func= guis_Gint();
     construct_widget(widget_no, cons_func);
     break;
 case GUII_INTARRAY:
     BD(-1,"INTARRAY");
     iV= guis_Gint();
     iA= NULL;
     if (iV) iA= malloc(sizeof(int)*iV);
     for(i=0;i<iV;i++) iA[iV-1-i]= guis_Gint();
     guis_Pptr(iA);
     break; 
 case GUII_FONTFILE:
     BD(-1,"FONTFILE");
     iV= guis_Gint();
     iS= guis_Gint();
     pV= guis_Gptr();
     gui_fontfiles[iV].data= pV;
     gui_fontfiles[iV].size= iS;
     break;
 case GUII_FONT:
     BD(-1,"FONT");
     iV= guis_Gint();
     gui_fontdefs[iV].file= gui_fontfiles + guis_Gint();
     gui_fontdefs[iV].size= guis_Gint();
     break;
 case GUII_ICON:
     BD(-1,"ICON");
     iV= guis_Gint();
     gui_icons[iV].nimages= guis_Gint();
     gui_icons[iV].images= calloc(gui_icons[iV].nimages,
                                  sizeof(gui_icons[iV].images[0]));
     break;
 case GUII_ICONIMAGE:
     BD(-1,"ICONIMAGE");
     IC= &gui_icons[guis_Gint()];
     IM= IC->images + guis_Gint();
     IM->min_dpi= guis_Gint();
     IM->data_length= guis_Gint();
     IM->data= guis_Gptr();
     break;
 case GUII_BLOB:
     BD(-1, "BLOB");
     iV= guis_Gint();
     guis_Pptr((void*) (code+pc));
     pc+= iV;
     break;
 case GUII_SECTION:
     iV= code[pc++];
     BD(-2, "SECTION %d", iV);
     iS= guis_Gint();
     BC= &screen_defs[iS].code[iV];
     BC->nlocals= guis_Gint();
     BC->code_length= iS= guis_Gint();
     BC->code= code+pc;
     pc+= iS;
     break;
  case GUII_GLOBALS_SECTION:
     BD(-1, "GLOBALS_SECTION");
     iS= guis_Gint();
     guibc_global_init_code= code+pc;
     guibc_global_init_size= iS;
     pc+= iS;
     break;
  case GUII_SCRINFO:
     BD(-1,"SCRINFO");
     SD= screen_defs + guis_Gint();
     SD->nwidgets= guis_Gint();
     SD->ndirect= guis_Gint();
     SD->widget_ids= guis_Gptr();
     break;
  case GUII_SUBLAYOUT:
     BD(-1,"SUBLAYOUT");
     pindex= guis_Gint();
     start= guis_Gint();
     end= guis_Gint();
     nlocals= guis_Gint();
     code_size= guis_Gint();
     gui_layout_widgets(layout[pindex*4+2], layout[pindex*4+3],
                       nlocals, code_size, code+pc, layout+4*start); 
     for(i=start;i<end;i++)
     {
       layout[4*i+0] += layout[4*pindex+0];
       layout[4*i+1] += layout[4*pindex+1];
     }
     pc+= code_size;
     break;
  case GUII_NOP: BD(-1,"NOP"); break;
  case GUII_SETG: BD(-1,"SETG"); gui_setg(code[pc++]); break;
  case GUII_DONE: BD(-1,"DONE"); return pc;
  }
  goto next;
arith:
  guis_pop(&A2);
  guis_pop(&A1);
  if (A1.type==GUIC_FLO || A2.type==GUIC_FLO) 
  {
    if (A1.type==GUIC_FLO) f1= A1.d.flo;
                      else f1= A1.d.sint;
    if (A2.type==GUIC_FLO) f2= A2.d.flo;
                      else f2= A2.d.sint;
    switch(C)
    {
    case GUII_ADD: f1= f1 + f2; BD(-1,"fADD"); break;
    case GUII_SUB: f1= f1 - f2; BD(-1,"fSUB"); break;
    case GUII_MUL: f1= f1 * f2; BD(-1,"fMUL"); break;
    case GUII_DIV: if (f2!=0) f1= f1 / f2; BD(-1,"fDIV"); break;
    case GUII_MOD: i1= f1; i2= f2; goto intarith;
    }
    guis_Pflo(f1);
    goto next;
  }
  i1= A1.d.sint;
  i2= A2.d.sint;
intarith:
  switch(C)
  {
    case GUII_ADD: i1= i1 + i2; BD(-1,"iADD"); break;
    case GUII_SUB: i1= i1 - i2; BD(-1,"iSUB"); break;
    case GUII_MUL: i1= i1 * i2; BD(-1,"iMUL"); break;
    case GUII_DIV: if (i2!=0) i1= i1 / i2; BD(-1,"iDIV"); break;
    case GUII_MOD: if (i2!=0) i1= i1 % i2; BD(-1,"iMOD"); break;
  } 
  guis_Pint(i1);
  goto next;
}


static void gui_layout_widgets
  (int WIDTH, int HEIGHT, int nlocals, int length,
   const unsigned char *code, int *Wcoor)
{
  int *locals;
  locals= malloc(sizeof(*locals)*nlocals);
  locals[GUILV_WIDTH]= WIDTH;
  locals[GUILV_HEIGHT]= HEIGHT;
  gui_run_bytecode(code, length, locals, Wcoor);                
  free(locals);
}

static void gui_compute_screen_dimensions
  (int parent_w,int parent_h,
   int n_locals, int length, const unsigned char *code,
   int *Scoor)
{
  int *locals;
  locals= malloc(sizeof(*locals)*n_locals);
  locals[GUILV_WIDTH]= parent_w;
  locals[GUILV_HEIGHT]= parent_h;
  Scoor[0]= Scoor[1]= Scoor[2]= Scoor[3]= -1;
  gui_run_bytecode(code, length, locals, Scoor);
  free(locals);
}

static void gui_construct_widgets
  (const unsigned char *code, int length)
{
  int locals[1];
  locals[0]= 0;
  gui_run_bytecode(code, length, locals, locals);
}

static void gui_screen(int num)
{
  screen_def_t *DD;
  if (num<0 || num>=n_screen_defs) return ;
  DD= screen_defs + num;

  int Scoor[4];
  int *Wcoor;
  int i;

  gui_compute_screen_dimensions(screen?screen->w:x_window_width,
                                screen?screen->h:x_window_height,
                      DD->code[GUISEC_SCREENDIM].nlocals,
                      DD->code[GUISEC_SCREENDIM].code_length,
                      DD->code[GUISEC_SCREENDIM].code, Scoor);
  // FIXME
  enter_screen(DD->nwidgets, DD->widget_ids, Scoor);

  gui_construct_widgets(DD->code[GUISEC_CONSTRUCTOR].code,
                        DD->code[GUISEC_CONSTRUCTOR].code_length);

  //
  // now, laying out the stuff
  // At this point, we could store the definition object within the
  // screen struct. This would allow us to re-layout widgets when
  // the window gets resized. I'm not going to do that for now.
  //
  Wcoor= malloc(sizeof(int)*DD->nwidgets*4);
  gui_layout_widgets(Scoor[2],Scoor[3], 
                     DD->code[GUISEC_WIDGETLAYOUT].nlocals,
                     DD->code[GUISEC_WIDGETLAYOUT].code_length,
                     DD->code[GUISEC_WIDGETLAYOUT].code, Wcoor);

  for(i=0;i<DD->nwidgets;i++)
  {
    Wcoor[4*i]+= Scoor[0];
    Wcoor[4*i+1]+= Scoor[1];
  }

  for(i=0;i<DD->nwidgets;i++)
     widget_move(widgets[DD->widget_ids[i]], 
                               Wcoor[4*i], Wcoor[4*i+1],
                               Wcoor[4*i+2], Wcoor[4*i+3]);
  for(i=0;i<DD->ndirect;i++)
     widget_show(widgets[DD->widget_ids[i]],1);
  free(Wcoor);
  gfx_update_all(); 
}

static void gfx_init_module(const unsigned char *module, size_t module_len)
{
  /**
  The top-level code is run without any variables or coordinate storage.
  This is because it's just used for parsing the data.
   **/
  gui_run_bytecode(module, module_len, NULL, NULL);
  /**
  When the code ends, everything has been stored in appropriate global
  variables. Now, the display can be started and stuff can be computed
  using the DPI etc.
   **/
}


