/***********************************************************************
 *                       WIDGET CODE                                   *
 *                                                                     *
 * Here starts the code for widgets. In order to reach a widget,       *
 * search for #W.widgetname, like #W.BUTTON etc. Some widget types     *
 * are defined at the top of the file since they are used by other     *
 * widgets. Others are just like mini files, independent of each       *
 * other.                                                              *
 *                   .--.             .---.                            *
 *                  /:.  '.         .' ..  '._.---.                    *
 *                 /:::-.  \.-"""-;` .-:::.     .::\                   *
 *                /::'|  `\/  _ _  \'   `\:'   ::::|                   *
 *            __.'    |   /  (o|o)  \     `'.   ':/                    *
 *           /    .:. /   |   ___   |        '---'                     *
 *          |    ::::'   /:  (._.) .:\                                 *
 *          \    .='    |:'        :::|                                *
 *           `""`       \     .-.   ':/                                *
 *                 jgs   '---`|I|`---'                                 *
 *                            '-'                                      *
 *                                                                     *
 *                                                                     *
 ***********************************************************************/

/***********************************************************************
 *                                                                     *
 *                             LABEL                                   *
 *                                                                     *
 * #W.LABEL                                                            *
 ***********************************************************************/
typedef struct 
{
  widget_t widget;
  int font, fg, bg;
  int just;
  char *text;
} gui_label_t;

static void label_render(gui_label_t *G)
{
  int tw, th; unsigned char *td;
  int dst_x, dst_y, copy_w, copy_h, src_x, src_y;

  int x,y,w,h;

  if (!G->widget.shown) return ;

  x= G->widget.x; y= G->widget.y; w= G->widget.w; h= G->widget.h;
  
  get_text_data(G->font, G->text, &tw, &th, NULL, &td);
  gfx_fill_rect(0, x, y, w, h, G->bg);

  if (th<=h)
  {
    src_y= 0; dst_y= y + (h-th)/2; copy_h= th;
  }
  else
  {
    src_y= (th-h)/2; dst_y= y; copy_h= h;
  }

  if (tw<w)
  {
    src_x= 0; copy_w= tw;
    if (G->just=='C')
      dst_x= x+(w-tw)/2;
    else if (G->just=='R')
      dst_x= x+w-tw;
    else
      dst_x= x;
  }
  else
  {
    dst_x= x; copy_w= w;
    src_x= 0;
  }

  blend_graymap(GFX_BACKGND, G->fg, G->bg, td, tw, th,
                src_x, src_y, dst_x, dst_y, copy_w, copy_h);
  free(td);
  gfx_update(x,y,w,h);
}

static void gui_label_show(widget_t *W)
{
  label_render((gui_label_t*) W);
}

static void gui_label_set_position(gui_label_t *L,int x,int y,int w,int h)
{
  widget_t *W;
  W= &L->widget;
  W->x= x; W->y= y; W->w= w; W->h= h;
  label_render(L);
}

static void gui_label_destroy(widget_t *W)
{
  gui_label_t *G;
  G= (void*) W;
  if (G->text)
   free(G->text);
}

static void gui_label_set_text(gui_label_t *G,const char *str)
{
  if (G->text) free(G->text);
  G->text= str ? strdup(str) : NULL;
  label_render(G);
}

static void label_set_fgbg(gui_label_t *G, int fg,int bg)
{
  if (G->fg==fg && G->bg==bg) return ;
  G->fg= fg;
  G->bg= bg;
  label_render(G);
}

static widget_t* gfx_label(int font, int fg, int bg, int just, char *str)
{
  gui_label_t *G;
  G= calloc(1,sizeof(*G));
  G->font= font;
  G->fg= fg;
  G->bg= bg;
  G->just= just;
  G->text= strdup(str ? str:"");

  G->widget.klass= GFXCLASS_LABEL;
  G->widget.show= gui_label_show;
  G->widget.destroy= gui_label_destroy;
  return &G->widget;
}


/***********************************************************************
 *                                                                     *
 *                               BUTTON                                *
 * #W.BUTTON                                                           *
 ***********************************************************************/

typedef struct
{
  widget_t widget;
  char *text;
  gui_icon_t *icon;
  int iconpos; // 'T' 'B' 'L' 'R'
  int state;  //0 = normal, 1= depressed 
  int enabled; 
  int click_handler;

  int (*click_func)(void *,gfx_event_t *);
  int (*release_func)(void *,gfx_event_t *);
  void *click_arg;
} gui_button_t;


static void button_render(gui_button_t *B)
{
  int x,y,w,h;
  colorpix_t pix;

  if (!B->enabled) return ;

  x= B->widget.x; y= B->widget.y; w= B->widget.w; h= B->widget.h;

  pix_of_widget(&B->widget, &pix);
  gui_draw_3dbox( &pix, B->state ? &RES(button.depressed)
                                     : &RES(button.normal) );

  gui_draw_icon_and_text(x,y,w,h, B->iconpos, B->icon,
                     RES(button.iconspc), 
                     RES(button.fg), B->text);
  gfx_update(x,y,w,h);
}

static void gui_button_set_text(gui_button_t *B, char *text)
{
  if (B->text) free(B->text);
  B->text= strdup(text);
  button_render(B);
}

static void gui_button_destroy(widget_t *W)
{
  gui_button_t *B;
  B= (void*) W;
  if (B->text) free(B->text);
  remove_click_handler(B->click_handler);
}

//
// This constructor is used internally. It doesn't register a click handler.
//
static gui_button_t *gui_button_new
  (const char *txt, int iconpos, int icon, int klass)
{
  gui_button_t *B;

  B= calloc(1,sizeof(*B));
  B->state= 0;
  B->enabled= 0;
  B->iconpos= iconpos;
  B->icon= gui_get_icon(icon);

  if (txt) B->text= strdup(txt);

  B->click_handler= -1;
  B->widget.klass= klass;
  B->widget.destroy= gui_button_destroy;
  return B;
}


static void button_set_state(gui_button_t *B, int state)
{
  if (state) state= 1;
  B->state= state;
  button_render(B);
}

static int gui_button_release(void *arg, gfx_event_t *event)
{
  gui_button_t *B;
  B= arg;
  event->type= B->widget.number;
  if (B->widget.klass==GFXCLASS_BUTTON)
    button_set_state(B,0);
  else
    event->key= B->state;
  if (B->release_func)
    return (*B->release_func)(B->click_arg,event);
  return 0;
}

static int gui_button_click(void *arg, gfx_event_t *event)
{
  gui_button_t *B;
  B= arg;
  if (!B->enabled) return 0; // shouldn't happen, but you never know
  if (B->widget.klass==GFXCLASS_BUTTON)
    button_set_state(B, 1);
  else
    button_set_state(B,B->state^1);
  install_release_handler(arg, gui_button_release);
  if (B->click_func)
     return (*B->click_func)(B->click_arg,event); 
  return 1;
}


// This one is the normal constructor called by user screens.
static widget_t* gfx_button
     (const char *txt,int iconpos, int icon,int klass)
{
  gui_button_t *B;
  B= gui_button_new(txt,iconpos,icon,klass);
  B->click_handler= install_click_handler(0,0,0,0, &B->enabled, 
                                           B, gui_button_click,1);
  return &B->widget;
}

static void button_set_position( gui_button_t *B, int x,int y,int w,int h)
{
  set_click_handler(B->click_handler, x,y,w,h);
  B->widget.x= x;
  B->widget.y= y;
  B->widget.w= w;
  B->widget.h= h;
}

// this is again for internal use. it sets the position,
// installs the click handler and then shows the button
static void gui_button_activate(widget_t *W, int x,int y,int w,int h)
{
  gui_button_t *B;
  B= (gui_button_t *)W;
  B->widget.x= x;
  B->widget.y= y;
  B->widget.w= w;
  B->widget.h= h;
 
  B->click_handler= install_click_handler(x,y,w,h, &B->enabled, 
                                           B, gui_button_click,1);
  B->enabled= 1;
  button_render(B);
}

static void gui_button_show(gui_button_t *B,int show)
{
  if (show) show= 1;
  if (B->enabled==show) return ;
  B->enabled= show;
  if (B->enabled) button_render(B);
}

//  This constructor doesn't allocate space. Uses space that is already
//  allocated. It's for internal use.
static void gui_button_init
  (gui_button_t *B,const char *txt, int iconpos, gui_icon_t* icon, 
   int (*click_func)(void *,gfx_event_t*), 
   int (*release_func)(void *,gfx_event_t*), void *click_arg)
{
  memset(B,0,sizeof(*B));
  B->state= 0;
  B->enabled= 0;
  if (txt) B->text= strdup(txt);
  B->iconpos= iconpos;
  B->icon= icon;

  B->click_func= click_func;
  B->release_func= release_func;
  B->click_arg= click_arg;

  B->widget.klass= GFXCLASS_BUTTON;
  
  B->click_handler= install_click_handler(0,0,0,0, &B->enabled, 
                                           B, gui_button_click,1);
}

/***********************************************************************
 *                                                                     *
 *                              CHECK BOX                              *
 * #W.CHECKBOX                                                         *
 ***********************************************************************
 The type is at the top, because other widgets use this.
 ***********************************************************************/

// #W.CHECKBOX
typedef struct checkbox
{
  widget_t widget;
  int state; 
  int enabled;
  int click;
  char *text;
  void *evarg; 
  event_function_t clickf;
} checkbox_t;

static void init_checkbox_images()
{
  if (checkbox_ok) return ;
  checkbox_ok= R_GRAYIMG_toserver(RES(checkbox.fg), RES(checkbox.bg),
                                  &RES(checkbox.ok));
  checkbox_nok= R_GRAYIMG_toserver(RES(checkbox.fg), RES(checkbox.bg),
                                  &RES(checkbox.nok));
}

static void checkbox_draw_button(checkbox_t *R,int state,int bg, int upd)
{
  int w,h,x,y;
  unsigned int *img;

  if (!R->enabled) return ;
  init_checkbox_images();
  img= state ? checkbox_ok : checkbox_nok;
  w= img[0];
  h= img[1];

  x= R->widget.x+RES(checkbox.pad_left);
  y= R->widget.y;
  if (R->widget.h>h) y+= (R->widget.h-h)/2;

  if (bg) gfx_fill_rect(0, x,y,w,h, RES(checkbox.bg));
  gfx_put(upd? GFX_UPDATE :0, img+2,w,h, 0,0, x,y,w,h);
}

static void gui_checkbox_destroy(widget_t *W)
{
  checkbox_t *R= (void*) W;
  if (R->text) free(R->text);
  remove_click_handler(R->click); 
}

static widget_t* gui_checkbox_new(const char *text,
    void *evarg, event_function_t clickf)
{
  checkbox_t *R;
  R= calloc(1,sizeof(*R));
  R->click= -1;
  if (text) R->text= strdup(text);
  R->widget.klass= GFXCLASS_CHECKBOX;
  R->widget.destroy= gui_checkbox_destroy;
  R->clickf= clickf;
  R->evarg= evarg;
  return &R->widget;
}

static void checkbox_draw_text(checkbox_t *R)
{
  unsigned char *td; int tw, th,cw,tx,ty;
  int x,y,w,h;
  if (!R->text || !R->text[0]) return ;

  x= R->widget.x; y= R->widget.y; w= R->widget.w; h= R->widget.h;
  
  get_text_data(1, R->text, &tw, &th, NULL, &td);
  tx= x + RES(checkbox.pad_left)+ gfx_checkbox_width + RES(checkbox.pad_mid);
  ty= y;
  if (h >= th)
    ty+= (h-th)/2;
  cw= (x+w)-tx;
  if (cw>tw) cw= tw;
  blend_graymap(GFX_BACKGND | GFX_UPDATE, RES(checkbox.fg), RES(checkbox.bg),
                td, tw, th, 0,0, tx,ty, cw, th);
  free(td);
}

static void checkbox_set(checkbox_t *R,int value)
{
  if (value) value= 1;
  if (R->state==value) return ;
  R->state= value;
  if (R->enabled)
    checkbox_draw_button(R,R->state,1,1);
}

static int checkbox_release(void *arg, gfx_event_t *event)
{
  checkbox_t *R;
  R= arg;
  event->type= R->widget.number;
  event->key= R->state;
  if (R->clickf && (*R->clickf)(R->evarg, event)) return 1;
  return 0;
}

static int checkbox_click(void *arg, gfx_event_t *event)
{
  checkbox_t *R;
  R= arg;
  checkbox_set(R, R->state^1);
  install_release_handler(arg, checkbox_release);
  return 1;
}


static void checkbox_draw(checkbox_t *R)
{
  int x,y,w,h;
  x= R->widget.x;
  y= R->widget.y;
  w= R->widget.w;
  h= R->widget.h;
  if (!R->enabled) return ;
  gfx_fill_rect(0,x,y,w, h, RES(checkbox.bg));
  checkbox_draw_button(R, R->state, 0,0);
  checkbox_draw_text(R);
  gfx_update(x,y,w, h);
}

static void checkbox_show(checkbox_t *R, int show)
{
  if (show)
  {
    if (R->enabled) return ;
    R->enabled= 1;
    if (R->click==-1)
      R->click= install_click_handler
          (R->widget.x,R->widget.y, R->widget.w, R->widget.h,
             &R->enabled, R, checkbox_click, 1);
    checkbox_draw(R);
  }
  else
  {
    if (!R->enabled) return ;
    // disable the click handler
    R->enabled= 0;
  }
}



/***********************************************************************
 *                                                                     *
 *                             LOG VIEWER                              *
 * #W.LOGVIEW                                                          *
 ***********************************************************************/
typedef struct
{
  widget_t widget;
  int nlines,slines;
  char **lines;
  int last_line_incomplete;

  int fh;
  int cx, cy;
  int px, py, pw, ph;

  screen_t *screen;
} gui_logview_t;

static void logview_print_line
    (gui_logview_t *L,char *line, int nonl,int *Ruys,int *Ruye);
static void logview_scroll_up(gui_logview_t *L, int yof);
static widget_t *gui_logview();
static void gui_logview_destroy(widget_t *W);
static void logview_render(gui_logview_t *L);
static void gui_logview_show(widget_t *W);
static void gui_logview_hide(widget_t *W);
static void gui_logview_position(widget_t *W, int x,int y,int w,int h);
static void logview_write(gui_logview_t *L, const char *txt, size_t len);
static void logview_put_line
    (gui_logview_t *L, const char *txt, size_t len, int nonl);

static int logwidget= -1;


static widget_t *gui_logview()
{
  gui_logview_t *B;

  B= calloc(1,sizeof(*B));
  // B->click_handler= -1;

  B->widget.klass= GFXCLASS_LOGVIEW;
  B->widget.show= gui_logview_show;
  B->widget.hide= gui_logview_hide;
  B->widget.destroy= gui_logview_destroy;
  B->widget.position= gui_logview_position;
  B->screen= screen;
  return &B->widget;
}


static void gui_logview_destroy(widget_t *W)
{
  int i;
  gui_logview_t *L= (gui_logview_t *)W;
  if (L->lines)
  {
    for(i=0;i<L->nlines;i++) free(L->lines[i]);
    free(L->lines);
    L->lines= NULL;
  }
  // remove_click_handler(L->click_handler);
}

static void logview_scroll_up(gui_logview_t *L, int yof)
{
  int x,y;
  int cx, cy, cw, ch;
  unsigned int color;
 
  colorpix_t pix;

  cx= L->px;
  cy= L->py;
  cw= L->pw;
  ch= L->ph;

  pix= get_screen_pix(L->screen);

  if (clip_for_rect(&cx, &cw, pix.w)) return ;
  if (clip_for_rect(&cy, &ch, pix.h)) return ;
  for(y=cy;y<cy+ch-yof;y++)
   for(x=cx;x<cx+cw;x++)
     PIX(pix,x,y)= PIX(pix, x,y+yof);

  color= color_u_to_s(RES(logview.bg));
  
  for(;y<cy+ch;y++)
   for(x=cx;x<cx+cw;x++)
     PIX(pix,x,y)= color;
}


static void logview_print_line
   (gui_logview_t *L,char *line, 
    int nonl,int *Ruys,int *Ruye)
{
  graypix_t text;
  int delta;

  text_to_graypix(1, line, &text, NULL);

  delta= L->cy+text.h+2 - L->ph;
  if (delta>0)
  {
    *Ruys= 0; *Ruye= L->ph;
    logview_scroll_up(L, delta);
    L->cy-= delta;
  }
  else { *Ruys= L->cy; *Ruye= L->cy+text.h; }

  int tw;
  tw= text.w;
  text.w= imin2(text.w, L->pw - L->cx);
  if (text.w>0)
  {
    colorpix_t pix;
    pix= get_screen_pix(L->screen);

    blend_graypix_bg(RES(logview.fg), RES(logview.bg),
           text, pix, L->cx+L->px, L->cy+L->py);
    if (nonl) L->cx+= tw; 
  } 
  free(text.data);
  if (!nonl) { L->cy+= font1height + font1height/10+ 2; L->cx= 0; }
}


static void logview_render(gui_logview_t *L)
{
  int i;
  int dum;
  colorpix_t pix;

  L->cx= 0;
  L->cy= 0;

  pix= get_screen_pix(L->screen);
  colorpix_fill_rect(pix, L->widget.x,L->widget.y, L->widget.w,L->widget.h,
                  RES(logview.bg));

  for(i=0;i<L->nlines-1;i++)
   logview_print_line(L,L->lines[i],0,&dum,&dum);
  if (L->nlines) logview_print_line(L,L->lines[L->nlines-1],
                                    L->last_line_incomplete,&dum,&dum);

  update_screen_pixels(L->screen,
         L->widget.x, L->widget.y,L->widget.w,L->widget.h);
}


static void gui_logview_show(widget_t *W)
{
  logview_render((gui_logview_t*)W);
}

static void gui_logview_hide(widget_t *W) { }

static void gui_logview_position(widget_t *W, int x,int y,int w,int h)
{
  gui_logview_t *L= (gui_logview_t*) W;
  int nl;
  char **ll;
  int i;

  widgetpos(W,x,y,w,h);
  if (!L->fh) get_text_data(1,"M",NULL,&L->fh,NULL,NULL);
  nl= (w+L->fh-1)/L->fh;
  if (nl!=L->slines)
  {
    ll= calloc(sizeof(*ll),nl);
    if (nl>L->nlines)
      for(i=0;i<L->nlines;i++) ll[i]= L->lines[i];
    else
    {
      for(i=0;i<nl;i++)
         ll[nl-1-i]= L->lines[L->nlines-i-1];
      for(i=0;i<L->nlines-nl;i++) free(L->lines[i]);
      L->nlines= nl;
    }
    if (L->lines) free(L->lines[i]);
    L->lines= ll;
    L->slines= nl;
  }
  L->pw= w-2*RES(logview.pad);
  L->ph= h-2*RES(logview.pad);
  L->px= x+RES(logview.pad);
  L->py= y+RES(logview.pad);
  if (W->shown) logview_render(L);
}

static void logview_write(gui_logview_t *L, const char *txt, size_t len)
{
  char *E;

  while(len)
  {
    E= memchr(txt, '\n', len);
    if (!E) 
    {
       logview_put_line(L, txt, len, 1);
       len= 0;
    }
    else if (E==txt)
    {
       logview_put_line(L, " ",1, 0);
       txt= E+1;
       len--;
    }
    else
    {
       logview_put_line(L,txt,E-txt,0);
       len-= E-txt+1;
       txt= E+1;
    }
  }
}

static void logview_put_line
   (gui_logview_t *L, const char *txt, size_t len, int nonl)
{
  char *ll,*pp,*oo;
  size_t lo;
  int sy,ey;
  int i;

again:
  if (L->last_line_incomplete)
  {
    if (!L->nlines) { L->last_line_incomplete= 0; goto again; }
    oo= L->lines[L->nlines-1];
    lo= strlen(oo);
    ll= malloc(lo+len+1);
    memcpy(ll, oo, lo);
    memcpy(ll+lo, txt, len);
    ll[len+lo]= 0;
    free(oo);
    L->lines[L->nlines-1]= ll;
    pp= ll+lo;
  }
  else 
  {
    if (L->nlines==L->slines)
    {
      if (L->widget.shown && L->nlines)
      {
        free(L->lines[0]);
        for(i=0;i<L->nlines-1;i++) L->lines[i]= L->lines[i+1];
        L->nlines--;
      }
      else 
      {
        if (!L->slines) 
          { L->slines= 50; L->lines= malloc(sizeof(L->lines[0])*L->slines); }
        else
          { L->slines+= L->slines/2; L->lines= realloc(L->lines,
                                             sizeof(L->lines[0])*L->slines); }
      }
    }
    ll= malloc(len+1);
    memcpy(ll,txt,len);
    ll[len]= 0;
    L->lines[L->nlines]= ll;
    L->nlines++;
    pp= ll;
  }
  L->last_line_incomplete= nonl;
  if (L->widget.shown)
  {
    logview_print_line(L, pp, nonl,&sy,&ey);
    update_screen_pixels(L->screen,L->px, L->py+sy, L->pw, ey-sy);
  }
}

void gfx_logview_clear(int wno)
{
  widget_t *W;
  int i;
  gui_logview_t *L;
  W= get_widget(wno);
  if (!W || W->klass!=GFXCLASS_LOGVIEW) return ;
  L= (gui_logview_t*) W;
  if (L->nlines==0) return ;
  for(i=0;i<L->nlines;i++) free(L->lines[i]);
  L->nlines= 0;
  if (W->shown) logview_render(L);
}

void gfx_set_log(int wno)
{
  logwidget= wno;
}

void loprintv(const char *fmt, va_list ap)
{
  widget_t *W; 
  char *txt;
  W= get_widget(logwidget);
  if (!W || W->klass!=GFXCLASS_LOGVIEW) return ;
  txt= vmprintf(fmt, ap);
  logview_write((gui_logview_t*) W, txt, strlen(txt));
  free(txt);
}

void loprintf(const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  loprintv(fmt,ap);
  va_end(ap);
}

void gfx_logview_printf(int wno, const char *fmt, ...)
{
  widget_t *W; 
  char *txt;
  va_list ap;
  W= get_widget(wno);
  if (!W || W->klass!=GFXCLASS_LOGVIEW) return ;
  va_start(ap,fmt);
  txt= vmprintf(fmt, ap);
  va_end(ap);
  
  logview_write((gui_logview_t*) W, txt, strlen(txt));
  free(txt);
}

void gfx_logview_write(int wno, const char *txt, size_t len)
{
  widget_t *W;
  W= get_widget(wno);
  if (!W || W->klass!=GFXCLASS_LOGVIEW) return ;
  logview_write((gui_logview_t*) W, txt, len);
}

/***********************************************************************
 *                                                                     *
 *                               SPINNER                               *
 * #W.SPINNER                                                          *
 ***********************************************************************/

typedef struct
{
  widget_t widget;
  int type; // 'F' = float 'I'= integer
  int frac; // fractional digits

  int intv, intq, intmin, intmax;
  double flov, floq, flomin, flomax;
  int limflags;

  int enabled;
  int timer, count;

  gui_button_t less,more;
  widget_t *label;
} gui_spinner_t;

#define SPINNER_MIN 1
#define SPINNER_MAX 2

static int spinner_lessb_width, spinner_moreb_width;


static void gui_spinner_destroy(gui_spinner_t *SS)
{
  gui_button_destroy((widget_t*) &SS->less);
  gui_button_destroy((widget_t*) &SS->more);
  widget_destroy(SS->label);
  if (SS->timer!=-1) gui_timer_cancel(SS->timer);
  SS->timer= -1;
}

static void gui_spinner_set_position
   (gui_spinner_t* LL,int x,int y,int w,int h)
{
  LL->widget.x= x;
  LL->widget.y= y;
  LL->widget.w= w;
  LL->widget.h= h;
    // FIXME: paddings should be resources
  button_set_position(&LL->less, x,y, spinner_lessb_width, h);
  button_set_position(&LL->more, x+w-spinner_moreb_width,y,
                                 spinner_moreb_width, h);
  widget_move(LL->label, x+spinner_lessb_width+5, y, 
      w - 10 - spinner_lessb_width-spinner_moreb_width , h);
}


static void gui_spinner_render(gui_spinner_t *S)
{
  char buf[40];
  if (S->type=='F')
  {
    sprintf(buf,"%f",S->flov);
  }
  else
    sprintf(buf,"%d", S->intv);
  gui_label_set_text((gui_label_t*) S->label,buf);
}

static void gui_spinner_show(gui_spinner_t *S, int show)
{
  if ((S->enabled && show) || (!S->enabled && !show)) return ;

  if (!show)
  {
    S->enabled= 0;
    gui_button_show(&S->less,0);
    gui_button_show(&S->more,0);
    widget_show(S->label, 0);
    return ;
  }

  S->enabled= 1;
  gui_button_show(&S->less,1);
  gui_button_show(&S->more,1);
  gui_spinner_render(S);
  widget_show(S->label, 1);
}


static void spinner_apply_limits(gui_spinner_t* S)
{
  if (S->type=='F')
  {
     if ((S->limflags&SPINNER_MIN) && S->flov<S->flomin) S->flov= S->flomin;
     if ((S->limflags&SPINNER_MAX) && S->flov>S->flomax) S->flov= S->flomax;
  }
  else
  {
     if ((S->limflags&SPINNER_MIN) && S->intv<S->intmin) S->intv= S->intmin;
     if ((S->limflags&SPINNER_MAX) && S->intv>S->intmax) S->intv= S->intmax;
  }
}

static int gui_spinner_delta(gui_spinner_t *S, gfx_event_t *event, int dir)
{
  int iV; double fV;
  if (S->type=='F')
  {
    fV= S->flov;
    S->flov+= dir*S->floq;
    spinner_apply_limits(S);
    if (S->flov!=fV)
    {
       gui_spinner_render(S);
       event->type= S->widget.number;
       event->flo0= S->flov;
       return 0; 
    }
    return 1;
  }
  iV= S->intv;
  S->intv += dir*S->intq;
  spinner_apply_limits(S);  
  if (S->intv!=iV)
  {
     gui_spinner_render(S);
     event->type= S->widget.number;
     event->key= S->intv;
     return 0;
  }
  return 1;
}

static gui_spinner_t *get_spinner_widget(int wno)
{
  widget_t *W;
  if (!(W=get_widget(wno))) return NULL;
  return W->klass==GFXCLASS_SPINNER ? (gui_spinner_t*) W : NULL;
}

static void gui_spinner_remove_timer
  (gui_spinner_t *SS)
{
  if (SS->timer!=-1) gui_timer_cancel(SS->timer);
  SS->timer= -1;
}

static void gui_spinner_reset_timer
  (gui_spinner_t *SS,int (*F)(void*,gfx_event_t*))
{
  int delay;
  SS->count++;
  if (SS->count>50) delay= 85;
  else if (SS->count>10) delay= 150;
  else delay= 200;

  SS->timer= gui_timer_schedule(delay, F, SS);
}

static void gui_spinner_init_timer
  (gui_spinner_t *SS,int (*F)(void*,gfx_event_t*))
{
  if (SS->timer!=-1) gui_timer_cancel(SS->timer);
  SS->count= 0;
  gui_spinner_reset_timer(SS,F);
}

static int gui_spinner_less_release(void *arg,gfx_event_t *event)
{
  gui_spinner_remove_timer(arg);
  return 0;
}

static int gui_spinner_less_timeout(void *arg,gfx_event_t *event)
{
  gui_spinner_reset_timer(arg, gui_spinner_less_timeout);
  return !gui_spinner_delta(arg, event,-1);
}

static int gui_spinner_less_click(void *arg,gfx_event_t *event)
{
  gui_spinner_init_timer(arg, gui_spinner_less_timeout);
  return gui_spinner_delta(arg, event,-1);
}

static int gui_spinner_more_release(void *arg,gfx_event_t *event)
{
  gui_spinner_remove_timer(arg);
  return 0;
}

static int gui_spinner_more_timeout(void *arg,gfx_event_t *event)
{
  gui_spinner_reset_timer(arg, gui_spinner_more_timeout);
  return !gui_spinner_delta(arg, event,1);
}

static int gui_spinner_more_click(void *arg,gfx_event_t *event)
{
  gui_spinner_init_timer(arg, gui_spinner_more_timeout);
  return gui_spinner_delta(arg, event,1);
}

static widget_t* gui_spinner(int type, int frac)
{
  gui_spinner_t *SS;
  static gui_icon_t icon_less, icon_more;
  static int init_ok;

  SS= calloc(1,sizeof(*SS));
  SS->widget.klass= GFXCLASS_SPINNER;
  SS->type= type;
  SS->frac= frac;
    // FIXME: colors should be resources
  SS->label= gfx_label(1,0x00,0xffFFff,'R',"");
  if (!init_ok)
  {
    icon_less.w= RES(spinner.less).w;
    icon_less.h= RES(spinner.less).h;
    icon_less.data= RES(spinner.less).data;

    icon_more.w= RES(spinner.more).w;
    icon_more.h= RES(spinner.more).h;
    icon_more.data= RES(spinner.more).data;
    init_ok= 1;
  }
  gui_button_init(&SS->less,NULL,'T', &icon_less,
                    gui_spinner_less_click, gui_spinner_less_release, SS);
  gui_button_init(&SS->more,NULL,'T', &icon_more,
                    gui_spinner_more_click, gui_spinner_more_release, SS);
  spinner_lessb_width= RES(spinner.less).w;
  spinner_moreb_width= RES(spinner.more).w;
  SS->intq= 1;
  SS->floq= 1.0;
  SS->timer= -1;
  return &SS->widget;
}

void gfx_spinner_set_imin(int wno, int min)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  SS->limflags |= SPINNER_MIN;
  SS->intmin= min;
  spinner_apply_limits(SS);
}

void gfx_spinner_set_imax(int wno, int max)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  SS->limflags |= SPINNER_MAX;
  SS->intmax= max;
  spinner_apply_limits(SS);
}

void gfx_spinner_set_iquant(int wno, int qua)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  SS->intq= qua;
  spinner_apply_limits(SS);
}

void gfx_spinner_set_fmin(int wno, double min)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  SS->limflags |= SPINNER_MIN;
  SS->flomin= min;
  spinner_apply_limits(SS);
}

void gfx_spinner_set_fmax(int wno, double max)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  SS->limflags |= SPINNER_MAX;
  SS->flomax= max;
  spinner_apply_limits(SS);
}

void gfx_spinner_set_fquant(int wno, double qua)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  SS->floq= qua;
  spinner_apply_limits(SS);
}

void gfx_spinner_set_fvalue(int wno, double value)
{
  double V;
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  V= SS->flov;
  SS->flov= value;
  spinner_apply_limits(SS);
  if (V!=SS->flov) gui_spinner_render(SS);
}

void gfx_spinner_set_ivalue(int wno, int value)
{
  int V;
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return ;
  V= SS->intv;
  SS->intv= value;
  spinner_apply_limits(SS);
  if (V!=SS->intv) gui_spinner_render(SS);
}

int gfx_spinner_get_ivalue(int wno)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return 0;
  return SS->intv; 
}


double gfx_spinner_get_fvalue(int wno)
{
  gui_spinner_t *SS;
  if (!(SS= get_spinner_widget(wno))) return 0.0;
  return SS->flov; 
}

/***********************************************************************
 *                                                                     *
 *                            PROGRESS BAR                             *
 * #W.PROGRESS                                                         *
 ***********************************************************************/

typedef struct 
{
  widget_t widget;
  int total, current;
  int font;
  int fg, bg, fill;
} gui_progressbar_t;


static void progressbar_render(gui_progressbar_t *G)
{
  char buf[10];
  int tw, th; unsigned char *td;
  int x,y,w,h,tx;
  int pos;

  if (!G->widget.shown) return ;

  x= G->widget.x; y= G->widget.y; w= G->widget.w; h= G->widget.h;

  if (G->total==0) pos= 0;
              else pos= G->current*w / G->total;
  if (pos<0) pos= 0;
  if (pos>w) pos= w;

  gfx_fill_rect(0, x, y, pos, h, G->fill);
  gfx_fill_rect(0, x+pos, y, w-pos-1, h, G->bg);
  gfx_draw_rect(0, x, y, w, h, G->fill);

  if (G->total==0) sprintf(buf,"0%%");
  else sprintf(buf,"%d%%", G->current*100/G->total);
  get_text_data(G->font, buf, &tw, &th, NULL, &td);

  tx= (w-tw)/2;
  if (pos<tx) {
    blend_graymap(0, G->fg, G->bg, td, tw, th, 0, 0,
                     x+(w-tw)/2, y+(h-th)/2, tw, th);
  } else if (pos>=tx+tw)
  {
    blend_graymap(0, G->bg, G->bg, td, tw, th, 0, 0,
                     x+(w-tw)/2, y+(h-th)/2, tw, th);
  } else
  {
    blend_graymap(0, G->bg, G->bg, td, tw, th, 0, 0,
                     x+(w-tw)/2, y+(h-th)/2, pos-tx, th);

    blend_graymap(0, G->fg, G->bg, td, tw, th, pos-tx, 0,
                     x+(w-tw)/2+pos-tx, y+(h-th)/2, tw-pos+tx, th);
  }
  free(td);
  gfx_update(x,y,w,h);
}

static void gui_progressbar_show(widget_t *W)
{
  progressbar_render((gui_progressbar_t*) W);
}

static void gui_progressbar_position
   (widget_t *W,int x,int y,int w,int h)
{
  W->x= x; W->y= y; W->w= w; W->h= h;
  progressbar_render((gui_progressbar_t *) W);
}

static widget_t* gui_progressbar_new(int font, int fg, int bg, int fill)
{
  gui_progressbar_t *G;
  G= calloc(1,sizeof(*G));
  G->font= font;
  G->fg= fg;
  G->bg= bg;
  G->fill= fill;

  G->widget.klass= GFXCLASS_PROGRESSBAR;
  G->widget.show= gui_progressbar_show;
  G->widget.position= gui_progressbar_position;
  return &G->widget;
}

void gfx_progressbar_total(int wno, int total)
{
  gui_progressbar_t *G;
  G= get_widgetk(wno, GFXCLASS_PROGRESSBAR);
  if (!G) return ;
  G->total= total;
}

void gfx_progressbar_set(int wno, int value)
{
  gui_progressbar_t *G;
  G= get_widgetk(wno, GFXCLASS_PROGRESSBAR);
  if (!G) return ;
  G->current= value;
  progressbar_render(G);
}

/***********************************************************************
 *                                                                     *
 *                            RADIO BUTTONS                            *
 * #W.RADIO                                                            *
 ***********************************************************************
 Radio buttons belong to a group. The group which they belong to is 
 identified by another radio button. The widget doesn't have any images 
 in it because all radio buttons look exactly the same.
 ***********************************************************************/

typedef struct radio
{
  widget_t widget;
  int state; 
  int enabled;
  int click;
  int change;  
  int group;
  char *text;
} radio_t;

static unsigned int *radio_on, *radio_off;

static void init_radio_images()
{
  if (radio_on) return ;
  radio_on= R_GRAYIMG_toserver(RES(radio.fg), RES(radio.bg),
                                     &RES(radio.on));
  radio_off= R_GRAYIMG_toserver(RES(radio.fg), RES(radio.bg),
                                     &RES(radio.off));
}

static void radio_draw_button(radio_t *R,int state,int bg, int upd)
{
  int w,h,x,y;
  unsigned int *img;

  if (!R->enabled) return ;
  init_radio_images();
  img= state ? radio_on : radio_off;
  w= img[0];
  h= img[1];

  x= R->widget.x+RES(radio.pad_left);
  y= R->widget.y;
  if (R->widget.h>h) y+= (R->widget.h-h)/2;

  if (bg) gfx_fill_rect(0, x,y,w,h, RES(radio.bg));
  gfx_put(upd? GFX_UPDATE :0, img+2,w,h, 0,0, x,y,w,h);
}

static radio_group_t *get_rgroup(int ident)
{
  int i;
  if (!screen || ident==-1) return NULL;
  for(i=0;i<screen->n_rgroups;i++)
    if (screen->rgroups[i].ident==ident)
       return screen->rgroups+i;
  return NULL; 
}

static void rgroup_clear_except(int ident,widget_t *R)
{
  widget_t *P;
  radio_group_t *G;
  int i;

  G= get_rgroup(ident);
  if (!G) return ;
  
  for(i=0;i<G->n_widgets;i++)
  {
    P= G->widgets[i];
    if (P==R) continue;
    widget_set_state(P,0);
  }
}


static int radio_release(void *arg, gfx_event_t *event)
{
  radio_t *R;

  R= arg;
  if (!R->change) return 1;
 
  event->type= R->widget.number;
  R->change= 0;
  return 0;
}

static void radio_set(radio_t *R)
{
  if (R->state) return ;
  R->state= 1;
  if (R->enabled) radio_draw_button(R,R->state,1,1);
  rgroup_clear_except(R->group, &R->widget);
}

static int radio_click(void *arg, gfx_event_t *event)
{
  radio_t *R;
  R= arg;
  if (!R->state)
  {
    radio_set(arg);
    R->change= 1;
  }
  else
    R->change= 0;
  install_release_handler(arg, radio_release);
  return 1;
}

static void radio_draw_text(radio_t *R)
{
  unsigned char *td; int tw, th,cw,tx,ty;
  int x,y,w,h;
  if (!R->text || !R->text[0]) return ;

  x= R->widget.x; y= R->widget.y; w= R->widget.w; h= R->widget.h;
  
  get_text_data(1, R->text, &tw, &th, NULL, &td);
  tx= x + RES(radio.pad_left) + gfx_radio_width + RES(radio.pad_mid);
  ty= y;
  if (h >= th)
    ty+= (h-th)/2;
  cw= (x+w)-tx;
  if (cw>tw) cw= tw;
  blend_graymap(GFX_BACKGND | GFX_UPDATE, RES(radio.fg), RES(radio.bg),
                td, tw, th, 0,0, tx,ty, cw, th);
  free(td);
}

static void radio_draw(radio_t *R)
{
  int x,y,w,h;
  x= R->widget.x;
  y= R->widget.y;
  w= R->widget.w;
  h= R->widget.h;
  if (!R->enabled) return ;
  gfx_fill_rect(0,x,y,w, h, RES(radio.bg));
  radio_draw_button(R, R->state, 0,0);
  radio_draw_text(R);
  gfx_update(x,y,w, h);
}

static void register_radio_in_group(int ident, widget_t *W)
{
  radio_group_t *G;
  G= get_rgroup(ident);
  if (G) goto found;

  if (screen->n_rgroups==screen->s_rgroups)
  {
    screen->s_rgroups+= screen->s_rgroups/2 + 10;
    screen->rgroups= realloc(screen->rgroups, 
                              sizeof(screen->rgroups[0])*screen->s_rgroups);
  }
  G= screen->rgroups+screen->n_rgroups;
  memset(G,0,sizeof(*G));
  G->ident= ident;
  screen->n_rgroups++;
found:
  if (G->n_widgets==G->s_widgets)
  {
    G->s_widgets+= G->s_widgets/2+10;
    G->widgets= realloc(G->widgets, sizeof(G->widgets[0])*G->s_widgets);
  }
  G->widgets[G->n_widgets]= W;
  G->n_widgets++;
}

static void free_rgroups(screen_t *S)
{
  int i;
  for(i=0;i<S->n_rgroups;i++)
    if (S->rgroups[i].widgets) free(S->rgroups[i].widgets);
  if (S->rgroups) free(S->rgroups);
}

static int radio_group_get_selected(int ident)
{
  radio_group_t *G;
  int i;
  G= get_rgroup(ident);
  if (!G) return -1;
  for(i=0;i<G->n_widgets;i++)
    if (widget_get_state(G->widgets[i])) 
       return G->widgets[i]->number;
  return -1;
}

static void radio_show(radio_t *R, int show)
{
  if (show)
  {
    if (R->enabled) return ;
    R->enabled= 1;
    if (R->click==-1)
      R->click= install_click_handler
          (R->widget.x,R->widget.y, R->widget.w, R->widget.h,
             &R->enabled, R, radio_click, 1);
    radio_draw(R);
  }
  else
  {
    if (!R->enabled) return ;
    // disable the click handler
    R->enabled= 0;
  }
}

static widget_t* radio_new(const char *text, int group)
{
  radio_t *R;
  R= calloc(1,sizeof(*R));
  R->click= -1;
  if (text) R->text= strdup(text);
  R->group= group;
  R->widget.klass= GFXCLASS_RADIO;
  if (group!=-1) register_radio_in_group(group, &R->widget);
  return &R->widget;
}

/***********************************************************************
 *                                                                     *
 *                                PANEL                                *
 * #W.PANEL                                                            *
 ***********************************************************************
 The panel widget is the only container widget in the library. A panel
 contains several subscreen objects, which are not widgets by themselves.

 When we create a panel, everything about it is known at compile time. 
 ie. the number of subscreens and the contents of each subscreen. 

 When we're computing the layout for a subscreen, it's done relative to 
 the coordinates of the panel. It's all done in the creation routine.
 ***********************************************************************/

typedef struct
{
  int n_widgets;
  int *widgets;
} subscreen_t;

typedef struct
{
  widget_t widget;
  int enabled;
  int active;
  int n_subs;
  subscreen_t *subs;
} panel_t;

static void panel_show_widgets(panel_t *P,int sub, int show)
{
  int i;
  subscreen_t *S; 
  if (sub<0 || sub>=P->n_subs) return ;
  S= P->subs+sub;

  for(i=0;i<S->n_widgets;i++)
  {
    widget_t *W;
    int n= S->widgets[i];
    if (n<first_widget_no || n>last_widget_no) continue;
    W= widgets[n];
    if (W) widget_show(W,show);
  }
}

static void panel_activate(panel_t *P,int sel)
{
  if (sel<0 || sel>=P->n_subs || sel==P->active) return ;
  if (!P->enabled) { P->active= sel; return ; }
  panel_show_widgets(P,P->active,0);
  gfx_fill_rect(0, P->widget.x,P->widget.y,P->widget.w,P->widget.h, 0xededed); 
  panel_show_widgets(P,sel,1);
  P->active= sel;
  gfx_update(P->widget.x,P->widget.y,P->widget.w,P->widget.h);
}

static void panel_show(panel_t *P,int show)
{
  P->enabled= show;
  if (show)
    gfx_fill_rect(0, P->widget.x,P->widget.y,P->widget.w,P->widget.h, 0xededed); 
  panel_show_widgets(P,P->active,show);
}

static void panel_position(panel_t *P,int nx,int ny,int nw,int nh)
{
  P->widget.x= nx;
  P->widget.y= ny;
  P->widget.w= nw;
  P->widget.h= nh;
}

#if 0
static panel_t *panel_new(int nsubs)
{
  panel_t *P;
  P= calloc(1,sizeof(*P));
  P->n_subs= nsubs;
  P->subs= calloc(nsubs,sizeof(P->subs[0]));
  P->widget.klass= GFXCLASS_PANEL;
  return P;
}
#endif

/**
 D contains the following:
   <n_subs:int>
   <subscreen1>
   <subscreen2>
   ...
 each subscreen contains:
   <n_widgets:int>
   <widget1:int>
   <widget2:int>
   ..
**/
static widget_t *panel_new(int *D)
{
  panel_t *P;
  subscreen_t *SS;
  int *OD;
  OD= D;
  int s,w;

  P= calloc(1,sizeof(*P));
  P->n_subs= *(D++);
  P->subs= calloc(P->n_subs,sizeof(P->subs[0]));
  P->widget.klass= GFXCLASS_PANEL;

  for(s=0;s<P->n_subs;s++)
  {
    SS= P->subs + s;
    SS->n_widgets= *(D++);
    if (!SS->n_widgets) continue;
    SS->widgets= malloc(sizeof(int)*SS->n_widgets);
    for(w=0;w<SS->n_widgets;w++)
      SS->widgets[w]= *(D++);
  }

  free(OD);
  // we free OD here because it was constructed on the heap
  // from the bytecode stack. in contrast, we  don't free
  // strings because they directly refer to the code buffer.
  return (widget_t*)P;
}


/***********************************************************************
 *                                                                     *
 *                             TEXT ENTRY                              *
 * #W.TEXTENTRY                                                        *
 ***********************************************************************
  Things to do here:
    - selections
      - by shift-arrow
      - by mouse drag
      - double click
    - copy and paste
    - borders and 3d look 
 ***********************************************************************/

typedef struct { uint32_t chr; uint8_t  *d; int w,h,a; int x; } glyph_t;

typedef struct 
 { widget_t widget;
   int x,y,w,h; glyph_t *glyphs; int size, len;
   int fonth, fonta;
   int left; int cursor; 
   int (*key_filter)(void *,void *,gfx_event_t *);
   void *key_arg;
   } text_entry_t;

static uint32_t TE_fg        = 0;
static uint32_t TE_bg        = 0xffffff;
static uint32_t TE_cfg       = 0xffffff;
static uint32_t TE_cbg       = 0x000088;

static int draw_glyph(text_entry_t  *te, int gn, int expose)
{
  int ws, we, sx, sy, cw, ch, dx,dy, gs,ge;
  uint32_t fg, bg;
  glyph_t *G;

  ws= te->left; 
  we= te->left+te->w;
  G= te->glyphs+gn;

  if (te->fonth>te->h)
  {
    sy= (te->fonth-te->h)/2;
    ch= te->h;
    dy= 0;
  } else {
     dy= (te->h-te->fonth)/2 + te->fonta - G->a;
     sy=0; ch= te->fonth; 
  }

  gs= G->x;
  ge= gs + G->w;

  if (ge<=ws) return 0;
  if (gs>=we) return 1;
    
  sx= 0; cw= te->glyphs[gn].w;
  if (gs<ws) { sx= ws-gs; cw-= ws-gs; }
  if (ge>we) { cw-= ge-we; }

  dx= sx + gs - ws;

  expose= expose ? GFX_UPDATE : 0;
  if (te->widget.has_focus && gn==te->cursor) { fg= TE_cfg; bg= TE_cbg; }
                                         else { fg= TE_fg;  bg= TE_bg; }
  gfx_fill_rect(expose, dx+te->x, te->y + dy + G->a - te->fonta, G->w,
                       te->fonth, bg);
  blend_graymap( expose | GFX_BACKGND, 
                 fg, bg, G->d, G->w, G->h, sx, sy,
                         dx + te->x, dy + te->y, cw, ch );
  return 0;
}


static void make_cursor_visible(text_entry_t *te)
{
  int cx, cw;
  int i;
  cx= te->glyphs[te->cursor].x;
  cw= te->glyphs[te->cursor].w;

  if (cx<te->left) te->left= cx; 
  else if (cx+cw-te->left>te->w) te->left= cx+cw-te->w; 

  if (te->widget.shown)
  {
    gfx_fill_rect(0, te->x, te->y, te->w, te->h, TE_bg);
    gfx_draw_rect(0, te->x-1, te->y-1, te->w+2, te->h+2, 0x888888);
    for(i=0;i<te->len;i++) if (draw_glyph(te, i,0)) break;
    gfx_update(te->x, te->y, te->w, te->h);
  }
}

static void textentry_position(text_entry_t *te,int x,int y,int w,int h)
{
  widget_t *W;
  te->x= x; te->y= y; te->w= w; te->h= h;
  W= &te->widget;
  W->x= x; W->y= y; W->w= w; W->h= h;
  make_cursor_visible(te);
}

static void mk_glyph(glyph_t *G, uint32_t chr)
{
  font_render(gui_fonts+0,&chr, 1, (int*) &G->w, &G->h, &G->a, &G->d);
  G->chr= chr;
}

static void textentry_focus(void *va, int focus)
{
  text_entry_t *te;
  te= va;
  if (!te->widget.shown) return ;
  if (te->widget.has_focus==focus) return ;
  te->widget.has_focus= focus;
  draw_glyph(te, te->cursor, 1);
}

static void text_entry_move_cursor(text_entry_t *te, int newcurs)
{
  int ws, we, gs,ge;
  int C;

  if (newcurs<0 || newcurs>=te->len || newcurs==te->cursor) return ;  

  gs= te->glyphs[newcurs].x;
  ge= gs + te->glyphs[newcurs].w;

  ws= te->left; 
  we= te->left+te->w;

  if (gs<ws|| ge>we)
  {
    te->cursor= newcurs;
    make_cursor_visible(te);
    return ;
  }

  C= te->cursor;
  te->cursor= newcurs;
  draw_glyph(te, C,1);
  draw_glyph(te, te->cursor,1);
}

static void delete_cursor_char(text_entry_t *te)
{
  int i,x;
  free(te->glyphs[te->cursor].d);
  x= te->glyphs[te->cursor].x;
  for(i=te->cursor;i<te->len-1;i++)
  {
    te->glyphs[i]= te->glyphs[i+1];
    te->glyphs[i].x= x;
    x+= te->glyphs[i].w;
  }
  te->len--;
  te->glyphs[i].x= x;
  make_cursor_visible(te);
}

static void insert_char(text_entry_t *te, uint32_t chr)
{
  int i;
  if (te->len==te->size)
  {
    te->size+= te->size/2;
    te->glyphs= realloc(te->glyphs, sizeof(te->glyphs[0])*te->size);
  }
  for(i=te->len;i>te->cursor;i--)
    te->glyphs[i]= te->glyphs[i-1];
  mk_glyph(te->glyphs+ te->cursor, chr);
  
  te->cursor++;
  te->len++;
  for(i=te->cursor;i<te->len;i++)
    te->glyphs[i].x= te->glyphs[i-1].x + te->glyphs[i-1].w;

  make_cursor_visible(te);
}

static int textentry_key(void* va, gfx_event_t *event) 
{
  text_entry_t *te;
  te= va;
    // this will not happen, hopefully.
  if (!te->widget.shown) return 0;
  if (te->key_filter && (*te->key_filter)(te->key_arg,va,event))
    return 1;
  
  switch(event->key)
  {
   case GFX_K_LEFT:
     if (te->cursor!=0) text_entry_move_cursor(te, te->cursor-1);
     return 1;
   case GFX_K_RIGHT:
     if (te->cursor!=te->len-1) text_entry_move_cursor(te,te->cursor+1);
     return 1;
   case GFX_K_HOME: text_entry_move_cursor(te, 0); return 1;
   case GFX_K_END:  text_entry_move_cursor(te, te->len-1); return 1;
   case GFX_K_DELETE:
     if (te->cursor!=te->len-1) delete_cursor_char(te);
     return 1;
   case GFX_K_BACKSPACE:
     if (te->cursor!=0)
     {
       te->cursor--;
       delete_cursor_char(te);
     }
     return 1;
   case GFX_K_PRT:
     insert_char(te, event->wide);
     return 1;
  }
  return 0;
}

static char *textentry_get_text(text_entry_t *te)
{
  uint32_t *wide, widelen;
  uint8_t *str;
  int i;
  size_t slen;

  widelen= te->len;
  wide= malloc(sizeof(uint32_t)*widelen);
  for(i=0;i<te->len;i++)
    wide[i]= te->glyphs[i].chr;

  str= malloc(widelen+1);
  utf8_encode_string(wide, widelen, str, &slen);
  if (slen) str[slen-1]= 0; // eat up the space we added
  free(wide);
  return (char*) str;
}


static void textentry_free(text_entry_t *te)
{
  int i;
  for(i=0;i<te->len;i++)
   free(te->glyphs[i].d);
  free(te->glyphs);
}


static void text_entry_set(text_entry_t *te, char *str)
{
  int i;
  uint32_t *wide_str, wide_str_len;
  int x;

  if (!str) str="";
  utf8_decode_alloc(str, &wide_str, &wide_str_len);

  if ((int) wide_str_len>=te->size)
  {
    te->size= wide_str_len + wide_str_len/2 + 2;
    if (te->size<50) te->size= 50;
    te->glyphs= realloc(te->glyphs,sizeof(te->glyphs[0])*te->size);
  }

  for(i=0;i<te->len;i++) free(te->glyphs[i].d);

  te->len= wide_str_len+1;
  
  for(i=0;i<te->len-1;i++) mk_glyph(te->glyphs+i, wide_str[i]);
  mk_glyph(te->glyphs+i, ' ');

  x= 0;
  for(i=0;i<te->len;i++)
  {
    te->glyphs[i].x= x;
    x+= te->glyphs[i].w;
  }
  
  free(wide_str);
  
  te->cursor= te->len-1;
  te->left= 0;
  if (te->widget.shown)
    make_cursor_visible(te);
}

static void gui_textentry_show(widget_t *arg)
{
  text_entry_t *te;
  te= (text_entry_t*) arg;
  te->widget.can_focus= 1;
  make_cursor_visible(te);
}

static void gui_textentry_hide(widget_t *arg)
{
  text_entry_t *te;
  te= (text_entry_t*) arg;
  if (te->widget.has_focus) focus_lost();
  te->widget.can_focus= 0;
  te->widget.has_focus= 0; 
}


static widget_t *text_entry_new(char *initstr)
{
  text_entry_t *te;
  int i;
  uint32_t *wide_str, wide_str_len;
  int size;
  int x;

  te= calloc(1,sizeof(*te));

  if (!initstr) initstr= "";
  utf8_decode_alloc(initstr, &wide_str, &wide_str_len);

  size= wide_str_len + wide_str_len/2 + 2;
  if (size<50) size= 50;

  te->size= size;
  te->len= wide_str_len+1;
  te->glyphs= malloc(sizeof(te->glyphs[0])*te->size);
  
  for(i=0;i<te->len-1;i++) mk_glyph(te->glyphs+i, wide_str[i]);
  mk_glyph(te->glyphs+i, ' ');
  free(wide_str);

  x= 0;
  for(i=0;i<te->len;i++)
  {
    te->glyphs[i].x= x;
    x+= te->glyphs[i].w;
  }
  
  get_text_data(1, "My", NULL, &te->fonth, &te->fonta, NULL);
  
  te->cursor= te->len-1;
  te->left= 0;

  te->widget.can_focus= 1;
  te->widget.has_focus= 0;
  te->widget.focus= textentry_focus;
  te->widget.handle_key= textentry_key;

  te->widget.klass= GFXCLASS_TEXTENTRY;
  te->widget.show= gui_textentry_show;
  te->widget.hide= gui_textentry_hide;

  return &te->widget;
}

/***********************************************************************
 *                                                                     *
 *                             PIXEL GRID                              *
 * #W.PIXGRID                                                          *
 ***********************************************************************
 The pixgrid widget displays a pixel buffer on the screen. The buffer is
 managed by the user and he uses gfx_put to display pixels.
 ***********************************************************************/

typedef struct
{
  widget_t widget;
  int click_handler;
} gui_pixgrid_t;

static int gui_pixgrid_release(void *arg,gfx_event_t *event)
{
  gui_pixgrid_t *B;
  B= arg;
  event->type = GFXEV(B->widget.number,BUTTON_RELEASE);
  event->x -= B->widget.x;
  event->y -= B->widget.y;
  return 0;
}

static int gui_pixgrid_click(void *arg,gfx_event_t *event)
{
  gui_pixgrid_t *B;
  B= arg;
  if (!B->widget.shown) return 0;
  event->type = GFXEV(B->widget.number,BUTTON_PRESS);
  event->x -= B->widget.x;
  event->y -= B->widget.y;
  install_release_handler(arg,gui_pixgrid_release);
  return 0;
}

static void gui_pixgrid_destroy(gui_pixgrid_t *G);
static void gui_pixgrid_position(gui_pixgrid_t *G,int x,int y,int w,int h);
static void gui_pixgrid_show(gui_pixgrid_t *G);
static void gui_pixgrid_hide(gui_pixgrid_t *G);


static widget_t* gfx_pixgrid()
{
  gui_pixgrid_t *B;
  B= calloc(1,sizeof(*B));

  B->widget.klass= GFXCLASS_PIXGRID;
  B->widget.destroy= (void*) gui_pixgrid_destroy;
  B->widget.position= (void*) gui_pixgrid_position;
  B->widget.show= (void*) gui_pixgrid_show;
  B->widget.hide= (void*) gui_pixgrid_hide;

  B->click_handler= install_click_handler(0,0,0,0,&B->widget.shown,B, 
                                          gui_pixgrid_click,1);
  return &B->widget;
}

static void gui_pixgrid_destroy(gui_pixgrid_t *G)
{
  remove_click_handler(G->click_handler);                                   
}      

static void gui_pixgrid_position(gui_pixgrid_t *G,int x,int y,int w,int h)
{
  G->widget.x= x;
  G->widget.y= y;
  G->widget.w= w;
  G->widget.h= h;
  set_click_handler(G->click_handler,x,y,w,h);
}

static void gui_pixgrid_show(gui_pixgrid_t *G)
{
  // nothing to do except for the flag update,
  // which was done before calling this method
}
static void gui_pixgrid_hide(gui_pixgrid_t *G)
{
  // nothing to do except for the flag update,
  // which was done before calling this method
}

static void gui_pixgrid_attribs
   (int pg, int *Rshown, unsigned int *Rw, unsigned int *Rh)
{
  gui_pixgrid_t *B;
  B= (gui_pixgrid_t*) get_widget(pg);
  if (!B || B->widget.klass!=GFXCLASS_PIXGRID) 
  {
    if (Rshown) *Rshown= 0;
    if (Rw) *Rw= 1;
    if (Rh) *Rh= 1;
  }
  else
  {
    if (Rshown) *Rshown= B->widget.shown;
    if (Rw) *Rw= B->widget.w;
    if (Rh) *Rh= B->widget.h;
  }
}

  /***
The following function gives direct access to the window buffer.
order contains the position of the red,green and blue channels
within each pixel. 
  ***/
int gfx_pixgrid_buffer
  (int pg, int *Rbpl, int *Rbpp, int *Rorder, unsigned char **Rbuffer)
{
  gui_pixgrid_t *B;
  B= (gui_pixgrid_t*) get_widget(pg);
  if (!B || B->widget.klass!=GFXCLASS_PIXGRID)  return 1;
#ifdef GFX_ANDROID
  if (Rbpl) *Rbpl= x_window_width * 4;
#endif
#ifdef GFX_X11
  if (Rbpl) *Rbpl= x_image->bpl;
#endif
#ifdef GFX_FB
  if (Rbpl) *Rbpl= x_window_width * 4;
#endif
#ifdef GFX_WIN
  if (Rbpl) *Rbpl = x_window_width * 4;
#endif
  if (Rbpp) *Rbpp= 4; // this library only works if this is true..
  if (Rbuffer) 
  {
    unsigned char *P;
#ifdef GFX_ANDROID
    P= (unsigned char*) x_image_data[0];
    *Rbuffer= P + (B->widget.y*x_window_width+B->widget.x)*4;
#endif
#ifdef GFX_X11
    P= (unsigned char*) x_image->data;
    *Rbuffer= P+B->widget.y*x_image->bpl+B->widget.x*4;
#endif
#ifdef GFX_FB
    P= (unsigned char*) x_image;
    *Rbuffer= P+(B->widget.y*x_window_width+B->widget.x)*4;
#endif
#ifdef GFX_WIN
	P = (unsigned char*)x_image_data[0];
	*Rbuffer = P + (B->widget.y * x_window_width + B->widget.x) * 4;
#endif
  }
  if (Rorder)
  {
    Rorder[0]= x_red_shift/8;
    Rorder[1]= x_green_shift/8;
    Rorder[2]= x_blue_shift/8;
  }
  return 0;
}

void gfx_pixgrid_update(int pg, int x,int y,int w, int h)
{
  gui_pixgrid_t *B;
  B= (gui_pixgrid_t*) get_widget(pg);
  if (!B || B->widget.klass!=GFXCLASS_PIXGRID)  return ;
  gfx_update(B->widget.x+x, B->widget.y+y, w,h);
}

void gfx_pixgrid_put(int pg, int x,int y,
                     unsigned int *img, unsigned int iW,unsigned int iH,
                     int src_x, int src_y, 
                     int copy_w, int copy_h)
{
  gui_pixgrid_t *B;
  B= (gui_pixgrid_t*) get_widget(pg);
  if (!B || B->widget.klass!=GFXCLASS_PIXGRID || !B->widget.shown)  return ;

  clip_for_rect(&src_x, &copy_w, iW);
  clip_for_rect(&src_y, &copy_h, iH);
  clip_for_rect(&x, &copy_w, B->widget.w);
  clip_for_rect(&y, &copy_h, B->widget.h);
  if (copy_w<=0 || copy_h<=0) return ;
  
  gfx_put(GFX_UPDATE,img,iW,iH,src_x,src_y,
                B->widget.x+x,B->widget.y+y, copy_w, copy_h);
}

void gfx_pixgrid_put_gray(int pg, int x,int y,
                     unsigned char *img, unsigned int iW,unsigned int iH,
                     int src_x, int src_y, 
                     int copy_w, int copy_h)
{
  gui_pixgrid_t *B;
  B= (gui_pixgrid_t*) get_widget(pg);
  if (!B || B->widget.klass!=GFXCLASS_PIXGRID || !B->widget.shown)  return ;

  clip_for_rect(&src_x, &copy_w, iW);
  clip_for_rect(&src_y, &copy_h, iH);
  clip_for_rect(&x, &copy_w, B->widget.w);
  clip_for_rect(&y, &copy_h, B->widget.h);
  if (copy_w<=0 || copy_h<=0) return ;
  
  gfx_put_gray(GFX_UPDATE,img,iW,iH,src_x,src_y,
                B->widget.x+x,B->widget.y+y, copy_w, copy_h);
}

void gfx_pixgrid_putsimple(int pg, unsigned int *img)
{
  gui_pixgrid_t *B;
  B= (gui_pixgrid_t*) get_widget(pg);
  if (!B || B->widget.klass!=GFXCLASS_PIXGRID || !B->widget.shown)  return ;

  gfx_pixgrid_put(pg, 0,0,
                  img, B->widget.w, B->widget.h,
                  0,0, B->widget.w, B->widget.h);
}

void gfx_pixgrid_fill_rect(int pg, int x, int y, int w, int h, int color)
{
  gui_pixgrid_t *B;
  B= (gui_pixgrid_t*) get_widget(pg);
  if (!B || B->widget.klass!=GFXCLASS_PIXGRID || !B->widget.shown)  return ;

  clip_for_rect(&x, &w, B->widget.w);
  clip_for_rect(&y, &h, B->widget.h);
  if (w<=0 || h<=0) return ;
 
  gfx_fill_rect(GFX_UPDATE, B->widget.x + x, B->widget.y+y, w, h, color);
}



static int gui_pixgrid_report_motion(void *arg, gfx_event_t *ev)
{
  gui_pixgrid_t *CC;
  CC= arg;
  if (ev->x < CC->widget.x ||
      ev->x >= CC->widget.x+CC->widget.w ||
      ev->y < CC->widget.y ||
      ev->y >= CC->widget.y+CC->widget.h)
    return 1;
  ev->type= GFXEV(CC->widget.number,POINTER_MOVE);
  ev->x= ev->x - CC->widget.x;
  ev->y= ev->y - CC->widget.y;
  return 0;
}

void gfx_pixgrid_begin_report_mouse_position(int wno)
{
  gui_pixgrid_t *CC;
  if (!(CC= (gui_pixgrid_t*) get_widget(wno))) return ;
  if (CC->widget.klass!=GFXCLASS_PIXGRID) return ;
  install_motion_handler(CC, gui_pixgrid_report_motion);
}

void gfx_pixgrid_end_report_mouse_position(int wno)
{
  gui_pixgrid_t *CC;
  if (!(CC= (gui_pixgrid_t*) get_widget(wno))) return ;
  if (CC->widget.klass!=GFXCLASS_PIXGRID) return ;
  uninstall_motion_handler();
}

/***********************************************************************
 *                                                                     *
 *                             SCROLL BAR                              *
 * #W.SCROLL                                                           *
 ***********************************************************************
A vertical scrollbar has several areas:

   R= total length of the trough =  h - Uh - Dh
  
    ----
    |  |  Up button              Uy= 0
    ----
    ....  Trough                 Ky= Uh
    ....
    ====    ----------------->   Py0= start*R + Uh
    |  |  Position indicator     
    |  |
    ====    ----------------->   Py1= end*R + Uh
    ....
    ....  Trough
    ....
    ----
    |  |  Down button            Dy= h - Dh
    ----
 ***********************************************************************/

typedef struct
{
  widget_t widget;
  double start, end;          // start/end points for the sliding button
  double step;                // step amount for up/down left/right buttons
  int tro_c0, tro_c1, tro_dc; // start/end points for trough. can be X or 
                              // Y coordinates dc is the difference between
                              // them, but normalized to prevent divide by zero
  int but_c0, but_c1;         // start/end points for button

  int tro_x, tro_y, tro_w, tro_h; // used for updates
  int last_seen_c;

  int enabled;                // for enabling/disabling clicks

  int orient;

  int action;                 // action is used for reporting changes to user.
  void *reportarg;            // the following two are used for reporting
                              // to other widgets.
  void (*reportfunc)(void *, double, double);
  
  // stepper button position
  int SX[2], SY[2];
  // stepper button images, 0 = Up-normal 1= Up-pressed 2-Down-normal ..
  uint32_t *SD[4], SW[4], SH[4];

    // click handler entries
    // 0= up button 
    // 1= down button
    // 2= trough
  int ch0, ch1, ch2;
} scrollbar_t;


/* The following has two functions: when called with a NULL scrollbar,
   it sets the gui_scrollbar_width/_height variables so that the layout
   engines can do their stuff. If SB is not NULL, then it assigns the
   images to the scrollbar, depending on the orientation */

static void scrollbar_init_buttons (scrollbar_t *SB)
{
   int i;
   static uint32_t *button_up[2], *button_down[2],  // 0= normal 1=depressed
                   *button_left[2], *button_right[2];

   if (SB==NULL)
   {
     // the normal button is the up button, the width for that gives us the
     // width of a vertical scrollbar. When it's rotated, the width becomes
     // the height and the button becomes a right button. The height of the
     // right button is the height of a horizontal scrollbar.
     gui_scrollbar_width= gui_scrollbar_height= RES(scrollbar.normal_button).w;
     return ;
   }

   if (!button_up[0])
   {
     button_up[0]= gui_R_COLORIMG_toserver(&RES(scrollbar.normal_button));
     button_up[1]= gui_R_COLORIMG_toserver(&RES(scrollbar.depressed_button));

     button_right[0]= gui_rotate_icon_cw(button_up[0]);
     button_right[1]= gui_rotate_icon_cw(button_up[1]);

     button_down[0]= gui_rotate_icon_cw(button_right[0]);
     button_down[1]= gui_rotate_icon_cw(button_right[1]);

     button_left[0]= gui_rotate_icon_cw(button_down[0]);
     button_left[1]= gui_rotate_icon_cw(button_down[1]);
   }

   if (SB->orient=='V')
   {
     SB->SD[0]= button_up[0]+2;   SB->SD[1]= button_up[1]+2;
     SB->SD[2]= button_down[0]+2; SB->SD[3]= button_down[1]+2;
   }
   else
   {
     SB->SD[0]= button_left[0]+2;  SB->SD[1]= button_left[1]+2;
     SB->SD[2]= button_right[0]+2; SB->SD[3]= button_right[1]+2;
   }
   for(i=0;i<4;i++)
   {
     SB->SW[i]= SB->SD[i][-2];
     SB->SH[i]= SB->SD[i][-1];
   }
}

static void scrollbar_compute_trough(scrollbar_t *SB)
{
  if (SB->orient=='V')
  {
    SB->tro_h= SB->widget.h - SB->SH[0]-SB->SH[2];

    SB->tro_c0= SB->widget.y+ SB->SH[0];
    SB->tro_c1= SB->tro_c0+SB->tro_h-1;
    SB->tro_dc= SB->tro_h-1;

    SB->tro_x= SB->widget.x;
    SB->tro_y= SB->tro_c0;
    SB->tro_w= SB->widget.w;
  }
  else
  {
    SB->tro_w= SB->widget.w - SB->SW[0]-SB->SW[2];

    SB->tro_c0= SB->widget.x + SB->SW[0];
    SB->tro_c1= SB->tro_c0 + SB->tro_w - 1;
    SB->tro_dc= SB->tro_w-1;

    SB->tro_x= SB->tro_c0;
    SB->tro_y= SB->widget.y;
    SB->tro_h= SB->widget.h;
  }
  if (SB->tro_dc<1) SB->tro_dc= 1;
}

static void scrollbar_compute_button(scrollbar_t *SB)
{
  SB->but_c0= SB->tro_dc*SB->start + SB->tro_c0;
  SB->but_c1= SB->tro_dc*SB->end   + SB->tro_c0;
}

static void fill_with_horizontal_gradient
             (int x0,int y0, int y1, unsigned int *gra)
{
  int y;
  int graW;
  graW= gra[0];
  gra+= 2;
  for(y=y0;y<=y1;y++)
    put_server_pix(gra, x0, y, graW, 1);
}

static void three_part_assembly_vertical
    (int x0,int y0, int y1, 
     uint32_t *top, uint32_t *mid, uint32_t *bot)
{
  int mid_start_y, mid_end_y;

  uint32_t *topD, topW, topH;
  uint32_t *botD, botW, botH;

  topD= top+2; topW= top[0]; topH= top[1];
  botD= bot+2; botW= bot[0]; botH= bot[1];
  
  mid_start_y= y0 + topH;
  mid_end_y=   y1 - botH;
  put_server_pix(topD, x0, y0, topW, topH);
  fill_with_horizontal_gradient(x0,mid_start_y, mid_end_y, mid);
  put_server_pix(botD, x0, mid_end_y+1, botW, botH);
}

static void fill_with_vertical_gradient
             (int x0, int x1, int y0, unsigned int *gra)
{
  int x;
  int graH;
  graH= gra[1];
  gra+= 2; // skip the W+H
  for(x=x0;x<=x1;x++)
    put_server_pix(gra, x, y0, 1, graH);
}

static void three_part_assembly_horizontal
    (int x0, int x1, int y0, 
     uint32_t *lef, uint32_t *mid, uint32_t *rig)
{
  int mid_start_x, mid_end_x;

  uint32_t *lefD, lefW, lefH,
           *rigD, rigW, rigH;

  lefW= lef[0]; lefH= lef[1]; lefD= lef+2;
  rigW= rig[0]; rigH= rig[1]; rigD= rig+2;

  mid_start_x= x0 + lefW;
  mid_end_x=   x1 - rigW;
  put_server_pix(lefD, x0, y0, lefW, lefH);
  fill_with_vertical_gradient(mid_start_x,mid_end_x, y0, mid);
  put_server_pix(rigD, mid_end_x+1, y0, rigW, rigH);
}

static void scrollbar_draw_trough(scrollbar_t *SB)
{
  static uint32_t *Vind_start, *Vind_mid, *Vind_end, *Vtro_grad,
                  *Hind_start, *Hind_mid, *Hind_end, *Htro_grad;

  int y,x;
  if (!SB->enabled) return ;

  if (!Vind_start)
  {
    // The resource image is the top part of the vertical scrollbar.
    // when you rotate it, it becomes the right part of the horizontal.
    //     "                     "    the bottom part of the vertical
    //     "                     "    the left part of the horizontal
    //
    Vind_start= gui_R_COLORIMG_toserver(&RES(scrollbar.ind_start));
    Hind_end= gui_rotate_icon_cw(Vind_start);
    Vind_end= gui_rotate_icon_cw(Hind_end);
    Hind_start= gui_rotate_icon_cw(Vind_end);

    Vind_mid= gui_R_COLORIMG_toserver(&RES(scrollbar.ind_mid));
    Hind_mid= gui_rotate_icon_cw(Vind_mid);

    Vtro_grad= gui_R_COLORIMG_toserver(&RES(scrollbar.tro_grad));
    Htro_grad= gui_rotate_icon_cw(Vtro_grad);
  }

  if (SB->orient=='V')
  {
    x= SB->widget.x;
    fill_with_horizontal_gradient
             (x, SB->tro_c0, SB->but_c0-1, Vtro_grad);

    fill_with_horizontal_gradient
             (x, SB->but_c1+1, SB->tro_c1, Vtro_grad);

    three_part_assembly_vertical
             (x, SB->but_c0, SB->but_c1,
              Vind_start, Vind_mid, Vind_end);
  }
  else
  {
    y= SB->widget.y;
    fill_with_vertical_gradient
             (SB->tro_c0, SB->but_c0-1, y, Htro_grad);

    fill_with_vertical_gradient
             (SB->but_c1+1, SB->tro_c1, y, Htro_grad);

    three_part_assembly_horizontal
             (SB->but_c0, SB->but_c1, y,
              Hind_start, Hind_mid, Hind_end);
  }
}

static void scrollbar_draw_stepper(scrollbar_t *SB, int sno, int state)
{
  int img;
  img= sno*2 + state;
  if (!SB->enabled) return ;
  put_server_pix(SB->SD[img], SB->SX[sno], SB->SY[sno],
                              SB->SW[img], SB->SH[img]);
}

static void scrollbar_draw_all(scrollbar_t *SB)
{
  scrollbar_draw_stepper(SB, 0,0);
  scrollbar_draw_stepper(SB, 1,0);
  scrollbar_draw_trough(SB);
}
static void scrollbar_position_buttons(scrollbar_t *SB,int x,int y,int w,int h)
{
  if (SB->orient=='V')
  {
    SB->SX[0]= x; SB->SX[1]= x;
    SB->SY[0]= y; SB->SY[1]= y+h-SB->SH[2];
  }
  else
  {
    SB->SY[0]= y;  SB->SY[1]= y;
    SB->SX[0]= x;  SB->SX[1]= x+w-SB->SW[2];
  }
}


static void scrollbar_update(scrollbar_t *SB)
{
  scrollbar_compute_button(SB);
  scrollbar_draw_trough(SB);
  if (SB->enabled)
    gfx_update(SB->tro_x, SB->tro_y, SB->tro_w, SB->tro_h);
}

static int scrollbar_set
    (scrollbar_t *SB,double S, double E, int report, gfx_event_t* event)
{
  if (SB->start==S && SB->end==E) return 1;
  SB->start= S;
  SB->end= E;
  scrollbar_update(SB);
  if (!report) return 1;
  if (SB->action && event) 
  {
    event->type= SB->widget.number;
    event->flo0= S;
    event->flo1= E;
    return 0;
  }
  if (SB->reportfunc)
    (*SB->reportfunc)(SB->reportarg, S, E);
  return 1;
}


static int scrollbar_release_up(void *A, gfx_event_t *event)
{
  scrollbar_t *SB= A;
  scrollbar_draw_stepper(SB, 0, 0);
  gfx_update(SB->SX[0], SB->SY[0], SB->SW[0], SB->SH[0]);
  if (SB->start>0.0)
  {
    double S;
    S= dmax2(0.0, SB->start - SB->step);
    return scrollbar_set(SB, S, S-SB->start+SB->end, 1, event);
  }
  return 1;
}

static int scrollbar_click_up(void *A, gfx_event_t *E)
{
  scrollbar_t *SB= A;
  scrollbar_draw_stepper(SB, 0, 1);
  install_release_handler(A, scrollbar_release_up);
  gfx_update(SB->SX[0], SB->SY[0], SB->SW[0], SB->SH[0]);
  return 1;
}


static int scrollbar_release_down(void *A, gfx_event_t *event)
{
  scrollbar_t *SB= A;
  scrollbar_draw_stepper(SB, 1, 0);
  gfx_update(SB->SX[1],SB->SY[1], SB->SW[2], SB->SH[2]);
  if (SB->end<1.0)
  {
    double E;
    E= dmin2(1.0,SB->end + SB->step);
    return scrollbar_set(SB,E-SB->end+SB->start,E, 1, event);
  }
  return 1;
}

static int scrollbar_click_down(void *A, gfx_event_t *E)
{
  scrollbar_t *SB= A;
  scrollbar_draw_stepper(SB, 1, 1);
  gfx_update(SB->SX[1],SB->SY[1], SB->SW[2], SB->SH[2]);
  install_release_handler(A, scrollbar_release_down);
  return 1;
}

static int scrollbar_release_dummy(void *A, gfx_event_t *E)
{
  // this function is used to eat up the dummy release events
  // left over from clicks into the trough area. 
  // normally, I'd do the returning stuff within this
  // release handler, but there is too much information to
  // pass around.
  return 1;
}

static int scrollbar_release_tro(void *A, gfx_event_t *E)
{
  uninstall_motion_handler();
  return 1;
}

static int scrollbar_move_pointer(void *A, gfx_event_t *event)
{
  scrollbar_t *SB= A;
  int D;
  double L, S, E;
  int C;

  if (SB->orient=='V') C= event->y; else C= event->x;

  if (C>SB->tro_c1 || C<SB->tro_c0) return 1;

  D= C - SB->last_seen_c;
  if (D==0) return 1;

  SB->last_seen_c= C;
 
  L= SB->end - SB->start;
  if (D>0)
  {
    if (SB->end==1.0) return 1;
    E= dmin2(1.0,SB->end + (double) D / SB->tro_dc);
    S= E - L;
  } else {
    if (SB->start==0.0) return 1;
    S= dmax2(0.0, SB->start + (double) D / SB->tro_dc);
    E= S + L;
  }

  return scrollbar_set(SB, S,E, 1, event);
}

static int scrollbar_click_tro(void *A,gfx_event_t *event)
{
  scrollbar_t *SB= A;
  double L,K, S, E;
  int C;

  if (SB->orient=='V') C= event->y; else C= event->x;
  if (C>=SB->but_c0 && C<=SB->but_c1)
  {
    install_release_handler(A, scrollbar_release_tro);
    install_motion_handler(A, scrollbar_move_pointer);
    SB->last_seen_c= C;
    return 1;
  }
  if (C<SB->tro_c0 || C>SB->tro_c1) return 1;
  C -= SB->tro_c0;
  L= SB->end - SB->start;

  K= (double) C / SB->tro_dc;
  S= K-L/2;
  E= S+L;

  if (S<0.0) { S= 0.0; E= S+L; }
  if (E>1.0) { E= 1.0; S= E-L;  }

  install_release_handler(A, scrollbar_release_dummy);
  return scrollbar_set(SB,S,E,1,event);
}



static void scrollbar_set_all
  (scrollbar_t *SB, double start, double end, double step)
{
  SB->start= dmin2(1.0,dmax2(0.0, start));
  SB->end=   dmax2(SB->start+0.000001,dmin2(1.0,dmax2(0.0, end)));
  SB->step=  dmin2(1.0,dmax2(0.0,step));
  scrollbar_update(SB);
}

void gfx_scrollbar_set(int wno, double start, double end, double step)
{
  scrollbar_t *SB;
  SB= (scrollbar_t*) get_widget(wno);
  if (!SB|| SB->widget.klass!= GFXCLASS_SCROLLBAR) return ;
  scrollbar_set_all(SB, start, end, step);
}


static void scrollbar_position(scrollbar_t *SB,int x,int y,int w,int h)
{
  SB->widget.w= w;
  SB->widget.h= h;
  SB->widget.x= x;
  SB->widget.y= y;

  scrollbar_position_buttons(SB,x,y,w,h);
  scrollbar_compute_trough(SB);
  scrollbar_compute_button(SB);
  scrollbar_draw_all(SB);

  SB->ch0= install_or_modify_click
        (SB->ch0, &SB->enabled, scrollbar_click_up, SB,
        SB->SX[0], SB->SY[0], SB->SW[0], SB->SH[0],1);

  SB->ch1= install_or_modify_click
        (SB->ch1, &SB->enabled, scrollbar_click_down, SB,
        SB->SX[1], SB->SY[1], SB->SW[2], SB->SH[2],1);
 
  SB->ch2= install_or_modify_click
        (SB->ch2, &SB->enabled, scrollbar_click_tro, SB,
         SB->tro_x, SB->tro_y, SB->tro_w, SB->tro_h,1);
}

               /********************************
Now we can implement the move-resize function. This is necessary in case
of two scrollbars being used depending on the content size. If only the
horizontal one is present, then it should cover all the widget width.
If both are present, then the horizontal bar should cover the widget width
minus the vertical bar width. You get the idea.
               ********************************/

static void scrollbar_relayout
   (scrollbar_t *SB, int nx, int ny, int nw,int nh,
    double start, double end, double step)
{
  SB->start= dmin2(1.0,dmax2(0.0, start));
  SB->end= dmax2(0.0,dmin2(1.0,end));
  SB->step= step;
  if (step<0.0 || step>=1.0) SB->step= 0.1;

  if (nw!=SB->widget.w || nh!=SB->widget.h 
      || nx!=SB->widget.x || ny!=SB->widget.y)
  {
    scrollbar_position(SB,nx,ny,nw,nh);
  }
  else
  {
    scrollbar_compute_trough(SB);
    scrollbar_compute_button(SB);
    scrollbar_draw_all(SB);
  }
}

static void scrollbar_destroy(scrollbar_t *SB)
{
  remove_click_handler(SB->ch0);
  remove_click_handler(SB->ch1);
  remove_click_handler(SB->ch2);
}

static void scrollbar_show(scrollbar_t *SB,int show)
{
  SB->enabled= show;
  if (SB->enabled) 
  {
    scrollbar_draw_all(SB);
    gfx_update(SB->widget.x, SB->widget.y, SB->widget.w, SB->widget.h);
  }
}

static widget_t* scrollbar_new
    (int orient, double start, double end, double step,
     int action, void *reportarg, void *reportfunc)
{
  scrollbar_t *SB;

  SB= calloc(1,sizeof(*SB));

  SB->start= start;
  SB->end= end;
  SB->step= step;
  SB->reportarg= reportarg;
  SB->reportfunc= reportfunc;
  SB->action= action;

  SB->orient= orient ? 'V': 'H';
  SB->enabled= 0;
  SB->widget.klass= GFXCLASS_SCROLLBAR;

  SB->ch0= SB->ch1= SB->ch2= -1;

  scrollbar_init_buttons(SB);

  return &SB->widget;
}


/***********************************************************************
 *                                                                     *
 *                                LIST                                 *
 * #W.LIST                                                             *
 ***********************************************************************
 Icon list displays text with an icon next to it. I will fix the icon 
 position to the left of the item.

 Icons will be represented as mini buffers, ie. argb32 pictures. Managing
 the icon memory is the responsibility of the user. This makes it easier
 to provide a limited amount of icons to the list widget, without ever
 deallocating them etc.

 The list will contain the items in its own struct. This won't be visible
 from the outside.
 ***********************************************************************/

typedef struct
{
  char *str;
  const unsigned int *icon_data;
  unsigned int icon_width, icon_height;
  int selected;
  void *user;

  int y;
  int w,h;
  int icon_y,icon_x,text_y,text_x;
  int text_w, text_h;
} list_item_t;

typedef struct
{
  widget_t widget;
  int n_items, s_items;
  list_item_t *items;

  int posx, posy; // scrolled position

  int has_scrollv, has_scrollh;
  scrollbar_t *scrollv, *scrollh;

  int content_w, content_h;
  int port_w, port_h;
  int port_x, port_y;

  int max_icon_width;
  int uses_icons;
  int multiselect;

  int enabled;
  int ch0;    // ID of the click handler, we use this
              // to modify the click area when the
              // scrollbars appear/disappear
  int ch1,ch2; // mouse buttons 4 and 5

    // for double click
  unsigned long click_time;
  int click_index;
 
  int x,y,w,h;

  void *evarg;
  event_function_t clickf,dlickf;
} gui_list_t;

static void gui_list_set_position(gui_list_t* LL,int x,int y,int w,int h)
{
    LL->x= x; LL->y= y; LL->w= w; LL->h= h;
    LL->port_x= LL->x+1;
    LL->port_y= LL->y+1;
    LL->widget.x= x;
    LL->widget.y= y;
    LL->widget.w= w;
    LL->widget.h= h;
}



static int gui_list_pad_horz= 2,
           gui_list_pad_vert= 2,
           gui_list_icontext_sep= 2,
           gui_list_spacing= 2,
           gui_list_border= 0x888888;

static void compute_list_items
    (gui_list_t* LL, int start, int end)
{
  int tw,th;
  int i;

  start= imax2(0,start);
  end= imin2(end,LL->n_items-1);

  for(i=start;i<=end;i++)
  {
    list_item_t *item= LL->items+ i;

    get_text_data(1, item->str, &tw, &th, NULL, NULL);
    if (LL->uses_icons)
      item->text_x= gui_list_pad_horz+gui_list_icontext_sep + LL->max_icon_width;
    else
      item->text_x= gui_list_pad_horz;

    item->w= item->text_x + tw + gui_list_pad_horz;

    if (LL->uses_icons&&item->icon_data)
    {
      item->h= imax2(item->icon_height, th);
      item->icon_y= (item->h-item->icon_height)/2;
      item->icon_x= (LL->max_icon_width-item->icon_width)/2 + gui_list_pad_horz;
    }
    else
    {
      item->h= th;
    }
    item->text_y= (item->h-th)/2;

    item->text_w= tw;
    item->text_h= th;
  }
}

static void compute_list_content_size(gui_list_t *LL,int *Rw, int *Rh)
{
  int i;
  int w,h;
  w= h= 0;
  for(i=0;i<LL->n_items;i++)
  {
    h+= LL->items[i].h;
    if (i>0 && i<LL->n_items-1) h+= gui_list_spacing;
    w= imax2(w, LL->items[i].w);
  }
  h+= 2*gui_list_pad_horz;
  if (Rw) *Rw= w;
  if (Rh) *Rh= h;
}

static int list_max_icon_width(gui_list_t *LL)
{
  int i,w;
  w= 0;
  if (!LL->uses_icons) return 0;
  for(i=0;i<LL->n_items;i++)
    if (LL->items[i].icon_data)
      w= imax2(w, LL->items[i].icon_width);
  return w;
}

static void compute_list_port
  (gui_list_t* LL, int *Rportw, int *Rporth, int *Rhs, int *Rvs)
{
  int portw,porth, wA,wB, hA, hB;

  wA= LL->w - 2 ;
  wB= LL->w - 1 - gui_scrollbar_width;

  hA= LL->h - 2; 
  hB= LL->h - 1 - gui_scrollbar_height;

  portw= wA;
  porth= hA;
  if (portw<LL->content_w) porth= hB;
  if (porth<LL->content_h) portw= wB;
  if (portw<LL->content_w) porth= hB;
 
  if (Rportw) *Rportw= portw;
  if (Rporth) *Rporth= porth;
  if (Rhs)    *Rhs= (porth==hB);
  if (Rvs)    *Rvs= (portw==wB);
}

static int list_render_item(gui_list_t *LL,int ino,int bg)
{
  list_item_t *item= LL->items+ino;

  int src_x,src_y, copy_w, copy_h;
  unsigned char *tdata;
  
  int cW = item->icon_width,
      cH = item->icon_height,
      cX = item->icon_x,
      cY = item->icon_y + item->y;

  int tW= item->text_w,
      tH= item->text_h,
      tX= item->text_x,
      tY= item->text_y + item->y;

  int iY= item->y,
      iH= item->h;
      //  iX= 0, iW= item->w,

  int sY= LL->posy, 
      sX= LL->posx,
      sW= LL->port_w,
      sH= LL->port_h;

  if ((iY+iH)<=sY)  return -1;
  if (iY>=(sY+sH))  return 1;

  if (item->selected || bg)
  {
    int bY= iY-sY+LL->port_y,
        bW= LL->port_w,
        bX= LL->port_x,
        bH= iH;
    int color;

    if (bY<LL->port_y) 
    {
      bH-= LL->port_y-bY;
      if (bH<=0) goto bg_done;
      bY= LL->port_y;
    }

    if ((bY+bH)>(LL->port_y+LL->port_h))
    {
      bH= (LL->port_y+LL->port_h) - bY;
      if (bH<=0) goto bg_done;
    }
    if (item->selected) color= 0xaabbff;
             else color= 0xffffff;
    gfx_fill_rect(0, bX, bY, bW, bH, color);
  }

bg_done:
  if (!LL->uses_icons || !item->icon_data) goto icon_done;

  src_x= 0;
  src_y= 0;
  copy_w= cW;
  copy_h= cH;

  if (cX<sX)
  {
    src_x += sX-cX;
    copy_w-= sX-cX;
    cX= sX;
  }

  if (cY<sY)
  {
    src_y+= sY-cY;
    copy_h-= sY-cY;
    cY= sY;
  }

  if ((cX+copy_w)>(sX+sW))
    copy_w-= (cX+copy_w) - (sX+sW);

  if ((cY+copy_h)>(sY+sH))
    copy_h-= (cY+copy_h) - (sY+sH);

  if (copy_w<=0)   goto icon_done;
  if (copy_h<=0)   goto icon_done;
  if (cY>=(sY+sH)) goto icon_done;
  if (cX>=(sX+sW)) goto icon_done;

  // now, cX and cY contain the document coordinates for copy.

  gfx_put_alpha(item->icon_data, cW, cH, src_x, src_y,
                cX-sX+LL->port_x, cY-sY+LL->port_y,
                copy_w, copy_h);
icon_done:
  src_x= 0;
  src_y= 0;
  copy_w= tW;
  copy_h= tH;

  if (tX<sX)
  {
    src_x += sX-tX;
    copy_w-= sX-tX;
    tX= sX;
  }

  if (tY<sY)
  {
    src_y += sY-tY;
    copy_h-= sY-tY;
    tY= sY;
  }

  if ((tX+copy_w)>(sX+sW))
  {
    copy_w -= (tX+copy_w)-(sX+sW);
  }
  if ((tY+copy_h)>(sY+sH))
  {
    copy_h -= (tY+copy_h)-(sY+sH);
  }

  if (copy_h<=0) goto text_done;
  if (copy_w<=0) goto text_done;
  if (tY>=(sY+sH)) goto text_done;
  if (tX>=(sX+sW)) goto text_done;

  get_text_data(1, item->str, NULL, NULL, NULL,&tdata);
  blend_graymap(0, 0x00, 0xffffff, tdata, tW, tH, src_x, src_y,
                             tX-sX+LL->port_x, tY-sY+LL->port_y,
                             copy_w, copy_h);
  free(tdata);

text_done:
  return 0;
}



static void list_layout_items(gui_list_t *LL,int start)
{
  int i,y;
  start= imax2(0,start);
  if (start==0) y= gui_list_pad_vert;
           else y= LL->items[start-1].y+LL->items[start-1].h+gui_list_spacing;
 
  for(i=start;i<LL->n_items;i++)
  {
    LL->items[i].y= y;
    y+= LL->items[i].h + gui_list_spacing;
  }
}

static void list_draw_borders(gui_list_t *LL,int which)
{
  int x,y,w,h;
  unsigned int border;
  x= LL->x; y= LL->y; w= LL->w; h= LL->h;

  border= color_u_to_s(gui_list_border);
  if (which & 8) hline(x, x+w-1, y, border);
  if (which & 2) hline(x, x+w-1, y+h-1, border);
  if (which & 4) vline(x, y, y+h-1, border);
  if (which & 1) vline(x+w-1, y, y+h-1, border);
}

static void list_draw_contents(gui_list_t *L)
{
  int i,j,k;
  gfx_fill_rect(0,L->port_x,L->port_y,L->port_w,L->port_h, 0xffffff);
  if (L->n_items==0) return ;
  i= 0; 
  j= L->n_items-1;
  while(i<=j)
  {
    k= (i+j)/2;
    switch(list_render_item(L,k,0))
    {
     case 1: j= k-1; break;
     case 0: goto found;
     case -1: i= k+1; break;
    }
  }

  return;
found:
  for(i=k-1;i>=0;i--)         if (list_render_item(L,i,0)) break;
  for(j=k+1;j<L->n_items;j++) if (list_render_item(L,j,0)) break;
}

static void gui_list_scrollbar_click_horz(void *a, double S, double E)
{
  int posx;
  gui_list_t *L;
  L= a;

  posx= S*L->content_w;
  if (posx+L->port_w>L->content_w)
    posx= L->content_w - L->port_w;
  if (posx<0) posx= 0;
  if (posx!=L->posx)
  {
    L->posx= posx;
    list_draw_contents(L);
    gfx_update(L->port_x,L->port_y,L->port_w,L->port_h);
  }
}

static void gui_list_scrollbar_click_vert(void *a, double S, double E)
{
  int posy;
  gui_list_t *L;
  L= a;

  posy= S*L->content_h;
  if (posy+L->port_h>L->content_h)
    posy= L->content_h - L->port_h;
  if (posy<0) posy= 0;
  if (posy!=L->posy)
  {
    L->posy= posy;
    list_draw_contents(L);
    gfx_update(L->port_x,L->port_y,L->port_w,L->port_h);
  }
}


static void list_compute_scroll
    (int pos,int win,int con, double *Rstart, double *Rend, double *Rstep)
{
  if (con<=0) con= 1;
  *Rstart=  dmax2(0.0,dmin2((double)pos/con,1.0));
  *Rend=    dmin2(1.0,dmax2((double)(pos+win)/con,0.0));
  *Rstep=   dmax2(0.0,dmin2(1.0,(double)win/con*0.9));
}

static void list_arrange_scrollbars
   (gui_list_t *LL, int have_horz, int have_vert)
{
  int ver_scroll_h, hor_scroll_w;
  double start,end,step;

  hor_scroll_w= LL->w;
  if (have_vert) hor_scroll_w-= gui_scrollbar_width;
  ver_scroll_h= LL->h;
  if (have_horz) ver_scroll_h-= gui_scrollbar_height;

  switch((have_vert ? 2: 0) | (LL->has_scrollv ?1:0))
  {
  case 0: break;
  case 1: list_draw_borders(LL, 1); break;
  case 3:
  case 2: 
    list_compute_scroll(LL->posy, LL->port_h, LL->content_h,
                        &start,&end,&step);
    if (LL->scrollv)
      scrollbar_relayout(LL->scrollv, 
                   LL->scrollv->widget.x, LL->scrollv->widget.y,
                   gui_scrollbar_width, ver_scroll_h, start, end,step);
    else {
      LL->scrollv= (scrollbar_t*)
        scrollbar_new(1, start,end,step,0, LL, gui_list_scrollbar_click_vert);
      scrollbar_position(LL->scrollv, LL->x+LL->w-gui_scrollbar_width, LL->y,
                          gui_scrollbar_width, ver_scroll_h);
    }
    scrollbar_show(LL->scrollv,1);
    break;
  }

  switch( (have_horz?2:1)  | (LL->has_scrollh?1:0))
  {
  case 0: break;
  case 1: list_draw_borders(LL,2); break;
  case 3:
  case 2: 
    list_compute_scroll(LL->posx, LL->port_w, LL->content_w,
                        &start,&end,&step);
    if (LL->scrollh)
      scrollbar_relayout(LL->scrollh, 
                   LL->scrollh->widget.x, LL->scrollh->widget.y,
                   hor_scroll_w, gui_scrollbar_height, start,end,step);
    else 
    {
      LL->scrollh= (scrollbar_t*)
        scrollbar_new(0, start,end,step,0, LL, gui_list_scrollbar_click_horz);
      scrollbar_position(LL->scrollh, 
             LL->x,LL->y+LL->h-gui_scrollbar_height,
             hor_scroll_w, gui_scrollbar_height);
    }
    scrollbar_show(LL->scrollh,1);
    break;
  }

  if (LL->scrollv) LL->scrollv->enabled= have_vert;
  if (LL->scrollh) LL->scrollh->enabled= have_horz;
    
  if (have_vert && have_horz)
  {
    if (!LL->has_scrollv || !LL->has_scrollh)
      gfx_fill_rect(0, LL->scrollh->widget.x+ LL->scrollh->widget.w,
                       LL->scrollv->widget.y+ LL->scrollv->widget.h,
                       gui_scrollbar_width, gui_scrollbar_height, 0xdddddd);
  }

  if (have_horz)
    gfx_update(LL->x,LL->scrollh->widget.y, LL->w, gui_scrollbar_height);
  if (have_vert)
    gfx_update(LL->scrollv->widget.x, LL->y, gui_scrollbar_width, LL->h);
  LL->has_scrollh= have_horz;
  LL->has_scrollv= have_vert;
}



static int gui_list_click(void *a, gfx_event_t *ev)
{
  int i,j,k;
  gui_list_t *L;
  int upd;
  event_function_t f;
  
  L= a;
  i= 0; j= L->n_items-1;

  while(i<=j)
  {
    int iY, iH;
    k= (i+j)/2;
    iY= L->items[k].y - L->posy + L->port_y; 
    iH= L->items[k].h;

    if (ev->y<iY) { j= k-1; continue; } 
    if (ev->y>=(iY+iH)) { i= k+1; continue; }
    goto found;
  }
  return 1;
found:
  upd=1;
  if (!L->multiselect)
  {
    for(i=0;i<L->n_items;i++)
      if (i!=k && L->items[i].selected)
      {
        L->items[i].selected= 0;
        upd*= list_render_item(L, i, 1);
      }
    if (!L->items[k].selected)
    {
       L->items[k].selected= 1;
       upd*= list_render_item(L,k,1);
    }
  }
  else
  {
    L->items[k].selected= L->items[k].selected ? 0:1;
    upd= list_render_item(L,k,1);
  }

  if (!upd)
    gfx_update(L->port_x,L->port_y,L->port_w,L->port_h);

  if (L->click_time && ev->time>L->click_time && 
        (ev->time-L->click_time<500) && L->click_index==k)
  {
    ev->type= GFXEV(L->widget.number, DOUBLE_CLICK);
    L->click_time= 0;
    L->click_index= -1;
    f= L->dlickf;
  }
  else
  {
    ev->type= L->widget.number;
    L->click_time= ev->time;
    L->click_index= k;
    f= L->clickf;
  }
  ev->key= k;
  ev->string= L->items[k].str;
  ev->usr= L->items[k].user;

  if (f && (*f)(L->evarg, ev)) return 1; 
  return 0;
}

 // mouse button 4: scroll up
static int gui_list_click4(void *a, gfx_event_t *ev)
{
  gui_list_t *L;
  int posy;
  L= a;

  posy= imax2(L->posy - L->port_h/2,0);
  if (posy!=L->posy)
  {
    L->posy= posy;
    list_draw_contents(L);
    if (L->has_scrollv) 
       scrollbar_set( L->scrollv,
                      (double) L->posy/L->content_h, 
                      (double)(L->posy+L->port_h)/L->content_h,
                      0, NULL );
    gfx_update(L->port_x,L->port_y,L->port_w,L->port_h);
  }
  return 1;
}

 // mouse button 5: scroll down
static int gui_list_click5(void *a, gfx_event_t *ev)
{
  gui_list_t *L;
  int posy;
  L= a;

  posy= imax2(0, imin2(L->posy + L->port_h/2, L->content_h - L->port_h));
  if (posy!=L->posy)
  {
    L->posy= posy;
    list_draw_contents(L);
    if (L->has_scrollv) 
       scrollbar_set( L->scrollv,
                      (double) L->posy/L->content_h, 
                      (double)(L->posy+L->port_h)/L->content_h,
                      0, NULL );
    gfx_update(L->port_x,L->port_y,L->port_w,L->port_h);
  }
  return 1;
}


static void list_install_click_handlers(gui_list_t *LL)
{
  int x,y,w,h;
  if (!LL->enabled) return ;

  x= LL->port_x; y= LL->port_y;
  w= LL->port_w; h= LL->port_h;
  
  LL->ch0= install_or_modify_click
             (LL->ch0,&LL->enabled, gui_list_click, LL, x,y,w,h, 1);
  LL->ch1= install_or_modify_click
             (LL->ch1,&LL->enabled, gui_list_click4, LL, x,y,w,h, 4);

  LL->ch2= install_or_modify_click
             (LL->ch2,&LL->enabled, gui_list_click5, LL, x,y,w,h, 5);
}


static void list_relayout(int flags, gui_list_t *L)
{
  int hh, hv;
  compute_list_content_size(L, &L->content_w, &L->content_h);
  compute_list_port(L, &L->port_w, &L->port_h,&hh, &hv);

  L->posy= imax2(0, imin2(L->posy, L->content_h-L->port_h));
  L->posx= imax2(0, imin2(L->posx, L->content_w-L->port_w));

  list_arrange_scrollbars(L,hh,hv);

  list_install_click_handlers(L);

  list_draw_contents(L);

  if (flags&GFX_UPDATE && L->enabled)
    gfx_update(L->x,L->y,L->w,L->h);
}

 
static void gfx_list_insert
   (gui_list_t* L, int pos, char *str, void *usr,const unsigned int*icon) 
{
  list_item_t *item;
  unsigned int icon_w, icon_h;

  if (icon)
  {
  icon_w= icon[0] & ~(1<<31);
  icon_h= icon[1];
  icon= icon+2;
  }
  else icon_w= icon_h= 0;

  if (pos<0 || pos>=L->n_items) pos= L->n_items;

  item= array_make1space(pos,(void**) &L->items, L->n_items, &L->s_items,
                         sizeof(L->items[0]));
  item->str= strdup(str);
  item->user= usr;
  item->icon_data= icon;
  item->icon_width= icon_w;
  item->icon_height= icon_h;
  item->selected= 0;

  L->n_items++;
  compute_list_items(L, pos, pos);

  if (!L->enabled) return ;

  L->max_icon_width= list_max_icon_width(L);
  list_layout_items(L, pos);
  list_relayout(GFX_UPDATE,L);
}

static void gfx_list_delete(gui_list_t *L, int pos, void **Ruser)
{
  int i;

  if (Ruser) *Ruser= NULL;
  if (L->n_items==0) return ;
  if (pos<0) pos= L->n_items-1;
  if (pos>=L->n_items) return ;

  if (Ruser) *Ruser= L->items[pos].user;
  free(L->items[pos].str);

  for(i=pos;i<L->n_items-1;i++)
     L->items[i]= L->items[i+1];
  L->n_items--;

  if (L->enabled)
  {
    if (pos<L->n_items)
      list_layout_items(L, pos);
    list_relayout(GFX_UPDATE,L);
  }
}


static int gfx_list_selected(gui_list_t *L)
{
  int i;
  for(i=0;i<L->n_items;i++)
   if (L->items[i].selected) return i;
  return -1;
}


static void gfx_list_select(gui_list_t* L,int pos)
{
  int i;
  int ch;
  if (pos<0 || pos>=L->n_items) return ;
  if (L->items[pos].selected) return ;
  ch= 1;
  for(i=0;i<L->n_items;i++)
    if (L->items[i].selected)
    {
      L->items[i].selected= 0;
      if (L->enabled) ch&= list_render_item(L,i,1);
    }
  L->items[pos].selected= 1;
  if (L->enabled) 
    ch &=list_render_item(L,pos,1);
  if (L->enabled && !ch)
    gfx_update(L->port_x,L->port_y,L->port_w,L->port_h);
}

static void gfx_list_clear(gui_list_t *L)
{
  int i;
  for(i=0;i<L->n_items;i++)
  {
    free(L->items[i].str);
    L->items[i].str= NULL;
  }
  L->n_items= 0;
  L->posx= 0;
  L->posy= 0;
  if (L->enabled)
    list_relayout(GFX_UPDATE,L);
}

static void gfx_list_show(gui_list_t *L, int show)
{
  int hv,hh;

  if ((L->enabled && show) || (!L->enabled && !show)) return ;

  if (!show)
  {
    L->enabled= 0;
    if (L->scrollv) L->scrollv->enabled= 0;
    if (L->scrollh) L->scrollh->enabled= 0;
    return ;
  }

  L->enabled= 1;

  L->max_icon_width= list_max_icon_width(L);
  compute_list_items(L, 0, L->n_items-1);
  compute_list_content_size(L, &L->content_w, &L->content_h);
  compute_list_port(L, &L->port_w, &L->port_h, &hh, &hv);
  list_layout_items(L,0);
  list_draw_borders(L, 0xf);
  list_arrange_scrollbars(L, hh, hv);
  list_draw_contents(L);
  gfx_update(L->x,L->y,L->w,L->h);
  list_install_click_handlers(L);
}

static widget_t* gui_list_new
   (int noicon, int multisel, void *arg, 
       event_function_t clickf,
       event_function_t dlickf)
{
  gui_list_t *LL;

  LL= calloc(1,sizeof(*LL));

  LL->uses_icons= !noicon;
  LL->multiselect= multisel;

  LL->ch0= LL->ch1= LL->ch2= -1;
  LL->click_index= -1;

  LL->evarg= arg;
  LL->clickf= clickf;
  LL->dlickf= dlickf;

  LL->widget.klass= GFXCLASS_LIST;
  return &LL->widget;
}

/***********************************************************************
 *                                                                     *
 *                              ICON LIST                              *
 * #W.ICONLIST                                                         *
 ***********************************************************************/

typedef struct
{
  colorpix_t pix;
  int x,y;
  int iv; void *pv;
  int sel;
} gui_icoent_t;

typedef struct
{
  widget_t widget;
  int multi;
  
  int fg, bg; int hpad, vpad;
  int max_text_width;
  int disp_w, total_h;

  int sel;

  int pos_y;
  scrollbar_t *vscroll;
  int has_scroll;
  int click_handler;
  int click_handler4;
  int click_handler5;

  gui_icoent_t *ent; int sent,nent;
  int layout_ok;
} gui_iconlist_t;


static void gui_iconlist_render(gui_iconlist_t *IC);

static void gui_iconlist_arrange_click_handler
   (gui_iconlist_t *IC, int need_scroll)
{
  set_click_handler(IC->click_handler, IC->widget.x,IC->widget.y,
                            IC->disp_w,IC->widget.h);
  set_click_handler(IC->click_handler4, IC->widget.x,IC->widget.y,
                            IC->disp_w,IC->widget.h);
  set_click_handler(IC->click_handler5, IC->widget.x,IC->widget.y,
                            IC->disp_w,IC->widget.h);
}

static void render_icoent
   (gui_iconlist_t *IC, gui_icoent_t *EE,
    char *str, int iw, int ih, unsigned int *id)
{
  graypix_t text;
  colorpix_t pix;
  int x,y,sx;

  text_to_graypix(1,str,&text,NULL);
  pix.w= imax2(iw, imin2(text.w, IC->max_text_width));

// FIXME: 5 here should be a resource
  pix.h= ih + 5 + text.h;
  pix.stride= pix.w;

  pix.data= malloc(pix.w*pix.h*sizeof(pix.data[0]));
  colorpix_fill_rect(pix, 0,0, pix.w, pix.h, IC->bg);

  if (pix.w>iw) sx= (pix.w-iw)/2; 
           else sx= 0;

  for(y=0;y<ih;y++)
   for(x=0;x<iw;x++)
     pix.data[(sx+x)+y*pix.stride]= color_u_to_s(id[x+y*iw]);

  if (text.w>pix.w) text.w= pix.w; 
  blend_graypix(IC->fg,text,pix,(pix.w-text.w)/2,ih+5);
  free(text.data);
  EE->pix= pix;
}


static void gui_iconlist_set_scrollbar(gui_iconlist_t *IC)
{
  scrollbar_set_all(IC->vscroll, 
       (double) IC->pos_y/IC->total_h, 
       (double) (IC->pos_y+IC->widget.h)/IC->total_h, 
       0.8*IC->widget.h/IC->total_h);
}


static void  gui_iconlist_scrollbar_click(void *_arg,double start,double end)
{
  gui_iconlist_t *IC;
  IC= _arg;
  int py;
  if (start<0.0) start= 0.0;
  py= start*IC->total_h;
  if (py+IC->widget.h>IC->total_h) py= IC->total_h-IC->widget.h;
  if (py<0) py= 0;
  if (py==IC->pos_y) return ;
  IC->pos_y= py;
  gui_iconlist_render(IC);
}

static void gui_iconlist_make_scrollbar(gui_iconlist_t *IC)
{
  if (IC->vscroll) return ;
 
  IC->vscroll= (scrollbar_t*) scrollbar_new(1, 
              0.0, 1.0, 0.02,
              0, IC, gui_iconlist_scrollbar_click);
  scrollbar_position(IC->vscroll,
              IC->widget.x+IC->widget.w-gui_scrollbar_width, IC->widget.y,
              gui_scrollbar_width, IC->widget.h);
}


static void gui_iconlist_layout(gui_iconlist_t *IC)
{
  int i, s,e;
  int maxw, avail_w, need_scroll, x, y, maxh;

  if (IC->layout_ok) return;
  avail_w= IC->widget.w;
  need_scroll= 0;
  IC->layout_ok= 1;
  IC->total_h= IC->vpad;
  if (!IC->nent) goto finish; 

  maxw= IC->ent[0].pix.w;
  for(i=1;i<IC->nent;i++) maxw= imax2(maxw,IC->ent[i].pix.w);
  if (maxw<=0) return ;

again:
  y= IC->vpad;
  x= IC->hpad;
  maxh= 0;
  for(i=0;i<IC->nent;i++)
  {
    if (maxh && x+maxw+IC->hpad>avail_w)
    {
      y+= maxh+IC->vpad;
      IC->total_h= y;
      maxh= 0;
      x= IC->hpad;
      if (i!=IC->nent-1 && (y+IC->ent[i].pix.h)>IC->widget.h && !need_scroll)
      {
        need_scroll= 1;
        avail_w -= gui_scrollbar_width;
        goto again;
      }
    }

    maxh= imax2(maxh, IC->ent[i].pix.h);
    IC->ent[i].y= y; IC->ent[i].x= x;
    x+= maxw+IC->hpad;
  } 
  IC->total_h= y + maxh + IC->vpad;

  s= 0;
tapdown:
  for(e=s+1;e<IC->nent;e++)
    if (IC->ent[e].y!=IC->ent[s].y) break;
  e--;
  maxh= 0;
  for(i=s;i<=e;i++)
    maxh= imax2(maxh, IC->ent[i].pix.h);
  for(i=s;i<=e;i++)
  {
    IC->ent[i].y += (maxh - IC->ent[i].pix.h)/2;
    IC->ent[i].x += (maxw-IC->ent[i].pix.w)/2;
  }
  s= e+1;
  if (s<IC->nent) goto tapdown;
  
finish:
  IC->disp_w= avail_w;

  if (need_scroll)
  {
    gui_iconlist_make_scrollbar(IC);
    gui_iconlist_set_scrollbar(IC);
  }
  if (IC->vscroll) 
    scrollbar_show(IC->vscroll,need_scroll);
  if (need_scroll!=IC->has_scroll)
  {
    gui_iconlist_arrange_click_handler(IC, need_scroll);
    IC->has_scroll= need_scroll;
  }
}

static int gui_icoent_paint(gui_iconlist_t *IC, gui_icoent_t *EE)
{
  if (EE->y+EE->pix.h<=IC->pos_y) return -1;
  if (EE->y>=IC->pos_y+IC->widget.h) return 1;

  put_pix_in_area(EE->pix, 
          EE->x+IC->widget.x, EE->y - IC->pos_y + IC->widget.y,
       IC->widget.x, IC->widget.y, IC->disp_w, IC->widget.h);

  return 0;
}

static void gui_iconsel_draw_selrect
  (gui_iconlist_t *IC,int item, int color,int upd)
{
  int a,sy,sx,i,sw,sh;
  if (item<0 || item>=IC->nent) return ;
  if (!IC->widget.shown) return ;
  if (color) color= IC->fg; else color= IC->bg;

  a= imin2(IC->hpad/2-1,IC->vpad/2-1);

  sx= IC->ent[item].x + IC->widget.x;
  sy= IC->ent[item].y + IC->widget.y - IC->pos_y;
  sw= IC->ent[item].pix.w;
  sh= IC->ent[item].pix.h;

  for(i=1;i<=a;i++)
    gfx_draw_rect_in(color, sx-i, sy-i, sw+2*i, sh+2*i,
                     IC->widget.x, IC->widget.y, IC->disp_w, IC->widget.h);
  i--;
  if (upd)
    gfx_update(sx-i,sy-i, sw+2*i, sh+2*i);
}


static void gui_iconlist_render(gui_iconlist_t *IC)
{
  int i;
  if (!IC->widget.shown) return ;
  gui_iconlist_layout(IC);
  gfx_fill_rect(0,IC->widget.x, IC->widget.y, IC->disp_w, IC->widget.h,IC->bg);
  gfx_draw_rect(0,IC->widget.x-1, IC->widget.y-1, 
                  IC->widget.w+2, IC->widget.h+2, IC->fg);

  if (!IC->nent) goto done;
  for(i=0;i<IC->nent;i++)
  {
    if (gui_icoent_paint(IC,IC->ent+i)==1) break; 
    if (IC->multi && IC->ent[i].sel)
       gui_iconsel_draw_selrect(IC, i, 1, 0);
  }

done:
  if (!IC->multi)
    gui_iconsel_draw_selrect(IC, IC->sel, 1, 0);
  gfx_update(IC->widget.x-1, IC->widget.y-1, IC->disp_w+2, IC->widget.h+2);
}

static void gui_iconlist_destroy(widget_t *W)
{
  gui_iconlist_t *IC;
  int i;
  IC= (gui_iconlist_t*) W;
  for(i=0;i<IC->nent;i++)
   free(IC->ent[i].pix.data);
  free(IC->ent);
  remove_click_handler(IC->click_handler);
  remove_click_handler(IC->click_handler4);
  remove_click_handler(IC->click_handler5);
}


static void gui_iconlist_position(widget_t *W, int x,int y,int w,int h)
{
  gui_iconlist_t *IC; 
  IC= (void*) W;
  set_click_handler(IC->click_handler, x,y,w,h);
  IC->widget.x= x;
  IC->widget.y= y;
  IC->widget.w= w;
  IC->widget.h= h;
}

static void gui_iconlist_show(widget_t *W)
{
  gui_iconlist_render((gui_iconlist_t*) W);
}

static void gui_iconlist_select_item(gui_iconlist_t *IC, int sel)
{
  gui_iconsel_draw_selrect(IC, IC->sel, 0, 1);
  if (IC->sel>=0 && IC->sel<=IC->nent) 
    IC->ent[IC->sel].sel= 0;
  IC->sel= sel;
  gui_iconsel_draw_selrect(IC, IC->sel, 1, 1);
  if (IC->sel>=0 && IC->sel<=IC->nent) 
    IC->ent[IC->sel].sel= 1;
}

static int gui_iconlist_click(void *_arg, gfx_event_t *ev)
{
  gui_iconlist_t *IC;
  int x,y,i;
  IC= _arg;
  x= ev->x - IC->widget.x;
  y= ev->y - IC->widget.y + IC->pos_y;
  for(i=0;i<IC->nent;i++)
  {
    if (x>=IC->ent[i].x && y>=IC->ent[i].y &&
        x<IC->ent[i].x+IC->ent[i].pix.w && y<IC->ent[i].y+IC->ent[i].pix.h)
    break;
  }
  if (i==IC->nent) return 1;
  if (IC->multi)
  {
    IC->ent[i].sel ^= 1;
    if (IC->sel==i) IC->sel= -1;
    else if (IC->sel==-1) IC->sel= i;
    gui_iconsel_draw_selrect(IC, i, IC->ent[i].sel, 1);
  }
  else
  {
    if (i==IC->sel) return 1;
    gui_iconlist_select_item(IC, i);
  }

  ev->type= IC->widget.number;
  ev->key= i;
  ev->usr= IC->ent[i].pv;
  ev->wide= IC->ent[i].iv;
  return 0;
}

static int gui_iconlist_click4(void *_arg, gfx_event_t *ev)
{
  gui_iconlist_t *IC;
  IC= _arg;
  if (!IC->vscroll || !IC->has_scroll || IC->pos_y==0) return 1;

  IC->pos_y = imax2(0, IC->pos_y - IC->widget.h * 8 /10);
 
  gui_iconlist_set_scrollbar(IC);
  gui_iconlist_render(IC); 
  
  return 1;
}

static int gui_iconlist_click5(void *_arg, gfx_event_t *ev)
{
  gui_iconlist_t *IC;
  IC= _arg;
  if (!IC->vscroll || !IC->has_scroll || IC->pos_y+IC->widget.h>=IC->total_h)
        return 1;

  IC->pos_y= imax2(imin2(IC->pos_y+IC->widget.h * 8 /10, 
                         IC->total_h-IC->widget.h),0);
 
  gui_iconlist_set_scrollbar(IC);
  gui_iconlist_render(IC); 
  
  return 1;
}

static void gui_iconlist_clear(gui_iconlist_t *IC)
{
  int i;
  if (IC->nent==0) return ;
  IC->sel= -1;
  for(i=0;i<IC->nent;i++)
   free(IC->ent[i].pix.data);
  IC->nent= 0;
  IC->layout_ok= 0;
  gui_iconlist_render(IC);
}

static void gui_iconlist_remove(gui_iconlist_t *IC,int pos, void **Ruser)
{
  int i;
  if (IC->nent==0 || pos<0 || pos>=IC->nent)  return ;
  if (Ruser) *Ruser= IC->ent[pos].pv;
  free(IC->ent[pos].pix.data);
  for(i=pos;i<IC->nent-1;i++)
     IC->ent[i]= IC->ent[i+1];
  if (IC->sel==pos) IC->sel= -1; 
  else if (IC->sel>pos) IC->sel--;
  IC->nent--;
  IC->layout_ok= 0;
  gui_iconlist_render(IC);
}

void gfx_iconlist_insert
   (int wno, int pos,
    char *text, 
    unsigned int w, unsigned int h, unsigned int *data,
    int ival, void *pval)
{
  gui_iconlist_t *IC;
  gui_icoent_t *EE;

  if (!(IC=get_widgetk(wno,GFXCLASS_ICONLIST))) return ;

  if (pos<0 || pos>=IC->nent) pos= IC->nent;
  EE= array_make1space(pos, (void**) &IC->ent, IC->nent, &IC->sent,
                             sizeof(IC->ent[0]));
  IC->nent++;
  if (IC->sel>=pos) IC->sel++;

  EE->iv= ival; EE->pv= pval; EE->sel= 0;
  IC->layout_ok= 0;
  render_icoent(IC, EE,text,w,h,data);
  gui_iconlist_render(IC);
}

void gfx_iconlist_updates(int wno, int upd)
{
  gui_iconlist_t *IC;
  if (!(IC=get_widgetk(wno,GFXCLASS_ICONLIST))) return ;
  IC->widget.shown= upd;
  if (upd) gui_iconlist_render(IC);
}

void gfx_iconlist_area(int wno, int *Rw, int *Rh)
{
  gui_iconlist_t *IC;
  int w, h;
  if ((IC=get_widgetk(wno,GFXCLASS_ICONLIST)))
  { w=  IC->widget.w;
    h=  IC->widget.h; }
  else { w= h= 10; }
  if (Rw) *Rw= w;
  if (Rh) *Rh= h;
}

void gfx_iconlist_select(int wno, int item, int sel)
{
  gui_iconlist_t *IC;
  int S;
  if (!(IC=get_widgetk(wno,GFXCLASS_ICONLIST))) return ;
  if (wno<0 || wno>=IC->nent) return ;
  if (sel) sel= 1;
  S=IC->ent[wno].sel;
  IC->ent[wno].sel= sel;  
  if (S==sel) return ;
  gui_iconsel_draw_selrect(IC, item, sel, 1);
}

void gfx_iconlist_permute(int wno, int N, int *P)
{
  gui_iconlist_t *IC;
  int i;
  if (!(IC=get_widgetk(wno,GFXCLASS_ICONLIST))) return ;
  if (N!=IC->nent) return ;
  for(i=0;i<N;i++) if (P[i]==IC->sel) { IC->sel= i; break; }
  IC->layout_ok= 0;
  permute(N, P, IC->ent, sizeof(IC->ent[0]));
  gui_iconlist_render(IC);
}

void gfx_iconlist_clear_selection(int wno)
{
  gui_iconlist_t *IC;
  int i;
  if (!(IC=get_widgetk(wno,GFXCLASS_ICONLIST))) return ;
  for(i=0;i<IC->nent;i++)
    if (IC->ent[i].sel)
    {
      IC->ent[i].sel= 0; 
      gui_iconsel_draw_selrect(IC, i, 0, 1);
      if (i==IC->sel) IC->sel= -1;
    }
  if (IC->sel!=-1)
  {
    gui_iconsel_draw_selrect(IC, IC->sel, 0, 1);
    IC->sel= -1;
  }
}


void gfx_iconlist_get_selection(int wno,int *Rn, int **Rs)
{
  gui_iconlist_t *IC;
  int i, n, *s;
  if (Rn) *Rn= 0;
  if (!(IC=get_widgetk(wno,GFXCLASS_ICONLIST))) return ;
  n= 0;
  for(i=0;i<IC->nent;i++)
    if (IC->ent[i].sel) n++;
  if (Rn) *Rn= n;
  if (!Rs) return ;
  s= malloc((n+1)*sizeof(int));
  n= 0;
  for(i=0;i<IC->nent;i++)
    if (IC->ent[i].sel) s[n++]= i;
  s[n]= -1;
  *Rs= s;
}

static widget_t* gui_iconlist_new
  (int fg, int bg,int vpad, int hpad,int max_text_width,int multisel)
{
  gui_iconlist_t *IC;
  IC= calloc(1,sizeof(IC[0]));

  IC->widget.klass= GFXCLASS_ICONLIST;

  IC->fg= fg; IC->bg= bg; IC->vpad= vpad; IC->hpad= hpad;
  IC->max_text_width= max_text_width;
  IC->multi= multisel;

  IC->widget.show= gui_iconlist_show;
  IC->widget.hide= generic_dummy_hide;
  IC->widget.destroy= gui_iconlist_destroy;
  IC->widget.position= gui_iconlist_position;

  IC->click_handler= install_click_handler(0,0,0,0, &IC->widget.shown, 
                                           IC, gui_iconlist_click,1);
  IC->click_handler4= install_click_handler(0,0,0,0, &IC->widget.shown, 
                                           IC, gui_iconlist_click4,4);
  IC->click_handler5= install_click_handler(0,0,0,0, &IC->widget.shown, 
                                           IC, gui_iconlist_click5,5);
  IC->sel= -1;
  return &IC->widget;
}



/***********************************************************************
 *                                                                     *
 *                            DROPDOWN MENU                            *
 * #W.DROPDOWN                                                         *
 ***********************************************************************/
typedef struct { void* ptr; char *str; } TUPsp_t;

typedef struct
{
  widget_t widget;
  int shown;
  int n_items, s_items;
  TUPsp_t *items;
  int selected;

  int click_handler;
  int click_enabled;

  int list_active;

  // list stuff
  int list_x,list_y,list_w,list_h;
  int list_px, list_py, list_pw, list_ph; // port coords
  unsigned int *list_backup;
  int fh;
  scrollbar_t *scroll;
  int posy;
  int list_cont_h;  // content height

  int but1clicked;
  int selitem;

  void *evarg; event_function_t clickf;
} dropdown_t;



static int dropdown_item_pos(dropdown_t *DD,int item)
{
  return item*(RES(dropdown.list.spc)+DD->fh) + RES(dropdown.list.padv);
}


static void dropdown_render_button(dropdown_t *DD, int p)
{
  char *text;
  int ix,iy, tx,ty;
  int tcw, tch, tsx, tsy, tw,th;
  // int w,h; // W and H minus the border
  unsigned char *tdata;

  int W,H,X,Y;
  colorpix_t pix;


  if (!DD->shown) return ;

  tdata= NULL;

  W= DD->widget.w;
  H= DD->widget.h;
  X= DD->widget.x;
  Y= DD->widget.y;

  if (W<4 || H<4)
  {
    // even the border doesn't fit!
    gfx_fill_rect(GFX_UPDATE, DD->widget.x, DD->widget.y, W,H, 0xFFff0000);
    return ;
  }
  pix_of_widget(&DD->widget, &pix);
  gui_draw_3dbox(&pix, p ? &RES(dropdown.depressed) 
                         : &RES(dropdown.normal));

  int buW,buH;
  buW= RES(dropdown.button).w;
  buH= RES(dropdown.button).h;

  ix= W -2 - RES(dropdown.pad.right)-buW;
  iy= (H - buH)/2;
    // no space for button. in fact, the text won't fit here also
    // so we simply display an empty area.
  if (ix<2 || iy<2) return ;
  
  blend_graymap(GFX_ALPHA, RES(dropdown.fg), 0xffFFff,
               RES(dropdown.button).data, buW, buH, 0,0,
               X+ix, Y+iy+2, buW, buH);

  if (DD->selected<0 || DD->selected>=DD->n_items) goto done;
  text= DD->items[DD->selected].str;
  if (!text || !text[0]) goto done;

  get_text_data(1,text, &tw, &th, NULL, &tdata);

  tx= 2 + RES(dropdown.pad.left);
  tsx= 0;
  tcw= imax2(0,imin2(tw, ix-RES(dropdown.pad.mid)-tx));
  if (tcw==0) goto done;

  ty= 2 + imax2( 0, (H-th)/2 );
  tsy= imax2(0, (th-H)/2); 
  tch= imin2(th, H);

  if (tch==0) goto done;

  blend_graymap(GFX_ALPHA, 0x00, 0xffFFff, tdata, tw, th, tsx,tsy,
                             X+tx, Y+ty, tcw, tch);
done:
  if (tdata) free(tdata);
  gfx_update(X,Y,W,H);
}



static void dropdown_get_list_backup(dropdown_t *DD)
{
  if (DD->list_w<=0 || DD->list_h<=0) return ;
  DD->list_backup= get_server_pix(DD->list_x,DD->list_y,DD->list_w,DD->list_h);
}

static void dropdown_put_list_backup(dropdown_t *DD)
{
  if (!DD->list_backup || DD->list_w<=0 || DD->list_h<=0) return ;
  put_server_pix(DD->list_backup, DD->list_x,DD->list_y,DD->list_w,DD->list_h);
  free(DD->list_backup);
  DD->list_backup= NULL;
}


static void dropdown_draw_listbg(dropdown_t *DD)
{
  gfx_fill_rect(0,DD->list_px,DD->list_py,DD->list_pw,DD->list_ph, 0xffffffff);
}

static void dropdown_draw_listframe(dropdown_t *DD,int which)
{
  int l,r,t,b;
  l= DD->list_x;
  r= DD->list_x+DD->list_w-1;
  t= DD->list_y;
  b= DD->list_y+DD->list_h-1;
  if (which&1) vline(r, t,b, RES(dropdown.list.border));
  if (which&2) vline(l, t,b, RES(dropdown.list.border));
  if (which&4) hline(l, r,t, RES(dropdown.list.border));
  if (which&8) hline(l, r,b, RES(dropdown.list.border));
}

static int dropdown_draw_item(dropdown_t *DD,int idx)
{
  int P;
  P= dropdown_item_pos(DD,idx);
  if (P>=DD->posy+DD->list_ph) return 1;
  if (P+DD->fh<DD->posy) return -1;
  print_clipped(DD->list_px,DD->list_py,DD->list_pw,DD->list_ph,
           DD->list_px, P-DD->posy+DD->list_py, 1, DD->items[idx].str); 
  return 0;
}

static void dropdown_draw_contents(dropdown_t *DD)
{
  int i,j,k;
  if (DD->n_items==0) return ;
  i= 0; 
  j= DD->n_items-1;
  while(i<=j)
  {
    k= (i+j)/2;
    switch(dropdown_draw_item(DD,k))
    {
     case 1: j= k-1; break;
     case 0: goto found;
     case -1: i= k+1; break;
    }
  }

  return;
found:
  for(i=k-1;i>=0;i--)         if (dropdown_draw_item(DD,i)) break;
  for(j=k+1;j<DD->n_items;j++) if (dropdown_draw_item(DD,j)) break;
}



static int dropdown_item_at(dropdown_t *DD,int y)
{
  int K;
  if (DD->n_items<=0) return -1;
  
  y= y - DD->list_py + DD->posy;
  if (y<DD->fh+RES(dropdown.list.padv))
     K= 0;
  else
     K= (y-RES(dropdown.list.padv)+RES(dropdown.list.spc))
          /(DD->fh+RES(dropdown.list.spc));
  if (K>=DD->n_items) return -1;
  return K;
}

static void dropdown_scroll_to(dropdown_t *DD,int pos,int report)
{
  if (pos<0) pos= 0;
  if (pos+DD->list_ph>DD->list_cont_h) pos= DD->list_cont_h-DD->list_ph;
  if (pos<0) pos= 0;
  if (DD->posy==pos) return ;
  
  DD->posy= pos;
  dropdown_draw_listbg(DD);
  dropdown_draw_contents(DD);
  if (report && DD->scroll) 
    scrollbar_set(DD->scroll, (double) DD->posy/DD->list_cont_h, 
                              (double) (DD->posy+DD->list_ph)/DD->list_cont_h, 
                              0, NULL);
  gfx_update(DD->list_x,DD->list_y,DD->list_w,DD->list_h);
}

static void dropdown_scroll(dropdown_t *DD,double S,double E)
{
  dropdown_scroll_to(DD, DD->list_cont_h*S,0);
}

static void dropdown_hilight_item(dropdown_t *DD, int item)
{
  int x,y,w,h;

  if (item<0 || item>=DD->n_items) return ;
  x= DD->list_px;
  y= dropdown_item_pos(DD,item)-RES(dropdown.list.spc)+DD->list_py-DD->posy;
  w= DD->list_pw;
  h= DD->fh+RES(dropdown.list.spc);
  fill_clipped_rect(DD->list_px, DD->list_py, DD->list_pw, DD->list_ph,
                    x,y,w,h, RES(dropdown.list.hilight));
  dropdown_draw_item(DD,item);
  gfx_update(x,y,w,h);
}

static void dropdown_hide_list(dropdown_t *DD)
{
  if (!DD->list_active) return ;
  DD->list_active= 0;
  dropdown_put_list_backup(DD);
  gfx_update(DD->list_x,DD->list_y,DD->list_w,DD->list_h);
  dropdown_render_button(DD,0);
  DD->click_enabled= 1;
  if (DD->scroll) DD->scroll->enabled= 0;
}

static int dropdown_mouse(dropdown_t *DD, gfx_event_t *event)
{
  switch(event->type)
  {
  case GFXEV_BUTTON_PRESS:
    switch(event->button)
    {
    case 1: 
      if (!in_rect(DD->list_x,DD->list_y,DD->list_w,DD->list_h,
           event->x,event->y))
      {
        DD->but1clicked= 1;
        DD->selitem= -1;
        return 1;
      }
      if (in_rect(DD->list_px,DD->list_py,DD->list_pw,DD->list_ph,
           event->x,event->y))
      {
        DD->but1clicked= 1;
        DD->selitem= dropdown_item_at(DD,event->y);
        dropdown_hilight_item(DD, DD->selitem);
        return 1;
      }
      return 0;
    case 4:
      dropdown_scroll_to(DD, DD->posy-0.9*DD->list_ph, 1);
      return 1;
    case 5:
      dropdown_scroll_to(DD, DD->posy+0.9*DD->list_ph, 1);
      return 1;
    }
    return 0;
  case GFXEV_BUTTON_RELEASE:
    if (event->button!=1 || !DD->but1clicked)
       return 0;
    DD->but1clicked= 0;
    if (DD->selitem!=-1)
       DD->selected= DD->selitem;
    dropdown_hide_list(DD);
    ungrab_mouse();
    if (DD->selitem==-1)
      return 1;
    event->type= DD->widget.number;
    event->key= DD->selitem;
    event->usr= DD->items[DD->selitem].ptr;
    DD->selitem= -1;
    if (DD->clickf && (*DD->clickf)(DD->evarg, event)) 
       return 1;
    return 0;
  }
  return 0;
}

static void dropdown_show_list(dropdown_t *DD)
{
  if (DD->list_backup) { free(DD->list_backup); DD->list_backup= NULL; }
  DD->list_active= 1;

  DD->list_w= DD->widget.w;
  DD->list_x= DD->widget.x;

  DD->list_y= DD->widget.y + DD->widget.h;

  get_text_data(1,"My",NULL,&DD->fh,NULL,NULL);
  int ni;
  ni= imax2(DD->n_items, 1);
  DD->list_cont_h= RES(dropdown.list.padv)*2 + ni*DD->fh + (ni-1)*RES(dropdown.list.spc);

  DD->posy= 0;
  if (DD->list_y+DD->list_cont_h+2<=x_window_height)
  {
    // no scrollbar needed
    DD->list_ph= DD->list_cont_h;
    DD->list_h= DD->list_ph+2;
    DD->list_pw= DD->list_w-2;
    DD->list_px= DD->list_x+1;
    DD->list_py= DD->list_y+1;

    if (DD->scroll) DD->scroll->enabled= 0;
    dropdown_get_list_backup(DD);
    dropdown_draw_listframe(DD,0xf);
  }
  else
  {
    DD->list_px= DD->list_x+1;
    DD->list_pw= DD->list_w - 1 - gui_scrollbar_width;
    DD->list_h= x_window_height - DD->list_y;
    if (DD->list_h<(DD->fh+RES(dropdown.list.padv)*2+2))
    {
      DD->list_y= DD->widget.y;
      DD->list_h= x_window_height - DD->list_y;
    }
    DD->list_ph= DD->list_h-2;
    DD->list_py= DD->list_y+1;
    if (DD->list_ph<=0) goto impossible;
    if (DD->selected!=-1)
    {
      DD->posy= dropdown_item_pos(DD,DD->selected) - DD->list_ph/2;
      if (DD->posy<0) DD->posy= 0;
      else if (DD->posy+DD->list_ph > DD->list_cont_h)
      {
        DD->posy= DD->list_cont_h - DD->list_ph;
        if (DD->posy<0) DD->posy= 0;
      }
    }
    dropdown_get_list_backup(DD);
     // scrollbar parameters
    int Sx, Sy, Sw, Sh;  
    double Sb, Se, Ss;

    Sx= DD->list_x+DD->list_w-gui_scrollbar_width;
    Sy= DD->list_y;
    Sw= gui_scrollbar_width;
    Sh= DD->list_h;
    Sb= (double)DD->posy/DD->list_cont_h;
    Se= (double)(DD->posy+DD->list_ph)/DD->list_cont_h;
    Ss= DD->list_ph*0.9/DD->list_cont_h;
    if (DD->scroll)
    {
      DD->scroll->enabled= 1;
      scrollbar_relayout(DD->scroll, Sx, Sy,Sw, Sh, Sb, Se, Ss);
    }
    else
    {
      DD->scroll= (scrollbar_t*)
                    scrollbar_new(1, Sb, Se, Ss, 0, DD, dropdown_scroll);
      scrollbar_position(DD->scroll, Sx, Sy,Sw, Sh);
    }
    scrollbar_show(DD->scroll,1);
    dropdown_draw_listframe(DD,0xe);
  }
  dropdown_draw_listbg(DD);
  dropdown_draw_contents(DD);
  gfx_update(DD->list_x,DD->list_y,DD->list_w,DD->list_h);
  DD->selitem= -1;
  DD->but1clicked= 0;
  grab_mouse((void*)dropdown_mouse, DD);
  DD->click_enabled= 0;

impossible:
  return ; // FIXME
}

static int dropdown_release(void *arg, gfx_event_t *event)
{
  dropdown_t *DD= arg;
  // dropdown_render_button(DD,0);
  dropdown_show_list(DD);
  return 1;
}

static int dropdown_click(void *arg, gfx_event_t* event)
{
  dropdown_t *DD= arg;
  if (event->type!=GFXEV_BUTTON_PRESS || event->button!=1) return 0;
  dropdown_render_button(DD,1);
  install_release_handler(DD, dropdown_release);
  return 1;
}

static void dropdown_show(dropdown_t *DD,int show)
{
  if (show==DD->shown) return ;
  DD->shown= show;
  if (show)
  {
    if (DD->click_handler==-1)
      DD->click_handler= install_click_handler(DD->widget.x,DD->widget.y,
                           DD->widget.w,DD->widget.h,
                           &DD->click_enabled, DD, dropdown_click, 1);
    DD->click_enabled= 1;
    dropdown_render_button(DD,0);
  }
  else
  {
    DD->click_enabled= 0;
  }
}


static void dropdown_insert(dropdown_t *DD,
                            int pos, const char *str, void *usr)
{
  TUPsp_t *src, *dst;
  int i;
  
  if (pos<0 || pos>DD->n_items) pos= DD->n_items;
  src= DD->items;
  if (DD->n_items==DD->s_items)
  {
    if (DD->s_items==0) DD->s_items= 10;
    else DD->s_items+= DD->s_items/2;
    DD->items= dst= malloc(DD->s_items*sizeof(DD->items[0]));
  }
  else dst= DD->items;

  if (src)
  {
    for(i=DD->n_items;i>pos;i--) dst[i]= src[i-1];
    if (src!=dst)
      for(i=0;i<pos;i++) dst[i]= src[i];
    if (src!=dst) free(src);
  }
  dst[pos].str= strdup(str);
  dst[pos].ptr= usr;
  DD->n_items++;
  if (pos<=DD->selected) DD->selected++;
}

static void dropdown_remove(dropdown_t *DD,int pos, void **Rusr)
{
  int i;
  
  if (DD->n_items==0) return ; 
  if (pos<0 || pos>=DD->n_items) pos= DD->n_items-1;
  if (Rusr) *Rusr= DD->items[pos].ptr;

  // actually, the string is always gotten from strdup, if is unnecessary
  if (DD->items[pos].str) free(DD->items[pos].str); 
  for(i=pos;i<DD->n_items-1;i++)
    DD->items[i]= DD->items[i+1];
  if (pos==DD->selected) DD->selected= -1;
  else if (pos<DD->selected) DD->selected--;
  DD->n_items--;
}

static void gui_dropdown_clear(dropdown_t *DD)
{
  int i;
  if (DD->n_items==0) return ; 
  for(i=0;i<DD->n_items;i++)
   free(DD->items[i].str);
  DD->selected= -1;
  DD->n_items= 0;  
  if (DD->shown)
    dropdown_render_button(DD,0);
}

static void gfx_dropdown_select(dropdown_t *DD, int pos)
{
  if (pos<0 || pos>=DD->n_items) pos= -1;
  if (DD->selected==pos) return ;
  DD->selected= pos;
  if (DD->shown)
    dropdown_render_button(DD,0);
}

static void gui_dropdown_destroy(widget_t *W)
{
  int i;
  dropdown_t *DD;
  DD= (void*) W;

  if (DD->scroll)
     scrollbar_destroy(DD->scroll);
  remove_click_handler(DD->click_handler);
  freeNnull(DD->list_backup);

  for(i=0;i<DD->n_items;i++)
    free(DD->items[i].str);
  freeNnull(DD->items);
}

static void dropdown_position(dropdown_t *DD,int x,int y,int w,int h)
{
  DD->widget.x= x;
  DD->widget.y= y;
  DD->widget.w= w;
  DD->widget.h= h;
  dropdown_render_button(DD,0);
}

static widget_t *dropdown_new(void *evarg, event_function_t clickf)
{
  dropdown_t *DD;
  DD= calloc(1,sizeof(*DD));
  DD->selected= -1;
  DD->click_handler= -1;
  DD->widget.klass= GFXCLASS_DROPDOWN;
  DD->widget.destroy= gui_dropdown_destroy;

  DD->evarg= evarg;
  DD->clickf= clickf;

  return &DD->widget;
}


/***********************************************************************
 *                                                                     *
 *                               SLIDER                                *
 * #W.SLIDER                                                           *
 ***********************************************************************/
typedef struct
{
  widget_t widget;
  int orient;
  double pos;
  int click_handler;
  int pressed;

  int button_x,button_y,button_w,button_h;
  int tx1, tx2, ty1, ty2;

} gui_slider_t;

static void slider_layout_vertical(gui_slider_t *B)
{
  int w, linew, x, h;
  B->button_w= RES(slider.vertical.normal_button).w;
  B->button_h= RES(slider.vertical.normal_button).h;

  linew= RES(slider.vertical.tro_start).w;

  h= B->widget.h;
  w= B->widget.w;
  if (linew<w) x= (w-linew)/2;
          else x= 0;
  B->tx1= B->tx2= B->widget.x + x;

  B->ty1= B->widget.y + B->button_h/2;
  B->ty2= B->ty1 + h - B->button_h-1;

  B->button_y= B->ty2-B->pos* (B->ty2-B->ty1)- B->button_h/2;
  B->button_x= B->tx1 - (B->button_w-linew)/2;
}

static void slider_layout_horizontal(gui_slider_t *B)
{
  int h, lineh, y, w;
  B->button_w= RES(slider.horizontal.normal_button).w;
  B->button_h= RES(slider.horizontal.normal_button).h;

  lineh= RES(slider.horizontal.tro_start).h;

  h= B->widget.h;
  w= B->widget.w;
  if (lineh<h) y= (h-lineh)/2;
          else y= 0;
  B->ty1= B->ty2= B->widget.y + y;

  B->tx1= B->widget.x + B->button_w/2;
  B->tx2= B->tx1 + w - B->button_w-1;

  B->button_x= B->pos* (B->tx2-B->tx1) + B->tx1 - B->button_w/2;
  B->button_y= B->ty1 - (B->button_h-lineh)/2;
}

static void slider_layout(gui_slider_t *B)
{
  if (B->orient=='V') slider_layout_vertical(B);
                 else slider_layout_horizontal(B);
}

static void slider_render_horizontal(gui_slider_t *B)
{
  static uint8_t *normal_button_mask, *depressed_button_mask;
  static uint32_t *normal_button, *depressed_button,
                  *tro_start, *tro_mid, *tro_end;
  uint32_t *img; uint8_t *mask;

  if (!normal_button_mask)
  {
    gui_R_COLORIMG_getmask(&RES(slider.horizontal.normal_button),
                           &normal_button_mask);
    gui_R_COLORIMG_getmask(&RES(slider.horizontal.depressed_button), 
                           &depressed_button_mask);
    tro_start= gui_R_COLORIMG_toserver(&RES(slider.horizontal.tro_start));
    tro_mid= gui_R_COLORIMG_toserver(&RES(slider.horizontal.tro_mid));
    tro_end= gui_R_COLORIMG_toserver(&RES(slider.horizontal.tro_end));

    normal_button= gui_R_COLORIMG_toserver
                           (&RES(slider.horizontal.normal_button));
    depressed_button= gui_R_COLORIMG_toserver
                           (&RES(slider.horizontal.depressed_button));
  }

  gfx_fill_rect(0,B->widget.x,B->widget.y,B->widget.w,B->widget.h,
                 RES(general.bg));

  three_part_assembly_horizontal(B->tx1,B->tx2,B->ty1, 
                                 tro_start, tro_mid, tro_end); 

  if (B->pressed)
  {
    img= depressed_button;
    mask= depressed_button_mask;
  }
  else
  {
    img= normal_button;
    mask= normal_button_mask;
  }

  put_server_pix_alphamask(img+2, mask, B->button_x,B->button_y,img[0],img[1]);
  gfx_update(B->widget.x,B->widget.y,B->widget.w,B->widget.h);
}

static void slider_render_vertical(gui_slider_t *B)
{
  static uint8_t *normal_button_mask, *depressed_button_mask;
  static uint32_t *normal_button, *depressed_button,
                  *tro_start, *tro_mid, *tro_end;
  uint32_t *img; uint8_t *mask;

  if (!normal_button_mask)
  {
    gui_R_COLORIMG_getmask(&RES(slider.vertical.normal_button),
                           &normal_button_mask);
    gui_R_COLORIMG_getmask(&RES(slider.vertical.depressed_button), 
                           &depressed_button_mask);
    tro_start= gui_R_COLORIMG_toserver(&RES(slider.vertical.tro_start));
    tro_mid= gui_R_COLORIMG_toserver(&RES(slider.vertical.tro_mid));
    tro_end= gui_R_COLORIMG_toserver(&RES(slider.vertical.tro_end));

    normal_button= gui_R_COLORIMG_toserver
                           (&RES(slider.vertical.normal_button));
    depressed_button= gui_R_COLORIMG_toserver
                           (&RES(slider.vertical.depressed_button));
  }

  gfx_fill_rect(0,B->widget.x,B->widget.y,B->widget.w,B->widget.h,
                     RES(general.bg));

  three_part_assembly_vertical(B->tx1,B->ty1,B->ty2, 
      tro_start, tro_mid, tro_end);

  if (B->pressed)
  {
    img= depressed_button;
    mask= depressed_button_mask;
  }
  else
  {
    img= normal_button;
    mask= normal_button_mask;
  }

  put_server_pix_alphamask(img+2, mask, B->button_x,B->button_y,img[0],img[1]);
  gfx_update(B->widget.x,B->widget.y,B->widget.w,B->widget.h);
}

static void slider_render(gui_slider_t *B)
{
  if (!B->widget.shown) return ;
  if (B->orient=='V') slider_render_vertical(B);
                 else slider_render_horizontal(B);
}

static void slider_move_button(gui_slider_t *B, gfx_event_t *event)
{
  if (B->orient=='V')
  {
    int y;
    y= imax2(B->ty1, imin2(event->y, B->ty2));
    B->pos= 1.0 - (y-B->ty1)/(double)(B->ty2-B->ty1);
  }
  else
  {
    int x;
    x= imax2(B->tx1, imin2(event->x, B->tx2));
    B->pos= (x-B->tx1)/(double)(B->tx2-B->tx1);
  }
  slider_layout(B);
  slider_render(B);
  event->type= B->widget.number;
  event->flo0= B->pos;
}


static int gui_slider_motion(void *arg, gfx_event_t *event)
{
  slider_move_button(arg,event);
  return 0;
}

static int gui_slider_release(void *arg, gfx_event_t *event)
{
  gui_slider_t *B;
  B= arg;
  B->pressed= 0;
  slider_move_button(B, event);
  uninstall_motion_handler();
  return 0;
}

static int gui_slider_click(void *arg, gfx_event_t *event)
{
  gui_slider_t *B;
  B= arg;

  if (!B->widget.shown) return 0; // shouldn't happen, but you never know

  B->pressed= 1;
  slider_move_button(B,event);
  install_release_handler(arg, gui_slider_release);
  install_motion_handler(B, gui_slider_motion);
  return 0;
}

static void gui_slider_destroy(widget_t *B)
{
  gui_slider_t *S;
  S= (gui_slider_t*) B;
  remove_click_handler(S->click_handler);
}


static void gui_slider_position(widget_t *W, int x,int y,int w,int h)
{
  gui_slider_t *B;
  B= (gui_slider_t*) W;
  set_click_handler(B->click_handler, x,y,w,h);
  B->widget.x= x;
  B->widget.y= y;
  B->widget.w= w;
  B->widget.h= h;
  slider_layout(B);
  if (B->widget.shown) slider_render(B);
}

static void gui_slider_show(widget_t *B)
{
  slider_render((gui_slider_t*) B);
}


static widget_t *gui_slider_new(int orient)
{
  gui_slider_t *B;

  B= calloc(1,sizeof(*B));
  B->widget.klass= GFXCLASS_SLIDER;
  B->orient= orient;
  B->widget.show= gui_slider_show;
  B->widget.hide= NULL;
  B->widget.destroy= gui_slider_destroy;
  B->widget.position= gui_slider_position;

  B->click_handler= install_click_handler(0,0,0,0, &B->widget.shown,
                                           B, gui_slider_click,1);
  return &B->widget;
}

void gfx_slider_set(int wno, double pos)
{
  widget_t *W;
  gui_slider_t *B;
  W= get_widget(wno);
  if (!W || W->klass!=GFXCLASS_SLIDER) return ;
  B= (gui_slider_t*) W;
  if (pos<0.0) pos= 0.0;
  if (pos>1.0) pos= 1.0;
  B->pos= pos;
  slider_layout(B);
  slider_render(B);
}

double gfx_slider_get(int wno)
{
  widget_t *W;
  gui_slider_t *B;
  W= get_widget(wno);
  if (!W || W->klass!=GFXCLASS_SLIDER) return 0.0;
  B= (gui_slider_t*) W;
  return B->pos;
}

/***********************************************************************
 *                                                                     *
 *                         CONFIRMATION DIALOG                         *
 * #D.ASK                                                              *
 ***********************************************************************
 The confirmation dialog asks the user a question and returns an answer.
 If the user tries to close the window, -1 is returned. Otherwise, the
 index of the clicked button is returned.

 The button string consists of a series of button labels, seperated
 by a particular character. This character is the first byte in the
 button string. For example:

  "!OK!Cancel"

 describes a two-button dialog. Return value 0 corresponds to the OK
 button and 1 corresponds to the Cancel button.

 fmt is a printf format, and the rest of the arguments are printf
 arguments.
 ***********************************************************************/


static void split_string_bychar_d(char *str, int chr, char ***Ra,int *Rn)
{
  int d,n;
  char *x,*k,**R;
 
  R= NULL;
retry:
  n= 0;
  d= 0;
  x= str;
  while(!d)
  {
    k= strchr(x,chr);
    if (!k)
    {
      k= x+strlen(x);
      d= 1;
    }
    if (R)
    {
      R[n]= x;
      k[0]= 0;
    }
    n++;
    if (!d) { x= k+1; }
  }

  if (!R)
  {
    R= malloc(sizeof(*R)*(n+1));
    goto retry;
  }
  R[n]= NULL;
  *Ra= R;
  *Rn= n;
}

static void multi_line_str_dim
  (int font,char **lines, int nlines,int *Rw,int *Rh)
{
  int fh,tw;
  int i,w,h;

  get_text_data(font, "My",NULL,&fh,NULL,NULL);
  w= 0; 
  h= 0;
  for(i=0;i<nlines;i++)
  {
    get_text_data(font, lines[i],&tw,NULL,NULL,NULL);
    if (tw>w) w= tw;
    if (i) h+= fh/10;
    h+= fh;
  }
  *Rw= w;
  *Rh= h;
}


static void gfxask_make_buttons
  (char *butstr,int *coor,int *Rnbut, int *Rbut_h,int *Rbut_total_w)
{
  int i,x,dummy,nr;
  char **buttons; int nbuttons;
  int button_h, button_w; // , button_total_w;
  int fh1;             // font height for font 1


  butstr= strdup(butstr);
  split_string_bychar_d(butstr+1,butstr[0],&buttons,&nbuttons);

  if (nbuttons>8) nbuttons= 8;

  multi_line_str_dim(1,buttons,nbuttons, &button_w, &dummy);
  button_w += RES(confirm.button_padding);
  *Rbut_total_w= button_w*nbuttons + (nbuttons-1)*RES(confirm.button_spacing);

  get_text_data(1,"M",NULL,&fh1,NULL,NULL);
  button_h= fh1 + RES(confirm.button_padding);

  x= 0;
  for(i=0;i<nbuttons;i++)
  {
    nr= GUIRaskB1+i;
    if (!widgets[nr]) 
    {
      widgets[nr]= (widget_t*) gui_button_new(buttons[i], 0, 0, 
                                               GFXCLASS_BUTTON);
      widgets[nr]->number= nr;
    }
    else
    {
      gui_button_set_text((gui_button_t*) widgets[nr], buttons[i]);
    }
    coor[4*i+0]= x;
    coor[4*i+2]= button_w;
    coor[4*i+3]= button_h;
    x+= button_w + RES(confirm.spacing);
  }
  free(buttons);
  free(butstr); 
  *Rnbut= nbuttons;
  *Rbut_h= button_h;
}

static void gfxask_show_buttons
    (int *coor_widgets, int nbuttons,int X,int Y)
{
  int i;
  for(i=0;i<nbuttons;i++)
  {
    int nr;
    nr= GUIRaskB1+i;
    coor_widgets[4*i+0] += X;
    coor_widgets[4*i+1] = Y;
    gui_button_activate(widgets[nr], 
           coor_widgets[4*i+0], coor_widgets[4*i+1],
           coor_widgets[4*i+2], coor_widgets[4*i+3]);
  }
}

static void gfxask_draw_border(int X,int Y,int W,int H)
{
  int i;
  int border;
  border= color_u_to_s(RES(confirm.border_color));
  for(i=1;i<=RES(confirm.border_width);i++)
  {
    hline(X-i,    X+W-1+i, Y-i     ,border);
    hline(X-i,    X+W-1+i, Y+H-1+i ,border);
    vline(X-i,    Y-i,     Y+H-1+i ,border);
    vline(X+W-1+i,Y-i,     Y+H-1+i ,border);
  }
  gfx_fill_rect(0, X,Y,W,H, RES(confirm.bg));
}

static void format_multi_line_str(char *fmt, va_list ap, char ***Rl, int *Rn)
{
  char *all;
  all= vmprintf(fmt, ap);
  split_string_bychar_d(all,'\n',Rl, Rn);
}

int gfx_ask(char *title, char *butstr, char *fmt, ...)
{
  va_list ap;
  int buttons_h, buttons_w, nbuttons;
  int titw, tith, titx, tity;
  int R;

  int W,H,X,Y; // for the whole dialog
  char **lines; int nlines, lw, lh,lx,ly;

  int i;

  static int dlg_widgets[]= 
    { GUIRaskB1, GUIRaskB2, GUIRaskB3, GUIRaskB4, GUIRaskB5,
      GUIRaskB6, GUIRaskB7, GUIRaskB8 };
  int coor_screen[4];
  int coor_widgets[4*8];

  // LAYOSCR_SCREEN_0(LL,x_window_width,x_window_height);
  gfxask_make_buttons(butstr, coor_widgets, 
                       &nbuttons, &buttons_h, &buttons_w);

  if (title)
    get_text_data(2,title,&titw,&tith,NULL,NULL);

  va_start(ap,fmt);
  format_multi_line_str(fmt, ap, &lines, &nlines);
  va_end(ap);
  multi_line_str_dim(1,lines,nlines,&lw,&lh);

  W= imax2(lw,buttons_w)+2*RES(confirm.spacing);
  H= 3*RES(confirm.spacing) + lh + buttons_h;

  if (title) H+= RES(confirm.spacing) + tith;
  if (title) W= imax2(W, titw+2*RES(confirm.spacing));

  if (H>x_window_height) Y= x_window_height - H;
                    else Y= (x_window_height-H)/2;
  if (W>x_window_width)  X= 0;
                    else X= (x_window_width-W)/2;

  coor_screen[0]= X; coor_screen[1]= Y;
  coor_screen[2]= W; coor_screen[3]= H;

  enter_screen(8, dlg_widgets, coor_screen);

  // FIXME: make it possible to reposition the dialog
  // when the containing window gets resized..
  // LAYOWID_SCREEN_0(x_window_width,x_window_height,LC);

  gfxask_draw_border(X,Y,W,H);
  gfxask_show_buttons(coor_widgets, nbuttons,
          X+(W - buttons_w)/2,
          Y+H - buttons_h - RES(confirm.spacing));

  if (lw>W) lx= X+0;
       else lx= X+(W-lw)/2;
  ly= Y+ RES(confirm.spacing);
  if (title) ly+= tith+RES(confirm.spacing);

  int fh;
  get_text_data(1, "My",NULL,&fh,NULL,NULL);
  for(i=0;i<nlines;i++)
  {
    txt_print(0, 1, RES(confirm.fg), RES(confirm.bg), lx,ly, lines[i], NULL,&fh);
    ly+= fh + 1 + fh/10;
  } 
  free(lines[0]);
  free(lines);

  if (title)
  {
    if (titw>W) titx= X+0;
           else titx= X+(W-titw)/2;
    tity= Y+RES(confirm.spacing);
    txt_print(0, 2, RES(confirm.fg), RES(confirm.bg),
              titx,tity, title, NULL,NULL);
  }

  gfx_update_all();

  R= -1;
  while(1)
  {
    gfx_event_t event;
    if (!gfx_event(&event)) continue;
    if (event.type==GFXEV_QUIT) { R= -1; break; }
    R= event.type- GUIRaskB1;
    if (R<0 || R>=nbuttons) continue;
    break; 
  }
  gui_leave_screen();
  return R;
}

/***********************************************************************
 *                                                                     *
 *                        FILE SELECTOR DIALOG                         *
 * #D.FILE                                                             *
 ***********************************************************************
The file selector will consist of:
 - a list of files & directories
 - a text entry 
 - some buttons
 - a file type filter
 - possibly another list of bookmarks
 - an optional preview area

I will initially do the first four. Later on, I can make the bookmarks and
preview stuff.

File type filters will be hierarchial lists. For example, we can have:

  All files
    Image Files
      JPEG images
      PNG  images
    Audio Files
      WAV files
      MP3 files
    Text Files
      Plain Text
      HTML

etc. So, an element of the list will contain:
 - a display string
 - a file extension list (.jpg .jpeg etc.)
 - a level

For non-leaf elements, there will be no separate extension list but the 
extensions of the child nodes will be used. For a simpler approach, I
will make a list parser for this and there won't be any other data
structure etc.

The grammar will be as follows:

  name\n
    name (extensionlist)\n

The number of spaces leading the name will determine how deep the element
will be nested. extensions in the list will be parsed as needed, there
won't be any intermediate structure for this. In fact, I can use an array
of strings for this so that each entry is a separate string. So, an example
list will be:

  "Image Files",
  "  JPEG images (*.jpg *.jpeg)",
  "  PNG images (*.png)",
  "Audio Files",
  "  Microsoft WAVE format (*.wav)",
  "  MPEG Layer 3 (*.mp3)",
  NULL

Yeah, that's much better. So, the only thing I need is to parse the extension
list.

For higher level entries, we need to make the list from the lower level
entries.
 ***********************************************************************/

typedef struct
{
  char *label;
  char *dir;
  int width;
} fsent_t;

typedef struct
{
  int n_dirs;
  fsent_t *dirs;
  int dir_start, dir_end;

  int nfiles, shown;
  char **files;

  char **types;
  char **ext;

  char *dir;
  int x,y,w,h;
  int button_h;
  int dirst_y;
} filesel_t;

                 /*******
   Here is some code which parses the input filter. 
                  *******/

static char **parse_ext_list(char *inp)
{
  char **R;
  char *K,*E,*S;
  int N;

  inp= strchr(inp, '(');
  if (!inp) return NULL;
  R= NULL;

again:
  N= 0;
  K= inp+1;
  while(K[0]!=0 && K[0]!=')')
  {
    for(S=K;S[0]!='.' && S[0]!=0 && S[0]!=')';S++)  ;
    if (S[0]!='.') { K=S; continue; }
    for(E=S+1;E[0]!=0 && E[0]!=')' && E[0]!=' ' && E[0]!=',';E++)  ;
    if (E==S+1) { K=E; continue; }
    if (R)
    {
      R[N]= malloc(E-S+1);
      memcpy(R[N],S, E-S);
      R[N][E-S]= 0;
    }
    N++;
    K= E;
  }
  if (!R)
  {
    R= malloc(sizeof(*R)*(N+1));
    goto again;
  }
  R[N]= NULL;
  return R;
}

static char **make_file_extension_list(char **inp)
{
  int i,j;
  char **R;
  int nent,ne;
  char ***L;

  if (!inp || !inp[0] || !inp[0][0]) return NULL;
  if (strchr(inp[0],'('))
  {
    R= parse_ext_list(inp[0]);
    goto done;
  }

  {
    int first_spc;
    for(i=0;inp[0][i]==' ';i++) ;
    first_spc= i;

    for(i=1;inp[i];i++)
    {
      for(j=0;inp[i][j]==' ';j++) ;
      if (j<=first_spc) break;
    }
    nent= i;
  }
  if (nent==1) return NULL;

  ne= 0;
  L= calloc(nent,sizeof(*L));
  for(i=0;i<nent;i++)
  {
   if (!strchr(inp[i],'(')) continue;
    L[i]= parse_ext_list(inp[i]);
    if (L[i])
    {
      for(j=0;L[i][j];j++) ;
      ne+= j;
    }
  }
  R= malloc(sizeof(*R)*(ne+1));
  ne= 0;
  for(i=0;i<nent;i++)
  {
    if (L[i])
    {
     for(j=0;L[i][j];j++)
       R[ne++]= L[i][j];
     free(L[i]);
    } 
  }
  R[ne]= NULL;
  free(L);

done:
  return R; // maybe sort the list, and unify it??
}

static int file_matches_ext_set(char *fn,char **ext)
{
#ifdef GFX_UNIX
  int i;
  int lf,le;
  if (!ext) return 1;
  lf= strlen(fn);
  for(i=0;ext[i];i++)
  {
    if (!strcmp(ext[i],".*")) return 1;
    le= strlen(ext[i]);
    if (lf<=le) continue;
    if (!strcasecmp(fn+lf-le,ext[i])) return 1;
  }
#endif
  return 0;
}


                /******
Now the extension stuff is done. Next thing is to find the file list.
I will first discover all the files and then apply the filters. Each entry
will be represented as a string. The first character will be the file type
and the rest will be the name. 
                 ******/

static char **list_files(char *dn)
{
#ifdef GFX_UNIX
  DIR *dir;
  struct dirent *de;
  char **ptr;
  char *str;
  size_t i, ptrsize, strsize, L;
  buffer_t *buffer;

  dir= opendir(dn);
  if (!dir)
  {
    ptr= malloc(sizeof(*ptr));
    ptr[0]= NULL;
    return ptr;
  }

  buffer= buffer_new(1024);
  ptrsize= strsize= 0;
  while((de=readdir(dir)))
  {
    buffer_write(buffer, "U", 1);
    L= strlen(de->d_name)+1;
    buffer_write(buffer, de->d_name, L);
    strsize+= L+1;
    ptrsize++;
  }
  closedir(dir);

  ptr= malloc( (ptrsize+1)*sizeof(char*) + strsize );
  str= (char*) ptr + (ptrsize+1)*sizeof(char*) ;

  buffer_read(buffer, str, strsize);
  buffer_free(buffer, NULL, NULL);

  for(i=0;i<ptrsize;i++)
  {
    ptr[i]= str;
    str+= strlen(str)+1;
  }
  ptr[i]= NULL;
  
  return ptr;
#else
	return NULL;
#endif
}

static void file_types(char *dn, char **files)
{
#ifdef GFX_UNIX
  size_t i;
  size_t mlen,L, dlen;
  char *path,*name;

  mlen= 0;
  for(i=0;files[i];i++)
  {
    L= strlen(files[i]);
    if (L>mlen) mlen= L;
  }

  dlen= strlen(dn);
  path= malloc(dlen+2+mlen); // slash + null
  name= path+dlen+1;
  memcpy(path, dn, dlen);
  path[dlen]= '/';

  for(i=0;files[i];i++)
  {
    struct stat sbuf;
    strcpy(name, files[i]+1);
    if (stat(path,&sbuf)) continue;
    if (S_ISDIR(sbuf.st_mode))      files[i][0]= 'D';
    else if (S_ISREG(sbuf.st_mode)) files[i][0]= 'F';
  } 

  free(path);
#endif
}

static int filter_files_byext(int nfiles,char **files, char **suffixes)
{
  int i;
  char *tmp;
  if (!nfiles) return 0;
  if (!suffixes) return nfiles;
  for(i=0;i<nfiles;i++)
  {
    if (files[i][0]!='F') continue;
    if (file_matches_ext_set(files[i],suffixes)) continue;
    tmp= files[nfiles-1];
    files[nfiles-1]= files[i];
    files[i]= tmp;
    nfiles--;
    i--;
  } 
  return nfiles;
}

static int filter_files_byhid(int nfiles,char **files,int showhid)
{
  int i, last;
  char *tmp;
  if (!nfiles) return 0;
  if (showhid) return nfiles;
  last= nfiles-1;
  for(i=0;i<last;i++)
  {
    if (files[i][1]!='.') continue;
    tmp= files[last];
    files[last]= files[i];
    files[i]= tmp;
    last--;
    i--;
  } 
  if (files[last][1]!='.') return last+1;
  return last;
}

static int comparstr(const void *a,const void* b)
{
  return strcmp(*(const char**) a, *(const char**) b);
}

static void sort_files(int nfiles, char **files)
{
  qsort(files, nfiles, sizeof(files[0]), comparstr);
}


static char *com3(char *x,char *y,char *z)
{
  char *R;
  int Lx,Ly,Lz;
  Lx= strlen(x); Ly= strlen(y); Lz= strlen(z);
  R= malloc(Lx+Ly+Lz+1);
  memcpy(R, x, Lx); memcpy(R+Lx,y,Ly); memcpy(R+Lx+Ly,z,Lz);
  R[Lx+Ly+Lz]= 0;
  return R;
}

static char *com2(char *x,char *y)
{
  char *R;
  int Lx,Ly;
  Lx= strlen(x); Ly= strlen(y);
  R= malloc(Lx+Ly+1);
  memcpy(R, x, Lx); memcpy(R+Lx,y,Ly);
  R[Lx+Ly]= 0;
  return R;
}

static char *fs_combine(char *x,char *y)
{
  if (!x) return strdup(y);
  if (x[0]=='/' && x[1]==0) return com2(x,y);
  return com3(x,"/",y);
}

static char *sau_getcwd()
{
#ifdef GFX_UNIX
  char *buf; size_t siz;

  siz= 0;
  
  while(1)
  {
    if (!siz) { siz= 256;  }
         else { siz+= siz/2; free(buf); }
    buf= malloc(siz);
    if (getcwd(buf,siz)) return buf;
    if (errno!=ERANGE)   return NULL;
  }
#else
	return "/";
#endif
}

static char *sau_gethome()
{
  char *x;
  x= getenv("HOME");
  if (!x) x= "/";
  return strdup(x);
}

static char* combine_path(char *dir,char *fn)
{
  char *R;
  int Ld;
  Ld= strlen(dir);
  R= malloc(Ld+strlen(fn)+2);
  if (Ld>0 && dir[Ld-1]=='/')
    sprintf(R,"%s%s",dir,fn);
  else
    sprintf(R,"%s/%s",dir,fn);
  return R;
}

static void split_initial_fn(char *init, char **Rdir, char **Rfil)
{
  char *dir; // , *fil;
  char *pfil,*pdir;
  char *dir_pfx;

  if (!init)
  {
    if (!(dir= sau_getcwd())) dir= sau_gethome();
    *Rdir= dir;
    *Rfil= NULL;
    return ;
  }
  
  pfil= strrchr(init,'/');
  if (pfil) 
  {
    pfil++;  
    pdir= strdup(init); 
    pdir[pfil-init]= 0; 
    if (!pfil[0]) pfil= NULL;
             else pfil= strdup(pfil);
  }
  else 
  { 
    pdir= NULL; 
  }

  if (init[0]=='/')
  {
    *Rdir= pdir;
    *Rfil= pfil;
    return ;
  }

  dir_pfx= sau_getcwd();
  if (!dir_pfx)
  {
    // we can't get the current directory, the rest of the 
    // path is meaningless except for the file name, if any.
    *Rdir= sau_gethome();
    *Rfil= pfil; 
    if (pdir) free(pdir);
    return ;
  }

  if (pdir) 
  {
    dir= malloc(strlen(pdir) + strlen(dir_pfx)+1);
    sprintf(dir,"%s/%s",dir_pfx,pdir);
    free(pdir);
    free(dir_pfx);
  }
  else
  {
    dir= dir_pfx;
  }

  *Rdir= dir;
  *Rfil= pfil;
}


RES_STRING(filesel_lbl_ok,"OK","Label for OK button");
RES_STRING(filesel_lbl_cancel, "Cancel", "Label for Cancel button");
RES_STRING(filesel_lbl_refresh, "Refresh", "Label for Refresh button");
RES_STRING(filesel_lbl_home, "Home", "Label for Home button");
RES_STRING(filesel_lbl_load, "Select File to Load", "File selector load mode title");
RES_STRING(filesel_lbl_save, "Select File to Save", "File selector write mode title");



static void filesel_make_screen(int *RX,int *RY,int *RW,int *RH,int title_w)
{
  int W,H,X,Y;
  int coor_screen[4];
  // this list should be the same order as in the enum definition
  static int dlg_widgets[]= {
    GUIRfilD1, GUIRfilD2, GUIRfilD3, GUIRfilD4, GUIRfilD5, GUIRfilD6,
    GUIRfilD7, GUIRfilD8, GUIRfilD9, GUIRfilD10, 
    GUIRfilHOME, GUIRfilREFRESH, GUIRfilOK, GUIRfilCANCEL, GUIRfilLIST,
    GUIRfilNAME,
  };

  W= imin2(imax2(imax2(400,x_window_width *8/10),title_w+2*RES(fileselector.spacing)),
           x_window_width);
  H= imin2(imax2(400,x_window_height*8/10),x_window_height);
  X= (x_window_width-W)/2;
  Y= (x_window_height-H)/2;

  coor_screen[0]= X; coor_screen[1]= Y;
  coor_screen[2]= W; coor_screen[3]= H;

  enter_screen(15, dlg_widgets, coor_screen);
  *RX= X; *RY= Y; *RW= W; *RH= H;

  gfx_fill_rect(0,X,Y,W, H,RES(fileselector.bg));
}

static int filesel_tab_complete(void *arg,void *te,gfx_event_t *event)
{
  if (event->type!=GFXEV_KEY_PRESS) return 0;
  switch(event->key)
  {
  case GFX_K_TAB:
     break;
  case GFX_K_RETURN:
     event->type= GUIRfilOK;
     break;
  case GFX_K_ESCAPE:
     event->type= GUIRfilCANCEL;
     break;
  }
  return 0; 
}


static void file_select_dialog
  (filesel_t *fs,int mode,char *ifil)
{
  char *title;
  text_entry_t *entry;
  int i;

  int W,H,X,Y;

  int button_w, button_h;
  int title_w, title_h, title_x, title_y;
  int list_w, list_h, list_x, list_y;
  // int dirst_w, dirst_h;
  int dirst_x, dirst_y;

  int homB_x, homB_y, refB_x, refB_y, okB_x, okB_y, canB_x, canB_y, 
      tex_x, tex_y, tex_w, tex_h;

  int fh1,fh3;

  gfx_text_size(1,"M",NULL,&fh1);
  gfx_text_size(3,"M",NULL,&fh3);

  title= (char*) (mode ? filesel_lbl_save : filesel_lbl_load);
  gfx_text_size(3, title, &title_w, &title_h);

  filesel_make_screen(&X,&Y,&W,&H, title_w);

  button_w= max_text_width(1, filesel_lbl_cancel, filesel_lbl_ok,
                              filesel_lbl_refresh, filesel_lbl_home, NULL) 
           + RES(fileselector.padding);
  button_h= fh1 + RES(fileselector.padding);

  title_x= (W - title_w)/2;
  title_y= RES(fileselector.spacing);

  dirst_y= title_y + title_h + RES(fileselector.spacing);
  dirst_x= RES(fileselector.dirstack.left);

  homB_x= RES(fileselector.spacing);
  homB_y= dirst_y + button_h + RES(fileselector.spacing);

  refB_x= W - RES(fileselector.spacing) - button_w;
  refB_y= homB_y;

  tex_x= homB_x + button_w + RES(fileselector.spacing);
  tex_y= homB_y;
  tex_w= refB_x - RES(fileselector.spacing) - tex_x;
  tex_h= button_h;

  list_y= tex_y + tex_h + RES(fileselector.spacing);
  list_x= RES(fileselector.spacing);
  list_w= imax2(1,W-list_x-RES(fileselector.spacing));

  canB_x= refB_x;
  canB_y= H - RES(fileselector.spacing) - button_h;
  okB_x= canB_x - RES(fileselector.spacing) - button_w;
  okB_y= canB_y;

  list_h= okB_y - RES(fileselector.spacing) - list_y;


  widgets[GUIRfilHOME]= gfx_button(filesel_lbl_home,0,0,GFXCLASS_BUTTON);
    button_set_position((gui_button_t*)widgets[GUIRfilHOME],
          X+homB_x, Y+homB_y, button_w, button_h); 
    gui_button_show((gui_button_t*)widgets[GUIRfilHOME],1);

  widgets[GUIRfilREFRESH]= gfx_button(filesel_lbl_refresh,0,0,
                                          GFXCLASS_BUTTON);
    button_set_position((gui_button_t*)widgets[GUIRfilREFRESH],
          X+refB_x, Y+refB_y, button_w, button_h); 
    gui_button_show((gui_button_t*)widgets[GUIRfilREFRESH],1);

  widgets[GUIRfilOK]= gfx_button(filesel_lbl_ok, 0, 0,GFXCLASS_BUTTON);
    button_set_position((gui_button_t*)widgets[GUIRfilOK],
          X+okB_x, Y+okB_y, button_w, button_h); 
    gui_button_show((gui_button_t*)widgets[GUIRfilOK],1);

  widgets[GUIRfilCANCEL]= gfx_button(filesel_lbl_cancel,0,0,GFXCLASS_BUTTON);
    button_set_position((gui_button_t*)widgets[GUIRfilCANCEL],
          X+canB_x, Y+canB_y, button_w, button_h); 
    gui_button_show((gui_button_t*)widgets[GUIRfilCANCEL],1);

  widgets[GUIRfilLIST]= gui_list_new(0,0,NULL,NULL,NULL);
    gui_list_set_position((gui_list_t*)widgets[GUIRfilLIST],
          X+list_x, Y+list_y, list_w, list_h);
  widgets[GUIRfilNAME]= text_entry_new(ifil);
    entry= (text_entry_t*) widgets[GUIRfilNAME];
    entry->key_filter= filesel_tab_complete;
    entry->key_arg= fs;
    textentry_position(entry, X+tex_x, Y+tex_y, tex_w, tex_h);
    widget_show((widget_t*)entry, 1);
    set_focus(widgets[GUIRfilNAME]);

  for(i=GUIRfilD1;i<=GUIRfilD10;i++)
  {
    widgets[i]= gfx_button("", 0, 0, GFXCLASS_TOGGLE);
    button_set_position((gui_button_t*)widgets[i], dirst_x, dirst_y, 
                        100,button_h);
  }

  for(i=GUIRfilD1;i<=GUIRfilNAME;i++)
    widgets[i]->number= i;

  fs->dirst_y= dirst_y;
  fs->x= X;
  fs->y= Y;
  fs->w= W;
  fs->h= H;
  fs->button_h= button_h;

  txt_print(0,3,RES(fileselector.fg), RES(fileselector.bg),
                X+title_x, Y+title_y, title, NULL, NULL);
}

static void file_list_icons(const unsigned int **icons)
{
  if (icons[0]) return ;
  icons[0]= gui_R_COLORIMG_toicon(&RES(fileselector.file));
  icons[1]= gui_R_COLORIMG_toicon(&RES(fileselector.dir));
  icons[2]= gui_R_COLORIMG_toicon(&RES(fileselector.alert));
}

static void populate_file_list (filesel_t *fs)
{
  char **files;
  int nfiles,shown,i;

  static const unsigned int *icons[3];

  file_list_icons(icons);

  if (fs->files) free(fs->files); // the whole array is one piece.
  files= list_files(fs->dir);
  file_types(fs->dir, files);

  gfx_list_show((gui_list_t*) widgets[GUIRfilLIST],0);
  gfx_list_clear((gui_list_t*) widgets[GUIRfilLIST]);
  for(nfiles=0;files[nfiles];nfiles++) ;

  shown= filter_files_byext(nfiles, files,fs->ext);
  shown= filter_files_byhid(shown, files, 0);
  sort_files(shown, files);
  for(i=0;i<shown;i++)
  {
    int isel;
    switch(files[i][0])
    {
    case 'F': isel= 0; break;
    case 'D': isel= 1; break;
    default:  isel= 2; break;
    }
    gfx_list_insert((gui_list_t*) widgets[GUIRfilLIST], -1,
            files[i]+1, files[i], icons[isel]);
  }
  fs->files= files;
  fs->nfiles= nfiles;
  fs->shown= shown;
  gfx_list_show((gui_list_t*) widgets[GUIRfilLIST],1);
}

static char **sepdir(char *name)
{
  char **R;
  int N,done;
  char *k,*str;

  R= NULL;
/* 
    /home/dodo/foo   : 4 components /,home,dodo,foo
     home/dodo/foo   : 3 components home,dodo,foo
*/
again:
  N= 0;
  str= name;

  if (str[0]=='/')
  {
    if (R) R[N]= strdup("/"); 
    N++;
    while(str[0]=='/') str++;
  }

  done= 0;
  if (!str[0]) done= 1;
  while(!done)
  {
    k= strchr(str,'/');
    if (!k) { k= str+strlen(str); done= 1; }
    if (R)
    {
      R[N]= malloc(k-str+1);
      memcpy(R[N], str, k-str);
      R[N][k-str]= 0;
    }
    N++;
    str= k+1;
  }

  if (R) { R[N]= NULL; return R; }
  R= malloc(sizeof(*R)*(N+1));
  goto again;
}

                       /*************** 
I need to make the widgets. The stack will consist of active and
inactive widgets. As I browse thru the directories, the buttons will be
re-used.  I will have 10 buttons, more buttons will simply clutter the
area and won't contribute anything.
                        ***************/



static void filesel_hide_buttons()
{
  int i;
  for(i=GUIRfilD1;i<=GUIRfilD10;i++)
    gui_button_show((gui_button_t*) widgets[i], 0);
}

static void filesel_show_buttons(filesel_t *fs)
{
  int i;
  for(i=fs->dir_start;i<=fs->dir_end;i++)
    gui_button_show((gui_button_t*) widgets[i-fs->dir_start+GUIRfilD1], 1);
}

static char* remove_trailing_slash(char *dn)
{
  int L;
  L= strlen(dn);
  while(L>1 && dn[L-1]=='/') { dn[L-1]= 0; L--; }
  return dn;
}

static void filesel_free_buttons(filesel_t *fs)
{
  int i;
  if (!fs->dirs) return ;
  for(i=0;i<fs->n_dirs;i++)
  {
    free(fs->dirs[i].label);
    free(fs->dirs[i].dir);
  }
  free(fs->dirs);
  fs->dirs= NULL;
  fs->n_dirs= 0;
}


static void filesel_create_buttons(filesel_t *fs)
{
  int i;
  char *dn;
  char **comp;
  int ncomp;
  char *Z;

  dn= remove_trailing_slash(strdup(fs->dir));

  comp= sepdir(dn);
  for(ncomp=0;comp[ncomp];ncomp++) ;

  filesel_free_buttons(fs);
  fs->n_dirs= ncomp;
  fs->dirs= calloc(ncomp,sizeof(fs->dirs[0]));
  Z= NULL;
  for(i=0;i<fs->n_dirs;i++)
  {
    fs->dirs[i].dir= fs_combine(Z, comp[i]);
    fs->dirs[i].label= comp[i];
    Z= fs->dirs[i].dir;
  }
  free(comp);
  free(dn);
}

static void filesel_activate_button(filesel_t *fs,int idx)
{
  int i;
  for(i=0;i<=fs->dir_end-fs->dir_start;i++)
    button_set_state((gui_button_t*)widgets[GUIRfilD1+i], i==idx);
}

static void filesel_render_buttons(filesel_t *fs)
{
  int i;
  int x,y;

  y= fs->y+fs->dirst_y;
  x= fs->x+RES(fileselector.dirstack.left);

  gfx_fill_rect(1, fs->x, y, fs->w, fs->button_h, RES(fileselector.bg));
  
  for(i=fs->dir_start;i<=fs->dir_end;i++)
  {
    gui_button_t *B;
    B= (gui_button_t*) widgets[i-fs->dir_start+GUIRfilD1];
    gui_button_set_text(B, fs->dirs[i].label);
    if (!fs->dirs[i].width)
    {
      gfx_text_size(1, fs->dirs[i].label, &fs->dirs[i].width, NULL);
      fs->dirs[i].width+= 2*RES(fileselector.dirstack.padding);
    }
    button_set_position(B, x, y, fs->dirs[i].width, fs->button_h);
    x+= RES(fileselector.dirstack.spacing)+fs->dirs[i].width;
    button_set_state(B, 0);
  }
}
           /***********
The following will switch to the given directory when clicked thru the
directory stack.
            ***********/

static void filesel_setdir(filesel_t *fs,char *dir)
{
  if (fs->dir) free(fs->dir);
  fs->dir= strdup(dir);
}

static void filesel_activate_dir(filesel_t *fs, int idx)
{
  filesel_activate_button(fs,idx);
  filesel_setdir(fs,fs->dirs[idx+fs->dir_start].dir);
  populate_file_list(fs);
}

static void filesel_render_dirstack(filesel_t *fs)
{
  int i;
  char *dn;

  filesel_hide_buttons();
  dn= remove_trailing_slash(strdup(fs->dir));

again:

  for(i=0;i<fs->n_dirs;i++)
    if (!strcmp(dn,fs->dirs[i].dir)) break;
  if (i!=fs->n_dirs)
  {
    int st,en,K;
    K= i;
    st= imax2(0,K-4);
    en= imin2(K+5,fs->n_dirs-1);
    if (fs->dir_start==st && fs->dir_end==en)
    {
      filesel_activate_button(fs, K-st);
      filesel_show_buttons(fs);
      return ;
    }
    fs->dir_start= st;
    fs->dir_end= en;
    filesel_render_buttons(fs);
    filesel_activate_button(fs, K-st);
    filesel_show_buttons(fs);
    return ;
  }

  filesel_create_buttons(fs);
  fs->dir_start= fs->dir_end= -1;
  goto again;
}



                      /*****************

at any time, the current directory needs to be visible, along with both
its immediate predecessor and successor in the directory stack.

If we have more directory components than slots, then we need to cut
it off somehow. It's best to position it in the middle. So, if our
current directory is at position K, and the total number of positions
is N, then we need to show positions

    K-4 K-3 K-2 K-1 K K+1 K+2 K+3 K+4 K+5

so, the minimum entry is:

  max(0,K-4)

maximum entry is

  min(N-1,K+5)
                        *****************/


char* gfx_select_file(int mode,char *init,char **types)
{
#ifdef GFX_UNIX
  char *idir, *ifil;
  char *x;
  filesel_t *fs;
  int q;
  char *result;

  if (mode) mode= 1;

  split_initial_fn(init, &idir, &ifil);
  fs= calloc(1,sizeof(*fs));
  fs->dir_start= fs->dir_end= -1;
  filesel_setdir(fs,idir);
  free(idir);
  fs->types= types;
  fs->ext= make_file_extension_list(types);
  file_select_dialog(fs,mode,ifil);
  filesel_render_dirstack(fs);
  populate_file_list(fs);
  gfx_update_all();

  if (ifil) free(ifil);
  ifil= NULL;

retry:
  q= -1;

  while(q<0)
  {
    gfx_event_t event;
    if (gfx_event(&event))
     switch(event.type)
     {
     case GUIRfilCANCEL:
     case GFXEV_QUIT:   q= 0; break;
     case GUIRfilOK: q= 1; break;
     case GUIRfilREFRESH:
       populate_file_list(fs);
       break;
     case GUIRfilHOME:
       filesel_setdir(fs, sau_gethome());
       populate_file_list(fs);
       filesel_render_dirstack(fs);
       break;
    case GUIRfilLIST:
       x= event.usr;
       if (!x) continue;
       if (x[0]=='D')
       {
         filesel_setdir(fs, combine_path(fs->dir,x+1));
         populate_file_list(fs);
         filesel_render_dirstack(fs);
         break;
       }
       if (x[0]=='F')
       {
         text_entry_set((text_entry_t*) widgets[GUIRfilNAME],x+1);
         break;
       }
       break;
     case GUIRfilD1: case GUIRfilD2: case GUIRfilD3: case GUIRfilD4:
     case GUIRfilD5: case GUIRfilD6: case GUIRfilD7: case GUIRfilD8:
     case GUIRfilD9: case GUIRfilD10:
       filesel_activate_dir(fs, event.type-GUIRfilD1);
       break;
     }
  }
  if (q)
  {
    char *fn, *path;
    int exists, isfile;
    int valid;

    struct stat sbuf;

    fn= textentry_get_text((text_entry_t*)widgets[GUIRfilNAME]);
    if (!fn[0]) { free(fn); goto retry; }
    path= combine_path(fs->dir, fn);

    if (stat(path,&sbuf))                 exists=0; else exists=1;
    if (exists && S_ISREG(sbuf.st_mode)) isfile= 1; else isfile= 0;

    valid= 1;
    if (exists & !isfile)
    {
      gfx_ask("Invalid Selection", "!OK",
              "Selected entry\n\n%s\n\nis not a regular file", path);
      valid= 0; 
    }
    else if (mode==0 && !exists)
    {
      gfx_ask("Invalid Selection", "!OK",
              "Entered path\n\n%s\n\ndoes not exist.", path); 
      valid= 0; 
    }
    else if (mode==1 && exists)
    {
       if ((!init ||strcmp(init,path)) && 
               gfx_ask("Confirm Overwrite", "!OK!Cancel",
                   "%s\n\nDo you want to overwrite this file?", path))
          valid= 0; 
    }
    if (!valid) { free(path); free(fn); goto retry; }
    result= path;
    free(fn);
  }
  else
    result= NULL;

  gui_leave_screen();

  if (fs->ext)
  {
    int i;
    for(i=0;fs->ext[i];i++) free(fs->ext[i]);
    free(fs->ext[i]);
  }
  filesel_free_buttons(fs);
  free(fs->dir);
  free(fs->files);
  free(fs);
  
  return result;
#endif
#ifdef GFX_WIN
  return NULL;
#endif
}



/***********************************************************************
 *                                                                     *
 *                              FILE LIST                              *
 * #W.FILELIST                                                         *
 ***********************************************************************/

// #W.FILELIST
typedef struct
{
  widget_t widget;

  char *dir;
  int nfiles;
  char **files;

  char **extensions;

  gui_list_t *w_list;
  dropdown_t *w_type;
  gui_label_t *w_typelab;
  checkbox_t *w_hidden;

  int show_hidden;

  void *evarg;
  event_function_t clickf,dlickf;
} gui_filelist_t;


static int filter_files_byspecial(int nfiles,char **files)
{
  int i;
  char *t;
  for(i=0;i<nfiles;i++)
  {
    if (!strcmp(files[i]+1,".") || !strcmp(files[i]+1,".."))
    {
     t= files[nfiles-1];
     files[nfiles-1]= files[i];
     files[i]= t;
     nfiles--;
     i--;
    }
  }
  return nfiles;
}

static void gui_filelist_set_directory(gui_filelist_t *FL,char *dir)
{
  int i;
  static const unsigned int *icons[3];
  static char fuu[]= "..";

  file_list_icons(icons);

  dir= path_absolute(dir);
  if (FL->dir) free(FL->dir);
  if (FL->files) free(FL->files);
  FL->nfiles= 0; FL->files= NULL; FL->dir= NULL;

  FL->dir= dir;
  FL->files= list_files(FL->dir);
  file_types(FL->dir, FL->files);

  for(i=0;FL->files[i];i++) ;
  FL->nfiles= i;

  FL->nfiles= filter_files_byspecial(FL->nfiles,FL->files);
  FL->nfiles= filter_files_byhid(FL->nfiles,FL->files,FL->show_hidden);
  FL->nfiles= filter_files_byext(FL->nfiles,FL->files,FL->extensions);
  sort_files(FL->nfiles,FL->files);

  if (FL->widget.shown) gfx_list_show(FL->w_list, 0);
  gfx_list_clear(FL->w_list);

  if (strcmp(FL->dir,"/"))
    gfx_list_insert(FL->w_list, -1, "<PARENT DIRECTORY>", fuu, icons[1]);
  for(i=0;i<FL->nfiles;i++)
  {
    gfx_list_insert( FL->w_list, -1, FL->files[i]+1, FL->files[i], 
           icons[ FL->files[i][0]=='F' ? 0: (FL->files[i][0]=='D'?1:2) ] );
  }
  if (FL->widget.shown) gfx_list_show(FL->w_list,1);
}

static void gui_filelist_show_all(gui_filelist_t *FL,int show)
{
  widget_show(&FL->w_list->widget,show);
  if (FL->w_typelab) widget_show(&FL->w_typelab->widget,show);
  if (FL->w_type) widget_show(&FL->w_type->widget, show);
  if (FL->w_hidden) widget_show(&FL->w_hidden->widget, show);
}

static void gui_filelist_show(widget_t *W)
{
  gui_filelist_t *FL;
  FL= (gui_filelist_t*) W;
  if (W->shown)
    gfx_fill_rect(0, W->x, W->y, W->w, W->h, 0xededed);
  gui_filelist_show_all(FL, 1);
  if (W->shown)
    gfx_update(W->x, W->y, W->w, W->h);
}
static void gui_filelist_hide(widget_t *W)
{
  gui_filelist_t *FL;
  FL= (gui_filelist_t*) W;
  gui_filelist_show_all(FL, 0);
}


static void gui_filelist_position(widget_t *W, int x,int y,int w,int h)
{
  gui_filelist_t *FL;
  int PAD,BUTH,LABW,XPAD;
  int bot, top,left;

// FIXME: convert these to resources
  PAD= 8;
  XPAD= 0;
  BUTH= 28;
  LABW= 120;
  FL= (gui_filelist_t*) W;

  W->x= x; W->y= y; W->w= w; W->h= h;
  if (W->shown)
    gfx_fill_rect(0, x,y,w,h, 0xededed);

  bot= h-PAD;

  if (FL->w_hidden)
  {
    top= bot - BUTH;
    widget_move(&FL->w_hidden->widget, W->x+XPAD, W->y+top, w-2*XPAD, BUTH);
    bot= top-PAD;
  }
  if (FL->w_typelab)
  {
   left= XPAD;
   top= bot-BUTH;
   widget_move(&FL->w_typelab->widget, W->x+left, W->y+top, LABW, BUTH);
   left+= LABW+PAD;
   widget_move(&FL->w_type->widget, W->x+left, W->y+top, w-left-PAD, BUTH);
   bot= top-PAD;
  }

  top= PAD;
  widget_move(&FL->w_list->widget,W->x+XPAD, W->y+top, w-2*XPAD, bot-top);
  
  if (W->shown)
    gfx_update(x,y,w,h);
}

static int gui_filelist_ipret_click(gui_filelist_t *FL, gfx_event_t *event)
{
  int sel;
  char *f;
  char *d;

  sel= event->key;
  if (!FL->dir || sel>FL->nfiles || sel<0) return 1;

  if (strcmp(FL->dir,"/"))
  {
    if (sel==0) // go up
    {
      gui_filelist_set_directory(FL,path_child(FL->dir, ".."));
      return 1;
    }
    sel--;
  }
  f= FL->files[sel];
  if (f[0]=='D')
  {
    gui_filelist_set_directory(FL,d=path_child(FL->dir, f+1));
    free(d);
    return 1;
  }
  event->string= path_absolute(d= path_child(FL->dir,f+1));
  free(d);
  return 0;
}

static int gui_filelist_listclick(void *arg, gfx_event_t *event)
{
  gui_filelist_t *FL;
  FL= arg;
  if (gui_filelist_ipret_click(FL,event)) return 1;
  event->type= FL->widget.number;
  if (FL->clickf && (*FL->clickf)(FL->evarg,event)) return 1;
  return 0;
} 

static int gui_filesel_listdlick(void *arg, gfx_event_t *event)
{
  gui_filelist_t *FL;
  FL= arg;
  if (gui_filelist_ipret_click(FL,event)) return 1;
  event->type= GFXEV(FL->widget.number, DOUBLE_CLICK);
  if (FL->dlickf && (*FL->dlickf)(FL->evarg,event)) return 1;
  return 0;
} 

static int gui_filelist_typeselect(void *arg, gfx_event_t *event)
{
  gui_filelist_t *FL;
  FL= arg;
  FL->extensions= event->usr;
  gui_filelist_set_directory(FL, FL->dir);
  return 1;
} 
static int gui_filelist_hiddenselect(void *arg, gfx_event_t *event)
{
  gui_filelist_t *FL;
  FL= arg;
  FL->show_hidden= event->key;
  gui_filelist_set_directory(FL, FL->dir);
  return 1;
} 

static void free_extension_list(char **E)
{
  int i;
  for(i=0;E[i];i++) free(E[i]);
  free(E);
}
static void gui_filelist_set_types(gui_filelist_t *FL, char **types)
{
  int i;
  if (!FL->w_type)
  {
    if (FL->extensions) free_extension_list(FL->extensions);
    FL->extensions= make_file_extension_list(types);
    return ;
  }
  while(1)
  {
    char **K;
    K= NULL;
    dropdown_remove(FL->w_type,-1,(void**) &K);
    if (!K) break;
    free_extension_list(K);
  }
  gui_dropdown_clear(FL->w_type);

  for(i=0;types[i];i++)
  {
    char **el;
    dropdown_insert(FL->w_type, -1, types[i], 
              el= make_file_extension_list(types+i));
    if (i==0) FL->extensions= el;
  }
  gfx_dropdown_select(FL->w_type, 0);
  if (FL->dir)
    gui_filelist_set_directory(FL,FL->dir);
}





static void gui_filelist_destroy(widget_t *W)
{
  gui_filelist_t *FL;
  FL= (gui_filelist_t*) W;
  // gui_list_destroy(&FL->w_list->widget); FIXME: implement this
  if (FL->w_hidden) gui_checkbox_destroy(&FL->w_hidden->widget);
  if (FL->w_typelab) gui_label_destroy(&FL->w_typelab->widget);
  if (FL->w_type) gui_dropdown_destroy(&FL->w_type->widget);
  freeNnull(FL->w_list); 
  freeNnull(FL->w_hidden); 
  freeNnull(FL->w_typelab); 
  freeNnull(FL->w_type); 

  freeNnull(FL->dir);
  freeNnull(FL->files);
}

static widget_t *gui_filelist_new
  (int s_ext, int s_hidbox,
        void *evarg,event_function_t clickf, event_function_t dlickf)
{
  gui_filelist_t *FL;
  FL= calloc(1,sizeof(*FL));

  // s_ext= 1; s_hidbox= 1;

  FL->widget.klass= GFXCLASS_FILELIST;
  FL->widget.show= gui_filelist_show;
  FL->widget.hide= gui_filelist_hide;
  FL->widget.destroy= gui_filelist_destroy;
  FL->widget.position= gui_filelist_position;

  FL->evarg= evarg;
  FL->clickf= clickf;
  FL->dlickf= dlickf;

  FL->w_list= (gui_list_t*) gui_list_new(0,0, FL, 
                               gui_filelist_listclick, 
                               gui_filesel_listdlick);

  if (s_ext)
  {
    FL->w_typelab= (gui_label_t*) gfx_label(1, 0, 0xededed, 'L', "File type");
    FL->w_type= (dropdown_t*) dropdown_new(FL, gui_filelist_typeselect);
  }
  if (s_hidbox)
  {
    FL->w_hidden= (checkbox_t*) gui_checkbox_new("Show hidden files", FL, 
                                    gui_filelist_hiddenselect);
  }

  return &FL->widget;
}

void gfx_filelist_set_directory(int wno,char *dir)
{
  gui_filelist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_FILELIST);
  if (!FL) return ;
  gui_filelist_set_directory(FL, dir);
}

void gfx_filelist_set_types(int wno, char **types)
{
  gui_filelist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_FILELIST);
  if (!FL) return ;
  gui_filelist_set_types(FL, types);
}

char** gfx_filelist_get_files(int wno)
{
  char **R,*s;
  gui_filelist_t *FL;
  int ssize,psize,L,i;
  FL= get_widgetk(wno, GFXCLASS_FILELIST);
  if (!FL || !FL->nfiles) 
  {
    R= malloc(sizeof(void*)); 
    R[0]= NULL;
    return R;
  }
  ssize= 0;
  for(i=0;i<FL->nfiles;i++)
    ssize+= strlen(FL->files[i])+1;
  psize= (FL->nfiles+1)*sizeof(char*);
  s= malloc(psize+ssize);
  R= (void*) s;
  s+= psize;
  
  for(i=0;i<FL->nfiles;i++)
  {
    R[i]= s;
    L= strlen(FL->files[i]);
    memcpy(s, FL->files[i], L+1);
    s+= L+1;
  }
  R[i]= NULL;
  return R;
}


char *gfx_filelist_get_directory(int wno)
{
  gui_filelist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_FILELIST);
  if (!FL || !FL->dir)
    return strdup(".");
  return strdup(FL->dir);
}

void gfx_filelist_refresh(int wno)
{
  gui_filelist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_FILELIST);
  if (!FL || !FL->dir) return ;
  gui_filelist_set_directory(FL, FL->dir);
}

char *gfx_filelist_get_selected(int wno)
{
  gui_filelist_t *FL;
  int n;
  FL= get_widgetk(wno, GFXCLASS_FILELIST);
  if (!FL || !FL->dir) return NULL;
  n= gfx_list_selected(FL->w_list);
  if (n<1 || n>FL->nfiles) return NULL;
  return path_child(FL->dir, FL->files[n-1]+1);
}


/***********************************************************************
 *                                                                     *
 *                         DIRECTORY LIST                              *
 * #W.DIRLIST                                                          *
 ***********************************************************************
  This is a simplified version of the file list. It can be used to 
  select directories. I didn't include any widgets other than a list,
  but maybe I could put some buttons and what not so that it becomes
  a builtin dialog.
 ***********************************************************************/

// #W.DIRLIST
typedef struct
{
  widget_t widget;

  char *dir;
  int nfiles;
  char **files;

  gui_list_t *w_list;

  void *evarg;
  event_function_t clickf,dlickf;
} gui_dirlist_t;



static void gui_dirlist_set_directory(gui_dirlist_t *FL,char *dir)
{
  int i;
  static const unsigned int *icons[3];
  static char fuu[]= "..";

  file_list_icons(icons);

  dir= path_absolute(dir);
  if (FL->dir) free(FL->dir);
  if (FL->files) free(FL->files);
  FL->nfiles= 0; FL->files= NULL; FL->dir= NULL;

  FL->dir= dir;
  FL->files= list_files(FL->dir);
  file_types(FL->dir, FL->files);

  for(i=0;FL->files[i];i++) ;
  FL->nfiles= i;

  FL->nfiles= filter_files_byspecial(FL->nfiles,FL->files);
  sort_files(FL->nfiles,FL->files);

  gfx_list_show(FL->w_list, 0);
  gfx_list_clear(FL->w_list);

  if (strcmp(FL->dir,"/"))
    gfx_list_insert(FL->w_list, -1, "<PARENT DIRECTORY>", fuu, icons[1]);
  for(i=0;i<FL->nfiles;i++)
  {
    gfx_list_insert( FL->w_list, -1, FL->files[i]+1, FL->files[i], 
           icons[ FL->files[i][0]=='F' ? 0: (FL->files[i][0]=='D'?1:2) ] );
  }
  gfx_list_show(FL->w_list,1);
}

static void gui_dirlist_show_all(gui_dirlist_t *FL,int show)
{
  widget_show(&FL->w_list->widget,show);
}

static void gui_dirlist_show(widget_t *W)
{
  gui_dirlist_t *FL;
  FL= (gui_dirlist_t*) W;
  if (W->shown)
    gfx_fill_rect(0, W->x, W->y, W->w, W->h, 0xededed);
  gui_dirlist_show_all(FL, 1);
  if (W->shown)
    gfx_update(W->x, W->y, W->w, W->h);
}
static void gui_dirlist_hide(widget_t *W)
{
  gui_dirlist_t *FL;
  FL= (gui_dirlist_t*) W;
  gui_dirlist_show_all(FL, 0);
}


static void gui_dirlist_position(widget_t *W, int x,int y,int w,int h)
{
  gui_dirlist_t *FL;
  int PAD;

// FIXME: convert these to resources
  PAD= 0;
  FL= (gui_dirlist_t*) W;

  W->x= x; W->y= y; W->w= w; W->h= h;
  if (W->shown) gfx_fill_rect(0, x,y,w,h, 0xededed);
  widget_move(&FL->w_list->widget,W->x+PAD, W->y+PAD, w-2*PAD, h-2*PAD);
  if (W->shown) gfx_update(x,y,w,h);
}

static int gui_dirlist_ipret_click(gui_dirlist_t *FL, gfx_event_t *event)
{
  int sel;
  char *f;
  char *d;

  sel= event->key;
  if (!FL->dir || sel>FL->nfiles || sel<0) return 1;

  if (strcmp(FL->dir,"/"))
  {
    if (sel==0) // go up
    {
      gui_dirlist_set_directory(FL,path_child(FL->dir, ".."));
      return 0;
    }
    sel--;
  }
  f= FL->files[sel];
  if (f[0]!='D') return 1; // not a directory, don't generate an event.
  
  gui_dirlist_set_directory(FL,d=path_child(FL->dir, f+1));
  free(d);
  return 0;
}

static int gui_dirlist_listclick(void *arg, gfx_event_t *event)
{
  gui_dirlist_t *FL;
  FL= arg;
  if (gui_dirlist_ipret_click(FL,event)) return 1;
  event->type= FL->widget.number;
  event->string= FL->dir;
  if (FL->clickf && (*FL->clickf)(FL->evarg,event)) return 1;
  return 0;
} 

static int gui_dirsel_listdlick(void *arg, gfx_event_t *event)
{
  gui_dirlist_t *FL;
  FL= arg;
  if (gui_dirlist_ipret_click(FL,event)) return 1;
  event->type= GFXEV(FL->widget.number, DOUBLE_CLICK);
  event->string= FL->dir;
  if (FL->dlickf && (*FL->dlickf)(FL->evarg,event)) return 1;
  return 0;
} 

static void gui_dirlist_destroy(widget_t *W)
{
  gui_dirlist_t *FL;
  FL= (gui_dirlist_t*) W;
  // gui_list_destroy(&FL->w_list->widget); FIXME: implement this
  freeNnull(FL->w_list); 
  freeNnull(FL->dir);
  freeNnull(FL->files);
}

static widget_t *gui_dirlist_new
  (void *evarg,event_function_t clickf, event_function_t dlickf)
{
  gui_dirlist_t *FL;
  FL= calloc(1,sizeof(*FL));

  FL->widget.klass= GFXCLASS_DIRLIST;
  FL->widget.show= gui_dirlist_show;
  FL->widget.hide= gui_dirlist_hide;
  FL->widget.destroy= gui_dirlist_destroy;
  FL->widget.position= gui_dirlist_position;

  FL->evarg= evarg;
  FL->clickf= clickf;
  FL->dlickf= dlickf;

  FL->w_list= (gui_list_t*) gui_list_new(0,0, FL, 
                               gui_dirlist_listclick, 
                               gui_dirsel_listdlick);
  return &FL->widget;
}

void gfx_dirlist_set_directory(int wno,char *dir)
{
  gui_dirlist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_DIRLIST);
  if (!FL) return ;
  gui_dirlist_set_directory(FL, dir);
}

char *gfx_dirlist_get_directory(int wno)
{
  gui_dirlist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_DIRLIST);
  if (!FL || !FL->dir) return strdup(".");
  return strdup(FL->dir);
}

void gfx_dirlist_refresh(int wno)
{
  gui_dirlist_t *FL;
  FL= get_widgetk(wno, GFXCLASS_DIRLIST);
  if (!FL || !FL->dir) return ;
  gui_dirlist_set_directory(FL, FL->dir);
}

/***********************************************************************
 *                                                                     *
 *                          VIRTUAL KEYBOARD                           *
 * D.VIRTKEYB                                                          *
 ***********************************************************************/


static vkb_t *active_vk;

static void vkey_render(vkb_t *V, int kn, int state,int upd)
{
  graypix_t text;
  colorpix_t button;
  char *s;
  vkey_t *K;
  K= V->keys + kn;

  s= K->str[V->state];
  if (s) text_to_graypix(1, s, &text, NULL);

  button.data= x_image_data[K->y] + K->x;
  button.stride= x_window_width;
  button.w= K->w; button.h= K->h;

  gui_draw_3dbox(&button, state ? &RES(button.depressed)
                                    : &RES(button.normal));
  if (s)
  {
    colorpix_put_graypix_centered(&button, &text, 0x00);
    free(text.data);
  }
  
  if (upd)
    gfx_update(K->x, K->y, K->w, K->h);
}

static void vk_textdim(char* str, int *Rw, int *Rh)
{
  int w,h;
  if (!str) return ;
  get_text_data(1,str, &w, &h, NULL, NULL);
  if (*Rw<w) *Rw= w;
  if (*Rh<h) *Rh= h;
}

static void vkey_getdesc(vkb_t *V, gfx_vkbdef_t *D)
{
  int nk, i, r, c, x;
  char *t;

  nk= 0;
  for(r=0;r<15;r++) 
  {
    for(c=0;c<50;c++)
      if (D->keys[r][c].func) nk++;
                         else break;
    if (!c) break;
  }
  V->nrows= r;
  V->nkeys= nk;
  V->keys= calloc(nk,sizeof(V->keys[0]));
  i= 0;
  for(r=0;r<V->nrows;r++)
  {
    x= 0;
    for(c=0;c<50;c++)
    {
      if (!D->keys[r][c].func) break;
      if (D->keys[r][c].func==GFX_VK_NOP) 
      {
        x+= D->keys[r][c].span;
        continue;
      }
      V->keys[i].x= x;
      V->keys[i].y= r;
      V->keys[i].w= imax2(1,D->keys[r][c].span);
      V->keys[i].h= 1;
      V->keys[i].func= D->keys[r][c].func;
      x+= V->keys[i].w;
      t= NULL;
      switch(V->keys[i].func)
      {
      case GFX_K_PRT:
        V->keys[i].str[0]= nstrdup(D->keys[r][c].str[0]);
        V->keys[i].str[1]= nstrdup(D->keys[r][c].str[1]);
        V->keys[i].str[2]= nstrdup(D->keys[r][c].str[2]);
        break;
      case GFX_K_BACKSPACE: t= "BKSP"; break;
      case GFX_K_RETURN: t= "OK"; break;
      }
      if (t)
      {
        V->keys[i].str[0]= strdup(t);
        V->keys[i].str[1]= strdup(t);
        V->keys[i].str[2]= strdup(t);
      }
      i++;
    }
  }
}

static void vkey_layout(vkb_t *V)
{
  int i, s, tw, th, maxtw, maxth, maxrw, cellw, cellh, j, ox, sx, ex, w;
  int vs, vp;
  maxrw= maxtw= maxth= 0;
  for(i=0;i<V->nkeys;i++)
  {
    tw= th= 0;
    for(s=0;s<3;s++) vk_textdim(V->keys[i].str[s], &tw, &th);
    if (tw%V->keys[i].w) tw= tw/V->keys[i].w+1;
                    else tw= tw/V->keys[i].w;
    maxtw= imax2(tw, maxtw);
    maxth= imax2(th, maxth);
    maxrw= imax2(V->keys[i].x+V->keys[i].w, maxrw);
  }

  vp= RES(virtkb.vpadding);
  vs= RES(virtkb.vspacing);

  V->h= V->nrows*(maxth+vp+vs)+vs;
  while(vs>0 && V->h>x_window_height/2)
  {
    vs--;
    V->h= V->nrows*(maxth+vp+vs)+vs;
  }
  while(vp>0 && V->h>x_window_height/2)
  {
    vp--;
    V->h= V->nrows*(maxth+vp+vs)+vs;
  }
  V->vs= vs;

  cellw= (x_window_width-RES(virtkb.hspacing)) / maxrw - RES(virtkb.hspacing);

  cellw= imin2(maxtw+RES(virtkb.hpadding), cellw);
  cellh= vp+ maxth;
  
  for(i=0;i<V->nkeys;i++) 
  {
    V->keys[i].x = V->keys[i].x*(cellw + RES(virtkb.hspacing)) 
                                         + RES(virtkb.hspacing);
    V->keys[i].y = V->keys[i].y*(cellh + vs) + vs;
    V->keys[i].w = cellw + (cellw+RES(virtkb.hspacing))*(V->keys[i].w-1);
    V->keys[i].h = cellh;
  }

  for(i=0;i<V->nkeys;)
  {
    for(j=i+1;j<V->nkeys;j++)
      if (V->keys[j].x<=V->keys[i].x) break;
    sx= V->keys[i].x;
    ex= V->keys[j-1].x + V->keys[j-1].w;
    w= ex-sx;
    ox= (x_window_width - w) /2;
    for(;i<j;i++) V->keys[i].x += ox; 
  }
  V->y= 0;
  V->x= 0;
  V->w= x_window_width;
}

static void vkey_hide(vkb_t *V)
{
  if (active_vk!=V) return ;
  active_vk= NULL;
  if (!V->backstore) return ;
  put_server_pix(V->backstore, V->x, V->y, V->w, V->h);
  free(V->backstore);
  V->backstore= NULL;
  gfx_update(V->x, V->y, V->w, V->h);
}

static void vkey_show(vkb_t *V)
{
  int i;
  if (active_vk==V) return ;
  if (active_vk) vkey_hide(active_vk);
  V->backstore= get_server_pix(V->x,V->y, V->w, V->h);
  active_vk= V;
  gfx_fill_rect(0, V->x, V->y, V->w, V->h,0xccCCcc);
  V->state= 0;
  for(i=0;i<V->nkeys;i++) vkey_render(V, i, 0, 0);
  gfx_update(V->x, V->y, V->w, V->h);
}

static void vkey_ypos(vkb_t *V, int yp)
{
  int i, D, h;
  if (yp==V->y) return ;
  h= 0;
  if (active_vk==V) { h=1; vkey_hide(V); }
  D= V->keys[0].y;
  for(i=0;i<V->nkeys;i++)
    V->keys[i].y= V->keys[i].y - D + yp + V->vs;
  V->y= yp;
  if (h) vkey_show(V);
}

static int vkey_find_button(vkb_t *V, int x, int y)
{
  int i,j,t;
  i= 0;
  j= V->nkeys-1;
  while(i<=j)
  {
    t= (i+j)/2;
    if ((y>V->keys[t].y+V->keys[t].h)||(x>V->keys[t].x+V->keys[t].w)) i= t+1;
    else if ((y<V->keys[t].y) || (x<V->keys[t].x)) j= t-1;
    else return t;
  }
  return -1;
}

static int vkey_event(gfx_event_t *ev)
{
  vkey_t *K;
  int k;
  char *s;
  if (!active_vk) return 0;
  if (ev->type==GFXEV_BUTTON_PRESS) goto button_press;
  if (ev->type==GFXEV_BUTTON_RELEASE) goto button_release;
  return 0;
  
button_press:
  if (active_vk->pressed!=-1) return 0;
  k= vkey_find_button(active_vk,ev->x,ev->y);
  if (k==-1) return 0;
  active_vk->pressed= k;
  vkey_render(active_vk, k, 1, 1);
  return 1;

button_release:
  if (active_vk->pressed==-1) return 0;
  k= active_vk->pressed;
  active_vk->pressed= -1;
  vkey_render(active_vk, k, 0, 1);
  K= active_vk->keys+k;
  ev->string= NULL;
  s= K->str[ active_vk->state ];
  switch(K->func)
  {
  case GFX_K_PRT:
    if (!s) return 0;
    ev->string= s;
    ev->wide= utf8_decode_single_char(s);
    break;
  case GFX_K_RETURN:
  case GFX_K_BACKSPACE:
    ev->string= NULL;
    break;
  default:
    return 0;
  }
  ev->key= K->func;
  ev->type= GFXEV_KEY_PRESS;
  return 0;
}

void gfx_virtual_keyboard(int num, gfx_vkbdef_t *D)
{
  vkb_t *V;
  V= calloc(1,sizeof(*V));
  vkey_getdesc(V, D);
  vkey_layout(V);
  V= obj_put(OBJ_VKB, num, V);
  // FIXME: free V
}

void gfx_edit_field(int widget, int kbnum)
{
  vkb_t *V;
  text_entry_t *W;
  int vtop;
  int off;
  V= obj_get(OBJ_VKB, kbnum);
  if (!V) return ;
  W= get_widgetk(widget, GFXCLASS_TEXTENTRY);
  if (!W) return ;
  vtop= x_window_height - V->h;
  if (W->y+W->h<=vtop) off= vtop;
                  else off= 0;
  vkey_ypos(V, off);
  vkey_show(V);
}

void gfx_hide_virtual_keyboard()
{
  if (!active_vk) return ;
  vkey_hide(active_vk);
}


/***********************************************************************
 *                                                                     *
 *                               CANVAS                                *
 * #W.CANVAS                                                           *
 ***********************************************************************
The canvas is a scrollable drawing area with its own buffer. You just need
to draw into it once, it will update the display from its buffer when it's
later hidden and shown. This buffer can be dynamically resized at run time. 
 ***********************************************************************/

typedef struct
{
  int w, h;
  int x,y; // these are relative to the canvas buffer origin
  int func; 
  int shown;
#define SPRITE_FUNC_SET  0
#define SPRITE_FUNC_XOR  1
#define SPRITE_FUNC_AND  2
#define SPRITE_FUNC_NEG  3
  uint32_t *data;
} sprite_t;

typedef struct
{
  widget_t widget;
  int buffer_w, buffer_h;
  uint32_t *buffer;
  int enabled;

  int pos_x, pos_y,port_w,port_h;
  int have_scrollh, have_scrollv;
  int can_scrollh, can_scrollv;
  scrollbar_t *scrollh, *scrollv;
  int too_small; // the display is too small for the
                 // scrollbars to fit.

  int click1, click3;

  int recsel;
  int rs_x0, rs_y0, rs_x1, rs_y1;

  int have_zoom, z_factor;
     // zoom window coordinates
  int z_wx, z_wy, z_ww, z_wh;
  unsigned int *z_buffer;
  int z_grid, z_gridcolor; 
     // the point to be zoomed around
  int z_cx, z_cy;

  int nsprites, ssprites;
  sprite_t **sprites;
} gui_canvas_t;


static void gui_sprite_free(sprite_t *S)
{
  free(S->data);
  free(S);
}

static gui_canvas_t *gui_canvas_get(int wno)
{
  widget_t *W;
  W= get_widget(wno);
  if (W && W->klass==GFXCLASS_CANVAS) return (gui_canvas_t*) W;
  return NULL;
}

static void gui_canvas_paint_scrollbars(gui_canvas_t *CC)
{
  if (!CC->enabled) return ;
  if (CC->have_scrollv) scrollbar_show(CC->scrollv,1);
  if (CC->have_scrollh) scrollbar_show(CC->scrollh,1);
}

static void gui_canvas_paint_buffer(gui_canvas_t *CC)
{
  if (!CC->enabled) return ;
  if (!CC->buffer)
  {
//
 //   gfx_fill_rect(GFX_UPDATE,CC->widget.x,CC->widget.y,
  //                           CC->port_w, CC->port_h, 0xfF00ff);
    return ;
  }
  gfx_put(GFX_UPDATE, CC->buffer, CC->buffer_w, CC->buffer_h,
            CC->pos_x, CC->pos_y,
           CC->widget.x, CC->widget.y, CC->port_w, CC->port_h);
}

static void gui_canvas_draw_all(gui_canvas_t *CC)
{
  gui_canvas_paint_scrollbars(CC);
  gui_canvas_paint_buffer(CC);
  if (!CC->enabled) return ;
  if (CC->port_w>CC->buffer_w)
    gfx_fill_rect(GFX_UPDATE, 
        CC->widget.x+CC->buffer_w, CC->widget.y, 
                    CC->port_w-CC->buffer_w, CC->port_h,0xFFffff);
  if (CC->port_h>CC->buffer_h)
    gfx_fill_rect(GFX_UPDATE, 
        CC->widget.x, CC->widget.y+CC->buffer_h, 
                    CC->port_w, CC->port_h-CC->buffer_h,0xffFFff);
}

static void gui_canvas_scrollv(gui_canvas_t *CC,double start,double end)
{
  CC->pos_y= start*CC->buffer_h;
  if (CC->pos_y+CC->port_h>CC->buffer_h)
   CC->pos_y=imax2(0,CC->buffer_h-CC->port_h);
  gui_canvas_paint_buffer(CC);
}

static void gui_canvas_scrollh(gui_canvas_t *CC,double start,double end)
{
  CC->pos_x= start*CC->buffer_w;
  if (CC->pos_x+CC->port_w>CC->buffer_w)
   CC->pos_x=imax2(0,CC->buffer_w-CC->port_w);
  gui_canvas_paint_buffer(CC);
}

static void gui_canvas_invert_hline(gui_canvas_t *CC, int x1, int x2, int y)
{
  int x;
  x1= imax2(0,x1);
  x2= imin2(CC->buffer_w-1, x2);
  if (range_test(y, 0, CC->buffer_h))
    for(x=x1;x<=x2;x++)
      CC->buffer[CC->buffer_w*y+x] ^= 0xFFffFF;
}

static void gui_canvas_invert_vline(gui_canvas_t *CC,int x,int y1, int y2)
{
  int y;
  y1= imax2(0,y1);
  y2= imin2(CC->buffer_h-1,y2);
  if (range_test(x, 0, CC->buffer_w))
    for(y=y1;y<=y2;y++)
      CC->buffer[CC->buffer_w*y+x] ^= 0xFFffFF;
}

static void gui_canvas_clipped_hline
   (gui_canvas_t *CC,int color,int x1,int x2,int y)
{
  int x, bw;
  if (y<0 || y>=CC->buffer_h) return ;
  bw= CC->buffer_w;
  x1= imax2(0,x1);
  x2= imin2(x2,bw-1);
  for(x=x1;x<=x2;x++) CC->buffer[y*bw+x]= color;
}


static void gui_canvas_clipped_vline
   (gui_canvas_t *CC,int color,int x,int y1,int y2)
{
  int y, bw;
  if (x<0 || x>=CC->buffer_w) return ;
  bw= CC->buffer_w;
  y1= imax2(0,y1);
  y2= imin2(y2,CC->buffer_h-1);
  for(y=y1;y<=y2;y++) CC->buffer[y*bw+x]= color;
}

static int gui_canvas_release(void *arg,gfx_event_t *ev)
{
  gui_canvas_t *CC;
  CC= arg;
  ev->type= GFXEV(CC->widget.number,BUTTON_RELEASE);
  ev->x= ev->x - CC->widget.x + CC->pos_x;
  ev->y= ev->y - CC->widget.y + CC->pos_y;
  return 0;
}


static int gui_canvas_click(void *arg,gfx_event_t *ev)
{
  gui_canvas_t *CC;
  CC= arg;
  ev->type= GFXEV(CC->widget.number,BUTTON_PRESS);
  ev->x= ev->x - CC->widget.x + CC->pos_x;
  ev->y= ev->y - CC->widget.y + CC->pos_y;
  install_release_handler(CC, gui_canvas_release);
  return 0;
}


static void gui_canvas_update_click_handlers(gui_canvas_t *CC)
{
  CC->click1= install_or_modify_click
       (CC->click1, &CC->enabled, gui_canvas_click, CC,
        CC->widget.x, CC->widget.y, CC->port_w, CC->port_h, 1);
  CC->click3= install_or_modify_click
       (CC->click3, &CC->enabled, gui_canvas_click, CC,
        CC->widget.x, CC->widget.y, CC->port_w, CC->port_h, 3);
}

static void gui_canvas_arrange_scrollbars
   (gui_canvas_t *CC)
{
  int dw,dh,bw,bh;
  int hor,ver;
  widget_t *W;

  dw= CC->widget.w;
  dh= CC->widget.h;

  bw= CC->buffer_w;
  bh= CC->buffer_h;

  hor= 0;
  ver= 0;
  if (bw>dw && CC->can_scrollh) { hor= 1; dh-= gui_scrollbar_height; }
  if (bh>dh && CC->can_scrollv) { ver= 1; dw-= gui_scrollbar_width; }
  if (!hor && bw>dw && CC->can_scrollh) { hor= 1; dh-= gui_scrollbar_height; }

  if (dw<=0 || dh<=0) { CC->too_small=1; return ; }

  CC->port_w= dw;
  CC->port_h= dh;

  if (hor)
  {
    if (CC->pos_x+CC->port_w>bw)
      CC->pos_x= imax2(0,bw-CC->port_w);

    W= &CC->scrollh->widget;
    scrollbar_relayout(CC->scrollh, W->x, W->y, CC->port_w, W->h,
          (double)CC->pos_x/CC->buffer_w,
          (double)imin2(CC->pos_x+CC->port_w,CC->buffer_w)/CC->buffer_w, 0.1);
    scrollbar_show(CC->scrollh, 1);
  }
  else
  {
    if (CC->have_scrollh)
      scrollbar_show(CC->scrollh, 0);
  }
  CC->have_scrollh= hor;

  if (ver)
  {
    if (CC->pos_y+CC->port_h>bh)
       CC->pos_y= imax2(0,bh-CC->port_h);
    W= &CC->scrollv->widget;
    scrollbar_relayout(CC->scrollv, W->x, W->y, W->w, CC->port_h,
          (double)CC->pos_y/CC->buffer_h,
          (double)imin2(CC->pos_y+CC->port_h,CC->buffer_h)/CC->buffer_h, 0.1);
    scrollbar_show(CC->scrollv, 1);
  }
  else
  {
    if (CC->have_scrollv)
       scrollbar_show(CC->scrollv,0);
  }
  CC->have_scrollv= ver;

  gui_canvas_update_click_handlers(CC);
}

static void gui_canvas_position(gui_canvas_t *CC,int x,int y,int w,int h)
{
  widget_t *W;
  W= &CC->widget;
  W->x= x;
  W->y= y;
  W->w= w;
  W->h= h;

  if (CC->can_scrollh)
    scrollbar_position(CC->scrollh, x, y+h-gui_scrollbar_height,w, 
                     gui_scrollbar_height);
  if (CC->can_scrollv)
    scrollbar_position(CC->scrollv, x+w-gui_scrollbar_width, y, 
                     gui_scrollbar_width, h);
  gui_canvas_arrange_scrollbars(CC);
}




/*****************************************************************
Now we have the following rectangles:

  update rectangle: (ux,uy,uw,uh)
  buffer rectangle: the whole canvas buffer  (0,0,buffer_w,buffer_h)
  port   rectangle: visible part (pos_x,pos_y,port_w,port_h)
  sprite rectangle: (S->x,S->y,S->w,S->h)

If I were to draw all of the sprite, I'd clip the sprite rectangle
using the buffer rectangle and then paint away.

But now, I need to clip using the update rectangle as well.
So, I will clip the update rectangle by the buffer rectangle
and then clip the sprite using the result.       
******************************************************************/
static void gui_canvas_paint_sprite
   (gui_canvas_t *C,sprite_t *S,int ux,int uy,int uw,int uh)
{
  int x,y,bx,by,sx,sy;

  if (!S || !S->shown) return ;

  if (rect_intersect(&ux,&uy, &uw, &uh, 
              0,0, C->buffer_w, C->buffer_h)) return ;
  if (rect_intersect(&ux,&uy, &uw, &uh, 
              C->pos_x, C->pos_y, C->port_w, C->port_h)) return ;
  if (rect_intersect(&ux,&uy, &uw, &uh,
              S->x, S->y, S->w, S->h)) return ;

  // now, we have the update coordinates relative to the buffer origin.
  // in order to convert it to screen coordinates, we need to subtract
  // pos_x and pos_y, then add widget.x and widget.y

  bx=  ux - C->pos_x + C->widget.x;
  by=  uy - C->pos_y + C->widget.y;

  // the coordinates within the sprite data are: update-sprite, that's it

  sx= ux - S->x;
  sy= uy - S->y;

#define ALPHABYTE  (0xff<<24)
  for(y=0;y<uh;y++)
    for(x=0;x<uw;x++)
    {
      uint32_t D;
      uint32_t B;
      D= S->data[(x+sx)+(y+sy)*S->w];
      if ( (D&ALPHABYTE)==0 ) continue;  
      D&= ~ALPHABYTE;
      B= x_image_data[y+by][x+bx];
      switch(S->func)
      {
      case SPRITE_FUNC_XOR:  B ^= D; break;
      case SPRITE_FUNC_AND:  B &= D; break;
      case SPRITE_FUNC_NEG:  B = ~B; break;
      case SPRITE_FUNC_SET:  
                   default:   B = D; break;
      }
      x_image_data[y+by][x+bx]= B;
    }
#undef ALPHABYTE
}

/***********************************************************************
 *                             ZOOM WINDOW                             *
 ***********************************************************************
When the user starts zooming, a small virtual window is created at the
bottom right corner of the widget's display area which excludes the 
scrollbars.  The user makes a zooming request at a specific point and the 
area surrounding this point is shown zoomed in the window. When the 
specified point falls within the zoom window, the window is put somewhere
else.

The window has its own buffer. This protects the main buffer 
from contamination. The constructor of the zoom window should
look like this:

   zoom(int width,int height,int zoom_factor,int grid,int gridlinecolor)

grid and gridlinecolor should be taken into account if zoom_factor is
more than 5, maybe.

Zoom window contents are displayed in the canvas_update() function.
The window has two lines of border, one black and one white.

First of all, I should render the thing. 

When rendering the buffer, I will loop on the zoom buffer instead
of the main buffer. Doing the inverse requires a lot of calculations
and sub-loops, which can cause a lot of subtle bugs.

So, when a pixel of the zoom buffer is to be updated, I should calculate
at which main buffer pixel it corresponds to.

Let be:
  z - the zoom factor
  w - width of the zoom window
  a - length to the left of the center pixel
  b - length to the right of the center pixel
  c - x(main) of the center pixel 
  s - x(main) of the leftmost pixel displayed 
  e - x(main) of the rightmost pixel displayed

Obviously,  z+a+b = w, if we set a=b, then we get

   a= (w-z)/2  and b = w-a-z,

the second one is due to the fact that w-z might be odd.
Then we get:

  s= c - ceil(a/z) and e= c + ceil(b/z)

The leftmost and rightmost pixels may not be displayed 
completely. Let be:

  m - number of pixels displayed for the leftmost pixel
  n - number of pixels displayed for the rightmost pixel

 m =  a - ( ceil(a/z) - 1)*z
 n =  b - ( ceil(b/z) - 1)*z

So, when given a range of pixels to be painted in the
main window, we can calculate which parts of the zoom
window should be updated. I will use the above formulas
to determine the boundaries of the zoom window with
respect to the main buffer and then intersect that rectangle
with the update range. The intersection will give me the
range of pixels to be updated within the zoom window. For
a given pixel p, the start and end coordinates are:

  x0=  {         0,   p=s
         m+(p-s)*z,   p> s }

  x1=  x0+z-1
 ***********************************************************************/

static void gui_canvas_zupdate
   (gui_canvas_t *CC,int ux0,int uy0,int uw,int uh)
{
  int ux1,uy1,ux,uy;
  int Xa,Xb,Xc,Xm,Xs,Xe; //int Xn,Yn;
  int Ya,Yb,Yc,Ym,Ys,Ye;
  int z, K, P;
  int ww,hh;
  int zx0,zx1,zy0,zy1,zx,zy,ex0,ey0,ex,ey;
  int bw, sk;

  if (!CC->have_zoom) return ;

  ux1= ux0 + uw-1;
  uy1= uy0 + uh-1;
  bw= CC->z_grid;

  if (!CC->z_buffer)
  {
    CC->z_buffer= calloc(CC->z_ww*CC->z_wh,sizeof(CC->z_buffer[0]));
    //int i;
    //for(i=0;i<CC->z_ww*CC->z_wh;i++) CC->z_buffer[i]= 0xff0000;
  }

  z= CC->z_factor;
  ww= CC->z_ww;             hh= CC->z_wh;
  Xc= CC->z_cx;             Yc= CC->z_cy;

  Xa= (ww-z)/2;            Xb= ww-z-Xa;
  Xs= Xc-GFXceil(Xa,z);       Xe= Xc+GFXceil(Xb,z);
  Xm= Xa-(GFXceil(Xa,z)-1)*z; //Xn= Xb-(GFXceil(Xb,z)-1)*z;

  Ya= (hh-z)/2;            Yb= hh-z-Ya;
  Ys= Yc-GFXceil(Ya,z);       Ye= Yc+GFXceil(Yb,z);
  Ym= Ya-(GFXceil(Ya,z)-1)*z; //Yn= Yb-(GFXceil(Yb,z)-1)*z;

  ux0= imax2(ux0, Xs);     ux1= imin2(ux1, Xe);
  uy0= imax2(uy0, Ys);     uy1= imin2(uy1, Ye);

  if (ux0>Xe || ux1<Xs || uy0>Ye || uy1<Ys) return ;

  zx0= (ux0==Xs) ? 0 : (Xm+(ux0-Xs-1)*z); 
  zx1= imin2(ww-1, z-1+((ux1==Xs) ? 0 : (Xm+(ux1-Xs-1)*z)));
  ex0= (ux0==Xs) ? (z-Xm) : 0;

  zy0= (uy0==Ys) ? 0 : (Ym+(uy0-Ys-1)*z);
  zy1= imin2(hh-1, z-1+((uy1==Ys) ? 0 : (Ym+(uy1-Ys-1)*z)));
  ey0= (uy0==Ys) ? (z-Ym) : 0;
  
  zx= zx0; ex= ex0; ux= ux0; 
  zy= zy0; ey= ey0; uy= uy0;

loop_y:
  if (zy<2 || zy>=hh-2)  goto border_horz; 
  if (ey<bw) goto pixborder_horz;
loop_x:
  P= CC->buffer[ux+uy*CC->buffer_w];
  if (zx<2 || zx>=ww-2) { K= (zx==0 || zx==ww-1) ? 0xffffff : 0; }
  else if (ex<bw) { K= CC->z_gridcolor; }
  else K= P;
  
  CC->z_buffer[zx+zy*ww]= K;
  zx++;
  ex= (ex+1)%z;
  if (ex==0) ux++; 
  if (zx<=zx1) goto loop_x;

loop_x_end:
  zx= 0;
  ux= ux0;
  ex= ex0;
  zy++;
  ey= (ey+1)%z;
  if (ey==0) uy++;
  if (zy<=zy1) goto loop_y;

  gfx_put(GFX_UPDATE,CC->z_buffer,CC->z_ww,CC->z_wh,zx0,zy0,
                     CC->widget.x+CC->z_wx+zx0,
                     CC->widget.y+CC->z_wy+zy0, zx1-zx0+1, zy1-zy0+1);

  return ;
border_horz:
  if (zy==0 || zy==hh-1) { K= 0xffffff; sk= 0; } else { K= 0; sk= 1; }
  for(zx=sk;zx<ww-sk;zx++) 
    CC->z_buffer[zx+zy*ww]= K;
  if (sk) { CC->z_buffer[0+zy*ww]= CC->z_buffer[ww-1+zy*ww]= 0xffFFff; }
  goto loop_x_end;
pixborder_horz:
  K= CC->z_gridcolor;
  CC->z_buffer[0+zy*ww]= CC->z_buffer[ww-1+zy*ww]=  0xffFFff;
  CC->z_buffer[1+zy*ww]= CC->z_buffer[ww-2+zy*ww]=  0x0;
  for(zx=zx0+2;zx<=zx1-2;zx++) 
    CC->z_buffer[zx+zy*ww]= K;
  goto loop_x_end;
}

static void gui_canvas_update(gui_canvas_t *CC,int x,int y,int w,int h)
{
  int sx,sy,ex,ey;
  rect_t sub[4], ev, block;
  int nsu,i;

  if (w<=0 || h<=0 || !CC->enabled) return ;
  sx= imax2(imax2(x, 0), CC->pos_x);
  sy= imax2(imax2(y, 0), CC->pos_y);

  ex= imin2(imin2(x+w,CC->buffer_w), CC->pos_x+CC->port_w);
  ey= imin2(imin2(y+h,CC->buffer_h), CC->pos_y+CC->port_h);

  if (ex<=sx || ey<=sy) return ;
  if (CC->have_zoom)
  {
    block.left= CC->z_wx; block.right= CC->z_wx+CC->z_ww-1;
    block.top= CC->z_wy;  block.bottom= CC->z_wy+CC->z_wh-1;

    ev.left= sx-CC->pos_x; ev.right= ex-CC->pos_x;
    ev.top= sy-CC->pos_y;  ev.bottom= ey-CC->pos_y;

    nsu= subtract_rectangle(&ev,&block,sub);
  }
  else
  {
    sub[0].left= sx-CC->pos_x; sub[0].right= ex-CC->pos_x;
    sub[0].top= sy-CC->pos_y;  sub[0].bottom= ey-CC->pos_y;
    nsu= 1;
  }

  for(i=0;i<nsu;i++)
  {
    int j;
    gfx_put(0, CC->buffer, CC->buffer_w, CC->buffer_h,
                sub[i].left+CC->pos_x, sub[i].top+CC->pos_y,
                sub[i].left+CC->widget.x, sub[i].top+CC->widget.y,
                sub[i].right-sub[i].left+1, sub[i].bottom-sub[i].top+1);

    for(j=0;j<CC->nsprites;j++)
      gui_canvas_paint_sprite(CC, CC->sprites[j],
            sub[i].left+CC->pos_x, sub[i].top+CC->pos_y,
            sub[i].right-sub[i].left+1, sub[i].bottom-sub[i].top+1);

    gfx_update(sub[i].left+CC->widget.x, sub[i].top+CC->widget.y,
               sub[i].right-sub[i].left+1, sub[i].bottom-sub[i].top+1);
  }
  if (CC->have_zoom)
    gui_canvas_zupdate(CC,x,y,w,h);
}




/***********************************************************************
 *                       Rectangular Selections                        *
 ***********************************************************************
  There are some things to do:
    - only update the borders of the rectangle, not the whole thing.
 ***********************************************************************/


static void gui_canvas_draw_recsel(gui_canvas_t *CC)
{
  int x,y,w,h,C;

  x= imin2(CC->rs_x0, CC->rs_x1);
  y= imin2(CC->rs_y0, CC->rs_y1);

  w= imax2(CC->rs_x0, CC->rs_x1) - x+1;
  h= imax2(CC->rs_y0, CC->rs_y1) - y+1;

  C=  (w<2) ? 0 : 1;
  C|= (h<2) ? 0 : 2;
  switch(C)
  {
  case 0: break ;
  case 1: gui_canvas_invert_hline(CC,x, x+w-1, y); break ;
  case 2: gui_canvas_invert_vline(CC,x, y, y+h-1); break ;
  case 3: gui_canvas_invert_hline(CC,x, x+w-1, y);
          gui_canvas_invert_hline(CC,x, x+w-1, y+h-1);
          gui_canvas_invert_vline(CC,x, y+1, y+h-2);
          gui_canvas_invert_vline(CC,x+w-1, y+1, y+h-2);
          break;
  }
}

static int gui_canvas_motion(void *arg, gfx_event_t *ev)
{
  gui_canvas_t *CC;
  int x,y, minx, maxx, miny, maxy;

  CC= arg;
  x= ev->x;
  y= ev->y;
  if (!range_test(x,CC->widget.x, CC->port_w) ||
      !range_test(y,CC->widget.y, CC->port_h))
    return 1;
  x= x - CC->widget.x + CC->pos_x;
  y= y - CC->widget.y + CC->pos_y;

  if (CC->rs_x0==x && CC->rs_y0==y) return 1;

  minx= imin2(imin2( CC->rs_x0, CC->rs_x1 ), x);
  maxx= imax2(imax2( CC->rs_x0, CC->rs_x1 ), x);
  miny= imin2(imin2( CC->rs_y0, CC->rs_y1 ), y);
  maxy= imax2(imax2( CC->rs_y0, CC->rs_y1 ), y);

  gui_canvas_draw_recsel(CC);
  CC->rs_x1= x;
  CC->rs_y1= y;
  gui_canvas_draw_recsel(CC);
  gui_canvas_update(CC, minx, miny, maxx-minx+1, maxy-miny+1);
  return 1;
}


void gfx_canvas_recsel_begin(int wno, gfx_event_t *ev)
{
  gui_canvas_t *CC;
  if (!(CC= gui_canvas_get(wno))) return ;
  if (!CC->buffer) return ;
  CC->rs_x0= CC->rs_x1= ev->x;
  CC->rs_y0= CC->rs_y1= ev->y;
  CC->recsel= 1;
  install_motion_handler(CC, gui_canvas_motion);
}

void gfx_canvas_recsel_end(int wno,int *Rx1, int *Ry1, int *Rx2, int *Ry2)
{
  int minx,miny,maxx,maxy,Z;
  gui_canvas_t *CC;
  if (!(CC= gui_canvas_get(wno))) return ;

  uninstall_motion_handler();
  gui_canvas_draw_recsel(CC);
  minx= imin2( CC->rs_x0, CC->rs_x1 );
  maxx= imax2( CC->rs_x0, CC->rs_x1 );
  miny= imin2( CC->rs_y0, CC->rs_y1 );
  maxy= imax2( CC->rs_y0, CC->rs_y1 );
  gui_canvas_update(CC, minx, miny, maxx-minx+1, maxy-miny+1);

  Rx1= Rx1 ? Rx1 : &Z;
  Ry1= Ry1 ? Ry1 : &Z;
  Rx2= Rx2 ? Rx2 : &Z;
  Ry2= Ry2 ? Ry2 : &Z;

  *Rx1= CC->rs_x0;
  *Ry1= CC->rs_y0;
  *Rx2= CC->rs_x1;
  *Ry2= CC->rs_y1;
  CC->recsel= 0;
}


/***********************************************************************
 *                       Sprites                                       *
 ***********************************************************************
  A canvas can also display sprites. These are painted over the canvas 
  buffer contents and don't affect the buffer. 
 ***********************************************************************/
int gfx_sprite_new(int wno,int func, int w,int h, uint32_t *data)
{
  gui_canvas_t *CC;
  int i,RV;

  if (w<=0 || h<=0 || !data) return -1;
  if (!(CC=gui_canvas_get(wno))) return -1;
  for(i=0;i<CC->ssprites;i++)
    if (CC->sprites[i]==NULL) break;

  if (i==CC->ssprites)
  {
    sprite_t **ns; int s;
    s= CC->ssprites+CC->ssprites/2+10;
    ns= calloc(s, sizeof(ns[0]));
    if (CC->ssprites)
      memcpy(ns, CC->sprites, CC->ssprites*sizeof(CC->sprites[0]));
    CC->ssprites= s;
    if (CC->sprites) free(CC->sprites);
    CC->sprites= ns;
  }
  RV= i;             
  switch(func)
  {
    case 'A': func= SPRITE_FUNC_AND; break;
    case 'N': func= SPRITE_FUNC_NEG; break;
    case 'X': func= SPRITE_FUNC_XOR; break;
    case 'S': default: func= SPRITE_FUNC_SET; break;
  }

  sprite_t *S;
  S= calloc(1,sizeof(*S));
  S->data= calloc(w*h,sizeof(S->data[0]));
  S->w= w;
  S->h= h;
  S->x= 0;
  S->y= 0;
  S->func= func;
  CC->sprites[RV]= S;

#define ALPHABYTE  (0xff<<24)
  for(i=0;i<w*h;i++)
   if (data[i]&ALPHABYTE)
    S->data[i]= color_u_to_s(data[i]) | ALPHABYTE;
#undef ALPHABYTE             
  for(i=CC->ssprites;i>=1;i--)
   if (CC->sprites[i-1]!=NULL) break;
  CC->nsprites= i;

  return RV;
}

void gfx_sprite_show(int wno,int sp,int sho)
{
  gui_canvas_t *CC;
  sprite_t *S;

  if (!(CC=gui_canvas_get(wno))) return ;
  if (sp>=CC->ssprites || sp<0) return ;
  S= CC->sprites[sp];
  if (!S) return ;
  if (sho==S->shown) return ;

  S->shown= sho;
  gui_canvas_update(CC, S->x, S->y, S->w, S->h);
} 

void gfx_sprite_remove(int wno, int sp)
{
  gui_canvas_t *CC;
  sprite_t *S;
  int i;

  if (!(CC=gui_canvas_get(wno))) return ;
  if (sp>=CC->ssprites || sp<0) return ;
  S= CC->sprites[sp];
  if (!S) return ;
  if (S->shown)
  {
    S->shown= 0;
    gui_canvas_update(CC, S->x, S->y, S->w, S->h);
  }
  gui_sprite_free(S);
  CC->sprites[sp]= NULL;

  for(i=CC->ssprites;i>=1;i--)
   if (CC->sprites[i-1]!=NULL) break;
  CC->nsprites= i;
} 

void gfx_sprite_move(int wno,int sp, int x,int y)
{
  gui_canvas_t *CC;
  sprite_t *S;

  if (!(CC=gui_canvas_get(wno))) return ;
  if (sp>=CC->ssprites || sp<0) return ;
  S= CC->sprites[sp];
  if (!S) return ;
  if (!S->shown) { S->x= x; S->y= y; return ; }

  int ux, uy, uw, uh;
  ux= S->x; uy= S->y; uw= S->w; uh= S->h;

  if (x<S->x) { ux= x; uw+= S->x-x; }
  else if (x>S->x) { uw+= x-S->x; }

  if (y<S->y) { uy= y; uh+= S->y-y; }
  else if (y>S->y) { uh+= y-S->y; }

  S->x= x; S->y= y;
  gui_canvas_update(CC, ux, uy, uw, uh);
} 

static widget_t *gui_canvas_new(int shor, int sver)
{
  gui_canvas_t *CC;
  CC= calloc(1,sizeof(*CC));
  CC->widget.klass=GFXCLASS_CANVAS;
  CC->can_scrollv= sver;
  CC->can_scrollh= shor;
  if (CC->can_scrollv)
    CC->scrollv= (scrollbar_t*) scrollbar_new(1, 0.0,1.0,0.1,0,
                                    CC,gui_canvas_scrollv);
  if (CC->can_scrollh)
  CC->scrollh= (scrollbar_t*) scrollbar_new(0, 0.0,1.0,0.1,0,
                                    CC,gui_canvas_scrollh);

  CC->click1= CC->click3= -1;

  CC->widget.position= (void*) gui_canvas_position;
  return &CC->widget;
} 


void gfx_canvas_draw_rectangle(int wno,int color, int x,int y,int w,int h)
{
  gui_canvas_t *CC;
  if (!(CC=gui_canvas_get(wno))) return ;
  if (w<=0 || h<=0 || !CC->buffer) return ;
  color= color_u_to_s(color);

  gui_canvas_clipped_hline(CC, color, x,x+w-1, y);
  gui_canvas_clipped_hline(CC, color, x,x+w-1, y+h-1);
 
  gui_canvas_clipped_vline(CC, color, x,y,y+h-1);
  gui_canvas_clipped_vline(CC, color, x+w-1,y, y+h-1);
  gui_canvas_update(CC,x,y,w,h);
}

void gfx_canvas_fill_rectangle(int wno,int color, int x,int y,int w,int h)
{
  gui_canvas_t *CC;
  int sx,sy,ex,ey,bw;

  if (!(CC=gui_canvas_get(wno))) return ;
  if (w<=0 || h<=0 || !CC->buffer) return ;
  color= color_u_to_s(color);

  sx= imin2(imax2(0,x),CC->buffer_w-1);
  sy= imin2(imax2(0,y),CC->buffer_h-1);

  ex= imin2(imax2(0,x+w), CC->buffer_w);
  ey= imin2(imax2(0,y+h), CC->buffer_h);

  bw= CC->buffer_w;
  for(y=sy;y<ey;y++)
    for(x=sx;x<ex;x++)
      CC->buffer[y*bw+x]= color;

  gui_canvas_update(CC,sx,sy,ex-sx,ey-sy);
}

void gfx_canvas_print(int wno, int font, int color, int x, int y, char *str)
{
  gui_canvas_t *CC;
  int bw;
  uint32_t *B; 
  int dp,sp;
  unsigned char *td; int tw, th, ta;
  int sx,sy,dx,dy,w,h;

  CC= (gui_canvas_t*) get_widget(wno);
  
  if (!CC || CC->widget.klass!=GFXCLASS_CANVAS || !CC->buffer
      || !str || !str[0]) return ;
  
  get_text_data(font,str,&tw,&th,&ta,&td);
 
  w= tw; h= th;
  sx= 0; sy= 0;
  dx= x; dy= y;
  if (clip_for_rect(&sx, &w, tw) ||
      clip_for_rect(&sy, &h, th) ||
      clip_for_rect(&dx, &w, CC->buffer_w) ||
      clip_for_rect(&dy, &h, CC->buffer_h)) 
    goto free_td;

  bw= CC->buffer_w;
  B= CC->buffer;
  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
    {
      dp= (dy+y)*bw + (dx+x);
      sp= (sy+y)*tw + (sx+x);
      B[dp]= blend_usS(td[sp], color, B[dp]);
    }

  gui_canvas_update(CC, dx, dy, w,h);
free_td:
  free(td);
}

void gfx_canvas_put_graymap
       (int wno, unsigned int color, int dx,int dy, int w, int h,
        unsigned char *source, int sw, int sh, int sx, int sy)
{
  gui_canvas_t *CC;
  int x,y;
  int bw;
  uint32_t *B; 
  int dp,sp;

  CC= (gui_canvas_t*) get_widget(wno);
  
  if (!CC || CC->widget.klass!=GFXCLASS_CANVAS || !CC->buffer ||
      clip_for_rect(&sx, &w, sw) ||
      clip_for_rect(&sy, &h, sh) ||
      clip_for_rect(&dx, &w, CC->buffer_w) ||
      clip_for_rect(&dy, &h, CC->buffer_h)) 
    return ;

  bw= CC->buffer_w;
  B= CC->buffer;
  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
    {
      dp= (dy+y)*bw + (dx+x);
      sp= (sy+y)*sw + (sx+x);
      B[dp]= blend_usS(source[sp], color, B[dp]);
    }

  gui_canvas_update(CC, dx, dy, w,h);
}

void gfx_canvas_put
       (int wno, int dx,int dy, int w, int h,
        unsigned int *source, int sw, int sh, int sx, int sy)
{
  gui_canvas_t *CC;
  int x,y;
  int bw;
  uint32_t *B; 
  int dp,sp;

  CC= (gui_canvas_t*) get_widget(wno);
  
  if (!CC || CC->widget.klass!=GFXCLASS_CANVAS || !CC->buffer ||
      clip_for_rect(&sx, &w, sw) ||
      clip_for_rect(&sy, &h, sh) ||
      clip_for_rect(&dx, &w, CC->buffer_w) ||
      clip_for_rect(&dy, &h, CC->buffer_h)) 
    return ;


  bw= CC->buffer_w;
  B= CC->buffer;

  for(y=0;y<h;y++)
    for(x=0;x<w;x++)
    {
      dp= (dy+y)*bw + (dx+x);
      sp= (sy+y)*sw + (sx+x);
      B[dp]= source[sp];
    }

  gui_canvas_update(CC, dx, dy, w,h);
}

int gfx_canvas_get_buffer
   (int wno, unsigned int *W, unsigned int *H, unsigned int **B)
{
  gui_canvas_t *CC;
  if (!(CC=gui_canvas_get(wno)))
  { 
    static unsigned int K;
    *W= *H= 1; *B= &K;
    return 1;
  }
  if (!CC->buffer)
    gfx_canvas_setsize(wno, CC->widget.w, CC->widget.h);
  *W= CC->buffer_w;
  *H= CC->buffer_h;
   // this is wrong, FIXME
  *B= (unsigned int*) CC->buffer;
  return 0;
}

void gfx_canvas_update(int wno, int x, int y, int w, int h)
{
  gui_canvas_t *CC;
  if ((CC=gui_canvas_get(wno))) gui_canvas_update(CC,x,y,w,h);
}


void gfx_canvas_set_zoom_params
  (int wno, int zf, int grid, unsigned int gridcolor,int w,int h)
{
  gui_canvas_t* CC;
  if (!(CC=gui_canvas_get(wno))) return ;
  w= imin2(CC->port_w/2,w);
  h= imin2(CC->port_h/2,h);
  CC->z_factor= zf;
  CC->z_ww= w;
  CC->z_wh= h;
  CC->z_grid= grid;
  CC->z_gridcolor= gridcolor;
  gui_canvas_zupdate(CC,0,0,CC->buffer_w,CC->buffer_h);
}


void gfx_canvas_zoom(int wno,int x,int y)
{
  gui_canvas_t* CC;
  int Xa,Xb,Xc,Xs,Xe;
  int Ya,Yb,Yc,Ys,Ye;
  int z, ww, hh, D, wx, wy;

  if (!(CC=gui_canvas_get(wno))) return ;
  if (!CC->buffer) return ;
  if (x<0 || y<0 || x>=CC->buffer_w || y>=CC->buffer_h) return ;
  if (!CC->z_factor || !CC->z_ww || !CC->z_wh) return ;

  z= CC->z_factor;
  ww= CC->z_ww;            hh= CC->z_wh;
  Xc= x;                   Yc= y;

  Xa= (ww-z)/2;            Xb= ww-z-Xa;
  Xs= Xc-GFXceil(Xa,z);       Xe= Xc+GFXceil(Xb,z);

  Ya= (hh-z)/2;            Yb= hh-z-Ya;
  Ys= Yc-GFXceil(Ya,z);       Ye= Yc+GFXceil(Yb,z);

  if (Xs<0) { Xe-= Xs; Xc-= Xs; Xs= 0; }
  if (Ys<0) { Ye-= Ys; Yc-= Ys; Ys= 0; }
  D= Xe-CC->buffer_w-1;
  if (D>0) { Xs-= D; Xc-= D; Xe= CC->buffer_w-1; }
  D= Ye-CC->buffer_h-1;
  if (D>0) { Ys-= D; Yc-= D; Ye= CC->buffer_h-1; }
  if (Xs<0 || Ys<0) return ; // FIXME: the buffer is too small, we should
                             // still display a zoom window, with decorations
                             // for the missing pixels.
  if (Xc-CC->pos_x<CC->buffer_w/2)
    wx= CC->port_w-CC->z_ww;
  else
    wx= 0; 

  if (Yc-CC->pos_y<CC->buffer_h/2)
    wy= CC->port_h-CC->z_wh;
  else
    wy= 0;

  if (CC->have_zoom && (wx!=CC->z_wx || wy!=CC->z_wy))
  {
    CC->have_zoom= 0;
    gui_canvas_update(CC,CC->z_wx+CC->pos_x,CC->z_wy+CC->pos_y,
                         CC->z_ww, CC->z_wh);
  }
  CC->have_zoom= 1;
  CC->z_cx= Xc;
  CC->z_cy= Yc;
  CC->z_wx= wx;
  CC->z_wy= wy;

  gui_canvas_zupdate(CC,0,0,CC->buffer_w,CC->buffer_h);
}

void gfx_canvas_hide_zoom(int wno)
{
  gui_canvas_t* CC;
  if (!(CC=gui_canvas_get(wno))) return ;
  if (!CC->have_zoom) return ;
  CC->have_zoom= 0;
  gui_canvas_update(CC,CC->z_wx+CC->pos_x,CC->z_wy+CC->pos_y,
                         CC->z_ww, CC->z_wh);
}

/***********************************************************************
 *                   CANVAS MOUSE POSITION REPORTING                   *
 ***********************************************************************/

static int gui_canvas_report_motion(void *arg, gfx_event_t *ev)
{
  gui_canvas_t *CC;
  CC= arg;
  if (ev->x < CC->widget.x ||
      ev->x >= CC->widget.x+CC->widget.w ||
      ev->y < CC->widget.y ||
      ev->y >= CC->widget.y+CC->widget.h)
    return 1;
  ev->type= GFXEV(CC->widget.number,POINTER_MOVE);
  ev->x= ev->x - CC->widget.x + CC->pos_x;
  ev->y= ev->y - CC->widget.y + CC->pos_y;
  return 0;
}


void gfx_canvas_begin_report_mouse_position(int wno)
{
  gui_canvas_t *CC;
  if (!(CC= gui_canvas_get(wno))) return ;
  if (!CC->buffer) return ;
  install_motion_handler(CC, gui_canvas_report_motion);
}

void gfx_canvas_end_report_mouse_position(int wno)
{
  gui_canvas_t *CC;
  if (!(CC= gui_canvas_get(wno))) return ;
  if (!CC->buffer) return ;
  uninstall_motion_handler();
}

static int gui_canvas_setbuffer(gui_canvas_t *CC,int new_w,int new_h)
{
  uint32_t *nb;

  if (new_w<=0) new_w= CC->buffer_w;
  if (new_h<=0) new_h= CC->buffer_h;

  if (new_w==CC->buffer_w && new_h==CC->buffer_h) return 1;

  nb= calloc(new_w*new_h, sizeof(uint32_t));
  if (CC->buffer)
  {
    int y,x;
    int cw,ch;
    cw= imin2(CC->buffer_w, new_w);
    ch= imin2(CC->buffer_h, new_h);
    for(y=0;y<ch;y++)
      for(x=0;x<cw;x++)
        nb[y*new_w + x]= CC->buffer[y*CC->buffer_w+x];
    free(CC->buffer);
  }
  CC->buffer= nb;
  CC->buffer_w= new_w;
  CC->buffer_h= new_h;

  return 0;
}

void gfx_canvas_setsize(int wno, int new_w, int new_h)
{
  gui_canvas_t *CC;
  CC= (gui_canvas_t*) get_widget(wno);
  if (!CC || CC->widget.klass!=GFXCLASS_CANVAS) return ;

  if (gui_canvas_setbuffer(CC,new_w,new_h)) return ;
  CC->pos_x= 0;
  CC->pos_y= 0;
  gui_canvas_arrange_scrollbars(CC);
  gui_canvas_draw_all(CC);
}

